/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/
///////////////////////////////////////////////////////////////////////////////
// @doc     CECScriptFile
// @module  CECScriptFile | File Access class
///////////////////////////////////////////////////////////////////////////////
#include "CECScriptFile.h"
#include "mdbexcept.h"
#include "mdbtypes2.h"

#ifndef CYRANO_TP_STANDALONE
#   include "RemoteRepository.h"
extern CRemoteRepository* Repos;
#endif

string enginesPath;
list<string> CECScriptFile::filename_cache;
#ifndef CYR_NO_STD_NAMESPACE
using namespace std;
#endif

///////////////////////////////////////////////////////////////////////////////
//@mfunc    constructor, define a file for use
//@parm	    FileId of file
CECScriptFile::CECScriptFile(const string& fileid):
        m_fileid(fileid),
        m_filename(""),
        m_options(0),
        m_access(0),
        m_File()
{
	enginesPath = getenv("OPENSTA_HOME");
	enginesPath += "Engines\\";
}
///////////////////////////////////////////////////////////////////////////////
//@mfunc Destructor

CECScriptFile::~CECScriptFile()
{}
///////////////////////////////////////////////////////////////////////////////
//@mfunc Open the file
//@parm  local file name for file (no pathname as will be retrieved from repository)
//@parm  access type
//@parm  other options (e.g. binary)

int CECScriptFile::Open(const string& filename,int access,int options)
{
    m_access=access;
    m_options=options;
    int openoptions=0;
    openoptions|=(options&TOF_FILOPT_BINFIL)?ios::binary:0;
    openoptions|=(access&TOF_FILACC_INPUT)?ios::in:0;
    openoptions|=(access&TOF_FILACC_OUTPUT)?ios::out:0;
    openoptions|=(access&TOF_FILACC_APPEND)?(ios::app|ios::out):0;
    bool have_file=false;
    if(m_access&TOF_FILACC_INPUT)//if for read only then only get if required
    {
        for(list<string>::iterator i=filename_cache.begin();i!=filename_cache.end();i++)
        {
            if(*i==filename)
                have_file=true;
        }
    }
    if(!have_file)
        GetFile(filename);
    m_File.open(m_filename.c_str(),openoptions);
    if (!m_File.good())
    {
        if(!m_File.eof()) throw mdbParseException("Unable to open file " + filename);
    }

#ifndef HPUX
    return m_File.is_open();
#else
    return 1;
#endif
}

string CECScriptFile::Read(int bytes)
{
    if(!(m_access&TOF_FILACC_INPUT))
        throw mdbParseException("Attempt to read file opened for output");

    string retstr("");
    char* buf=new char[bytes];
    char* nlpos;
    int bytes_read;

    long oldpos = m_File.tellg();      // Get file position
    if (oldpos == -1)
    {
        delete[] buf;
        throw mdbParseException(string("Error Reading from File ")+m_filename);
    }

    // Read in required number of bytes
    m_File.read(buf, bytes);

    // If at EOF, re-read character by character to end of record.
    if(m_File.eof())
    {
        m_File.clear();         // Reset stream state
        m_File.seekg(oldpos);   // Reset read position
        char nxtchr;
        nlpos = buf;
        do
        {
            m_File.read (&nxtchr, 1);
            if (m_File.eof())
            {
                m_File.clear();             // Reset stream state
                m_File.seekg(0L,ios::beg);  // Reset reading position to beginning of file
                *nlpos = '\0';
                retstr = buf;
                delete[] buf;
                return retstr;              // If end of file, return buffer now
            }

            if (!m_File.good())
            {
                delete[] buf;
                throw mdbParseException(string("Error Reading from File ")+m_filename);
            }

            *nlpos++ = nxtchr;
        }
        while (nxtchr != '\n');
        bytes_read = nlpos - buf;
    }
    else
    {
        nlpos = (char*) memchr (buf, '\n' ,bytes);
        bytes_read = bytes;
    }

    if(!(m_options&TOF_FILOPT_BINFIL))//not binary file reading
    {
        if(nlpos == NULL) //no newline found read until newline
        {
            char throw_away_char;
            do
            {
                m_File.read (&throw_away_char, 1);
                //Read the next byte from the file
                if(m_File.eof())
                {
                    m_File.clear();             // Reset stream state
                    m_File.seekg(0L,ios::beg);
                    retstr= buf;
                    delete[] buf;
                    return retstr;//if end of file return buffer now
                }

                if(!m_File.good())
                {
                    delete[] buf;
                    throw mdbParseException(string("Error Reading from File ")+m_filename);
                }
            }
            while (throw_away_char != '\n');
        }
        else//rewind to newline
        {
            size_t nllength=nlpos-buf;       //length to new line is address of new line-address of buf[0]
            long rewind = -(long)(bytes_read - nllength); //length to rewind is the length read in - the length to the new line
            m_File.seekg(rewind,ios::cur);
            buf[nllength]='\0';
        }

    }
    retstr=buf;
    delete[] buf;
    return retstr;
}

int CECScriptFile::Write(const string& data)
{
    string strData(data);
    if(!(m_options & TOF_FILOPT_BINFIL))
        strData+="\n";
    if(m_access & TOF_FILACC_INPUT)
        throw mdbParseException("Attempt to write to file opened for input");
    m_File.write(strData.c_str(), strData.length());
    if(!m_File.good())
        throw mdbParseException(string("Error Writing to File ")+m_filename);

    return 0;
}

int CECScriptFile::Close()
{
    m_File.close();
    if(!m_options&TOF_FILACC_INPUT)
        PutFile();
    return 0;
}

int CECScriptFile::Rewind()
{
    if(m_options&TOF_FILACC_INPUT)
    {
        m_File.clear();             // Reset stream state
        m_File.seekg(0L,ios::beg);
    }
    else
    {
        m_File.clear();             // Reset stream state
        m_File.seekp(0L,ios::beg);
    }
    return 0;
}

void CECScriptFile::GetFile(const string& Filename)
{
    string FilenamePath = enginesPath + FN_PATHSEP + Filename;
#ifdef CYRANO_TP_STANDALONE
	HKEY hKey;
	if(RegOpenKeyEx(HKEY_LOCAL_MACHINE,"SOFTWARE\\OpenSTA\\Repository",0,KEY_READ,&hKey)==ERROR_SUCCESS)
    {
	    DWORD dwType;
	    DWORD dwLen = MAX_PATH;
	    char szPath[MAX_PATH];
	    if(RegQueryValueEx(hKey,"Path",0,&dwType,(LPBYTE)szPath,&dwLen)==ERROR_SUCCESS)
            FilenamePath = string(szPath) + "Repository\\Data\\" + Filename;
	    RegCloseKey(hKey);
    }
#else
    try
    {
        m_OrbFile = new CECOrbFile(Repos->GetDataFolder(), Filename.c_str());
    }
    catch(...)//assume this is beacuse it is not there and try and create a new one
    {
        m_OrbFile = new CECOrbFile(Repos->GetDataFolder(), Filename.c_str(),"Data File");
    }
    // Validate the file exchange pointer
    if (m_OrbFile == NULL)
    {
        throw mdbParseException("Cannot get file");
    }
    // Copy the file requested to the local disk space.
    m_OrbFile->CheckOut(FilenamePath.c_str(),true);
#endif
    m_filename=FilenamePath;
}
void CECScriptFile::PutFile()
{
#ifndef CYRANO_TP_STANDALONE
    m_OrbFile->CheckIn(m_filename.c_str(),true);
    delete m_OrbFile;
#endif
}

