/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CyrFFApi
// @module  SCL File Wrapper | This module contains the implementation of
//          the CECFFSCLFile wrapper class.  This class simply extends the
//          CECFFFile wrapper class.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include "CECFFSCLFile.h"
#include "CECFFImpactFolders.h"
#include "CECFFImpactProject.h"
#include <Process.hpp>

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor retrieves a specified source file from the
//          OpenSTARep repository.
// @parm    Pointer to the parent Folder object.
// @parm    The name of the File to retrieve.
CECFFSCLFile::CECFFSCLFile( CECFFFolder* parent, const char* name )
        : CECFFFile( parent, name )
{}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor creates a new File object in the specified Folder.
// @parm    Pointer to the parent Folder.
// @parm    The name of the File to create.
// @parm    The description of the File to create.

CECFFSCLFile::CECFFSCLFile( CECFFFolder* parent, const char* name, const char* desc )
        : CECFFFile( parent, name, desc )
{}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method compiles the source file into an object code file.
// @rdesc   Returns true if the compilation was successful, otherwise false.

bool CECFFSCLFile::Compile()
{
    // Error information
    const char* where = "CECFFSCLFile::Compile";

    // Create command line for spawn process object and pass it
    string cmdLine = BuildCommandLine();
    CTProcess spawn( cmdLine );
    long pid = spawn.start();

    // Wait for process to terminate
    long status = spawn.wait();

    // Get file name of generated file
    string path = GetTOFName();

    // Get required parameters
    CECFFFolder* rootFolder = new CECFFFolder( GetParentProject() );
    CECFFObjectCodeFolder* codeFolder = new CECFFObjectCodeFolder( rootFolder );
    size_t index = path.last( DIR_SLASH );
    string name = path( index + 1, path.length() - index - 1 );

    // Create new OpenSTARep file
    try
    {
        CECFFFile* tofFile = new CECFFFile( codeFolder, name );
        tofFile->Update();
        delete tofFile;
    }
    catch( CyrRepositoryException ex )
    {
        if ( ex.status() != CYR_FILE_NOT_FOUND )
            throw CyrRepositoryException( ex.what(), where, ex.status() );
        else
        {
            CECFFFile* tofFile = new CECFFFile( codeFolder, name, "Compiled TOF file" );
            delete tofFile;
        }
    }

    // Tidy memory
    delete rootFolder;
    delete codeFolder;

    // Return success or failure
    if ( status != -1 )
        return true;
    else
        return false;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method constructs the command line string that will compile
//          the source file into an object code file.
// @rdesc   The command line string.
string CECFFSCLFile::BuildCommandLine()
{
    // Error information
    const char* where = "CECFFSCLFile::BuildCommandLine";

    // Get project for this object
    CECFFImpactProject* project = static_cast<CECFFImpactProject*>(GetParentProject());
    if ( project == NULL )
        throw CyrRepositoryException( "Project not an OpenSTA project", where,
                                      CYR_PROJECT_WRONG_TYPE );

    // Construct command line to execute
    string cmdLine( "scl.exe -o \"" );
    cmdLine += GetTOFName();
    cmdLine += "\" ";

    if ( project != NULL )
    {
        // Add build settings parameters
        if ( project->GetBuildList() )
            cmdLine += "-l on ";
        if ( project->GetBuildDebug() )
            cmdLine += "-d ";
        if ( project->GetBuildCheckSteps() )
        {
            char buf[20];
            sprintf( buf, "-m %d ", project->GetBuildMaxSteps() );
            cmdLine += buf;
        }
    }

    // Add source file name
    cmdLine += "-v \"";
    cmdLine += GetSCLName();
    cmdLine += "\"";

    return cmdLine;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Returns the full path name of the underlying SCL source file on the
//          remote OpenSTARep machine.  This is required for the command line
//          to generate the TOF file.
// @rdesc   The full path name of the remote file.
string CECFFSCLFile::GetSCLName()
{
    // Get current file path
    return GetFullPath();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Returns the full path name of the TOF file generated when the
//          compile method is called.
// @rdesc   The full path name of the remote file.
string CECFFSCLFile::GetTOFName()
{
    // Get current file path
    string path = GetSCLName();

    // Extract file name and source code folder name
    size_t index = path.last( DIR_SLASH );
    string filename = path( index + 1, path.length() - index - 1 );
    path.resize( index );
    index = path.last( DIR_SLASH );
    path.resize( index + 1 );

    // Convert filename from scl (or whatever) to tof
    index = filename.last( '.' );
    filename.resize( index + 1 );
    filename += "tof";

    // Add new directory and filename
    path += CECFFObjectCodeFolder::FOLDER_NAME;
    path += "\\";
    path += filename;

    return path;
}



