/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CyrStRepository
// @module  Server Implementation | This module contains the implementation of
//          the CyrStServer CORBA object.  This CORBA object is simply a CORBA
//          wrapper around the OpenSTARep COM interface wrapper.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#ifdef _WIN32
#   pragma warning (disable : 4503)
#endif

#include <Repository/CyrStServer_i.h>
#include <CyrException.h>

#ifdef __VDK002__
#include <Logging/CECLog.h>
extern CECLog* v2logger;
#endif

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor extracts the relevant parameters from the
//          parameter list and creates an instance of the relevant COM wrapper
//          class.
// @parm    Pointer to parent Provider module
// @parm    List of parameters to be used to construct server object
CyrStServerImpl::CyrStServerImpl( CyrStProvider* pProvider, ParamList_ptr params )
{
    // And add some tracing
#ifdef __VDK002__
    v2logger->AddTrace( CYR_VDK002, 50, "Constructing a CyrStServerImpl." );
#endif
    // Assign member variables
    m_provider = pProvider;

    // Extract parameters from list
    CORBA::String_var user     = params->GetValue( CyrStServer::PARAM_USER );
    CORBA::String_var password = params->GetValue( CyrStServer::PARAM_PASSWORD );

#ifdef __VDK002__
    string user_msg("User: " );
    user_msg+= user;
    v2logger->AddTrace( CYR_VDK002, 50,  user_msg.c_str() );
#endif


    // Create COM wrapper class
    m_com = new CECFFServer( user, password );

    // Register COM pointer
    m_regName = m_provider->StorePtr( m_com, m_com->GetAddress().c_str(), user );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The destructor destroys the underlying COM object as part of its
//          tidying up.  It also disconnects from the OpenSTARep database.
CyrStServerImpl::~CyrStServerImpl()
{
    // And add some tracing
#ifdef __VDK002__
    v2logger->AddTrace( CYR_VDK002, 50, "Destroying a CyrStServerImpl." );
#endif

    // Destroy underlying server reference
    m_provider->RemovePtr( m_regName );
#ifdef __VDK002__
    string reg_msg = "Removing Registered name: " + m_regName;
    v2logger->AddTrace( CYR_VDK002, 100, reg_msg.c_str() );
#endif

    // Disconnect from server
    delete m_com;
    m_com = NULL;

#ifdef __VDK002__
    v2logger->AddTrace( CYR_VDK002, 50, "Deleted Com Member." );
#endif
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Within the Reaper mechanism of the Factory, each object must be
//          given a unique name.  This method returns that name for this
//          particular object.  The name itself is generated by the Provider
//          when the underlying COM object is stored.  This name can be used
//          to increase the lifetime of the CORBA object.  The name itself is
//          generated from the name of the Server, plus the name of the current
//          user.  This should guarantee uniqueness unless the same user wishes
//          to create the same object again - in which case they are
//          responsible for producing a different name.
// @rdesc   The registered name.
char* CyrStServerImpl::GetRegisteredName()
{
    // And add some tracing
#ifdef __VDK002__
    string msg = "Get Registered Server Name: " + m_regName;
    v2logger->AddTrace( CYR_VDK002, 100, msg.c_str() );
#endif
    return CORBA::string_dup( m_regName.c_str() );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method retrieves a list of Project names currently available
//          through the underlying COM interface.  This list is separated by
//          a single '\|' character and the list is terminated by two '\|'
//          characters.
// @devnote This list is only a snap shot and there is no guarantee that two
//          consecutive requests will return either the same list or the same
//          ordering within the list.
// @rdesc   The bar separated list.
// @parm    Returns the number of elements in the list.
char* CyrStServerImpl::GetProjects( long& count )
{
    // And add some tracing
#ifdef __VDK002__
    v2logger->AddTrace( CYR_VDK002, 100, "Get List of Project Names." );
#endif
    return CORBA::string_dup( m_com->GetProjects( count ).c_str());
}

