/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CyrStRepository
// @module  Test Run Implementation | This module contains the implementation
//          of the CyrStTestRun CORBA object.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#ifdef _WIN32
#   pragma warning (disable : 4503 4786)
#endif

#include <CyrHelper.inl>
#include <Repository/CyrStTestRun_i.h>
#include <Repository/CyrStHelper.inl>
#ifdef __VDK002__
#include <Logging/CECLog.h>
extern CECLog* v2logger;
#endif
///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor extracts the relevant information from the
//          parameter list and either uses it to construct a new Test Run or
//          retrieve an existing one.
// @parm    Pointer to the parent Provider object.
// @parm    List of construction parameters.
CyrStTestRunImpl::CyrStTestRunImpl( CyrStProvider* pProvider, ParamList_ptr params )
        : CyrStFolderImpl( pProvider )
{
    // And add some tracing
#ifdef __VDK002__
    v2logger->AddTrace( CYR_VDK002, 50, "Constructing a CyrStTestRunImpl." );
#endif
    const char* where = "CyrStTestRunImpl::CyrStTestRunImpl";

    // Retrieve parameters
    CORBA::String_var paramParent = params->GetValue( CyrStTestRun::PARAM_STORE );
    CORBA::String_var paramName = params->GetValue( CyrFactory::PARAM_NAME );
    if ( strlen( paramParent )== 0 )
        throw CyrRepositoryException( "Invalid parent parameter", where, CYR_FOLDER_NEW_FAILED );

    // Get correct parent folder and server interfaces from array
    CECFFFolder* pParent = (CECFFFolder*)m_provider->GetPtr( paramParent.in() );

    // Determine if we are retrieving or creating a Folder
    if ( strlen( paramName ) != 0 )
    {
        // And add some tracing
#ifdef __VDK002__
        string msg("Retrieving an existing TestRun:");
        msg+= paramName;
        v2logger->AddTrace( CYR_VDK002, 100, msg.c_str() );
#endif
        // Retrieve existing Test Run
        m_com = new CECFFTestRun( pParent, paramName );
    }
    else
    {
        // And add some tracing
#ifdef __VDK002__
        v2logger->AddTrace( CYR_VDK002, 100, "Creating a New TestRun." );
#endif
        // Extract remaining parameters
        CORBA::String_var paramDesc = params->GetValue( CyrStTestRun::PARAM_DESC );
        CORBA::String_var paramNode = params->GetValue( CyrStTestRun::PARAM_NODENAME );
        CORBA::String_var paramExec = params->GetValue( CyrStTestRun::PARAM_EXECNAME );

        // And add some tracing
#ifdef __VDK002__
        string msg ( "Creating a New TestRun on node:" );
        msg+= paramNode;
        v2logger->AddTrace( CYR_VDK002, 100, msg.c_str() );
#endif

        if ( strlen( paramDesc ) == 0 )
            throw CyrRepositoryException( "Invalid description parameter", where, CYR_FOLDER_NEW_FAILED );
        if ( strlen( paramNode ) == 0 )
            throw CyrRepositoryException( "Invalid node name parameter", where, CYR_FOLDER_NEW_FAILED );
        if ( strlen( paramExec ) == 0 )
            throw CyrRepositoryException( "Invalid executor name parameter", where, CYR_FOLDER_NEW_FAILED );
#ifdef __VDK002__
        msg = "Creating a New CECComTestRun ";
        v2logger->AddTrace( CYR_VDK002, 100, msg.c_str() );
#endif

        // Create new test run
        m_com = new CECFFTestRun( pParent, paramDesc, paramNode, paramExec );
    }
#ifdef __VDK002__
    string msg = "Have a New CECComTestRun Get Server";
    v2logger->AddTrace( CYR_VDK002, 100, msg.c_str() );
#endif

    // Get current user name
    CECFFServer* pServer = m_com->GetParentServer();
#ifdef __VDK002__
    msg = "Got Server get username";
    v2logger->AddTrace( CYR_VDK002, 100, msg.c_str() );
#endif

    string user = pServer->GetUserName();

#ifdef __VDK002__
    msg = "Created by user:" + user;
    v2logger->AddTrace( CYR_VDK002, 100, msg.c_str() );
#endif

    // Build registered Folder name and store pointer with parent Provider
    m_regName = BuildItemPath( m_com ).c_str() + string(m_com->GetName());
    m_regName = m_provider->StorePtr( m_com, m_regName.c_str(), user.c_str() );
}

CyrStTestRunImpl::~CyrStTestRunImpl()
{
    // And add some tracing
#ifdef __VDK002__
    v2logger->AddTrace( CYR_VDK002, 50, "Destroying a CyrStTestRunImpl." );
#endif
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the name of the Test being run.
// @rdesc   The name of the Test.
char* CyrStTestRunImpl::GetTestName()
{
    CECFFTestRun* pTestRun = static_cast<CECFFTestRun*>( m_com );

    // And add some tracing
#ifdef __VDK002__
    string msg = "Get Test Name: " + pTestRun->GetTestName();
    v2logger->AddTrace( CYR_VDK002, 100, msg.c_str() );
#endif

    if ( pTestRun != NULL )
        return CORBA::string_dup( pTestRun->GetTestName().c_str() );
    else
        return CORBA::string_dup( "" );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns a description of the test run.
// @rdesc   The Test Run description.
char* CyrStTestRunImpl::GetDescription()
{
    CECFFTestRun* pTestRun = static_cast<CECFFTestRun*>( m_com );

    // And add some tracing
#ifdef __VDK002__
    string msg = "Get Test Description: " + pTestRun->GetDescription();
    v2logger->AddTrace( CYR_VDK002, 100, msg.c_str() );
#endif

    if ( pTestRun != NULL )
        return CORBA::string_dup( pTestRun->GetDescription().c_str() );
    else
        return CORBA::string_dup( "" );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the name of the node that is currently
//          executing the Test.
// @rdesc   The node name.
char* CyrStTestRunImpl::GetNodeName()
{
    CECFFTestRun* pTestRun = static_cast<CECFFTestRun*>( m_com );

    // And add some tracing
#ifdef __VDK002__
    string msg = "Get Node Name: " + pTestRun->GetNodeName();
    v2logger->AddTrace( CYR_VDK002, 100, msg.c_str() );
#endif

    if ( pTestRun != NULL )
        return CORBA::string_dup( pTestRun->GetNodeName().c_str() );
    else
        return CORBA::string_dup( "" );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the name of the executor that is currently
//          executing the Test.
// @rdesc   The executor name.
char* CyrStTestRunImpl::GetExecutorName()
{
    CECFFTestRun* pTestRun = static_cast<CECFFTestRun*>( m_com );

    // And add some tracing
#ifdef __VDK002__
    string msg = "Get Executer Name: " + pTestRun->GetExecutorName();
    v2logger->AddTrace( CYR_VDK002, 100, msg.c_str() );
#endif

    if ( pTestRun != NULL )
        return CORBA::string_dup( pTestRun->GetExecutorName().c_str() );
    else
        return CORBA::string_dup( "" );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the version number of the Test currently being
//          executed.
// @devnote This might need changing to better reflect OpenSTARep's version
//          numbers.
// @rdesc   The version number.
long CyrStTestRunImpl::GetTestVersion()
{
    /*CECComTestRun* pTestRun = dynamic_cast<CECComTestRun*>( m_com );

    // And add some tracing
    #ifdef __VDK002__
    string msg = "Get Test Version: " + pTestRun->GetTestVersion();
    v2logger->AddTrace( CYR_VDK002, 100, (const char*) msg );
    #endif

    if ( pTestRun != NULL )
        return pTestRun->GetTestVersion();
    else*/
    return 0;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the time the Test Run was started.  The time
//          format is suitable for creation of a CTTime object.  It
//          is a long value that represents the number of seconds since
//          midnight 01/01/1901.
// @rdesc   The start time.
long CyrStTestRunImpl::GetTimeStarted()
{
    // And add some tracing
#ifdef __VDK002__
    v2logger->AddTrace( CYR_VDK002, 100, "Obtaining time when the test was started." );
#endif
    CECFFTestRun* pTestRun = static_cast<CECFFTestRun*>( m_com );
    if ( pTestRun != NULL )
        return pTestRun->GetTimeStarted().getSeconds();
    else
        return 0;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the time the Test Run was stopped.  The time
//          format is suitable for creation of a CTTime object.  It
//          is a long value that represents the number of seconds since
//          midnight 01/01/1901.
// @rdesc   The stop time.
long CyrStTestRunImpl::GetTimeStopped()
{
    // And add some tracing
#ifdef __VDK002__
    v2logger->AddTrace( CYR_VDK002, 100, "Obtaining time when the test finished." );
#endif

    CECFFTestRun* pTestRun = static_cast<CECFFTestRun*>( m_com );
    if ( pTestRun != NULL )
        return pTestRun->GetTimeStopped().getSeconds();
    else
        return 0;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   As the time a Test Run stops cannot be determined at creation, this
//          method is required.  It necessarily causes an update to the
//          underlying COM interface.
// @parm    The time the Test Run stopped.
void CyrStTestRunImpl::StopTestRun()
{
    // And add some tracing
#ifdef __VDK002__
    v2logger->AddTrace( CYR_VDK002, 100, "Obtaining Test Run in order to stop it." );
#endif
    CECFFTestRun* pTestRun = static_cast<CECFFTestRun*>( m_com );
    if ( pTestRun != NULL )
    {
        // And add some tracing
#ifdef __VDK002__
        v2logger->AddTrace( CYR_VDK002, 100, "Attempting to stop Test Run." );
#endif
        pTestRun->StopTestRun();
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method retrieves a sequence of test output data as a
//          collection of strings.  This method assumes that all test output
//          can be stored as a single line in a text file as this is how the
//          data will be eventually stored.
// @rdesc   The sequence of strings.
// @parm    The type of output to retrieve.
// @parm    The start index of the output to retrieve.
// @parm    The stop index of the ouptut to retrieve.
char* CyrStTestRunImpl::GetFirstOutput( const char* outType )
{
    // Get correct entry from output table
    CECFFTestRun* pTestRun = static_cast<CECFFTestRun*>( m_com );
    if ( pTestRun != NULL )
        return CORBA::string_dup(  pTestRun->GetFirstOutputString( outType ).c_str());
    else
        return CORBA::string_dup( "" );
}

char* CyrStTestRunImpl::GetNextOutput( const char* outType )
{
    // Get correct entry from output table
    CECFFTestRun* pTestRun = static_cast<CECFFTestRun*>( m_com );
    if ( pTestRun != NULL )
        return CORBA::string_dup(  pTestRun->GetNextOutputString( outType ).c_str());
    else
        return CORBA::string_dup( "" );
}
char*  CyrStTestRunImpl::GetOutput(const char*  outType, CORBA::Long  Iter)
{
    CECFFTestRun* pTestRun = static_cast<CECFFTestRun*>( m_com );
    if ( pTestRun != NULL )
        return CORBA::string_dup(pTestRun->GetOutputString( outType, Iter ).c_str());
    else
        return CORBA::string_dup("");

}
CORBA::Long  CyrStTestRunImpl::GetNoOutputEntries(const char*  outType)
{
    CECFFTestRun* pTestRun = static_cast<CECFFTestRun*>( m_com );
    if ( pTestRun != NULL )
        return (  pTestRun->GetNoOutputEntries( outType));
    else
        return ( 0 );

}

CORBA::Long  CyrStTestRunImpl::GetOutputOffset(const char*  outType)
{
    CECFFTestRun* pTestRun = static_cast<CECFFTestRun*>( m_com );
    if ( pTestRun != NULL )
        return (  pTestRun->GetOutputOffset( outType));
    else
        return ( 0 );

}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method adds a new test output entry to the relevant file it
//          is stored in.  This method assumes that all test output can be
//          stored as a single line in a text file.
// @parm    The type of output to add.
// @parm    The stringified output value.
void CyrStTestRunImpl::AddTestOutput( const char* outType, const char* entry )
{
    CECFFTestRun* pTestRun = static_cast<CECFFTestRun*>( m_com );
    if ( pTestRun != NULL )
        pTestRun->AddTestOutput( outType, entry );
}

char*  CyrStTestRunImpl::GetNextOutputMsg(const char* outType, CORBA::Long& filePos)
{
    CECFFTestRun* pTestRun = static_cast<CECFFTestRun*>( m_com );
    if ( pTestRun != NULL )
        return CORBA::string_dup(pTestRun->GetNextOutputMsg( outType, filePos ).c_str());
	else
		return CORBA::string_dup( "" );
}
