/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// This class defines interprocess mutexes, for processes running on the same
// machine.
///////////////////////////////////////////////////////////////////////////////


#ifndef __IPMUTEX__
#define __IPMUTEX__

//=================================================================< Includes >
#ifndef __DEFINITION__
#  include "definition.h"
#endif  // __DEFINITION__

#ifdef _WIN32
#   include <windows.h>
#endif

#ifndef __MTSAFE__
#  include "mtsafe.hpp"
#endif  // __MTSAFE__

#ifndef __MUTEX__
#  include "mutex.hpp"
#endif  // __MUTEX__

#ifdef SUN
#   include <sys/time.h>
#   include <sys/stat.h>
#   include <fcntl.h>
#   include <errno.h>
#   include <sys/mman.h>
#   include <pthread.h>
#   define SHAREMEMORYNAME	"vdk10ShareMem"
#endif

#include <string>


//====================================================================< Class >
enum CtWaitResult {CT_WAIT_OK, CT_WAIT_TIMEOUT, CT_WAIT_ERROR};

class CTIPMutex
{
public:
    // This constructor creates a mutex object.
    // If zName is not NULL:
    //      - zName defines the mutex's name,
    //      - if the mutex object already exists, it is opened.
    //      - another process can open the mutex using its name.
    //      - The attach() method should never be called.
    //
    // If zName is NULL:
    //      - the mutex's name is generated (as an OS key or handle),
    //      - the generated name is unique: no collision can happen at any given
    //        time, but the names can be recycled after the mutex is deleted.
    //      - the generated name can be retrieved later with the getName() method,
    //      - a child process can open the mutex using the nameless constructor
    //        followed by a call to the attach() method, passing the name returned
    //        by getName() in the parent process.
    //      - NOTE: the generated name is only usable in a child process if this
    //        child process inherits inheritable kernel object handles from its
    //        parent.
    // The ulTimeOutMs argument specifies the timeout used while waiting for the
    // mutex to be acquired (when the lock() method is called).
    CTIPMutex(const char* zName, unsigned long ulTimeOutMs = 30000);

    // This constructor is to be used by a child process wishing to access an
    // already existing, inherited mutex. This mutex must have been created as
    // anonymous in the parent process (first constructor with zName set to NULL).
    // The call to this constructor must be followed by a call to attach().
    CTIPMutex(unsigned long ulTimeOutMs = 30000);

    ~CTIPMutex();

    // This method is to be called by a child process wishing to access an
    // already existing, inherited mutex. This mutex must have been created as
    // anonymous in the parent process (first constructor with zName set to NULL).
    // The call to this method must be preceded by a call to the second constructor.
    // zAutoName is the generated name returned by getName() in the parent process.
    bool         attach(const char* zAutoName);

    std::string  getName()    const {return strName_;}
    bool         isAutoName() const {return bAutoName_;}

    bool         isValid(void);
    CtWaitResult lock(void);
    bool         unlock (void);

protected:
    std::string strName_;   // The mutex's name.
    bool        bAutoName_; // The name was generated.
#ifdef _WIN32
    HANDLE      hMutex_;
    DWORD       dwMilliseconds_;
#endif

private:
    //// Forbidden:
    CTIPMutex(const CTIPMutex&);
    operator=(const CTIPMutex&);
};

///////////////////////////////////////////////////////////////////////////////
// This class defines interprocess signals, for processes running on the same
// machine.
///////////////////////////////////////////////////////////////////////////////

class CIPSignal
{
public:
    // This constructor creates a signal object, which is an object having 2
    // stable states. Thus, one thread can wait for another (possibly in
    // another process) to set the signal.
    // If zName is not NULL:
    //      - zName defines the signal's name,
    //      - if the signal object already exists, it is opened.
    //      - another process can open the signal using its name.
    //      - The attach() method should never be called.
    //
    // If zName is NULL:
    //      - the signal's name is generated (as an OS key or handle),
    //      - the generated name is unique: no collision can happen at any given
    //        time, but the names can be recycled after the signal is deleted.
    //      - the generated name can be retrieved later with the getName() method,
    //      - a child process can open the signal using the nameless constructor
    //        followed by a call to the attach() method, passing the name returned
    //        by getName() in the parent process.
    //      - NOTE: the generated name is only usable in a child process if this
    //        child process inherits inheritable kernel object handles from its
    //        parent.
    // The ulTimeOutMs argument specifies the timeout used while waiting for the
    // signal to be set (when the signal() method is called).
    CIPSignal(const char* zName, unsigned long ulTimeOutMs = 30000);

    // This constructor is to be used by a child process wishing to access an
    // already existing, inherited signal. This signal must have been created as
    // anonymous in the parent process (first constructor with zName set to NULL).
    // The call to this constructor must be followed by a call to attach().
    CIPSignal(unsigned long ulTimeOutMs = 30000);

    ~CIPSignal();

    // This method is to be called by a child process wishing to access an
    // already existing, inherited signal. This signal must have been created as
    // anonymous in the parent process (first constructor with zName set to NULL).
    // The call to this method must be preceded by a call to the second constructor.
    // zAutoName is the generated name returned by getName() in the parent process.
    bool         attach(const char* zAutoName);

    std::string  getName()    const {return strName_;}
    bool         isAutoName() const {return bAutoName_;}

    bool         isValid();
    CtWaitResult wait();
    bool         signal();

protected:
    std::string strName_;   // The event's name.
    bool        bAutoName_; // The name was generated.
#ifdef _WIN32
    DWORD       dwMilliseconds_;
    HANDLE      hEvent_;
#elif defined( SUN )
	static int iShareMemoryId_ ;
	static CTMutex signalMutex_ ;
	int ipcFd_ ;
	pthread_cond_t *pCondition_ ;
	CTMutex condMutex_ ;
	int iMilliseconds_ ;
#endif

private:
    //// Forbidden:
    CIPSignal(const CIPSignal&);
    operator=(const CIPSignal&);
};

#endif
 
