/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef __MTHREAD__
#define __MTHREAD__

//
//  class CTThread : Should be used to create, to access to and to control
//  --------------   a thread which is not the current thread.
//                 - constructor : You construct a new thread by specifying
//                   the start routine and its arguments pointer. You can
//                   also specify an abort routine with its arguments, or
//                   the initial state of the thread (running or suspended).
//                 - getThreadId : Get the thread identifier.
//                 - getState : Get the thread state.
//                 - suspend : Suspend the thread.
//                 - resume : Resume the thread.
//                 - kill : Kill the thread. Specify optionnaly if the
//                   current thread has to wait for the termination of
//                   the specified thread.
//
//  class CTCurrentThread : Should be used to control the current thread.
//  --------------------- - getThreadId : Get the thread identifier.
//                        - getState : Get the thread state.
//                        - getData : Get the data pointer associated to the
//                          current thread.
//                        - setData : Associate a data pointer to the current
//                          thread.
//                        - setAbortRoutine : Specify the user abort routine to
//                          call when the current thread aborts.
//                        - abort : Abort the current thread. The user abort
//                          routine will be called (if any).
//                        - sleep : Make the current thread to sleep for the
//                          specified time (in milli-seconds).
//                        - join : Blocks the current thread until the
//                          specified thread terminates. A pointer to the
//                          termination status of the specified thread can be
//                          passed as argument.
//


//==================================================================< Pragmas >
#ifdef  WIN32
#  pragma warning(disable : 4710)
#endif  // WIN32


//=================================================================< Includes >
#ifndef __DEFINITION__
#  include "definition.h"
#endif  // __DEFINITION__

#ifndef __EXCEPTION__
#  include "exception.hpp"
#endif  // __EXCEPTION__

#include <stddef.h>

#ifndef __FUNCTOR__
#  include "functor.hpp"
#endif  // __FUNCTOR__

#ifndef __MTSAFE__
#  include "mtsafe.hpp"
#endif  // __MTSAFE__


//====================================================================< Types >
enum enumThreadState                                    // Thread state.
{
    THREAD_CREATED,
    THREAD_RUNNING,
    THREAD_SUSPENDED,
    THREAD_KILLED,
    THREAD_TERMINATED
} ;

enum enumThreadPriority
{
    PRIORITY_HIGH,
    PRIORITY_NORMAL,
    PRIORITY_LOW,
    PRIORITY_ERROR
} ;


typedef void (*fnThreadAbortRoutine) (void*) ;


//====================================================================< Class >

///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////// Class CTThreadException //
///////////////////////////////////////////////////////////////////////////////
// Class used to keep threads exception.
class CTThreadException : public CTException
{
    //-----------------------------------------------------------------------------
public:                     // Constructor(s) and destructor.
    CTThreadException (const string& clMsg, const string& clFilename, short sLineno, long lNumber = -1L)
: CTException (clMsg, clFilename, sLineno, lNumber)
    { }


public:                     // Operators.
    void                    operator= (const CTThreadException& exception)
    {
        CTException::operator= (exception) ;
    }

    //-----------------------------------------------------------------------------
}

; // CTThreadException

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////// End of Class CTThreadException //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////// Class CTThread //
///////////////////////////////////////////////////////////////////////////////
// Class used to create and control others threads than the current thread.
class CTThread : virtual public CTMTSafe
{
    //-----------------------------------------------------------------------------
public:                         // Constructor(s) and destructor.
    CTThread (const CTFunctor& functorStartRoutine, enumThreadPriority ePriority = PRIORITY_NORMAL)
: pThreadInfo_ (NULL)
    {
        create (functorStartRoutine, NULL, THREAD_RUNNING, ePriority) ;
    }
    CTThread (const CTFunctor& functorStartRoutine, enumThreadState eInitialState, enumThreadPriority ePriority = PRIORITY_NORMAL)
: pThreadInfo_ (NULL)
    {
        create (functorStartRoutine, NULL, eInitialState, ePriority) ;
    }
    CTThread (const CTFunctor& functorStartRoutine, fnThreadAbortRoutine pAbortRoutine, enumThreadState eInitialState = THREAD_RUNNING, enumThreadPriority ePriority = PRIORITY_NORMAL)
: pThreadInfo_ (NULL)
    {
        create (functorStartRoutine, pAbortRoutine, eInitialState, ePriority) ;
    }
    CTThread (const void* pThreadId)
: pThreadInfo_ ( (void*) pThreadId)
    { }

    ~CTThread () ;

public:                         // Get methods.
    const void*                 getThreadId (void) const
    {
        return pThreadInfo_ ;
    }
    const enumThreadState       getState (void) const ;
    const enumThreadPriority    getPriority (void) const ;

public:                         // Control methods.
    unsigned int                suspend (void) ;
    unsigned int                resume (void) ;
    void                        kill (void* pStatus = NULL) ;
    void                        setPriority (enumThreadPriority) ;

    //-----------------------------------------------------------------------------
private:                        // Constructor(s) and destructor.
    CTThread (const CTThread&) ;

private:                        // Operators.
    CTThread&                   operator= (const CTThread&) ;

private:                        // Control methods.
    void                        create (const CTFunctor& functorStartRoutine, fnThreadAbortRoutine pAbortRoutine, enumThreadState eInitialState, enumThreadPriority ePriority) ;

private:                        // Members.
    void*                       pThreadInfo_ ;

    //-----------------------------------------------------------------------------
}

; // CTThread

int operator==(const CTThread& t1,const CTThread& t2);


///////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////// End of Class CTThread //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////// Class CTCurrentThread //
///////////////////////////////////////////////////////////////////////////////
// Class used to control the current thread.
class CTCurrentThread : virtual public CTMTSafe
{
    //-----------------------------------------------------------------------------
public:                         // Constructor(s) and destructor.

public:                         // Get methods.
    static void*                getThreadId (void) ;
    static enumThreadState      getState (void) ;
    static void*                getData (void) ;
    static enumThreadPriority   getPriority (void) ;

public:                         // Control methods.
    static void                 setAbortRoutine (fnThreadAbortRoutine pAbortRoutine) ;
    static void                 setData (void* pData) ;
    static void                 abort (void* pStatus = NULL) ;
    static void                 sleep (unsigned long ulMilliseconds = 0L) ;
    static void                 join (CTThread& thread, void** ppStatus = NULL) ;

    //-----------------------------------------------------------------------------
private:                        // Constructor(s) and destructor.
    CTCurrentThread () ;
    CTCurrentThread (const CTCurrentThread&) ;

private:                        // Operators.
// ### L.M.
//    void*                       operator new (size_t) ;
//    void                        operator delete (void*) ;
    CTCurrentThread&            operator= (const CTCurrentThread&) ;

    //-----------------------------------------------------------------------------
}

; // CTCurrentThread

///////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////// End of Class CTCurrentThread //
///////////////////////////////////////////////////////////////////////////////

#endif  // __MTHREAD__
