/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*===========================================================================*/
/*                                       */
/*               **** C O M M O N ****                   */
/*                                       */
/*===========================================================================*/
/*                                       */
/*          Module Name: BIT_ROUTINES_WIN32              */
/*          ============                         */
/*                                       */
/* Description:                                  */
/* ============                                  */
/*                                       */
/*  This module contains a number of routines for manipulating       */
/*  bit masks.                               */
/*                                       */
/*  Note that this module contains only a subset of the routines in      */
/*  BIT_ROUTINES_VMS, as required by the V-products.             */
/*                                       */
/*                                       */
/*===========================================================================*/


/*
 *  Standard ANSI header files.
 */
#include <stdio.h>
#include <limits.h>
#include <psl_macros.h>         /* General PSL Macros       */
#ifdef WIN32

/*
 *  Function Prototypes.
 */
#include psl_common_file (psl_rtndef.h) /* PSL routines         */

/*
 *  Structure Definitions.
 */
#include psl_common_file (shrmsg.h) /* PSL messages         */


/************************************************************************/
/*                                  */
/*      BIT_SET     Set bits in a bit mask              */
/*      =======                             */
/*                                  */
/*  This routine sets a specified number of bits in a supplied  */
/*  bit mask.                           */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  iret = BIT_SET (bitmask_ptr, offset_ptr, count_ptr)     */
/*                                  */
/*  Where:                              */
/*                                  */
/*  iret        = Unsigned long PSL routine status          */
/*  bitmask_ptr = (void *) Pointer to bitmask           */
/*  offset_ptr  = (unsigned long *) Pointer to offset in bitmask    */
/*  count_ptr   = (unsigned long *) Pointer to count of bits to set */
/*                                  */
/************************************************************************/
unsigned long   BIT_SET (void *bitmask_ptr,
                         unsigned long *offset_ptr,
                         unsigned long *count_ptr)
{
    /*
     *    Local data
     */
    unsigned long   count;          /* Number of bits to set */
    unsigned long   offset;         /* Offset */
    unsigned char   *mask_ptr;      /* Pointer to bitmask */

    mask_ptr = (unsigned char *) bitmask_ptr;
    /* Point to 1st char of bitmask */
    offset = *offset_ptr;       /* Get initial bit offset */

    /*
     *  Loop round, setting a bit at a time.  Note that we operate on chars
     *  rather than longs to maintain compatibility across big- and
     *  little-endian architectures
     */
    for (count = 0; count < *count_ptr; count++, offset++)
    {
        *(mask_ptr + (offset/8)) |= 1<<(offset%8);
        /* Set bit */
    }
    /*
     *  Return to the user with success
     */
    return PSL_NORMAL;
}


/************************************************************************/
/*                                  */
/*      BIT_CLEAR   Clear bits in a bit mask            */
/*      =========                           */
/*                                  */
/*  This routine clears a specified number of bits in a supplied    */
/*  bit mask.                           */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  iret = BIT_CLEAR (bitmask_ptr, offset_ptr, count_ptr)       */
/*                                  */
/*  Where:                              */
/*                                  */
/*  iret        = Unsigned long PSL routine status          */
/*  bitmask_ptr = (void *) Pointer to bitmask           */
/*  offset_ptr  = (unsigned long *) Pointer to offset in bitmask    */
/*  count_ptr   = (unsigned long *) Pointer to count of bits to set */
/*                                  */
/************************************************************************/
unsigned long   BIT_CLEAR (void *bitmask_ptr,
                           unsigned long *offset_ptr,
                           unsigned long *count_ptr)
{
    /*
     *    Local data
     */
    unsigned long   count;          /* Number of bits to clear */
    unsigned long   offset;         /* Offset */
    unsigned char   *mask_ptr;      /* Pointer to bitmask */

    mask_ptr = (unsigned char *) bitmask_ptr;
    /* Point to 1st char of bitmask */
    offset = *offset_ptr;       /* Get initial bit offset */

    /*
     *  Loop round, clearing a bit at a time.  Note that we operate on chars
     *  rather than longs to maintain compatibility across big- and
     *  little-endian architectures
     */
    for (count = 0; count < *count_ptr; count++, offset++)
    {
        *(mask_ptr + (offset/8)) &= ~(1<<(offset%8));
        /* Clear bit */
    }
    /*
     *  Return to the user with success
     */
    return PSL_NORMAL;
}


/************************************************************************/
/*                                  */
/*      BIT_BEST_FIT    Find Best Fit in Bit mask           */
/*      ============                            */
/*                                  */
/*  This routine finds the 'best fit' set of clear bits for a   */
/*  specified number of bits in a supplied bit mask.        */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  iret = BIT_BEST_FIT (bitmask_ptr, offset_ptr, size_ptr,     */
/*               count_ptr, retoff_ptr)         */
/*                                  */
/*  Where:                              */
/*                                  */
/*  iret        = Unsigned long PSL routine status          */
/*          = PSL_NORMAL if area located successfully       */
/*          = PSL_NOROOM if no room for specified bits      */
/*  bitmask_ptr = (void *) Pointer to bitmask           */
/*  offset_ptr  = (unsigned long *) Pointer to offset in bitmask    */
/*  size_ptr    = (unsigned long *) Pointer to count of bits to scan*/
/*  count_ptr   = (unsigned long *) Pointer to count of bits reqd   */
/*  retoff_ptr  = (unsigned long *) Pointer to returned offset  */
/*                                  */
/************************************************************************/
unsigned long   BIT_BEST_FIT (void *bitmask_ptr,
                              unsigned long *offset_ptr,
                              unsigned long *size_ptr,
                              unsigned long *count_ptr,
                              unsigned long *retoff_ptr)
{
    /*
     *    Local data
     */
    unsigned long   offset;         /* Offset within bitmask */
    unsigned long   end_offset;     /* Offset(+1) of end of scan */
    unsigned char   *mask_ptr;      /* Pointer to bitmask */
    unsigned long   best_size = ULONG_MAX;  /* Best size so far */
    unsigned long   saved_offset;       /* Saved offset */
    unsigned long   bitcount;       /* Count of contiguous bits */

    mask_ptr = (unsigned char *) bitmask_ptr;
    /* Point to 1st char of bitmask */
    offset = *offset_ptr;       /* Get initial bit offset */
    end_offset = offset + *size_ptr;
    /* Calculate end offset */

    /*
     *  Loop round, searching for the first clear bit
     */
find_clear:
    for (; offset < end_offset; offset++)
    {
        if ((*(mask_ptr + (offset/8)) & (1<<(offset%8))) == 0)
            /* If bit is clear */
        {
            goto found_clear;   /* Reached start of clear area */
        }
    }
    goto end_scan;          /* End of scan */

    /*
     *  We have now found a clear bit.   Check to see if remaining area
     *  is large enough. We do this so as not to waste time during our
     *  scanning loop.
     */
found_clear:
    if ((end_offset - offset) < *count_ptr) goto end_scan;
    /* Remaining area is insufficient */

    /*
     *  Now scan through the remaining buffer to see how much contiguous
     *  free space we have.  Note that we check on the way to see if we
     *  have gone too far so as to optimize the case where we are scanning
     *  to the end of the bit mask (might be very large).
     */
    saved_offset = offset;      /* Remember current offset */
    /* in case the area is large enough */

    for (; offset < end_offset; offset++)
    {
        if ((*(mask_ptr + (offset/8)) & (1<<(offset%8))) != 0)
            /* If bit is set */
        {
            goto end_clear;     /* Reached end of clear area */
        }
    }

    /*
     *  We have found the end of the free area - see if it is large enough
     */
end_clear:
    bitcount = offset - saved_offset;
    /* Bits in this area */
    if (bitcount >= *count_ptr) /* If area is large enough */
    {
        if (bitcount < best_size)   /* If < best size so far */
        {
            best_size = bitcount;   /* Save best size so far */
            *retoff_ptr = saved_offset;
            /* Return offset */
        }
        if (bitcount == *count_ptr) goto end_found;
        /* If exact size, exit with success */
    }

    /*
     *  Free area found is not an exact match - keep looking.
     */
    if (offset < end_offset) goto find_clear;
    /* If not at end, find next free area */

    /*
     *  End of scan - see if we found an area
     */
end_scan:
    if (best_size == ULONG_MAX) return PSL_NOROOM;
    /* Required area not found */

    /*
     *  We have found an area which is big enough
     */
end_found:
    return PSL_NORMAL;      /* Return success */
}


/************************************************************************/
/*                                  */
/*      BIT_FIRST_FIT   Find First Fit in Bit mask          */
/*      =============                           */
/*                                  */
/*  This routine finds the 'first fit' set of clear bits for a  */
/*  specified number of bits in a supplied bit mask.        */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  iret = BIT_FIRST_FIT (bitmask_ptr, offset_ptr, size_ptr,    */
/*                count_ptr, retoff_ptr)            */
/*                                  */
/*  Where:                              */
/*                                  */
/*  iret        = Unsigned long PSL routine status          */
/*          = PSL_NORMAL if area located successfully       */
/*          = PSL_NOROOM if no room for specified bits      */
/*  bitmask_ptr = (void *) Pointer to bitmask           */
/*  offset_ptr  = (unsigned long *) Pointer to offset in bitmask    */
/*  size_ptr    = (unsigned long *) Pointer to count of bits to scan*/
/*  count_ptr   = (unsigned long *) Pointer to count of bits reqd   */
/*  retoff_ptr  = (unsigned long *) Pointer to returned offset  */
/*                                  */
/************************************************************************/
unsigned long   BIT_FIRST_FIT (void *bitmask_ptr,
                               unsigned long *offset_ptr,
                               unsigned long *size_ptr,
                               unsigned long *count_ptr,
                               unsigned long *retoff_ptr)
{
    /*
     *    Local data
     */
    unsigned long   offset;         /* Offset within bitmask */
    unsigned long   end_offset;     /* Offset(+1) of end of scan */
    unsigned char   *mask_ptr;      /* Pointer to bitmask */
    unsigned long   saved_offset;       /* Saved offset */
    unsigned long   bitcount;       /* Count of contiguous bits */
    unsigned long   count_left;     /* Count of remaining bits */

    mask_ptr = (unsigned char *) bitmask_ptr;
    /* Point to 1st char of bitmask */
    offset = *offset_ptr;       /* Get initial bit offset */
    end_offset = offset + *size_ptr;
    /* Calculate end offset */

    /*
     *  Loop round, searching for the first clear bit
     */
find_clear:
    for (; offset < end_offset; offset++)
    {
        if ((*(mask_ptr + (offset/8)) & (1<<(offset%8))) == 0)
            /* If bit is clear */
        {
            goto found_clear;   /* Reached start of clear area */
        }
    }
    goto end_notfound;      /* End of scan */

    /*
     *  We have now found a clear bit.   Check to see if remaining area
     *  is large enough. We do this so as not to waste time during our
     *  scanning loop.
     */
found_clear:
    count_left = end_offset - offset;
    /* Calculate remaining count */
    if (count_left < *count_ptr) goto end_notfound;
    /* Remaining area is insufficient */

    /*
     *  Now scan through the remaining buffer to see how much contiguous
     *  free space we have.  Note that we check on the way to see if we
     *  have gone too far so as to optimize the case where we are scanning
     *  to the end of the bit mask (might be very large).
     */
    saved_offset = offset;      /* Remember current offset */
    /* in case the area is large enough */

    for (; offset < end_offset; offset++)
    {
        if ((*(mask_ptr + (offset/8)) & (1<<(offset%8))) != 0)
            /* If bit is set */
        {
            goto end_clear;     /* Reached end of clear area */
        }
    }

    /*
     *  We have found the end of the free area - see if it is large enough
     */
end_clear:
    bitcount = offset - saved_offset;
    /* Bits in this area */
    if (bitcount >= *count_ptr) /* If area is large enough */
    {
        *retoff_ptr = saved_offset; /* Return offset */
        return PSL_NORMAL;      /* Return success */
    }

    /*
     *  Free area found is not large enough - keep looking.
     */
    if (offset < end_offset) goto find_clear;
    /* If not at end, find next free area */

    /*
     *  We did not find a large enough area
     */
end_notfound:
    return PSL_NOROOM;      /* Required area not found */
}


/************************************************************************/
/*                                  */
/*      BIT_COUNT   Count number of clear bits          */
/*      =========                           */
/*                                  */
/*  This routine counts the number of clear bits in a       */
/*  supplied bit mask.                      */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  iret = BIT_COUNT (bitmask_ptr, offset_ptr, size_ptr, count_ptr  */
/*                                  */
/*  Where:                              */
/*                                  */
/*  iret        = Unsigned long PSL routine status          */
/*  bitmask_ptr = (void *) Pointer to bitmask           */
/*  offset_ptr  = (unsigned long *) Pointer to offset in bitmask    */
/*  size_ptr    = (unsigned long *) Pointer to count of bits to scan*/
/*  count_ptr   = (unsigned long *) Pointer to count of clear bits  */
/*                                  */
/************************************************************************/
unsigned long   BIT_COUNT (void *bitmask_ptr,
                           unsigned long *offset_ptr,
                           unsigned long *size_ptr,
                           unsigned long *count_ptr)
{
    /*
     *    Local data
     */
    unsigned long   offset;         /* Offset within bitmask */
    unsigned long   end_offset;     /* Offset(+1) of end of scan */
    unsigned char   *mask_ptr;      /* Pointer to bitmask */

    mask_ptr = (unsigned char *) bitmask_ptr;
    /* Point to 1st char of bitmask */
    offset = *offset_ptr;       /* Get initial bit offset */
    end_offset = offset + *size_ptr;
    /* Calculate end offset */
    *count_ptr = 0;         /* Initialize count of clear bits */

    /*
     *  Loop round, counting clear bits
     */
    for (; offset < end_offset; offset++)
    {
        if ((*(mask_ptr + (offset/8)) & (1<<(offset%8))) == 0)
            /* If bit is clear */
        {
            *count_ptr += 1;    /* Increment count */
        }
    }
    return PSL_NORMAL;      /* Return success */
}


/************************************************************************/
/*                                  */
/*      BIT_STATS   Return Statistics on Bitmap Usage       */
/*                                  */
/*  This routine returns statistics on the current usage of a   */
/*  specified bit map.                      */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  BIT_STATS (bitmask_ptr, size_ptr, freebits_ptr,         */
/*         freefrag_ptr, maxfrag_ptr                */
/*                                  */
/*  Where:                              */
/*                                  */
/*  iret        = Unsigned long PSL routine status          */
/*  bitmask_ptr = (void *) Pointer to bitmask           */
/*  size_ptr    = (unsigned long *) Pointer to count of bits in map */
/*  freebits_ptr= (unsigned long *) Pointer to count of clear bits  */
/*  freefrag_ptr= (unsigned long *) Ptr to count of clear fragments */
/*  maxfrag_ptr = (unsigned long *) Ptr to size of largest fragment */
/*                                  */
/************************************************************************/
void    BIT_STATS (void *bitmask_ptr,
                   unsigned long *size_ptr,
                   unsigned long *freebits_ptr,
                   unsigned long *freefrag_ptr,
                   unsigned long *maxfrag_ptr)
{
    /*
     *    Local data
     */
    unsigned long   offset;         /* Offset within bitmask */
    unsigned long   end_offset;     /* Offset(+1) of end of scan */
    unsigned char   *mask_ptr;      /* Pointer to bitmask */
    unsigned long   saved_offset;       /* Saved offset */
    unsigned long   bitcount;       /* Count of contiguous bits */

    mask_ptr = (unsigned char *) bitmask_ptr;
    /* Point to 1st char of bitmask */
    offset = 0;         /* Initial bit offset */
    end_offset = *size_ptr;     /* End offset */
    *freebits_ptr = 0;      /* Initialize free bit count */
    *freefrag_ptr = 0;      /* Initialize free fragment count */
    *maxfrag_ptr = 0;       /* Initialize max fragment size */

    /*
     *  Loop round, searching for the first clear bit
     */
find_clear:
    for (; offset < end_offset; offset++)
    {
        if ((*(mask_ptr + (offset/8)) & (1<<(offset%8))) == 0)
            /* If bit is clear */
        {
            goto found_clear;   /* Reached start of clear area */
        }
    }
    goto end_scan;          /* End of scan */

    /*
     *  We have now found a clear bit.
     */
found_clear:
    saved_offset = offset;      /* Remember current offset */

    for (; offset < end_offset; offset++)
    {
        if ((*(mask_ptr + (offset/8)) & (1<<(offset%8))) != 0)
            /* If bit is set */
        {
            goto end_clear;     /* Reached end of clear area */
        }
    }

    /*
     *  We have found the end of the free area
     */
end_clear:
    bitcount = offset - saved_offset;
    /* Bits in this area */
    *freebits_ptr += bitcount;  /* Increment free bit count */
    *freefrag_ptr += 1;     /* Increment free fragment count */
    if (bitcount > *maxfrag_ptr) *maxfrag_ptr = bitcount;
    /* Update maximum fragment size */
    if (offset < end_offset) goto find_clear;
    /* If not at end, find next free area */

    /*
     *  End of scan
     */
end_scan:
    return;
}
#endif
