/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*
 * @doc     scl_act_routines.c
 * @module  scl_act_routines.c | This module contains all the TPA action
 * routines that do not belong to any of the other SCL TPA action routine
 * modules (containing specific types of action routine).
 */

#include <psl_macros.h>
#include <vtest_macros.h>

/*
 *  Standard ANSI header files.
 */
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/*
 *  Function Prototypes.
 */
#include <vtest_macros.h>       /* V-TEST Macros */
#include <psl_rtndef.h>         /* PSL Routines */
#include <trc_rtndef.h>         /* Tracing Routines */
#include <vtest_rtndef.h>       /* V-TEST Routines */
#include "scl_rtndef.h"         /* SCL Routines */
#include "scl_def_tbl_rtndef.h"     /* SCL definitions routines */
#include "scl_env_tbl_rtndef.h"     /* SCL environment routines */
#include "scl_cod_tbl_rtndef.h"     /* SCL code routines */
#include "scl_sub_tbl_rtndef.h"     /* SCL subexpression routines */

static  void    ADD_FILE    (struct tpaarg *argblk_ptr);

static  void    ADD_PIPE    (struct tpaarg *argblk_ptr);

static void     ADD_SHRVAR  (struct tpaarg *argblk_ptr,
                             struct symtbl *var_ptr,
                             unsigned long var_size,
                             unsigned long addr);

static  void    ADD_THR     (struct tpaarg *argblk_ptr);

static  void    ADD_TST     (struct tpaarg *argblk_ptr);

static  void    ADD_TSTTHR  (struct tpaarg *argblk_ptr);

static Boolean  CHECK_FOR_DUPLICATES    (struct tpaarg *argblk_ptr);

/*
 *  Structure definitions.
 */
#include <shrmsg.h>         /* PSL Messages         */
#include <depprmdef.h>          /* Script dependency parameters */
#include <psldiag_prmdef.h>     /* Diagnostic Routine Symbols   */
#include <psl_vm_prmdef.h>      /* Linked list parameters   */
#include <psl_tpaargdef.h>      /* TPARSE Argument Block    */
#include <psl_trc_prmdef.h>     /* Tracing parameters       */
#include <scpdef.h>         /* SCP Header record        */
#include "scl_tbldef.h"         /* SCL Table Definitions    */
#include <tofprmdef.h>          /* TOF parameters       */
#include <vtest_prmdef.h>           /* VTEST Parameters         */
#include "scl_cmd.h"            /* Command Line Data Structure  */
#include "scl_comdef.h"         /* Common Data          */
#include "scl_limits.h"         /* SCL limits           */
#include "scl_tpacom.h"         /* Tparse common Data       */

/*======================================================================*/
/*                                  */

/*                                  */
/*    ADD_ARRAY_BOUND - Add an array boundary to variable definition    */
/*    ===============                           */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   INTEGER_VALUE - Integer array value         */
/*    Outputs:  Variable definition array bounds set            */
/*                                  */
/*    This routine is called after parsing an array bound. Its purpose  */
/*    is to update the variable definition in the work structure to     */
/*    hold the array bound value.                   */
/*                                  */
/*    Note that each array bound actually has 2 values. A start and an  */
/*    end. The boundary which we parse first (this one) is initially    */
/*    set as the END boundary (the start having been defaulted to 1).   */
/*    If we subsequently parse a second value we shuffl;e this value    */
/*    to be the starting value.                     */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   ADD_ARY_BOUND1  (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com *com_ptr;        /* Common Data          */
    struct scl_tpa *tpa_ptr;        /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    If we have greater than MAX_ARRAY_DIM dimensions then report that this
     *    is an invalid array declaration.
     */
    if ((*tpa_ptr).array_dimcntl >= MAX_ARRAY_DIM)
    {
        REPORT_ERROR (argblk_ptr, SCL_TOOMANYDIM);
        /* Pass to underline routine    */
    }

    else
    {
        (*tpa_ptr).def_array_bounds2l10 [(*tpa_ptr).array_dimcntl]
        = (*tpa_ptr).int_valuel;
        /* Save the array boundary  */
        (*tpa_ptr).array_dimcntl += 1;
    }

    return PSL_NORMAL;
}

/*======================================================================*/
/*                                  */

/*                                  */
/*    ADD_ARY_BOUND2 - Add the end array bound to variable definition   */
/*    ==============                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   INTEGER_VALUE - Integer array value         */
/*    Outputs:  Variable definition array bounds set            */
/*                                  */
/*    This routine is called after parsing the end value for an array   */
/*    boundary. This gives us an array definition as follows:       */
/*                                  */
/*        ARRAY [low_bound:high_bound]                  */
/*                                  */
/*    At this point we have parsed the low_bound but assumed it was     */
/*    the high boundary. We must now copy the old high boundary to the  */
/*    low boundary and the current value as the new high boundary.  */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   ADD_ARY_BOUND2  (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com *com_ptr;        /* Common Data          */
    struct scl_tpa *tpa_ptr;        /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Copy the value previously thought to be the HIGH boundary value into the
     *    low boundary value. Note that we have previously updated the dimension
     *    count field.
     */
    (*tpa_ptr).def_array_bounds1l10 [(*tpa_ptr).array_dimcntl - 1] =
        (*tpa_ptr).def_array_bounds2l10 [(*tpa_ptr).array_dimcntl - 1];

    /*
     *    Check the specified HIGH array bound which must be greater than or
     *    equal to the low boundary value but must NOT be zero.
     */
    if (((*tpa_ptr).int_valuel == 0)
            ||  ((*tpa_ptr).int_valuel <
                 (*tpa_ptr).def_array_bounds1l10 [(*tpa_ptr).array_dimcntl - 1]))
    {
        REPORT_ERROR (argblk_ptr, VTEST_INVSUBVAL);
        /* Pass to underline routine    */
    }

    else
    {
        (*tpa_ptr).def_array_bounds2l10 [(*tpa_ptr).array_dimcntl - 1]
        = (*tpa_ptr).int_valuel;
    }

    return PSL_NORMAL;
}

/*======================================================================*/
/*                                  */
/*    ADD_CHR_VAL - Add Character Value to List             */
/*    ============                          */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   OBJ_VALUES = Address of value list buffer       */
/*          OBJ_VALUES_PTR = Value list buffer pointer      */
/*          CHR_STRING = Character string               */
/*          CHR_STRLENL = Character string length           */
/*          DEF_SIZEL = Variable size in bytes          */
/*          DEF_VALUES_CNTL = Value count               */
/*    Outputs:  OBJ_VALUES_PTR = Value list buffer pointer updated  */
/*          DEF_VALUES_CNTL = Value count incremented       */
/*                                  */
/*    This routine attempts to add a character value to the value   */
/*    list buffer and increment the value count. If the value to be */
/*    added would cause the value list buffer to overflow OR the    */
/*    value is too long an error is reported.               */
/*                                  */
/*    On entry the character value is held in CHR_STRING and its    */
/*    length in CHR_STRING_LENGTH.                  */
/*                                  */
/*    Note that value list entries are "padded" to allow the "length"   */
/*    to be longword-aligned.                       */
/*----------------------------------------------------------------------*/

unsigned long   ADD_CHR_VAL (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com *com_ptr;        /* Common Data          */
    struct scl_tpa *tpa_ptr;        /* TPARSE Data          */
    char  *val_ptr;        /* Pointer to value list    */

#if defined(UNIX) || defined(WIN32)
    unsigned long   alignment_fault;    /* Alignment fault count    */
#endif

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Check the length of the character value to be added, reporting an
     *    error if it exceeds the length of the variable for which it is in the list
     *
     *    Note: As a temporary measure we do not use the REPORT_ERROR routine to
     *      report this error as we may have read in a new line and therefore
     *      will be unable to underline the error. We therefore manually report
     *      the problem. Hopefully we will return to this problem later.
     */
    if ((*tpa_ptr).chr_strlenl > (*tpa_ptr).def_sizel)
    {
        (*argblk_ptr).paraml = SCL_VALTOOLONG;
        /* Set up error status  */
        (void) ERROR_MESSAGE (argblk_ptr);
        /* ...and report the error */
    }

    else
    {
        val_ptr = (*tpa_ptr).values_nxtpc + (*tpa_ptr).chr_strlenl;
        /* Pointer to byte after this value */
        val_ptr += sizeof(long);    /* ...plus a (long) count */
        /*
         *  Ensure this value list entry is a whole number of longwords "long"
         *  for UNIX environments. This ensures that the string "length" is
         *  longword-aligned.
         */
#if defined(UNIX) || defined(WIN32)
        alignment_fault = ((*tpa_ptr).chr_strlenl % sizeof(long));
        /* Get any "remainder"      */
        alignment_fault = ((sizeof(long) - alignment_fault) % sizeof(long));
        /* Convert to "shortfall"   */

        val_ptr += alignment_fault; /* Point to next longword bound */
#endif

        /*
         *     Check to ensure that the value will fit into the remaining space in the
         *     value list buffer.
         */
        if (val_ptr > (*tpa_ptr).values_endpc)
        {
            REPORT_ERROR (argblk_ptr, SCL_VALOVERFLOW);
            /* Report error if overflow */
        }

        else
        {

            /*
             *    We have space for the value in the value list buffer. Add it to the buffer
             *    prefixing the entry with a long size
             */
            val_ptr = (*tpa_ptr).values_nxtpc;
            /* Point to FREE space in buffer */
            ENC_INTEGER_AT (val_ptr, (long) (*tpa_ptr).chr_strlenl);
            /* Copy SIZE into buffer    */
            val_ptr += sizeof(long);    /* ...and increment pointer */

            (void) memcpy (val_ptr,     /* ...and copy in value     */
                           (*tpa_ptr).chr_stringc65536,
                           (*tpa_ptr).chr_strlenl);

            (*tpa_ptr).values_nxtpc = val_ptr + (*tpa_ptr).chr_strlenl;
            /* ...and point beyond value in list */

#if defined(UNIX) || defined(WIN32)
            (*tpa_ptr).values_nxtpc += alignment_fault;
            /* Align to next longword bound */
#endif

            (*tpa_ptr).def_values_cntl +=1;
            /* Increment values count   */
        }
    }

    return PSL_NORMAL;
}

/************************************************************************/
/*                                  */
/*    ADD_DEF - Add Definition to Symbol Table              */
/*    =======                               */
/*                                  */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Tparse argument block.         */
/*      tpa_ptr - Common tparse data.               */
/*    Outputs:  tpa_ptr - Common tparse data reset for next definition  */
/*      Symbol Table - Entry added to appropriate Symbol    */
/*          Table for last definition (if valid).       */
/*                                  */
/*    This routine is called when the parsing of a definition is    */
/*    complete. It first checks that the definition is 'complete',  */
/*    and if it is, adds an entry for it to the appropriate symbol  */
/*    table. It then resets all the 'work' variables used to build  */
/*    a definition.                         */
/*                                  */
/*----------------------------------------------------------------------*/
unsigned long   ADD_DEF (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    Boolean  duplicate;      /* Duplicate definition flag    */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Check to see if we have a definition being parsed.
     */
    if ((*tpa_ptr).def_typl == 0) goto end;

    /*
     *    Check for duplicate definitions as in most cases we will not allow
     *    variables to be duplicately defined.
     */
    duplicate = CHECK_FOR_DUPLICATES (argblk_ptr);
    if (duplicate) goto end;

    /*
     *    We are parsing a definition - If we have a name then add the definition
     *    otherwise simply clear the variable parsing data.
     */
    if ((*tpa_ptr).def_namec64 [0] != '\0')
        /* If definition name specified */
    {

        if ((*tpa_ptr).def_typl == DEF_FILE)
        {               /* if a FILE definition */
            ADD_FILE (argblk_ptr);  /* ...add FILE definition */
        }

        else if ((*tpa_ptr).def_typl == DEF_PIPE)
        {               /* ...else if a PIPE definition */
            ADD_PIPE (argblk_ptr);  /* ...add PIPE definition */
        }

        else if (((*tpa_ptr).def_typl == DEF_SUBTEST)
                 ||   ((*tpa_ptr).def_typl == DEF_TEST))
        {               /* ...and defining a TEST or SUB-TEST */
            ADD_TST (argblk_ptr);   /* ...add a TEST definition */
        }

        else if ((*tpa_ptr).def_typl == DEF_TEST_THREAD)
        {               /* ...and defining a TEST THREAD */
            ADD_TSTTHR (argblk_ptr);
            /* ...add a TEST THREAD definition */
        }

        else if ((*tpa_ptr).def_typl == DEF_TIMER)
        {               /* ...and defining a TIMER */
            ADD_TIMER (argblk_ptr); /* ...add a TIMER definition */
        }

        else
        {               /* ...else a VARIABLE definition    */
            ADD_VARIABLE (argblk_ptr);  /* ...add a VARIABLE definition  */
        }

    }

    /*
     *    Clear the 'work' variables used to build a definition.
     */
end:
    ;
    INI_DEF (com_ptr);          /* Initialise definition data */

    return PSL_NORMAL;          /* ...and return success */
}

/************************************************************************/
/*                                  */
/*    ADD_DEP - Add an item to the script dependency list       */
/*    =======                               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   com_ptr - SCL data pointer              */
/*      depnam_ptr - Pointer to dependency name         */
/*      deptyp - Dependency type                */
/*      depdat_ptr - Dependency type specific data      */
/*    Outputs:  Item added to end of dependency list.               */
/*                                  */
/*    This routine adds an item to the end of the script dependency */
/*    linked list and increments the dependency count.          */
/*                                  */
/************************************************************************/
void ADD_DEP (struct scl_com *com_ptr,
              char *depnam_ptr,
              enum deptyp_t deptyp,
              char *depdat_ptr)
{
    /*
     *    Local data
     */
    struct sdep  newdep;    /* New script dependency list entry */
    struct sdep  olddep;    /* Old script dependency list entry */
    unsigned long    iret;      /* Status return */
    unsigned long    isub;      /* Loop control */
    unsigned long    mode;      /* PSL_VM_COPY mode */
    unsigned long    depnamlen; /* Dependency name length */
    char    *llb_ctx;  /* Linked list context */
    char    *pos_ptr;  /* Linked list position */

    /*
     *  Set up the dependency record. This consists of the dependency name, dependency type
     *  and dependency type specific data. Convert the dependency name to uppercase,
     *  where character case is not significant and is not 'handled' elsewhere, i.e. in
     *  CALL SCRIPT and EXECUTE THREAD ... USING SCRIPT dependencies - dependencies whose
     *  names are stored in tables will already have been converted to uppercase.
     */
    newdep.sdep_typl = deptyp;      /* Set dependency type */

    if ((deptyp == DEP_TYP_CALSCR)  /* If a CALL SCRIPT dependency */
            ||  (deptyp == DEP_TYP_EXESCR)) /* or an EXECUTE THREAD ... USING SCRIPT dependency */
    {
        depnamlen = strlen (depnam_ptr);
        for (isub = 0; isub <= depnamlen; isub++)
        {
            newdep.sdep_namec256 [isub] = toupper (depnam_ptr [isub]);
        }
    }
    else
    {
        (void) strcpy (newdep.sdep_namec256, depnam_ptr);
    }

    /*
     *  Add the dependency data (or a NULL string if there is none).
     */
    if (depdat_ptr != NULL)
    {
        (void) strcpy (newdep.sdep_datac256, depdat_ptr);
    }
    else
    {
        newdep.sdep_datac256[0] = '\0';
    }

    /*
     *  Check that this is not a duplicate entry. If it is, just return - nothing to add.
     *  A dependency is considered a duplicate if its name AND type AND data are identical
     *  to an existing dependency.
     */
    for (isub = 0, pos_ptr = NULL, mode = PSL_VM_MODE_FIRST;
            isub < (*com_ptr).dep_cntl;
            isub++, pos_ptr = llb_ctx, mode=PSL_VM_MODE_NEXT)
    {
        iret = PSL_VM_COPY ((*com_ptr).dep_llps ,(char *) &olddep
                            ,pos_ptr ,&llb_ctx ,mode);
        /* Remove the OPH pointer first */
        if  (!(psl_test_status (iret))) /* ...report any errors */
        {
            PSL_SIGDIAG ((*com_ptr).diag_ptrps, 2, PSL_SYSERVERR
                         ,"PSL_VM_COPY (in ADD_DEP)", iret);
        }

        if (newdep.sdep_typl == olddep.sdep_typl)
        {
            if (strcmp (newdep.sdep_namec256, olddep.sdep_namec256) == 0)
            {
                if (strcmp (newdep.sdep_datac256, olddep.sdep_datac256) == 0) return;
            }
        }
    }

    /*
     *  Add the dependency record to the end of the script dependency list.
     */
    iret = PSL_VM_INSERT ((*com_ptr).dep_llps
                          ,(char *) &newdep
                          ,NULL
                          ,NULL
                          ,PSL_VM_POS_END);
    if (!psl_test_status (iret))    /* If error inserting into list */
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps,
                     2,
                     PSL_SYSERVERR,
                     "PSL_VM_INSERT (in ADD_DEP)",
                     iret);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
    }

    (*com_ptr).dep_cntl += 1;       /* Increment dependency count */

    return;
}


/*======================================================================*/
/*                              */
/*    ADD_DEP_CALSCR - Add a 'call script' entry to the dependency list */
/*    ==============                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to TPARSE argument block       */
/*    Outputs:  None                            */
/*                                  */
/*    This routine adds an entry for a called script to the script  */
/*    dependency list. The script name is assumed to be held in the */
/*    first operand on the parsed operand linked list. It can only  */
/*    be a dependency if it is an immediate character value.        */
/*                                  */
/*----------------------------------------------------------------------*/
unsigned long ADD_DEP_CALSCR (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data.
     */
    struct   scl_com *com_ptr;      /* Common Data */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data */
    struct   oph     *oph_ptr;      /* Operand header pointer */
    struct   opd     *opd_ptr;      /* Operand descriptor pointer */
    unsigned long    iret;          /* Status return */
    char    scrnam[256];       /* Script name */
    char    *llb_ctx;      /* Linked list context */
    int     toftyp = DEP_TOFTYP_SCRIPT;
    /* Dependency TOF type - assume script */
    char depdat [DEPBUFLEN];   /* Dependency type specific data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *  Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *  Copy the operand containing the script name from the parsed operand stack.
     */
    iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps ,(char *) &oph_ptr
                        ,NULL ,&llb_ctx ,PSL_VM_MODE_FIRST);
    /* Remove the OPH pointer first */
    if  (!(psl_test_status (iret))) /* ...report any errors */
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_COPY OPH (in ADD_DEP_CALSCR)", iret);
    }

    iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps ,(char *) &opd_ptr
                        ,llb_ctx ,NULL ,PSL_VM_MODE_NEXT);
    /* ...then remove the OPD ptr */
    if  (!(psl_test_status (iret))) /* ...report any errors */
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_COPY OPD (in ADD_DEP_CALSCR)", iret);
    }

    /*
     *  Add script name to dependency list, if required.
     */
    if (((*oph_ptr).addrl == ADDR_IMMEDIATE)
            &&  ((*oph_ptr).dtypl == DTYP_CHARACTER))
    {
        (void) memcpy (scrnam, (*opd_ptr).uopd.imc.val_pc, (*oph_ptr).sizel);
        scrnam [(*oph_ptr).sizel] ='\0';

        /*
         *    Pick up the command options as this will tell us whether we have an excute script
         *    or an execute TOF which will need flagging in the generated dependency file.
         */
        if (((*tpa_ptr).cmnd_opt_specifiedl & CMND_OPT_TOF) != 0)
        {
            toftyp = DEP_TOFTYP_TOF;
        }

        (void) sprintf (depdat, "-%c%d" ,DEP_SEPCHR, toftyp);
        ADD_DEP (com_ptr, scrnam, DEP_TYP_CALSCR, depdat);
    }

end:
    ;
    return PSL_NORMAL;
}

/*======================================================================*/
/*                              */
/*    ADD_DEP_FILE - Add a file entry to the script dependency list */
/*    ============                          */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to TPARSE argument block       */
/*    Outputs:  None                            */
/*                                  */
/*    This routine adds an entry for a file to the script dependency    */
/*    list. The file name is assumed to be held in the first operand    */
/*    on the parsed operand linked list. It can only be a dependency    */
/*    if it is an immediate character value.                */
/*                                  */
/*----------------------------------------------------------------------*/
unsigned long ADD_DEP_FILE (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data.
     */
    struct   scl_com *com_ptr;      /* Common Data */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data */
    struct   oph     *oph_ptr;      /* Operand header pointer */
    struct   opd     *opd_ptr;      /* Operand descriptor pointer */
    unsigned long    iret;          /* Status return */
    char    filnam[256];       /* File name */
    char    *llb_ctx;      /* Linked list context */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *  Do not encode anything if we have had an error in the command or the file
     *  was NOT opened for input - either explicitly or by default.
     */
    if (((*tpa_ptr).error_in_commandn)
            ||  ((((*tpa_ptr).open_maskl & OPEN_INPUT) == 0)
                 &&   ((*tpa_ptr).open_maskl != 0))) goto end;

    /*
     *    Remove the operand defining the file name from the parsed operand stack.
     */
    iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps ,(char *) &oph_ptr
                        ,NULL ,&llb_ctx ,PSL_VM_MODE_FIRST);
    /* Remove the OPH pointer first */
    if  (!(psl_test_status (iret))) /* ...report any errors */
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_COPY OPH (in ADD_DEP_FILE)", iret);
    }

    iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps ,(char *) &opd_ptr
                        ,llb_ctx ,NULL ,PSL_VM_MODE_NEXT);
    /* ...then remove the OPD ptr */
    if  (!(psl_test_status (iret))) /* ...report any errors */
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_COPY OPD (in ADD_DEP_FILE)", iret);
    }

    /*
     *  Add file name to dependency list, if required.
     */
    if (((*oph_ptr).addrl == ADDR_IMMEDIATE)
            &&  ((*oph_ptr).dtypl == DTYP_CHARACTER))
    {
        (void) memcpy (filnam, (*opd_ptr).uopd.imc.val_pc, (*oph_ptr).sizel);
        filnam [(*oph_ptr).sizel] ='\0';
        ADD_DEP (com_ptr, filnam, DEP_TYP_FILE, NULL);
    }

end:
    ;
    return PSL_NORMAL;
}


/************************************************************************/
/*                                  */
/*    ADD_INT_VAL - Add Integer Value to List               */
/*    ===========                           */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Moved to end of integer value               */
/*    Inputs:   OBJ_VALUES = Address of value list buffer       */
/*          OBJ_VALUES_PTR = Value list buffer pointer      */
/*          INT_VALUES = Integer value              */
/*          DEF_VALUES_CNTL = Value count               */
/*    Outputs:  OBJ_VALUES_PTR = Value list buffer pointer updated  */
/*          DEF_VALUES_CNTL = Value count updated           */
/*                                  */
/*    This routine attempts to add an integer value to the value list   */
/*    buffer and increment the value count. If the value to be added    */
/*    would cause the value list buffer to overflow an error is     */
/*    reported.                             */
/*                                  */
/*    On entry the integer value is held in INT_VALUE.          */
/*                                  */
/************************************************************************/

unsigned long   ADD_INT_VAL (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Check to see if the addition of this integer value would cause the
     *    value list buffer to overflow and report an error if it would.
     */
    if ((*tpa_ptr).values_nxtpc > (*tpa_ptr).values_endpc)
    {
        REPORT_ERROR (argblk_ptr, SCL_VALOVERFLOW);
        /* Report error if overflow */
    }

    else
    {

        /*
         *    We have space for the value in the value list buffer. Add it to the buffer
         */
        ENC_INTEGER_AT ((*tpa_ptr).values_nxtpc, (long) (*tpa_ptr).int_valuel);
        /* Copy value into buffer   */
        (*tpa_ptr).values_nxtpc += sizeof(long);
        /* update value list pointer    */
        (*tpa_ptr).def_values_cntl += 1;/* Increment values count   */
    }
    return PSL_NORMAL;
}

/*======================================================================*/
/*                              */
/*    ADD_OPR - Save Operator Operand                   */
/*    =======                               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   com_ptr = Pointer to SCL common data            */
/*      opr = Operator code                     */
/*    Outputs:  Operand = Operator operand added to LL of parsed    */
/*                operands.                 */
/*                                  */
/*    This routine creates an operand descriptor for an operator    */
/*    operand and adds it to the LL of parsed operands.         */
/*                                  */
/*======================================================================*/
void ADD_OPR (struct scl_com *com_ptr,
              unsigned long opr)
{
    /*
     *    Local data
     */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct oph  *oph_ptr;       /* Operand header pointer   */
    struct opd  *opd_ptr;       /* Operand descriptor pointer   */

    /*
     *    Copy pointers for ease of reference.
     */
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Allocate sufficient space to hold an operand header record and save its
     *    address into the parsed operands linked list.
     */
    ALLOCATE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr);

    /*
     *    Now set up this operand descriptor.
     */
    (*oph_ptr).addrl = ADDR_OPERATOR;   /* Operator addressing mode */
    (*oph_ptr).dtypl = DTYP_INTEGER;    /* Integer data type */
    (*oph_ptr).sizel = sizeof (long);   /* Size = (long) bytes */
    (*oph_ptr).optl = 0;        /* Clear operand options */
    (*oph_ptr).idl = (*tpa_ptr).this_cmnd_optl;
    /* Set operand id. */
    (*opd_ptr).uopd.opr.operatorl = opr;
    /* Save operator */
    return;
}


/************************************************************************/
/*                                  */
/*    ADD_FILE - Add FILE definition to External Data file symbol Table */
/*    ========                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - TPARSE argument block          */
/*    Outputs:  Current token added to external variables symbol table  */
/*                                  */
/*    This routine is called after we have parsed a definition for an   */
/*    external data file.                       */
/*                                  */
/*    It adds an entry for this FILE to the symbol table of external    */
/*    data files referenced.                        */
/*                                  */
/*    If the /DIAGNOSE qualifier has been specified or we are an SQL    */
/*    mode script then this routine will also add an entry for the file */
/*    to the variables symbol table as in this situation a FILE is  */
/*    classed as a special type of variable.                */
/*                                  */
/************************************************************************/
static void ADD_FILE (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct   scl_com    *com_ptr;       /* Common Data          */
    struct   scl_tpa    *tpa_ptr;       /* TPARSE Data          */
    struct   fattbl     *fattbl_ptr;    /* Pointer File symbol table    */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    /* Get common data pointer  */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* ...and TPARSE data pointer   */
    fattbl_ptr = (*tpa_ptr).fat_begps + (*tpa_ptr).fat_cntl;
    /* Next FREE FAT entry      */

    /*
     *    Add a new File entry to the table at the current point
     */
    (void) memcpy ((*fattbl_ptr).fat_namec20
                   ,(*tpa_ptr).def_namec64
                   ,(*tpa_ptr).def_nlengthl);
    /*
     * .Reset TOF Source Line Number (unless processing an INCLUDED file).
     */
    if ((*com_ptr).include_depthl == 0)
    {
        (*fattbl_ptr).fat_tof_srclinl = (*com_ptr).srclinl10[0];
    }

    /* Copy variable name into table  */
    (*fattbl_ptr).fat_namec20 [(*tpa_ptr).def_nlengthl] = '\0';
    (*fattbl_ptr).fat_refflgl = DEFINED_FILE;
    /* File purely defined      */
    (*tpa_ptr).open_fididxl = (*tpa_ptr).fat_cntl + 1;
    /* Save index of this file  */

    /*
     *    If required add it also to the EXTERNAL variables symbol table.
     */
    ADD_FILE_SYMBOL (com_ptr ,fattbl_ptr);

    (*tpa_ptr).fat_cntl += 1;       /* ...and increment File count  */


    return;             /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    ADD_PIPE - Add PIPE definition to the PIPE table          */
/*    ========                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - TPARSE argument block          */
/*    Outputs:  Current token added to PIPE table           */
/*                                  */
/*    This routine is called after we have parse a definition for an    */
/*    pipe.                             */
/*                                  */
/*    It addes an entry to the PIPE definition table.           */
/*                                  */
/************************************************************************/

static  void    ADD_PIPE    (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com    *com_ptr;       /* Common Data          */
    struct   scl_tpa    *tpa_ptr;       /* TPARSE Data          */
    struct   fattbl     *fattbl_ptr;    /* Pointer File symbol table    */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    /* Get common data pointer  */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* ...and TPARSE data pointer   */
    fattbl_ptr = (*tpa_ptr).pipe_begps + (*tpa_ptr).pipe_cntl;
    /* Next FREE FAT entry      */

    /*
     *    Add a new Pipe entry to the table at the current point
     */
    (void) memcpy ((*fattbl_ptr).fat_namec20
                   ,(*tpa_ptr).def_namec64
                   ,(*tpa_ptr).def_nlengthl);

    /*
     * .Reset TOF Source Line Number (unless processing an INCLUDED file).
     */
    if ((*com_ptr).include_depthl == 0)
    {
        (*fattbl_ptr).fat_tof_srclinl = (*com_ptr).srclinl10[0];
    }
    
    /* Copy variable name into table  */
    (*fattbl_ptr).fat_namec20 [(*tpa_ptr).def_nlengthl] = '\0';
    if (((*tpa_ptr).def_optl & VOPT_PIPE_SYNC) != 0)
    {
        (*fattbl_ptr).fat_refflgl |= PIPATT_SYNC;
        /* Set SYNC/ASYNC flag      */
    }
    else
    {
        (*fattbl_ptr).fat_refflgl &= ~PIPATT_SYNC;
        /* Clear SYNC/ASYNC flag    */
    }

    (*tpa_ptr).open_fididxl = (*tpa_ptr).pipe_cntl + 1;
    /* Save index of this pipe  */
    (*tpa_ptr).pipe_cntl += 1;      /* ...and increment pipe count  */

    return;             /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    ADD_FILE_PRM - Add Script Parameter which is a file reference */
/*    =============                         */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   SCPPRM_CNT = Parameter count                */
/*          (*tpa_ptr).open_fididxl - Index of file entry       */
/*    Outputs:  Script parameter count incremented          */
/*      Script parameter for the file initialized.      */
/*                                  */
/*    This routine is called to add a script parameter to the list of   */
/*    defined script parameters when that parameter has been determined */
/*    to be a reference to an external data file.           */
/*                                  */
/*    We add the FILE to the parameter list remembering that we set */
/*    the file index to be a passed file index by adding on the base    */
/*    value as this will show this file as a passed file.       */
/*    In addition we set the referenced flag in the file access table   */
/*    entry to mark the file as having been passed to this script. This */
/*    will prevent the file from requiring an open statement and will   */
/*    serve to identify it as a passed file to the other file access    */
/*    routines which will act accordingly.              */
/*                                  */
/************************************************************************/

unsigned long   ADD_FILE_PRM  (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct fattbl   *fat_ptr;       /* File Access Table pointer    */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct prmtbl   *prm_ptr;       /* Parameter table      */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Check to see if the addition of this parameter would exceed the maximum
     *    number of permissable parameters.
     */
    if ((*tpa_ptr).scpprm_cntl >= MAX_PARAMETERS)
    {
        REPORT_ERROR (argblk_ptr, SCL_MAXPRM);
        /* Report error if overflow */
    }

    else
    {

        /*
         *    Check script parameter is unique, if not report an error.
         */
        (void) CHK_SCRPRMUNQ (argblk_ptr);

        /*
         *    Add an entry to the Script Parameters Table.
         */
        prm_ptr = (*tpa_ptr).parameter_begps + (*tpa_ptr).scpprm_cntl;
        /* Get pointer to next FREE         */
        /* entry in script parameters table */
        (*tpa_ptr).scpprm_cntl += 1;    /* Increment count of parameters    */

        /*
         *    Clear the script parameter record before using it in case there is data
         *    left over from a previous compilation.
         */
        (void) memset (prm_ptr ,'\0' ,sizeof (struct prmtbl));

        /*
         *    ...and copy in the FILE parameters data to the script parameter table.
         */
        (*prm_ptr).prm_addrl = ADDR_NONE;
        /* Null addressing mode */
        (*prm_ptr).prm_dtypl = DTYP_FILE;
        /* Data type = File     */
        (*prm_ptr).prm_sizel = (*tpa_ptr).open_fididxl + PASSED_FILE_BASE;
        /* Size = file index        */

        /*
         *    Pickup the file access table entry and mark this file as having been
         *    passed as a parameter.
         */
        fat_ptr = (*tpa_ptr).fat_begps + (*tpa_ptr).open_fididxl - 1;
        /* Point to current entry in table */
        (*fat_ptr).fat_refflgl = PASSED_FILE;
        /* ...and show passed file  */

        (*prm_ptr).prm_nlengthl = strlen ((*fat_ptr).fat_namec20);
        /* ...and copy file name length */
        (void) memcpy ((*prm_ptr).prm_namec16,
                       (*fat_ptr).fat_namec20,
                       (*prm_ptr).prm_nlengthl);
        /* ...and copy the file name */

    }

    return PSL_NORMAL;
}


/******************************************************************************
 * @func    This routine adds an entry to the Labels Symbol Table reporting an
 *      error, if the Symbol Table is full.
 *
 * @parm    Pointer to Tparse argument block.
 * @parm    Pointer to label name.
 * @parm    Label name length.
 */
void ADD_LABEL (struct tpaarg *argblk_ptr,
                char *label_ptr,
                unsigned long label_len)
{
    /*
     *  Local data.
     */
    struct   scl_com *com_ptr;      /* SCL common data */
    struct   scl_tpa *tpa_ptr;      /* SCL Tparse data */
    struct   labtbl  *labtbl_ptr;       /* Pointer to labels symbol table */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;

    /*
     *  If the Label Symbol Table is full report an error and return.
     */
    if ((*tpa_ptr).lab_cntl >= MAX_LABELS)
    {
        REPORT_ERROR (argblk_ptr, SCL_MAXLAB);
        return;
    }

    /*
     *  Get pointer to next free entry in Labels Symbol Table.
     */
    labtbl_ptr = (*tpa_ptr).label_begps + (*tpa_ptr).lab_cntl;

    /*
     *  Add entry to the Labels Symbol Table.
     */
    (void) memcpy (labtbl_ptr->lab_namec16, label_ptr, label_len);
    /* Set label name */
    labtbl_ptr->lab_nlengthl = label_len;
    /* Set label name length */
    labtbl_ptr->lab_defn = PTRUE;   /* Flag label is defined */

    tpa_ptr->lab_cntl += 1;     /* Increment label count */

    return;
}



/************************************************************************/
/*                                  */
/*    ADD_PIPE_PRM - Add Script Parameter which is a pipe reference */
/*    =============                         */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   SCPPRM_CNT = Parameter count                */
/*          (*tpa_ptr).open_fididxl - Index of pipe entry       */
/*    Outputs:  Script parameter count incremented          */
/*      Script parameter for the pipe initialized.      */
/*                                  */
/*    This routine is called to add a script parameter to the list of   */
/*    defined script parameters when that parameter has been determined */
/*    to be a reference to an external data pipe.           */
/*                                  */
/*    We add the PIPE to the parameter list remembering that we set */
/*    the pipe index to be a passed pipe index by adding on the base    */
/*    value as this will show this pipe as a passed pipe.       */
/*    In addition we set the referenced flag in the pipe access table   */
/*    entry to mark the pipe as having been passed to this script. This */
/*    will prevent the pipe from requiring an open statement and will   */
/*    serve to identify it as a passed pipe to the other pipe access    */
/*    routines which will act accordingly.              */
/*                                  */
/************************************************************************/

unsigned long   ADD_PIPE_PRM  (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct fattbl   *fat_ptr;       /* Pipe Access Table pointer    */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct prmtbl   *prm_ptr;       /* Parameter table      */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Check to see if the addition of this parameter would exceed the maximum
     *    number of permissable parameters.
     */
    if ((*tpa_ptr).scpprm_cntl >= MAX_PARAMETERS)
    {
        REPORT_ERROR (argblk_ptr, SCL_MAXPRM);
        /* Report error if overflow */
    }

    else
    {
        /*
         *    Check script parameter is unique, if not report an error.
         */
        (void) CHK_SCRPRMUNQ (argblk_ptr);

        /*
         *    Add an entry to the Script Parameters Table.
         */
        prm_ptr = (*tpa_ptr).parameter_begps + (*tpa_ptr).scpprm_cntl;
        /* Get pointer to next FREE         */
        /* entry in script parameters table */
        (*tpa_ptr).scpprm_cntl += 1;    /* Increment count of parameters    */

        /*
         *    Clear the script parameter record before using it in case there is data
         *    left over from a previous compilation.
         */
        (void) memset (prm_ptr ,'\0' ,sizeof (struct prmtbl));

        /*
         *    ...and copy in the PIPE parameters data to the script parameter table.
         */
        (*prm_ptr).prm_addrl = ADDR_NONE;
        /* Null addressing mode */
        (*prm_ptr).prm_dtypl = DTYP_PIPE;
        /* Data type = Pipe     */
        (*prm_ptr).prm_sizel = (*tpa_ptr).open_fididxl + PASSED_PIPE_BASE;
        /* Size = pipe index        */

        /*
         *    Pickup the pipe access table entry and mark this pipe as having been
         *    passed as a parameter.
         */
        fat_ptr = (*tpa_ptr).pipe_begps + (*tpa_ptr).open_fididxl - 1;
        /* Point to current entry in table */
        (*fat_ptr).fat_refflgl |= PIPATT_PRM;
        /* ...and show passed pipe  */

        (*prm_ptr).prm_nlengthl = strlen ((*fat_ptr).fat_namec20);
        /* ...and copy file name length */
        (void) memcpy ((*prm_ptr).prm_namec16,
                       (*fat_ptr).fat_namec20,
                       (*prm_ptr).prm_nlengthl);
        /* ...and copy the file name */

    }

    return PSL_NORMAL;
}


/************************************************************************/
/*                                  */
/*    ADD_PARAMETER - Add Script Parameter              */
/*    =============                         */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   SCPPRM_CNT = Parameter count                */
/*          VAR_PTR = Address of variable's entry in the        */
/*                Variables Symbol Table            */
/*    Outputs:  None                            */
/*                                  */
/************************************************************************/

unsigned long   ADD_PARAMETER (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com  *com_ptr;     /* Common Data          */
    struct   scl_tpa  *tpa_ptr;     /* TPARSE Data          */
    struct   prmtbl   *prm_ptr;     /* Parameter table      */
    struct   symtbl   *var_ptr;     /* Variable pointer     */
    unsigned long     saved_diagctx;    /* Saved diagnostics context    */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Add a diagnostic to the stack to detail where we are up to.
     */
    saved_diagctx = PSL_ADD_STATDIAG ((*com_ptr).diag_ptrps
                                      ,"During ADD_PARAMETER");

    /*
     *    Check to see if the addition of this integer value would cause the
     *    value list buffer to overflow and report an error if it would.
     */
    if ((*tpa_ptr).scpprm_cntl >= MAX_PARAMETERS)
    {
        REPORT_ERROR (argblk_ptr, SCL_MAXPRM);
        /* Report error if overflow */
    }

    else
    {
        /*
         *    Check script parameter is unique, if not report an error.
         */
        (void) CHK_SCRPRMUNQ (argblk_ptr);

        /*
         *    Add an entry to the Script Parameters Table.
         */
        prm_ptr = (*tpa_ptr).parameter_begps + (*tpa_ptr).scpprm_cntl;
        /* Get pointer to next FREE         */
        /* entry in script parameters table */
        (*tpa_ptr).scpprm_cntl += 1;    /* Increment count of parameters    */
        var_ptr = (*tpa_ptr).var_ptrps; /* Pointer to current variable  */

        /*
         *    Clear the script parameter record before using it in case there is data
         *    left over from a previous compilation.
         */
        (void) memset (prm_ptr ,'\0' ,sizeof (struct prmtbl));

        /*
         *    ...and copy in the script parameters data to the script parameter table.
         */
        (*prm_ptr).prm_addrl = (*var_ptr).var_addrl;
        /* Copy addressing mode     */
        (*prm_ptr).prm_dtypl = (*var_ptr).var_typel;
        /* ...and copy data type    */
        (*prm_ptr).prm_sizel = (*var_ptr).var_sizel;
        /* ...and copy data size    */
        (*prm_ptr).prm_offsetl = (*var_ptr).var_offl;
        /* ...and copy variable offset  */
        (*prm_ptr).prm_nlengthl = (*var_ptr).var_nlengthl;
        /* ...and copy variable name length */
        (*prm_ptr).prm_plengthl = (*var_ptr).var_plengthl;
        /* ...and copy prompt length    */

        (void) memcpy ((*prm_ptr).prm_namec16,
                       (*var_ptr).var_namec16,
                       (*var_ptr).var_nlengthl);
        /* ...and copy the variable name */
        (void) memcpy ((*prm_ptr).prm_promptc50,
                       (*var_ptr).var_promptc50,
                       (*var_ptr).var_plengthl);
        /* ...and copy the prompt string */

        /*
         *    Now handle the parameter options.
         */
        if (((*var_ptr).var_optl & VOPT_ECHO) != 0)
        {
            (*prm_ptr).prm_echol = 1;   /* Show echo if specified   */
        }

        if (((*var_ptr).var_optl & VOPT_UPPERCASE) != 0)
        {
            (*prm_ptr).prm_uppercasel = 1;
            /* Show uppercase if specified  */
        }

    }

    PSL_REMOVE_DIAGNOSTICS ((*com_ptr).diag_ptrps ,saved_diagctx);
    return PSL_NORMAL;
}

/************************************************************************/
/*                              */
/*    ADD_QUOTED - Place quoted character string in character buffer    */
/*    ==========                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   None.                           */
/*    Outputs:  tpa_ptr->chr_stringc65536 = String excluding quotes.    */
/*      tpa_ptr->chr_strlenl = String length.           */
/*                                      */
/*    This routine picks up the quoted character string identified by   */
/*    the token pointer and token count from the source line (i.e.  */
/*    the line in its original case) and copies it - excluding the  */
/*    quotes - to the string buffer in the SCL Tparse data structure.   */
/*                                  */
/*    This routine is invoked by the CHAR_QUOTED subexpression.     */
/*                                  */
/************************************************************************/
unsigned long ADD_QUOTED (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    unsigned long offset;           /* Offset in line */
    struct   scl_com *com_ptr;      /* Common Data */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;

    /*
     *  Get the length and offset of the string (excluding quotes) in the original
     *  source line and copy it to the character string buffer in the SCL Tparse
     *  data structure.
     */
    tpa_ptr->chr_strlenl = argblk_ptr->tokencntl - 2;
    offset = (long) (argblk_ptr->token_ptrpc - tpa_ptr->parsable_linec136 + 1);
    /* Current offset in line   */
    (void) memcpy (tpa_ptr->chr_stringc65536,
                   &tpa_ptr->source_linec136 [offset],
                   tpa_ptr->chr_strlenl);

    return PSL_NORMAL;
}


/************************************************************************/
/*                                  */
/*    ADD_RANGE_VAL - Add Range Value                   */
/*    ==============                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   PSLTPA$L_PARAM(AP) = Flag indicating whether value  */
/*                       is low or high range value     */
/*          INT_VALUE = Integer value               */
/*    Outputs:  DEF_LOW_RANGEL = Set if low range value         */
/*          DEF_HIGH_RANGEL = Set if high range value       */
/*                                  */
/*    This routine sets either the low or high range value variable */
/*    depending upon the setting of the supplied flag.          */
/*                                  */
/************************************************************************/

unsigned long   ADD_RANGE_VAL (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Check to see if we are entering a low or high range value and save the
     *    value into the relevant variable.
     */
    if ((*argblk_ptr).paraml == RANGE_LOW)
        /* If LOW range entered     */
    {
        (*tpa_ptr).def_low_rangel = (*tpa_ptr).int_valuel;
        /* Save into low range value    */
    }

    else                /* ...otherwise         */
    {
        (*tpa_ptr).def_high_rangel = (*tpa_ptr).int_valuel;
        /* Save into high range value   */
    }

    return PSL_NORMAL;
}

/************************************************************************/
/*                              */
/*    ADD_SHRVAR - Add a Shareable Variable             */
/*    ==========                            */
/*                                  */
/*    Type: Service routine                     */
/*    Status:   Not applicable                      */
/*    Pointers: Unchanged                       */
/*    Inputs:   OBJ_SHRVAR = Shareable Variables Symbol Table       */
/*      SHRVAR_CNT = Shareable variable count           */
/*      def_namec64 = Definition name               */
/*    Outputs:  SHRVAR_CNT = Shareable variable count incremented   */
/*                                  */
/*    This routine attempts to add an entry to the Shareable Variables  */
/*    Symbol Table using the data held in the 'work' variable       */
/*    def_namec64. If the Symbol Table is already full it reports an    */
/*    error. The Shareable Variables Symbol Table contains entries  */
/*    for variables that may be shared by different scripts, i.e.   */
/*    thread, global and testwide variables.                */
/*                                  */
/************************************************************************/
static void ADD_SHRVAR (struct tpaarg *argblk_ptr,
                        struct symtbl *var_ptr,
                        unsigned long var_size,
                        unsigned long addr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct shrvartbl
                *shrvar_ptr;        /* Shareable variable table entry pointer */
    unsigned long   alignment;      /* Alignment count      */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Check to ensure that we have not exceeded the maximum number of
     *    shareable variables and report an error if we have.
     */
    if ((*tpa_ptr).shrvar_cntl >= MAX_SHRVAR)
{
        REPORT_ERROR (argblk_ptr, SCL_MAXGBL);
        /* Pass to underline routine    */
    }

    else
    {

        if  ((*tpa_ptr).def_dtypl == DTYP_INTEGER)
        {
            (*tpa_ptr).shrvar_rec_offl += (*tpa_ptr).shrvar_align_offl;
            (*tpa_ptr).shrvar_data_offl += (*tpa_ptr).shrvar_align_offl;
            /* Add alignment offset for ints   */
        }

        /*
         *    Add an entry to the Shareable Variables Symbol Table.
         */
        shrvar_ptr = (*tpa_ptr).shrvar_begps + (*tpa_ptr).shrvar_cntl;
        /* Point to next FREE entry in table */
        /*
         *    Clear the shareable variables record before using it in case there is data
         *    left over from a previous compilation.
         */
        (void) memset (shrvar_ptr ,'\0' ,sizeof (struct shrvartbl));

        /*
         *    ...and copy in the shareable variable data.
         */
        (void) memcpy ((*shrvar_ptr).shrvar_namec16,
                       (*tpa_ptr).def_namec64,
                       (*tpa_ptr).def_nlengthl);
        /* Copy variable name into table  */
        (*shrvar_ptr).shrvar_dtypl = (*tpa_ptr).def_dtypl;
        /* ...and variable data type       */
        (*var_ptr).var_offl = (*tpa_ptr).shrvar_rec_offl;
        /* Field offset, if any */
        (*tpa_ptr).shrvar_cntl += 1;    /* Increment shareable variable count */


        /*
         *    We must be very careful regarding the size of the shareable variable as
         *    we must ensure that we set it to be the total size required by all
         *    elements of the variable if it is an array. This is the only variable type
         *    for which we need to do this as all other types have arrays taken into
         *    account in SCL, shareable variables need to inform LGR that they are an array so that
         *    sufficient space can be allocated for them at runtime.
         */
        (*shrvar_ptr).shrvar_sizel = var_size;/* ...and variable size          */
        (*tpa_ptr).shrvar_rec_offl += var_size;
        /* Increment shareable count          */
        /*
         *    Calculate the alignment offset - incase needed for alignment of following
         *    variable.
         */
        (*tpa_ptr).shrvar_data_offl += var_size;
        /* Keep track of data area used    */
        alignment = ((*tpa_ptr).shrvar_rec_offl % sizeof(long));
        /* Get any "remainder"      */

        alignment = ( ( sizeof(long) - alignment ) % sizeof(long));
        /* Get longword shortfall   */
        /*
         *    Save alignment
         */
        (*tpa_ptr).shrvar_align_offl = alignment;

        /*
         * If record header save shrvar_ptr entry so we can assign size
         * record.
         * This will be used bt set_end_record to save the size of
         * the entire record which is then used in LGR_INIT_GCA to set
         * up the record alignment properly.
         */

        if ((*shrvar_ptr).shrvar_dtypl == DTYP_RECORD)
        {
            (*tpa_ptr).last_shrvar_recps = shrvar_ptr;
        }

        /*
         *  Save addressing mode.
         */
        (*shrvar_ptr).shrvar_addrl = addr;
    }

    return;             /* ...and return */
}


/************************************************************************/
/*                              */
/*    ADD_STRING_CHR - Add character to string.                 */
/*    ===============                               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   tpa_ptr->string_chrc8 - Single character string to be   */
/*          added to consolidated character string.     */
/*      tpa_ptr->chr_stringc65536 = Consolidated character  */
/*          string.                     */
/*      tpa_ptr->chr_strlenl = Consolidated character string    */
/*          length.                     */

/*      FORMAT_FLAGL = Flag indicating if in FORMAT command     */
/*      FORMAT_ARGUML = Current number of FORMAT arguments  */
/*      FORMAT_CHAR_SET = Flag to indicate previous character   */
/*                was a "%" indicating possible     */
/*                format identifier         */
/*                                  */
/*    Outputs:  tpa_ptr->chr_stringc65536 = Character string added. */
/*      tpa_ptr->chr_strlenl = Incremented by length of single  */
/*          character string.                   */
/*      FORMAT_ARGUML = Incremented if F.I. Identified      */
/*                                  */
/*      FORMAT_CHAR_SET = Set if current char is "%"            */
/*                                  */
/*    This routine adds the single character string held in     */
/*    tpa_ptr->string_chrc8 to the consolidated character string    */
/*    tpa_ptr->chr_stringc65536 and updates the character string    */
/*    length tpa_ptr->chr_strlenl. The routine allows the string to */
/*    exceeded the string length limit by one character value,      */
/*    thereafter all characters are ignored. Any routine using      */
/*    tpa_ptr->chr_stringc65536 must check its length is in bounds !    */
/*                                  */
/*    Checks if Format identifiers are in char_value string.        */
/*    Keeps a count of them.                        */
/*                                  */
/************************************************************************/
unsigned long ADD_STRING_CHR (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */
    char     *wrk_ptr;       /* Working pointer */
    long chrlen;         /* Single character string length */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *  If there is space left in the character string store then copy the parsed
     *  character string into the consolidated character string, updating the total length.
     */
    chrlen = strlen ((*tpa_ptr).string_chrc8);
    if (((*tpa_ptr).chr_strlenl + chrlen) <= MAX_LITERAL_SIZE)
        /* If space left in string */
    {
        wrk_ptr = (*tpa_ptr).chr_stringc65536 + (*tpa_ptr).chr_strlenl;
        /* Calc next FREE address in string */
        memcpy (wrk_ptr, (*tpa_ptr).string_chrc8, chrlen);
        /* Copy string into consolidated string */
        (*tpa_ptr).chr_strlenl+= chrlen;    /* Increment length of string */

        /*
         * Checking if in format command and if we have format identifier.
         * Keep count of format identifiers so know how many arguments to expect.
         */
        if (((*tpa_ptr).format_flagl == 1)      /* If in format command */
                && ((*tpa_ptr).format_char_setl == 1)      /* ..and previous char = % */
                && ((*tpa_ptr).string_chrc8[1] == '\0')    /* ..and we have a single character */
                && (  ((*tpa_ptr).string_chrc8[0] == 'U')  /* And this character is */
                      || ((*tpa_ptr).string_chrc8[0] == 'L')  /* ...a valid format */
                      || ((*tpa_ptr).string_chrc8[0] == 'D')  /* ...identifier */
                      || ((*tpa_ptr).string_chrc8[0] == 'T')))
        {
            (*tpa_ptr).format_arguml += 1;        /* Increment number of      */
            /* format arguments         */
            (*tpa_ptr).format_char_setl = 0;      /* Reset Format char flag   */
        }

        if ((*tpa_ptr).format_flagl == 1         /* If in format command */
                && ((*tpa_ptr).string_chrc8[1] == '\0') /* ..and we have a single character */
                && ((*tpa_ptr).string_chrc8[0] == '%')) /* .. and the character is "%" */
        {
            (*tpa_ptr).format_char_setl = 1;      /* set flag indicating    */
            /* start of format ident. */
        }
    }

    /*  Else NOT enough space in consolidated character string for next character
     *  string, update the character length to show overflow.
     */
    else /* Else no room for next character string */
    {
        (*tpa_ptr).chr_strlenl = MAX_LITERAL_SIZE + 1;
    }

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/
/*                              */
/*    ADD_THR - Add a Thread Definition                 */
/*    =======                               */
/*                                  */
/*    Type: Service routine.                    */
/*    Status:   Not applicable.                     */
/*    Pointers: Unchanged.                      */
/*    Inputs:   argblk_ptr - Tparse argument block.         */
/*      tpa_ptr - Common tparse data.               */
/*    Outputs:  tpa_ptr - Common tparse data reset for next definition  */
/*      Threads Symbol Table - Entry added to Threads Symbol    */
/*          Table for last definition (if valid).       */
/*                                  */
/*    This routine attempts to add an entry to the Threads Symbol   */
/*    Table using the data held in the definition 'work' variables. If  */
/*    the Thread Symbol Table is already full it reports an error.  */
/*                                  */
/************************************************************************/
static void ADD_THR (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */
    struct thrtbl   *thr_ptr;       /* Pointer to thread definition */
    struct thrtbl   *thrtbl_ptr;        /* Pointer to Thread Symbol Table */
    /* ...entry address */
    unsigned long   isub;           /* Loop control */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;

    /*
     *    Check for the maximum number of threads having already been entered
     *    and report an error if this is so, otherwise add an entry to the
     *    Threads Symbol Table.
     */
    if ((*tpa_ptr).thr_cntl >= MAX_THR)
    {
        REPORT_ERROR (argblk_ptr, SCL_MAXTHR);
    }

    else
    {
        thrtbl_ptr = (*tpa_ptr).thr_begps + (*tpa_ptr).thr_cntl;
        /* Point to next FREE entry in table */
        /*
         *    Copy the THREAD data into the Threads Symbol Table.
         */
        *thrtbl_ptr = *thr_ptr;     /* Copy data from work record */

        /*
         *    Copy name from definition data, null terminate and convert to lowercase.
         */
        (void) memcpy ((*thrtbl_ptr).thr_namec20,
                       (*tpa_ptr).def_namec64,
                       (*tpa_ptr).def_nlengthl);
        /*
         * .Reset TOF Source Line Number (unless processing an INCLUDED file).
         */
        if ((*com_ptr).include_depthl == 0)
        {
            (*thr_ptr).thr_tof_srclinl = (*com_ptr).srclinl10[0];
        }

        (*thrtbl_ptr).thr_namec20[(*tpa_ptr).def_nlengthl] = '\0';

        for (isub = 0; isub < (*tpa_ptr).def_nlengthl; isub++)
        {
            (*thrtbl_ptr).thr_namec20 [isub] =
                tolower ((*thrtbl_ptr).thr_namec20 [isub]);
        }

        /*
         *    Show that we have finished with the structure in the work structure.
         */
        (*thr_ptr).thr_tdf_ptrps = NULL;

        (*tpa_ptr).thr_cntl += 1;   /* Increment thread count */
    }

    return;             /* ...and return */
}


/************************************************************************/
/*                              */
/*    ADD_THR_DEF - Add Thread Script Definition to Symbol Table    */
/*    ===========                           */
/*                                  */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Tparse argument block.         */
/*      tpa_ptr - Common tparse data.               */
/*    Outputs:  tpa_ptr - Common tparse data reset for next definition  */
/*      Symbol Table - Entry added to appropriate Symbol    */
/*          Table for last definition (if valid).       */
/*                                  */
/*    This routine is called when the parsing of a definition in a  */
/*    Thread script is complete. It first checks that the definition    */
/*    is 'complete', and if it is, adds an entry for it to the      */
/*    appropriate symbol table. It then resets all the 'work'       */
/*    variables used to build a definition.             */
/*                                  */
/************************************************************************/
unsigned long ADD_THR_DEF (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */
    Boolean  duplicate;      /* Duplicate definition flag */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Check to see if we have a definition being parsed.
     */
    if ((*tpa_ptr).def_typl == 0) goto end;

    /*
     *    Check the thread definition and report any errors.
     */
    (void) CHK_THR_DEF (argblk_ptr);

    /*
     *    Check for duplicate definitions - do not allow duplicate definitions.
     */
    duplicate = CHECK_FOR_DUPLICATES (argblk_ptr);
    if (duplicate) goto end;

    /*
     *    We are parsing a definition - If we have a name then add the definition
     *    otherwise simply clear the definition parsing data.
     */
    if ((*tpa_ptr).def_namec64 [0] != '\0')
        /* If definition name specified */
    {
        ADD_THR (argblk_ptr);       /* Add a THREAD definition */
    }

    /*
     *    Clear the 'work' variables used to build a Thread script definition.
     */
end:
    ;
    INI_THR_DEF (com_ptr);      /* Initialise definition data */

    return PSL_NORMAL;          /* Return success */
}

/************************************************************************/
/*                                  */
/*    ADD_TIMER - Add a Timer Definition                */
/*    =========                             */
/*                                  */
/*    Type: Service routine                     */
/*    Status:   Not applicable                      */
/*    Pointers: Unchanged                       */
/*    Inputs:   OBJ_TIMERS = Timer Symbol Table             */
/*          TIM_CNT = Timer count                   */
/*          def_namec64 = Definition name               */
/*    Outputs:  TIM_CNT = Timer count incremented           */
/*                                  */
/*    This routine attempts to add an entry to the Timers Symbol    */
/*    Table using the data held in the 'work' variable def_namec64. */
/*    If the Timers Symbol Table is already full it reports an error.   */
/*                                  */
/************************************************************************/

void        ADD_TIMER (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct timtbl   *tim_ptr;       /* Timer symbol table       */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Check for the maximum number of timers having already been entered and
     *    report an error if this is so.
     */
    if ((*tpa_ptr).tim_cntl >= MAX_TIMERS)
    {
        REPORT_ERROR (argblk_ptr, SCL_MAXTIM);
        /* Pass to underline routine    */
    }

    else
    {

        /*
         *    Add an entry to the Timers Symbol Table.
         */
        tim_ptr = (*tpa_ptr).timer_begps + (*tpa_ptr).tim_cntl;
        /* Point to next FREE entry in table */

        /*
         *    Clear the TIMER record before using it in case there is data
         *    left over from a previous compilation.
         */
        (void) memset (tim_ptr ,'\0' ,sizeof (struct timtbl));

        /*
         *    ...and copy in the TIMER data to the TIMERS table.
         */
        (void) memcpy ((*tim_ptr).tim_namec68,
                       (*tpa_ptr).def_namec64,
                       (*tpa_ptr).def_nlengthl); /* Copy the timer name into table */
        /*
         * .Reset TOF Source Line Number (unless processing an INCLUDED file).
         */
        if ((*com_ptr).include_depthl == 0)
        {
            (*tim_ptr).tim_tof_srclinl = (*com_ptr).srclinl10[0];
        }

        (*tim_ptr).tim_namec68 [(*tpa_ptr).def_nlengthl] = '\0';
        /* Null-terminate */
        (*tpa_ptr).tim_cntl += 1;   /* Increment timer count    */
    }

    return;             /* ...and return        */
}

/************************************************************************/
/*                              */
/*    ADD_TST - Add a Test Definition                   */
/*    =======                               */
/*                                  */
/*    Type: Service routine.                    */
/*    Status:   Not applicable.                     */
/*    Pointers: Unchanged.                      */
/*    Inputs:   argblk_ptr - Tparse argument block.         */
/*      tpa_ptr - Common tparse data.               */
/*    Outputs:  tpa_ptr - Common tparse data reset for next definition  */
/*      Tests Symbol Table - Entry added to Tests Symbol    */
/*          Table for last definition (if valid).       */
/*                                  */
/*    This routine attempts to add an entry to the Tests Symbol Table   */
/*    using the data held in the definition 'work' variables. If the    */
/*    Tests Symbol Table is already full it reports an error.       */
/*                                  */
/************************************************************************/
static void ADD_TST (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */
    struct tsttbl   *tst_ptr;       /* Tests Symbol Table */
    struct tsttbl   *tsttbl_ptr;        /* Pointer to Test Symbol Table */
    /* ...entry address */
    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    tst_ptr = (*tpa_ptr).def_tst_wrkps;

    /*
     *    Check for the maximum number of tests having already been entered
     *    and report an error if this is so, otherwise add an entry to the
     *    Tests Symbol Table.
     */
    if ((*tpa_ptr).tst_cntl >= MAX_TST)
    {
        REPORT_ERROR (argblk_ptr, SCL_MAXTST);
    }

    else
    {
        tsttbl_ptr = (*tpa_ptr).tst_begps + (*tpa_ptr).tst_cntl;
        /* Point to next FREE entry in table */
        /*
         *    Copy the TEST data into the Tests Symbol Table.
         */
        *tsttbl_ptr = *tst_ptr;     /* Copy data from work record */

        /*
         *    Copy name from definition data and null terminate.
         */
        (void) memcpy ((*tsttbl_ptr).tst_namec68,
                       (*tpa_ptr).def_namec64,
                       (*tpa_ptr).def_nlengthl);
        
        (*tsttbl_ptr).tst_namec68[(*tpa_ptr).def_nlengthl] = '\0';

        /*
         * .Reset TOF Source Line Number (unless processing an INCLUDED file).
         */
        if ((*com_ptr).include_depthl == 0)
        {
            (*tsttbl_ptr).tst_tof_srclinl = (*com_ptr).srclinl10[0];
        }

        /*
         *    Show that we have finished with the structures in the work structure.
         */
        (*tst_ptr).tst_prm_ptrps = NULL;
        (*tst_ptr).tst_tlh_ptrps = NULL;

        (*tpa_ptr).tst_cntl += 1;   /* Increment test count */
    }

    return;             /* Return */
}


/************************************************************************/
/*                              */
/*    ADD_TSTTHR - Add a Test Thread Definition             */
/*    ==========                            */
/*                                  */
/*    Type: Service routine.                    */
/*    Status:   Not applicable.                     */
/*    Pointers: Unchanged.                      */
/*    Inputs:   argblk_ptr - Tparse argument block.         */
/*      tpa_ptr - Common tparse data.               */
/*    Outputs:  tpa_ptr - Common tparse data reset for next definition  */
/*      Test Threads Symbol Table - Entry added to Test Threads */
/*          Symbol Table for last definition (if valid).    */
/*                                  */
/*    This routine attempts to add an entry to the Test Threads Symbol  */
/*    Table using the data held in the definition 'work' variables. If  */
/*    the Test Thread Symbol Table is already full it reports an error. */
/*                                  */
/************************************************************************/
static void ADD_TSTTHR (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com   *com_ptr;      /* Common Data */
    struct scl_tpa   *tpa_ptr;      /* TPARSE Data */
    struct tstthrtbl *tstthr_ptr;       /* Test Threads Symbol Table */
    struct tstthrtbl *tstthrtbl_ptr;    /* Pointer to Test Threads Symbol */
    /* ...Table entry address */
    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    tstthr_ptr = (*tpa_ptr).def_tstthr_wrkps;

    /*
     *    Check for the maximum number of test threads having already been entered
     *    and report an error if this is so, otherwise add an entry to the
     *    Test Threads Symbol Table.
     */
    if ((*tpa_ptr).tstthr_cntl >= MAX_TSTTHR)
    {
        REPORT_ERROR (argblk_ptr, SCL_MAXTSTTHR);
    }

    else
    {
        tstthrtbl_ptr = (*tpa_ptr).tstthr_begps + (*tpa_ptr).tstthr_cntl;
        /* Point to next FREE entry in table */
        /*
         *    Copy the TEST THREAD data into the Test Threads Symbol Table.
         */
        *tstthrtbl_ptr = *tstthr_ptr;   /* Copy data from work record */

        /*
         *    Copy name from definition data and null terminate.
         */
        (void) memcpy ((*tstthrtbl_ptr).tstthr_namec68,
                       (*tpa_ptr).def_namec64,
                       (*tpa_ptr).def_nlengthl);

        (*tstthrtbl_ptr).tstthr_namec68[(*tpa_ptr).def_nlengthl] = '\0';

        /*
         * .Reset TOF Source Line Number (unless processing an INCLUDED file).
         */
        if ((*com_ptr).include_depthl == 0)
        {
            (*tstthr_ptr).tstthr_tof_srclinl = (*com_ptr).srclinl10[0];
        }

        /*
         *    Show that we have finished with the structures in the work structure.
         */
        (*tstthr_ptr).tstthr_tthr_ptrps = NULL;
        (*tstthr_ptr).tstthr_tdf_ptrps = NULL;

        (*tpa_ptr).tstthr_cntl += 1;    /* Increment test thread count */
    }

    return;             /* Return */
}



/******************************************************************************
 * @func    This routine adds an 'undefined label' entry to the Labels Symbol
 *      Table reporting an error, if the Symbol Table is full.
 *
 * @parm    Pointer to Tparse argument block.
 * @parm    Pointer to label name.
 * @parm    Label name length.
 */
void ADD_UNDEF_LABEL (struct tpaarg *argblk_ptr,
                      char *label_ptr,
                      unsigned long label_len)
{
    /*
     *  Local data.
     */
    struct   scl_com *com_ptr;      /* SCL common data */
    struct   scl_tpa *tpa_ptr;      /* SCL Tparse data */
    struct   labtbl  *labtbl_ptr;       /* Pointer to labels symbol table */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;

    /*
     *  If the Label Symbol Table is full report an error and return.
     */
    if ((*tpa_ptr).lab_cntl >= MAX_LABELS)
    {
        REPORT_ERROR (argblk_ptr, SCL_MAXLAB);
        return;
    }

    /*
     *  Get pointer to next free entry in Labels Symbol Table.
     */
    labtbl_ptr = (*tpa_ptr).label_begps + (*tpa_ptr).lab_cntl;

    /*
     *  Add 'undefined label' entry to the Labels Symbol Table.
     */
    (void) memcpy (labtbl_ptr->lab_namec16, label_ptr, label_len);
    /* Set label name */
    labtbl_ptr->lab_nlengthl = label_len;
    /* Set label name length */
    labtbl_ptr->lab_defn = PFALSE;  /* Flag label not yet defined */

    tpa_ptr->lab_cntl += 1;     /* Increment label count */

    return;
}


/************************************************************************/
/*                                  */
/*    ADD_VARIABLE - Add a Variable Definition              */
/*    =============                         */
/*                                  */
/*    Type: Service routine                     */
/*    Status:   Not applicable                      */
/*    Pointers: Not changed                     */
/*    Inputs:   VARIABLES_TABLE = Variables Symbol Table        */
/*          OBJ_SHRVAR = Shareable Variables Symbol table       */
/*          OBJ_VALUES_PTR = Value List Buffer Pointer      */
/*          VAR_CNT = Variable count                */
/*          VAL_CNT = Value list/range count            */
/*          SCRIPT_VAL_CNT = Script variables with associated   */
/*               values count               */
/*          RANDOM_VAL_CNT = Random Value list/range count      */
/*          SHRVAR_CNT = Shareable variable count           */
/*          def_namec64 = Definition name               */
/*    Outputs:  VAR_CNT = Variable count incremented            */
/*          SCRIPT_VAL_CNT = Script variables with          */
/*                   associated values count updated    */
/*          VAL_CNT = Value list/range count incremented variable   */
/*                has an associated value list/range        */
/*          RANDOM_VAL_CNT = Random Value list/range count      */
/*               incremented if variable has a 'Random' */
/*               Value List/Range           */
/*          SHRVAR_CNT = Shareable variable count incremented if    */
/*                   variable is shareable.         */
/*                                  */
/*    This routine attempts to add an entry to the Variables Symbol */
/*    Table using the data held in the 'work' variables DEF_*. If   */
/*    the variable is shareable across multiple scripts (i.e. it is */
/*    a thread, global or testwide variable) it also attempts to add    */
/*    an entry to the Shareable Variables Symbol Table. If either of    */
/*    the required Symbol Tables is full an error is reported.      */
/*                                  */
/*    All integer data is longword aligned, character variables are */
/*    stored consecutively. Padding is only added as necessary      */
/*    between characters and integers.                  */
/*                                  */
/************************************************************************/
void ADD_VARIABLE (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct   var_symtbl *var_symtbl_ptr;    /* Pointer to DEBUG var table   */
    struct   var_symtbl var_symtbl;     /* variables symbol table entry */
    struct   scl_com *com_ptr;      /* Common data pointer      */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   symtbl  *var_ptr;      /* Variables symbol table   */
    struct   scl_cmd *scl_cmd_ptr;      /* Pointer to Command Line Struct */
    unsigned long    const_offset;      /* Offset of constant       */
    unsigned long    iret;          /* Status return variable.  */
    unsigned long    isub;          /* Loop control         */
    unsigned long    total_var_size;    /* Total length of variable*/
    long    *tmp_ptr;      /* Temp pointer to long     */
    char    *const_ptr;        /* Temp pointer for constants   */
    void    *void_ptr;     /* WORK POINTER TO VOID     */
    unsigned long   alignment;      /* Alignment count          */

    struct   scph    *header_ptr;       /* Pointer to header record */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    scl_cmd_ptr = (*com_ptr).cmd_ptrps;
    header_ptr = tpa_ptr->header_begps; /* Pointer to header */

    /*
     *    Check for the maximum number of variables having already been entered and
     *    report an error if this is so.
     */
    if ((*tpa_ptr).var_cntl >= MAX_VARIABLES)
    {
        REPORT_ERROR (argblk_ptr, SCL_MAXVAR);
        /* Pass to underline routine    */
        return;             /* ...and return        */
    }

    /*
     *     If we have an integer then set the size to be the size of a long
     */
    if ((*tpa_ptr).def_dtypl == DTYP_INTEGER)
    {
        (*tpa_ptr).def_sizel = INTEGER_SIZE;
        /* All integers are 4 bytes long */
    }

    /*
     *    If the variable is an array then we must calculate the total size of
     *    the array to account for the array dimensions.
     */
    total_var_size = (*tpa_ptr).def_sizel;
    for (isub = 0; isub < (*tpa_ptr).array_dimcntl; isub++)
    {
        total_var_size *= ((*tpa_ptr).def_array_bounds2l10[isub]
                           - (*tpa_ptr).def_array_bounds1l10[isub] + 1);
    }

    /*
     *    ...and now add the variable to the variables symbol table
     */
    var_ptr = (*tpa_ptr).variable_begps + (*tpa_ptr).var_cntl;
    /* Point to next FREE entry in table */

    /*
     *    Clear the VARIABLE TABLE record before using it in case there is data
     *    left over from a previous compilation.
     */
    (void) memset (var_ptr ,'\0' ,sizeof (struct symtbl));


    /*
     *    ...and copy in the VARIABLE data to the variables symbol table.
     */
    (void) memcpy ((*var_ptr).var_namec16,
                   (*tpa_ptr).def_namec64,
                   (*tpa_ptr).def_nlengthl);   /* Copy variable name into table  */

    /*
     * .Reset TOF Source Line Number (unless processing an INCLUDED file).
     */
    if ((*com_ptr).include_depthl == 0)
    {
        (*var_ptr).var_tof_srclinl = (*com_ptr).srclinl10[0];
    }

    (*var_ptr).var_nlengthl = (*tpa_ptr).def_nlengthl;
    /* Copy length of name into table  */

    (*var_ptr).var_addrl = (*tpa_ptr).def_addrl;
    /*     ...and addressing mode */
    (*var_ptr).var_typel = (*tpa_ptr).def_dtypl;
    /* ...and variable data type    */
    (*var_ptr).var_sizel = (*tpa_ptr).def_sizel;
    /* ...and variable size     */
    (*var_ptr).var_optl = (*tpa_ptr).def_optl;
    /* ...and variable options  */

    /*
     *  Unless the variable is a 'record' variable, set the parent record index.
     *  The record index will be zero for variables not within a record definition.
     */
    if ((*tpa_ptr).def_dtypl == DTYP_RECORD)
    {
        (*var_ptr).var_recidxl = 0;
    }
    else
    {
        (*var_ptr).var_recidxl = (*tpa_ptr).def_recidxl;
    }

    (*var_ptr).var_values_offl = (*tpa_ptr).def_values_offl;
    /* Value list start offset  */
    (*var_ptr).var_values_cntl = (*tpa_ptr).def_values_cntl;
    /* Count of values in list  */
    (*var_ptr).var_values_endl = ((unsigned long) (*tpa_ptr).values_nxtpc -
                                  (unsigned long) (*tpa_ptr).values_begpc);
    /* Offset of byte beyond end of */
    /* the value list       */

    (*var_ptr).var_low_rangel = (*tpa_ptr).def_low_rangel;
    /* Save the low range value */
    (*var_ptr).var_high_rangel = (*tpa_ptr).def_high_rangel;
    /* Save the high range value    */

    /*
     *    Copy the array dimensions into the variable symbol table.
     */
    for (isub = 0; isub < (*tpa_ptr).array_dimcntl; isub++)
    {
        (*var_ptr).var_array_bounds1l10[isub] =
            (*tpa_ptr).def_array_bounds1l10[isub];
        (*var_ptr).var_array_bounds2l10[isub] =
            (*tpa_ptr).def_array_bounds2l10[isub];
    }
    (*var_ptr).var_array_dimcntl = (*tpa_ptr).array_dimcntl;

    /*
     *    If we have specified a seed value then save this into the variable symbol
     *    Ensure that if we have specified a seed that the repeatable random option
     *    has also been specified. If not then this is an error.
     */
    if (((*tpa_ptr).def_optl & VOPT_SEED) != 0)
        /* If seed clause specified */
    {
        if (((*tpa_ptr).def_optl & VOPT_RANDOM) == 0)
            /* ...and if NOT RANDOM     */
        {
            REPORT_ERROR (argblk_ptr, SCL_NOTREPRND);
            /* ...then report an error  */
        }

        (*var_ptr).var_seedl = (*tpa_ptr).def_seedl;
        /* ...else save seed value  */
    }

    /*
     *   Add the prompt string if one has been specified.
     */
    if (((*tpa_ptr).def_optl & VOPT_PROMPT) != 0)
        /* If prompt specified      */
    {
        (void) memcpy ((*var_ptr).var_promptc50,
                       (*tpa_ptr).def_promptc50,
                       (*tpa_ptr).def_plengthl);
        /* Copy the prompt string   */
        (*var_ptr).var_plengthl = (*tpa_ptr).def_plengthl;
        /* ...and the length of prompt  */
    }

    else                /* ...else if no prompt     */
    {
        (void) memcpy ((*var_ptr).var_promptc50,
                       (*tpa_ptr).def_namec64,
                       (*tpa_ptr).def_nlengthl);
        /* Default prompt to variable name */
        (*var_ptr).var_plengthl = (*tpa_ptr).def_nlengthl;
        /* ...and the length of prompt  */
    }

    /*
     *   If this variable is a file variable, MALLOC space for the name and
     *   save a pointer to it, otherwise set the filename pointer to NULL.
     */
    if (((*tpa_ptr).def_optl & VOPT_FILE) != 0) /* If file variable */
    {
        (*var_ptr).var_filnam_pc =
            (char *) MALLOC (strlen ((*tpa_ptr).def_filnamc256) + 1);

        if ((*var_ptr).var_filnam_pc == NULL)
        {
            PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                         ,"MALLOC (allocating space for file variable file name)", 0);
        }

        (void) strcpy ((*var_ptr).var_filnam_pc, (*tpa_ptr).def_filnamc256);
    }
    else                /* ...else not a file variable */
    {
        (*var_ptr).var_filnam_pc = NULL;
    }
    /*
     *    Set Variable Data Area Offset(s).
     */
    if ((*tpa_ptr).def_addrl == ADDR_LOCAL)
        /* If this is a LOCAL variable  */
    {

        /*
         *    Add Local Variable Offset(s)
         *    ----------------------------
         */
        if (((*tpa_ptr).def_optl & VOPT_LIST) != 0)
            /* ...and we have a value list    */
        {
            (*tpa_ptr).local_data_offl += LIST_SIZE;
            /* ...allocate 12 extra bytes in LDA */
        }

        else if (((*tpa_ptr).def_optl & VOPT_RANGE) != 0)
            /* ...or we have a value range  */
        {
            (*tpa_ptr).local_data_offl += RANGE_SIZE;
            /* ...allocate 8 extra bytes in LDA */
        }

        /*
         * THIS APPLIES TO EACH OF DATA AREAS : LOCAL,SCRIPT
         *
         *    This section has been modified so padding is added only between
         *    character and integer data. This means all integer data will be
         *    longword aligned. This fixes the misaligned record problem in
         *    a platform independent way.
         *
         *    We ignore all of the alignment code if the user has specified that he is
         *    not interested in alignment of data (records especially)
         */
        if (((*tpa_ptr).def_dtypl == DTYP_INTEGER)
                ||  ((*tpa_ptr).def_dtypl == DTYP_RECORD)
                ||  ((*tpa_ptr).def_recidxl == 0))
        {
            (*tpa_ptr).local_data_offl += (*tpa_ptr).local_align_offl;
            /* Add alignment offset        */
        }

        (*var_ptr).var_offl = (*tpa_ptr).local_data_offl;
        /* ...and save the variable offset */

        (*tpa_ptr).local_data_offl += total_var_size;
        /* Point past this item in LDA   */

        /*
         * Calculate the alignment offset - incase needed for alignment of following
         * variable.
         */
        alignment = ((*tpa_ptr).local_data_offl % sizeof(long));
        /* Get any "remainder"      */

        alignment = ((sizeof(long) - alignment) % sizeof(long));
        /* Get longword shortfall   */
        /*
         * Save alignment
         */
        (*tpa_ptr).local_align_offl = alignment;
    }

    else if ((*tpa_ptr).def_addrl == ADDR_SCRIPT)
        /* ...else If a SCRIPT variable */
    {

        /*
         *    Add Script Variable Offset(s)
         *    -----------------------------
         */
        if (((*tpa_ptr).def_optl & VOPT_LIST) != 0)
            /* ...and we have a value list    */
        {
            (*tpa_ptr).script_data_offl += LIST_SIZE;
            /* ...allocate 12 extra bytes in SDA */
            /* for Value List pointer, SEED and  */
            /* SAVED SEED   */
        }

        else if (((*tpa_ptr).def_optl & VOPT_RANGE) != 0)
            /* ...or we have a value range  */
        {
            (*tpa_ptr).script_data_offl += RANGE_SIZE;
            /* ...allocate 8 extra bytes in SDA */
            /* for SEED and SAVED SEED  */
        }

        /*
         *     Saving of variable offset in SDA and align the data
         *
         *    We ignore all of the alignment code if the user has specified that he is
         *    not interested in alignment of data (records especially)
         */
        if (((*tpa_ptr).def_dtypl == DTYP_INTEGER)
                ||  ((*tpa_ptr).def_dtypl == DTYP_RECORD)
                ||  ((*tpa_ptr).def_recidxl == 0))
        {
            (*tpa_ptr).script_data_offl += (*tpa_ptr).script_align_offl;
        }

        (*var_ptr).var_offl = (*tpa_ptr).script_data_offl;
        /* ...and save the variable offset */

        (*tpa_ptr).script_data_offl += total_var_size;
        /* Point past this item in LDA  */

        /*
         *    Calculate the alignment offset - incase needed for alignment of following
         *    variable.
         */
        alignment = ((*tpa_ptr).script_data_offl % sizeof(long));
        /* Get any "remainder"      */
        alignment = ((sizeof(long) - alignment) % sizeof(long));
        /* Get longword shortfall   */
        /*
         *    Save alignment
         */
        (*tpa_ptr).script_align_offl = alignment;
    }


    else if (((*tpa_ptr).def_addrl == ADDR_GLOBAL) ||
             ((*tpa_ptr).def_addrl == ADDR_THREAD) ||
             ((*tpa_ptr).def_addrl == ADDR_TESTWIDE))
    {                   /* ...else If a Shareable variable */

        /*
         *    Add Shareable Variable Offset(s)
         *    --------------------------------
         *
         *    Add an entry to the Shareable Variables Symbol Table for this variable.
         */
        ADD_SHRVAR (argblk_ptr ,var_ptr ,total_var_size ,(*tpa_ptr).def_addrl);
    }

    else if ((*tpa_ptr).def_addrl == ADDR_CONSTANT)
        /* ...else If a CONSTANT    */
    {

        /*
         *    Add CONSTANT Variable offset
         *    ----------------------------
         *
         *    Add the constants address within the constants area to the Variables
         *    Symbol Table entry and move the value of this variable into the
         *    constants area.
         *
         *    We must firstly assure that there is sufficient space to add the constant
         *    to the constants area.
         *
         */
        const_offset = (*tpa_ptr).constant_data_offl + (*tpa_ptr).def_sizel;
        /* Calc end of constant in table. */
        if (const_offset > MAX_CONSTANT_SIZE)
            /* if this is past end of buffer  */
        {
            REPORT_ERROR (argblk_ptr, SCL_CONSOVERFLOW);
            /* ...then report an error  */
            return;         /* ...and return        */
        }

        /*
         *    check to see if we are within a record as we do not support constants 
         *    within a record definition.
         */
        if ((*tpa_ptr).def_recidxl != 0)
        {
            (*tpa_ptr).rec_error_flgl = 1;
            /* Flag record definition error */
            REPORT_ERROR (argblk_ptr, SCL_CONINVREC);
            /* ...then report an error  */
            return;         /* ...and return        */
        }

        /*
         *   The constant value will fit into the table. Move the value of the
         *   constant into the table.
         */
        if ((*tpa_ptr).def_dtypl == DTYP_CONINTEGER)
            /* If this is a constant integer */
        {

            /*
             *    Integer - Add its value to the constants data table.
             *
             *    NOTE: For UNIX environments, ensure that the integer is added
             *    to the table at the next available longword boundary.
             */
#if defined(UNIX) || defined(WIN32)
            alignment = ((*tpa_ptr).constant_data_offl % sizeof(long));
            /* Get any "remainder"      */
            alignment  = ((sizeof(long) - alignment) % sizeof(long));
            /* Get longword shortfall   */
            (*tpa_ptr).constant_data_offl += alignment;
            /* Jump data to longword bound  */
#endif

            const_ptr = (*tpa_ptr).constant_begpc +
                        (*tpa_ptr).constant_data_offl;
            /* Next FREE address in table   */
            void_ptr = (void *) const_ptr;
            /* CONVERT TO VOID * AS FRIG TO */
            tmp_ptr = (long *) void_ptr;/* GET A POINTER TO LONG... */
            *tmp_ptr = (*tpa_ptr).int_valuel;
            /* Move integer value to table  */
            (*var_ptr).var_offl = (*tpa_ptr).constant_data_offl;
            /* Save offset of constant in table */
            (*tpa_ptr).constant_data_offl += (*tpa_ptr).def_sizel;
            /* ...point past constant in table. */
        }

        else                /* ...else must be constant character */
        {

            /*
             *    Character constant - Add its value to the constants data table.
             */
            const_ptr = (*tpa_ptr).constant_begpc +
                        (*tpa_ptr).constant_data_offl;
            /* ...next FREE address in table */
            (void) memcpy (const_ptr,   /* Copy char constant value */
                           (*tpa_ptr).chr_stringc65536,
                           (*tpa_ptr).def_sizel);

            (*var_ptr).var_offl = (*tpa_ptr).constant_data_offl;
            /* Save offset of constant in table */
            (*tpa_ptr).constant_data_offl += (*tpa_ptr).def_sizel;
            /* ...point past constant in table. */
        }
    }

    /*
     *   Handle value lists now
     */
    if (((*tpa_ptr).def_optl & (VOPT_LIST | VOPT_RANGE)) != 0)
        /* If arange/list specified */
    {
        (*tpa_ptr).val_cntl += 1;   /* Increment value count    */

        if ((*tpa_ptr).def_addrl == ADDR_SCRIPT)
            /* If a script variable     */
        {
            (*tpa_ptr).script_val_cntl += 1;
            /* Increment script value count */
        }

        if (((*tpa_ptr).def_optl & VOPT_RANDOM) != 0)
            /* If a script variable     */
        {
            (*tpa_ptr).random_val_cntl += 1;
            /* Increment script value count */
        }

    }

    /*
     *    If the script is either MODE SQL then we
     *    will need to maintain the external variables symbol table. It is this
     *    table which we shall write to the SCP and which provides a lookup
     *    functionality in the Test Executer to obtain the location of the value
     *    of a variable from its name.
     *
     *    Add the VARIABLE (null terminated) to the EXTERNAL VARIABLES symbol table.
     *    This symbol table may be written out to the SCP if this is a mode SQL
     *    script.
     */
    if (((*header_ptr).scph_mode_b & MODE_IMPACT) != 0)
    {
        var_symtbl_ptr = &var_symtbl;   /* Point to record      */
        (void) memcpy ((*var_symtbl_ptr).symbol_namec20
                       ,(*var_ptr).var_namec16
                       ,(*var_ptr).var_nlengthl);
        /* ...copy name into symbol table */
        (*var_symtbl_ptr).symbol_namec20 [(*var_ptr).var_nlengthl] = '\0';
        /* ...and null terminate it.    */

        /*
         *    ...and add the other details about the symbol to the table.
         */
        (*var_symtbl_ptr).addrl = (*var_ptr).var_addrl;
        /* ...and addressing mode   */
        (*var_symtbl_ptr).typel = (*var_ptr).var_typel;
        /* ...and variable data type    */
        (*var_symtbl_ptr).sizel = (*var_ptr).var_sizel;
        /* ...and variable size     */
        (*var_symtbl_ptr).offl = (*var_ptr).var_offl;
        /* ...and variable options  */
        (*var_symtbl_ptr).optl = (*var_ptr).var_optl;
        /* ...and variable options  */
        (*var_symtbl_ptr).values_offl = (*var_ptr).var_values_offl;
        /* Value list start offset  */
        (*var_symtbl_ptr).values_endl = (*var_ptr).var_values_endl;
        /* Value list end offset    */
        (*var_symtbl_ptr).low_rangel = (*var_ptr).var_low_rangel;
        /* Low range value      */
        (*var_symtbl_ptr).high_rangel = (*var_ptr).var_high_rangel;
        /* High range value     */
        (*var_symtbl_ptr).values_cntl = (*var_ptr).var_values_cntl;
        /* Value list count */

        /*
         *    ...and insert into the LL of variables.
         */
        iret = PSL_VM_INSERT ((*com_ptr).variables_ctxps
                              ,(char *) var_symtbl_ptr
                              ,NULL
                              ,NULL
                              ,PSL_VM_POS_END);
        if (!psl_test_status (iret))    /* If error allocating list */
        {
            PSL_PUTDIAG ((*com_ptr).diag_ptrps
                         ,2
                         ,PSL_SYSERVERR
                         ,"PSL_VM_INSERT"
                         ,iret);
            PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        }
    }

    (*tpa_ptr).var_cntl += 1;       /* ...increment count of variables */

    return;             /* ...and return        */
}

/************************************************************************/
/*                                  */
/*    BLANKS_OFF - Disable blanks recognition               */
/*    ==========                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   PSLTPA$L_OPTIONS(AP) = Tparse options mask      */
/*    Outputs:  PSLTPA$L_OPTIONS(AP) = Blank recognition switched off   */
/*                                  */
/*    This routine clears the 'blank recognition' bit in the TPARSE */
/*    argument block to switch off blank recognition.           */
/*                                  */
/************************************************************************/

unsigned long   BLANKS_OFF (struct tpaarg *argblk_ptr)

{
    (*argblk_ptr).optionsw = ((*argblk_ptr).optionsw & (~TPAARG_BLANKSM));
    /* Switch blank recognition off */
    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    BLANKS_ON - Enable blanks recognition             */
/*    ==========                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   PSLTPA$L_OPTIONS(AP) = Tparse options mask      */
/*    Outputs:  PSLTPA$L_OPTIONS(AP) = Blank recognition switched on    */
/*                                  */
/*    This routine sets the 'blank recognition' bit in the TPARSE   */
/*    argument block to switch off blank recognition.           */
/*                                  */
/************************************************************************/

unsigned long   BLANKS_ON (struct tpaarg *argblk_ptr)

{
    (*argblk_ptr).optionsw = ((*argblk_ptr).optionsw | TPAARG_BLANKSM);
    /* Switch blank recognition on  */
    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    CLR_DELIM_CHR - Clear the delimiter character.            */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   None                            */
/*    Outputs:  DELIMITER_CHRB cleared to show no delimiter.        */
/*                                  */
/*    This routine clears the delimiter character to show that      */
/*    there is no character string being parsed.            */
/*                                  */
/************************************************************************/

unsigned long   CLR_DELIM_CHR (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    (*tpa_ptr).delimiter_chrb = 0;  /* Clear delimiter char */
    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    CLR_VAR_PTR - Clear the variable pointer.                 */
/*    ===============                               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   None                            */
/*    Outputs:  VAR_PTR cleared to show no variable being parsed    */
/*                                  */
/*    This routine clears the variable pointer VAR_PTR to show that     */
/*    there is no variable being parsed.                */
/*                                  */
/************************************************************************/

unsigned long   CLR_VAR_PTR (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    (*tpa_ptr).var_ptrps = NULL;    /* Clear variable pointer   */
    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                              */
/*    DEC_COND_LEVEL - Decrement complex conditional expression level   */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - TPARSE argument block pointer.     */
/*      tpa_ptr->cond_levell - Current nesting level        */
/*    Outputs:  tpa_ptr->cond_levell - Decremented by one.      */
/*                                  */
/*    This routine is called when we encounter a closing parentheses    */
/*    when parsing a complex conditional expression. If we are already  */
/*    at the top level an error is reported, otherwise an 'end      */
/*    expression' delimiter at the end of the default 'TOF' line    */
/*    buffer and the conditional expression nesting level decremented.  */
/*                                  */
/************************************************************************/
unsigned long DEC_COND_LEVEL (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct scl_com  *com_ptr;       /* SCL common data */
    struct scl_tpa  *tpa_ptr;       /* SCL Tparse data */
    char    *lin_ptr;       /* Pointer to line buffer */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;
    lin_ptr = com_ptr->toflinc1024; /* 'tof' line buffer pointer */

    /*
     *  If we are already at the top level then report an invalid structure,
     *  otherwise encode an 'end expression' delimiter at the end of the default
     *  'TOF' line buffer and decrement the nesting level.
     */
    if (tpa_ptr->cond_levell <= 0)
    {
        REPORT_ERROR (argblk_ptr, SCL_INVSTRUCT);
    }
    else
    {
        (void) strcat (lin_ptr, TOF_OPR_EEXPR);
        tpa_ptr->cond_levell -= 1;  /* Decrement stack level */
    }

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    DEC_IFDEF_LEVEL - Decrement the IFDEF nesting level       */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - TPARSE argument block pointer.     */
/*      IFDEF_LEVELL - Current IFDEF nesting level indicator    */
/*    Outputs:  IFDEF_LEVELL - decremented to show one less level.      */
/*                                  */
/*    This routine is called when we encounter an ENDIF         */
/*    command to show that we have one less level of condtional code    */
/*                                  */
/************************************************************************/

unsigned long   DEC_IFDEF_LEVEL (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Simply decrement the nesting level.
     */
    if ((*tpa_ptr).ifdef_levell != 0)
    {
        (*tpa_ptr).ifdef_levell -= 1;
    }

    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                              */
/*    DEC_EXPR_LEVEL - Decrement EXPRESSION nesting level       */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - TPARSE argument block pointer.     */
/*      tpa_ptr->expr_levell - Current expression nesting level */
/*    Outputs:  tpa_ptr->expr_levell - Decremented to show one less */
/*                              level   */
/*                                  */
/*    This routine is called when we encounter a closing parentheses    */
/*    when parsing an expression. If we are already at the top level    */
/*    an error is reported, otherwise an 'end expression' operator  */
/*    is added to the parsed operand linked list and the expression */
/*    nesting level decremented.                    */
/*                                  */
/************************************************************************/
unsigned long DEC_EXPR_LEVEL (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct   scl_com *com_ptr;      /* SCL common data */
    struct   scl_tpa *tpa_ptr;      /* SCL Tparse data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;

    /*
     *  If we are already at the top level then report an invalid structure,
     *  otherwise add an 'end expression' operator to the parsed operand linked
     *  list and decrement the nesting level.
     */
    if (tpa_ptr->expr_levell <= 0)  /* If already at top level */
    {
        REPORT_ERROR (argblk_ptr, SCL_INVSTRUCT);
    }
    else
    {
        ADD_OPR (com_ptr, OPR_EEXPR);
        tpa_ptr->expr_levell -= 1;  /* Decrement expression level */
    }

    return PSL_NORMAL;          /* Return success */
}


/*======================================================================*/
/*                                  */
/*    DEFINE_CMND_CHR - Define the command character            */
/*    ================                          */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block                   */
/*    Outputs:  Command character modified              */
/*                                  */
/*    This routine is called after parsing the DEFINE COMMAND CHARACTER */
/*    command. Its purpose is to change the SCL Command character   */
/*                                  */
/*======================================================================*/

unsigned long   DEFINE_CMND_CHR (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    unsigned long    iret;          /* Status return variable   */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;

    /*
     *    Check to see if this is a valid command character.
     */
    iret = CHK_LGN_CMND_CHR (&(*argblk_ptr).charb);
    /* Check the character      */
    if (!psl_test_status (iret))    /* If not valid command character */
    {
        REPORT_ERROR (argblk_ptr, SCL_INVCMNDCHR);
        /* ...report invalid command char */
    }

    else                /* If command character valid   */
    {
        if ((*argblk_ptr).charb == (*com_ptr).ctrl_chr_valuec256 [0])
            /* ...if command = control  */
        {
            REPORT_ERROR (argblk_ptr, SCL_DUPCTRLCHR);
            /* ...report duplicate command char */
        }

        else                /* ...else if not the same  */
        {
            (*com_ptr).cmnd_chr_valuec256 [0] = (*argblk_ptr).charb;
            /* Set new value for command char */
        }
    }

    return PSL_NORMAL;          /* ...and return success    */
}

/*======================================================================*/
/*                                  */
/*    DEFINE_CTRL_CHR - Define the control character            */
/*    ================                          */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block                   */
/*    Outputs:  Control character modified              */
/*                                  */
/*    This routine is called after parsing the DEFINE CONTROL CHARACTER */
/*    command. Its purpose is to change the SCL Control character   */
/*                                  */
/*======================================================================*/

unsigned long   DEFINE_CTRL_CHR (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    unsigned long    iret;          /* Status return variable   */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;

    /*
     *    Check to see if this is a valid control character.
     */
    iret = CHK_LGN_CMND_CHR (&(*argblk_ptr).charb);
    /* Check the character      */
    if (!psl_test_status (iret))    /* If not valid control character */
    {
        REPORT_ERROR (argblk_ptr, SCL_INVCMNDCHR);
        /* ...report invalid control char */
    }

    else                /* If control character valid   */
    {
        if ((*argblk_ptr).charb == (*com_ptr).cmnd_chr_valuec256 [0])
            /* ...if control = command  */
        {
            REPORT_ERROR (argblk_ptr, SCL_DUPCMNDCHR);
            /* ...report duplicate command char */
        }

        else                /* ...else if not the same  */
        {
            (*com_ptr).ctrl_chr_valuec256 [0] = (*argblk_ptr).charb;
            /* Set new value for control char */
        }
    }

    return PSL_NORMAL;          /* ...and return success    */
}

/*======================================================================*/
/*                              */
/*    DEFINE_UNDEF_TST  - Define an undefined test          */
/*    ================                          */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block                   */
/*    Outputs:  Test added to Tests symbol table.           */
/*                                  */
/*    This routine is entered on encountering a reference to a test */
/*    which has not been defined. The purpose of this routine is to     */
/*    define a null test at the end of the Tests Symbol Table       */
/*    in order that the undefined variable is only reported once.   */
/*                                  */
/*    This routine will also perform the underlining and the reporting  */
/*    of the error.                         */
/*                                  */
/*    Because we have actually parsed the test and found that it    */
/*    does not exist we should add an operand to the linked list of     */
/*    saved operands so that it appears that we have parsed a valid     */
/*    test reference.                                   */
/*                                  */
/*======================================================================*/
unsigned long DEFINE_UNDEF_TST (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Set up fields required by ADD_TST to set up 'dummy' test.
     */
    (*tpa_ptr).def_nlengthl = (*argblk_ptr).tokencntl;
    (void) memcpy ((*tpa_ptr).def_namec64,
                   (*argblk_ptr).token_ptrpc,
                   (*argblk_ptr).tokencntl);

    /*
     *    ...now add this test to the symbol table.
     */
    ADD_TST (argblk_ptr);       /* Add test definition */

    /*
     *    ...and add an operand to the linked list of parsed operands as if we
     *    had parsed an operand sucessfully.
     */
    (void) SAVE_NULL_OPD (argblk_ptr);

    /*
     *    Now call the underline routine to underline the variable
     */
    (void) ERROR_TOKEN (argblk_ptr);    /* Report Error */

    return PSL_NORMAL;          /* ...and return success */
}


/*======================================================================*/
/*                              */
/*    DEFINE_UNDEF_TSTTHR - Define an undefined thread          */
/*    ===================                       */
/*                                  */
/*    Type: Action routine.                     */
/*    Status:   Always succeeds (PSL_NORMAL).               */
/*    Pointers: Unchanged.                      */
/*    Inputs:   Tparse argument block.                  */
/*      argblk_ptr.paraml = Error to report.            */
/*    Outputs:  Thread added to Threads symbol table.           */
/*                                  */
/*    This routine is entered on encountering a reference to a thread   */
/*    which has not been declared. The purpose of this routine is to    */
/*    add a 'null'entry for the thread to the Threads Symbol Table  */
/*    in order that the undeclared thread is only reported once. It */
/*    also reports the 'undeclared thread' error.           */
/*                                  */
/*    Because we have actually parsed the thread and found that it      */
/*    does not exist we should add an operand to the linked list of     */
/*    saved operands so that it appears that we have parsed a valid     */
/*    thread reference.                                 */
/*                                  */
/*======================================================================*/
unsigned long DEFINE_UNDEF_TSTTHR (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Set up fields required by ADD_TSTTHR to set up 'dummy' test thread.
     */
    (void) memcpy ((*tpa_ptr).def_namec64,
                   (*argblk_ptr).token_ptrpc,
                   (*argblk_ptr).tokencntl);
    (*tpa_ptr).def_nlengthl = (*argblk_ptr).tokencntl;

    /*
     *    ...now add this test thread to the symbol table.
     */
    ADD_TSTTHR (argblk_ptr);        /* Add test thread declaration */

    /*
     *    ...and add an operand to the linked list of parsed operands as if we
     *    had parsed an operand sucessfully.
     */
    (void) SAVE_NULL_OPD (argblk_ptr);

    /*
     *    Now report the error passed in the TPARSE argument block.
     */
    (void) ERROR_TOKEN (argblk_ptr);    /* Report Error */

    return PSL_NORMAL;          /* ...and return success */
}


/*======================================================================*/
/*                                  */
/*    DEFINE_UNDEF_VAR  - Define an undefined variable          */
/*    ================                          */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block                   */
/*    Outputs:  DEF_ADDRB - Set to null type                */
/*          DEF_SIZEL - Cleared                 */
/*          DEF_OPTL  - Cleared                 */
/*          DEF_DTYPL - Cleared                 */
/*          Variable added to variables symbol table.       */
/*                                  */
/*    This routine is entered on encountering a reference to a variable */
/*    which has not been defined. The purpose of this routine is to     */
/*    define a null variable at the end of the variables table with a   */
/*    null datatype in order that the undefined variable is only    */
/*    reported once.                            */
/*                                  */
/*    This routine will also perform the underlining and the reporting  */
/*    of the error.                         */
/*                                  */
/*    Because we have actually parsed the variable and found that it    */
/*    does not exist we should add an operand to the linked list of     */
/*    saved operands so that it appears that we have parsed a valid     */
/*    variable reference.                                   */
/*                                  */
/*======================================================================*/

unsigned long   DEFINE_UNDEF_VAR (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Default the DEF_* variables to create a undefined variable.
     */

    if (((*tpa_ptr).def_recidxl != 0))      /* If in record scope = record scope*/

    {
        (*tpa_ptr).def_addrl = (*tpa_ptr).rec_addrl;
    }
    else
    {
        (*tpa_ptr).def_addrl = ADDR_NONE;  /* Set NONE addressing mode.    */
    }

    (*tpa_ptr).def_sizel = 0;       /* Set zero size        */
    (*tpa_ptr).def_optl = 0;        /* No options           */
    (*tpa_ptr).def_dtypl = 0;       /* Null data type       */
    (*tpa_ptr).def_nlengthl = (*argblk_ptr).tokencntl;
    /* Data name length     */

    (void) memcpy ((*tpa_ptr).def_namec64,
                   /* Copy name from dataname storage */
                   (*argblk_ptr).token_ptrpc,
                   (*argblk_ptr).tokencntl);

    /*
     *    ...now add this variable to the symbol table
     */
    ADD_VARIABLE (argblk_ptr);      /* Add variable definition  */

    /*
     *    ...and add an operand to the linked list of parsed operands as if we
     *    had parsed an operand sucessfully.
     */
    (void) SAVE_NULL_OPD (argblk_ptr);

    /*
     *    Now call the underline routine to underline the variable
     */
    (void) ERROR_TOKEN (argblk_ptr);    /* Report Error */

    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/
/*                                  */
/*    END_CHAR_LIST_VAL - End Character List Value Processing       */
/*    =================                         */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  tpa-ptr->char_list_cnt += 1             */
/*                                  */
/*    This routine performs all the processing required after       */
/*    parsing a value in a character list.              */
/*                                  */
/************************************************************************/
unsigned long END_CHAR_LIST_VAL (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    /* Common data pointer */
    tpa_ptr = com_ptr->tpa_ptrps;   /* TPARSE data pointer */

    /*
     *    Countthe last value in the character list.
     */
    tpa_ptr->char_list_val_cntl += 1;

    return PSL_NORMAL;          /* Return success */
}

/************************************************************************/
/*                                  */
/*    END_PRM_HEADER - End Command Parameter HEADER processing      */
/*    ==============                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Tparse argument block.                  */
/*    Outputs:  scl_tpa.this_cmnd_optl = zero                   */
/*                                  */
/*    This routine performs all processing required after parsing   */
/*    the command parameter HEADER - available on HTTP request      */
/*    commands.                             */
/*                                  */
/************************************************************************/
unsigned long END_PRM_HEADER (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer */

    /*
     *  Clear the command option ID.
     */
    (*tpa_ptr).this_cmnd_optl = 0;
    return PSL_NORMAL;
}

/************************************************************************/
/*                                  */
/*    END_SCRIPT_PROC - End of Script Processing            */
/*    ===============                           */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   None                            */
/*    Outputs:  None                            */
/*                                  */
/*    This routine performs the end of script processing. It is called  */
/*    at the end of each section of a script (ie before and after   */
/*    each subroutine).                         */
/*                                  */
/*    It checks for the following errors:               */
/*                                  */
/*    1. Incomplete structured IF and DO commands.          */
/*    2. References to undefined labels.                */
/*    3. Missing script description.                    */
/*    4. Missing END SWITCH commands.                   */
/*                                  */
/************************************************************************/
unsigned long END_SCRIPT_PROC (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   labtbl  *lab_ptr;      /* Labels table pointer     */
    unsigned long    isub;          /* Loop count variable      */
    char    wrkbuf [20];       /* Work Buffer */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */

    /*
     *    Check for and report any incomplete IF or DO structures.
     */
    if ((*tpa_ptr).ifstack_levell != 0) /* If IF count not 0        */
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps,
                     1,
                     SCL_MISENDIF,
                     (*tpa_ptr).ifstack_levell);
        /* Report incomplete IF's   */
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        /* Output diagnostic */
        (*tpa_ptr).error_countl += 1;   /* Increment error count    */
    }

    if ((*tpa_ptr).do_cntl != 0)    /* If DO count not 0        */
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps,
                     1,
                     SCL_MISENDDO,
                     (*tpa_ptr).do_cntl);
        /* Report incomplete DO's   */
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        /* Output diagnostic */
        (*tpa_ptr).error_countl += 1;   /* Increment error count    */
    }

    /*
     * Missing END SWITCH check
     */
    if ((*tpa_ptr).sw_depthl != 0)   /* If SWITCH count not 0    */
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps,
                     1,
                     SCL_MISENDSW,
                     (*tpa_ptr).sw_depthl);
        /* Report incomplete switches's */
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        /* Output diagnostic */
        (*tpa_ptr).error_countl += 1;   /* Increment error count    */
    }

    /*
     *  Check for a missing END TEST-CASE command.
     */
    if (((*tpa_ptr).testcase_flagsl & TCAS_ACTIVE_TC) != 0)
        /* If Test-case is active   */
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 1, SCL_MISENDTC);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        /* Output diagnostic */
        (*tpa_ptr).error_countl += 1;   /* Increment error count    */
    }


    /*
     *    Scan the Labels Symbol Table for undefined labels.
     */
    for (isub = 0, lab_ptr = (*tpa_ptr).label_begps;
            isub < (*tpa_ptr).lab_cntl;
            isub++, lab_ptr++)     /* Loop through the labels table */
    {
        if (!(*lab_ptr).lab_defn)   /* If unresolved references */
        {
            (void) memset (wrkbuf, '\0', sizeof (wrkbuf));
            (void) memcpy (wrkbuf,
                           (*lab_ptr).lab_namec16,
                           (*lab_ptr).lab_nlengthl);
            PSL_PUTDIAG ((*com_ptr).diag_ptrps,
                         1,
                         SCL_UNDEFLABEL,
                         wrkbuf);
            /* Set up diagnostic */
            PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
            /* Output diagnostic */
            (*tpa_ptr).error_countl += 1;
            /* Increment error count    */
        }
    }

    return PSL_NORMAL;
}


/************************************************************************/
/*                                  */
/*    END_SCRIPT_PROC2 - End of Source file Processing          */
/*    ================                          */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   None                            */
/*    Outputs:  None                            */
/*                                  */
/*    This routine completes the end of script processing. It is only   */
/*    ever called when the end of the source file has been encountered  */
/*    and is not called at the end of a module ie subroutine of main.   */
/*                                  */
/*    It checks for the following errors:               */
/*                                  */
/*    1. Missing script description.                    */
/*                                  */
/*    2. Undefined references to subroutines.               */
/*                                  */
/*    3. Unterminated Subroutines                   */
/*                                  */
/*    4. Undefined references to external files             */
/*                                  */
/*    5. Unterminated conditional compilation commands          */
/*                                  */
/*    6. Pipes that are accessed but neither opened nor passed as prm   */
/*                                  */
/*    If the error count is zero at the end of this routine then the    */
/*    total length of the script (minus the header) will be written to  */
/*    the header record and the script object file will be written. */
/*                                  */
/*    If there are no errors a summary of the variable used is printed  */
/*    at the end of the .PRT file.                  */
/*                                  */
/************************************************************************/

unsigned long   END_SCRIPT_PROC2 (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   subtbl  *sub_ptr;      /* Subroutines table pointer    */
    struct   fattbl  *fat_ptr;      /* File access table pointer    */
    struct   tsttbl  *tst_ptr;      /* Test table pointer */
    struct   tstthrtbl
                *tstthr_ptr;       /* Test thread table pointer */
    unsigned long    isub;          /* Loop count variable  */
    struct   scl_cmd *scl_cmd_ptr;      /* Pointer to Command Line Struct */

    char    itmnam [256];      /* Item name */
    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;   /* Common data pointer      */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* TPARSE data pointer      */
    scl_cmd_ptr = (*com_ptr).cmd_ptrps;

    /*
     *    Check that the mandatory script description command has been
     *    specified in the environment section.
     */
    if (((*tpa_ptr).env_opt_specl & ENV_DESC) == 0)
{
        PSL_PUTDIAG ((*com_ptr).diag_ptrps,
                     1,
                     SCL_MISSCRDSC);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;   /* Increment error count    */
    }

    /*
     *    Scan the Subroutines Symbol Table for undefined subroutines.
     */
    for (isub = 0, sub_ptr = (*tpa_ptr).subroutine_begps;
            isub < (*tpa_ptr).subrtn_cntl;
            isub++, sub_ptr++)     /* Loop through subroutines table */
    {
        if (!(*sub_ptr).sub_defn)   /* If unresolved references */
        {
            PSL_PUTDIAG ((*com_ptr).diag_ptrps,
                         1,
                         SCL_UNDEFSUB,
                         (*sub_ptr).sub_namec20);
            /* Set up diagnostic */
            PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
            /* Output diagnostic */
            (*tpa_ptr).error_countl += 1;
            /* Increment error count    */
        }
    }

    /*
     *    Check to see if we still have an 'open' subroutine ie have encounterd
     *    a SUBROUTINE clause but not a corresponding END SUBROUTINE'
     */
    if (((*tpa_ptr).subroutine_flagsl & SUB_START) != 0)
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps,
                     1,
                     SCL_MISENDSUB);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;   /* Increment error count    */
    }
    (*tpa_ptr).subroutine_flagsl = 0;   /* ...and clear subroutine flags */

    /*
     *    Check to see if we have any unterminated records
     */
    if ((*tpa_ptr).def_recidxl != 0)    /* If records still open    */
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps,
                     1,
                     SCL_MISENDREC);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;   /* Increment error count    */
    }

    /*
     *    Scan the External File Symbol Table for undefined (ie unopened) files.
     */
    for (isub = 0, fat_ptr = (*tpa_ptr).fat_begps;
            isub < (*tpa_ptr).fat_cntl;
            isub++, fat_ptr++)     /* Loop through File Access table */
    {
        if ((*fat_ptr).fat_refflgl == ACCESSED_FILE)
        {               /* If accessed but not opened   */
            PSL_PUTDIAG ((*com_ptr).diag_ptrps,
                         1,
                         SCL_UNDEFFIL,
                         (*fat_ptr).fat_namec20);
            /* Set up diagnostic */
            PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
            /* Output diagnostic */
            (*tpa_ptr).error_countl += 1; /* Increment error count  */
        }
    }

    /*
     *    Scan the Named Pipe Symbol Table for pipes that were accessed but
     *    were neither opened by the script nor passed as a parameter to the
     *    script.
     */
    for (isub = 0, fat_ptr = (*tpa_ptr).pipe_begps;
            isub < (*tpa_ptr).pipe_cntl;
            isub++, fat_ptr++)     /* Loop through Pipe Access table */
    {
        if ((((*fat_ptr).fat_refflgl & PIPATT_ACC) != 0)
                /* If accessed          */
                && (((*fat_ptr).fat_refflgl & (PIPATT_OPEN | PIPATT_PRM)) == 0))
        {               /* but neither opened nor passed */
            PSL_PUTDIAG ((*com_ptr).diag_ptrps,
                         1,
                         SCL_PIPENOTOPN,
                         (*fat_ptr).fat_namec20);
            /* Set up diagnostic */
            PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
            /* Output diagnostic */
            (*tpa_ptr).error_countl += 1; /* Increment error count  */
        }
    }


    /*
     *    Check to see if we still have an unterminated conditional compilation
     *    section (ie an IFDEF without an ENDIF)
     */
    if ((*tpa_ptr).ifdef_levell != 0)
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps ,1 ,PSL_MISCONDEND
                     ,(*tpa_ptr).ifdef_levell
                     ,(*com_ptr).scl_namec256);
        /* Report incomplete #IFDEF */
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        (*tpa_ptr).error_countl += 1;   /* Increment error count    */
    }

    /*
     *  Scan the test symbol table reporting a warning for any unreferenced tests and
     *  an error if there is a reference within a WAIT FOR TEST, but not in an EXECUTE.
     */
    for (tst_ptr = tpa_ptr->tst_begps,
            isub = 1;
            isub <= tpa_ptr->tst_cntl;
            tst_ptr++, isub++)
    {
        if (!(tst_ptr->tst_exe_refn))   /* If test/sub-test unreferenced by EXECUTE command */
        {
            if (tst_ptr->tst_subtstn)
            {
                (void) strcpy (itmnam, "sub-test");
            }
            else
            {
                (void) strcpy (itmnam, "test");
            }

            if (!(tst_ptr->tst_wait_refn))  /* If not referenced at all */
            {
                PSL_PUTDIAG (com_ptr->diag_ptrps, 1, PSL_NOREF, tst_ptr->tst_namec68, itmnam);
                PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
            }
            else                /* Else only referenced in a WAIT FOR TEST */
            {
                PSL_PUTDIAG (com_ptr->diag_ptrps, 1, SCL_NOEXEREF, tst_ptr->tst_namec68,
                             itmnam);
                PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
                (*tpa_ptr).error_countl += 1;   /* Increment error count */
            }
        }
    }

    /*
     *  Scan the test threads symbol table reporting a warning for any unreferenced threads.
     */
    for (tstthr_ptr = tpa_ptr->tstthr_begps,
            isub = 1;
            isub <= tpa_ptr->tstthr_cntl;
            tstthr_ptr++, isub++)
    {
        if (!(tstthr_ptr->tstthr_exe_refn))   /* If test thread unreferenced by EXECUTE */
        {
            PSL_PUTDIAG (com_ptr->diag_ptrps, 1, PSL_NOREF, tstthr_ptr->tstthr_namec68,
                         "thread");
            PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        }
    }

    /*
     *  If there were no compilation errors, and we are listing, list out
     *  the variable information.
     */
    if (((*tpa_ptr).error_countl == 0)  /* If no compilation errors */
            &&  ((*scl_cmd_ptr).listn))     /* ...and listing enabled */
    {
        LIST_VARIABLES (com_ptr);   /* List the variables used */
        LIST_TIMERS (com_ptr);      /* List the timers used */
    }

    return PSL_NORMAL;          /* ...and return success    */
}

/*======================================================================*/
/*                              */
/*    FLG_EXECUTE_THR - Flag test thread referenced in EXECUTE command  */
/*    ===============                           */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to TPARSE argument block       */
/*    Outputs:  None                            */
/*                                  */
/*    This routine sets the 'referencd in an EXECUTE command' flag  */
/*    for a test thread. The test thread operand is assumed to be held  */
/*    in the first operand on the parsed operand linked list.       */
/*                                  */
/*----------------------------------------------------------------------*/
unsigned long FLG_EXECUTE_THR (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data.
     */
    struct   scl_com *com_ptr;      /* Common Data */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data */
    struct   oph     *oph_ptr;      /* Operand header pointer */
    struct   opd     *opd_ptr;      /* Operand descriptor pointer */
    struct   tstthrtbl
                *tstthr_ptr;       /* Test thread table entry pointer */
    unsigned long    iret;          /* Status return */
    char    *llb_ctx;      /* Linked list context */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *  Copy the test operand from the parsed operand stack.
     */
    iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps ,(char *) &oph_ptr
                        ,NULL ,&llb_ctx ,PSL_VM_MODE_FIRST);
    /* Remove the OPH pointer first */
    if  (!(psl_test_status (iret))) /* ...report any errors */
{
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_COPY OPH (in FLG_EXECUTE_THR)", iret);
    }

    iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps ,(char *) &opd_ptr
                        ,llb_ctx ,NULL ,PSL_VM_MODE_NEXT);
    /* ...then remove the OPD ptr */
    if  (!(psl_test_status (iret))) /* ...report any errors */
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_COPY OPD (in FLG_EXECUTE_THR)", iret);
    }

    /*
     *  Flag test as referenced in an EXECUTE command.
     */
    tstthr_ptr = (*opd_ptr).uopd.tstthr.tstthrtbl_ptrps;
    tstthr_ptr->tstthr_exe_refn = PTRUE;

    return PSL_NORMAL;
}


/*======================================================================*/
/*                              */
/*    FLG_EXECUTE_TST - Flag test referenced in EXECUTE command     */
/*    ===============                           */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to TPARSE argument block       */
/*    Outputs:  None                            */
/*                                  */
/*    This routine sets the 'referencd in an EXECUTE command' flag  */
/*    for a test. The test operand is assumed to be held in the     */
/*    first operand on the parsed operand linked list.          */
/*                                  */
/*----------------------------------------------------------------------*/
unsigned long FLG_EXECUTE_TST (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data.
     */
    struct   scl_com *com_ptr;      /* Common Data */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data */
    struct   oph     *oph_ptr;      /* Operand header pointer */
    struct   opd     *opd_ptr;      /* Operand descriptor pointer */
    struct   tsttbl  *tst_ptr;      /* Test table entry pointer */
    unsigned long    iret;          /* Status return */
    char    *llb_ctx;      /* Linked list context */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *  Copy the test operand from the parsed operand stack.
     */
    iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps ,(char *) &oph_ptr
                        ,NULL ,&llb_ctx ,PSL_VM_MODE_FIRST);
    /* Remove the OPH pointer first */
    if  (!(psl_test_status (iret))) /* ...report any errors */
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_COPY OPH (in FLG_EXECUTE_TST)", iret);
    }

    iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps ,(char *) &opd_ptr
                        ,llb_ctx ,NULL ,PSL_VM_MODE_NEXT);
    /* ...then remove the OPD ptr */
    if  (!(psl_test_status (iret))) /* ...report any errors */
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_COPY OPD (in FLG_EXECUTE_TST)", iret);
    }

    /*
     *  Flag test as referenced in an EXECUTE command.
     */
    tst_ptr = (*opd_ptr).uopd.tst.tsttbl_ptrps;
    tst_ptr->tst_exe_refn = PTRUE;

    return PSL_NORMAL;
}


/*======================================================================*/
/*                              */
/*    FLG_WAIT_TST - Flag test referenced in WAIT command       */
/*    ============                          */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to TPARSE argument block       */
/*    Outputs:  None                            */
/*                                  */
/*    This routine sets the 'referencd in a WAIT FOR TEST command'  */
/*    flag for a test. The test operand is assumed to be held in the    */
/*    first operand on the parsed operand linked list.          */
/*                                  */
/*----------------------------------------------------------------------*/
unsigned long FLG_WAIT_TST (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data.
     */
    struct   scl_com *com_ptr;      /* Common Data */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data */
    struct   oph     *oph_ptr;      /* Operand header pointer */
    struct   opd     *opd_ptr;      /* Operand descriptor pointer */
    struct   tsttbl  *tst_ptr;      /* Test table entry pointer */
    unsigned long    iret;          /* Status return */
    char    *llb_ctx;      /* Linked list context */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *  Copy the test operand from the parsed operand stack.
     */
    iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps ,(char *) &oph_ptr
                        ,NULL ,&llb_ctx ,PSL_VM_MODE_FIRST);
    /* Remove the OPH pointer first */
    if  (!(psl_test_status (iret))) /* ...report any errors */
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_COPY OPH (in FLG_WAIT_TST)", iret);
    }

    iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps ,(char *) &opd_ptr
                        ,llb_ctx ,NULL ,PSL_VM_MODE_NEXT);
    /* ...then remove the OPD ptr */
    if  (!(psl_test_status (iret))) /* ...report any errors */
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_COPY OPD (in FLG_WAIT_TST)", iret);
    }

    /*
     *  Flag test as referenced in a WAIT command.
     */
    tst_ptr = (*opd_ptr).uopd.tst.tsttbl_ptrps;
    tst_ptr->tst_wait_refn = PTRUE;

    return PSL_NORMAL;
}


/*======================================================================*/
/*                              */
/*    FLG_WAIT_TSTS - Flag tests referenced in WAIT command     */
/*    =============                         */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to TPARSE argument block       */
/*    Outputs:  None                            */
/*                                  */
/*    This routine sets the 'referencd in a WAIT FOR TEST command'  */
/*    flag for all tests referenced by the current WAIT FOR TEST    */
/*    command. The test operands are assumed to be held in the      */
/*    parsed operand linked list.                   */
/*                                  */
/*----------------------------------------------------------------------*/
unsigned long FLG_WAIT_TSTS (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data.
     */
    struct   scl_com *com_ptr;      /* Common Data */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data */
    struct   oph     *oph_ptr;      /* Operand header pointer */
    struct   opd     *opd_ptr;      /* Operand descriptor pointer */
    struct   tsttbl  *tst_ptr;      /* Test table entry pointer */
    unsigned long    iret;          /* Status return */
    char    *llb_ctx;      /* Linked list context */
    char    *pos_ptr;      /* Linked list position */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *  Copy each of the operands from the parsed operand linked list and for
     *  each test operand, set its 'referenced in a WAIT FOR TEST command' flag.
     *
     *  We use an error status from PSL_VM_COPY to indicate 'end of list'.
     */
    for (iret = PSL_NORMAL, pos_ptr = NULL; psl_test_status (iret); pos_ptr = llb_ctx)
    {
        iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps ,(char *) &oph_ptr
                            ,pos_ptr ,&llb_ctx ,PSL_VM_MODE_NEXT);
        /* Remove the OPH pointer first */
        if (psl_test_status (iret))
        {
            pos_ptr = llb_ctx;
            iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps ,(char *) &opd_ptr
                                ,pos_ptr ,&llb_ctx ,PSL_VM_MODE_NEXT);
            /* ...then copy the OPD pointer */
            if (psl_test_status (iret))
            {
                if (oph_ptr->addrl == ADDR_TST)     /* If a test operand */
                {
                    tst_ptr = (*opd_ptr).uopd.tst.tsttbl_ptrps;
                    tst_ptr->tst_wait_refn = PTRUE;
                }
            }
        }
    }

    return PSL_NORMAL;
}

/************************************************************************/
/*                                                                      */
/*    INC_ARRAY_IDXCNT - Increment array index/subscript count          */
/*    ================                                                  */
/*																		*/
/*    Type:     Action routine                                          */
/*    Status:   Always PSL_NORMAL                                       */
/*    Pointers: Unchanged												*/
/*    Inputs:   argblk_ptr = Pointer to TPARSE Argument Block Structure */
/*    Outputs:  None													*/
/*																		*/
/*    This routine increments the array index/subscript count and		*/
/*    should be called after parsing an array index/subscript           */
/*    in an array reference.                                            */
/*																		*/
/************************************************************************/
unsigned long INC_ARRAY_IDXCNT (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct   scl_com  *com_ptr;     /* Common Data */
    struct   scl_tpa  *tpa_ptr;     /* TPARSE Data */
    struct   array_blk *array_blk_ptr;  /* Pointer to current array */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    array_blk_ptr = (*tpa_ptr).array_blk_ptrps;

    /*
     *    We have parsed an array index/subscript, so count it by updating
     *    the value in the array block.
     */
    (*array_blk_ptr).array_indicesl += 1;

    return PSL_NORMAL;
}

/************************************************************************/
/*                              */
/*    INC_COND_LEVEL - Increment complex conditional expression level   */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - TPARSE argument block pointer.     */
/*      tpa_ptr->cond_levell - Current nesting level        */
/*    Outputs:  tpa_ptr->cond_levell - Incremented by one.      */
/*                                  */
/*    This routine is called when we encounter an opening parentheses   */
/*    when parsing a complex conditional expression. A 'start       */
/*    expression' delimiter is encoded at the end of the default 'TOF'  */
/*    line buffer and the condition nesting level incremented.  There   */
/*    is no limit on the nesting level - the only reason for        */
/*    maintaining the count is to validate the expression syntax.   */
/*                                  */
/************************************************************************/
unsigned long INC_COND_LEVEL (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct scl_com  *com_ptr;       /* SCL common data */
    struct scl_tpa  *tpa_ptr;       /* SCL Tparse data */
    char    *lin_ptr;       /* Pointer to line buffer */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;
    lin_ptr = com_ptr->toflinc1024; /* 'tof' line buffer pointer */

    /*
     *  Encode a 'start expression' delimiter at the end of the default
     *  'TOF' line buffer and increment the nesting level.
     */
    (void) strcat (lin_ptr, TOF_OPR_SEXPR);
    (*tpa_ptr).cond_levell += 1;

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    INC_IFDEF_LEVEL - Increment the IFDEF nesting level       */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - TPARSE argument block pointer.     */
/*      IFDEF_LEVELL - Current IFDEF nesting level indicator    */
/*    Outputs:  IFDEF_LEVELL - incremented to show one more level.      */
/*                                  */
/*    This routine is called when we encounter an IFDEF or IFNDEF   */
/*    command to show that we have a further level of condtional code   */
/*                                  */
/************************************************************************/

unsigned long   INC_IFDEF_LEVEL (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Check to see if we have exceeded the maximum nesting level of IFDEF's
     *    and report an error if we have.
     */
    if (((*tpa_ptr).ifdef_levell + 1) == MAX_IFDEF)
    {
        REPORT_ERROR (argblk_ptr, SCL_MAXCONDNEST);
        /* ...then report invalid structure */
    }

    else
    {

        /*
         *    ...and increment the nesting level.
         */
        (*tpa_ptr).ifdef_levell += 1;
    }

    return PSL_NORMAL;          /* ...and return success    */
}


/************************************************************************/
/*                              */
/*    INC_EXPR_LEVEL - Increment EXPRESSION nesting level       */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - TPARSE argument block pointer.     */
/*      tpa_ptr->expr_levell - Current expression nesting level */
/*    Outputs:  tpa_ptr->expr_levell - Incremented to show one more level   */
/*                                  */
/*    This routine is called when we encounter an opening parentheses   */
/*    when parsing an expression. A 'start expression' operator is  */
/*    added to the parsed operand linked list and the expression    */
/*    nesting level incremented. There is no limit on the nesting   */
/*    level - the only reason for maintaining the count is to       */
/*    validate the expression syntax.                   */
/*                                  */
/************************************************************************/
unsigned long INC_EXPR_LEVEL (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data.
     */
    struct scl_com  *com_ptr;       /* SCL common data */
    struct scl_tpa  *tpa_ptr;       /* SCL Tparse data */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;

    /*
     *  Add a 'start expression' operator to the parsed operand linked list and
     *  increment the nesting level.
     */
    ADD_OPR (com_ptr, OPR_SEXPR);
    tpa_ptr->expr_levell += 1;

    return PSL_NORMAL;          /* Return success */
}


/*======================================================================*/
/*                                  */
/*    INCLUDE_SCRIPT - Include from a script file           */
/*    ===============                           */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to TPARSE argument block       */
/*    Outputs:  None                            */
/*                                  */
/*    This routine is called while parsing an INCLUDE command.      */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   INCLUDE_SCRIPT (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   oph     *oph_ptr;      /* Operand header pointer   */
    struct   opd     *opd_ptr;      /* Operand descriptor pointer   */
    unsigned long    iret;          /* Status return        */
    struct   scl_cmd *scl_cmd_ptr;      /* Pointer to Command Line Struct */

    char   filename[256];      /* Name of include file         */
    char   *file_ptr;      /* Pointer to filename          */
    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    scl_cmd_ptr = (*com_ptr).cmd_ptrps;
    file_ptr = filename;
    /*
     *    Check the current script nesting depth to ensure we do not exceed it
     */
    if (((*com_ptr).include_depthl + 1) >= MAX_INC_DEPTH)
    {                   /* If exceeded  include depth   */
        REPORT_ERROR (argblk_ptr, SCL_MAXINCNEST);
        /* Report it            */
        (*tpa_ptr).error_countl += 1;   /* ...increment error count */
        goto end;           /* ...and branch to return  */
    }

    /*
     *    Remove the operand defining the name of the include file from the
     *    parsed operand stack.
     */
    iret = PSL_VM_REMOVE ((*tpa_ptr).parsed_operand_llps ,(char *) &oph_ptr
                          ,NULL ,NULL ,PSL_VM_POS_BEG);
    /* Remove the OPH pointer first */
    if  (!(psl_test_status (iret))) /* ...report any errors     */
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_REMOVE (in INCLUDE_SCRIPT)", iret);
    }

    iret = PSL_VM_REMOVE ((*tpa_ptr).parsed_operand_llps ,(char *) &opd_ptr
                          ,NULL ,NULL ,PSL_VM_POS_BEG);
    /* ...then remove the OPD ptr   */
    if  (!(psl_test_status (iret))) /* ...report any errors     */
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_REMOVE (in INCLUDE_SCRIPT)", iret);
    }

    /*
     *    ...now open the include file remembering to increment the include depth
     *    if we succeed and initialize the source file number.
     */
    iret = SCL_OPEN_INCLUDE (
               com_ptr
               ,(*opd_ptr).uopd.imc.val_pc
               ,(*oph_ptr).sizel
               ,INCLUDE_START_UNIT + (*com_ptr).include_depthl);

    /* Open the include file on next unit */
    if (!(psl_test_status (iret)))  /* ...if an error opening file  */
    {
        (*tpa_ptr).error_countl += 1;   /* ...increment error count */
    }
    else                /* ...else file opened O.k. */
    {
        (*com_ptr).include_depthl += 1; /* ...increment nesting depth   */
        (*com_ptr).srclinl10 [(*com_ptr).include_depthl] = 0;
        /* initialize source line number */
        /*
         *  Add include file name to dependency list, if required.
         */
        ADD_DEP (com_ptr, (*com_ptr).exp_src_filc256, DEP_TYP_INCLUDE, NULL);

        (*com_ptr).srcfil_cntl += 1;    /* ...and count the extra source file */

        /*
         *    We store the current count of source files as the index of the file
         *    currently being compiled. This is used during the encoding of debug
         *    information when we need to know where to pick up the name of the file
         *    and are not interested in what nesting depth we are at.
         *
         *    We need to save the current value of this field so that it can be
         *    restored on return from the included script.
         */
        (*com_ptr).src_fil_indxl10 [(*com_ptr).include_depthl - 1] =
            (*tpa_ptr).src_fil_indxl;   /* Save current source file index */
        (*tpa_ptr).src_fil_indxl = (*com_ptr).srcfil_cntl - 1;
        /* ...and update to current index */
    }

    /* Free memory associated with the operand */
    FREE_OPH_OPD (oph_ptr, opd_ptr);

    /*
     *    END
     *    ===
     *
     *    ...and return (always success even if we have had problems with the
     *    include).
     */
end:
    ;
    return PSL_NORMAL;
}

/******************************************************************************
 * @func    This Tparse action routine resets the fields within the Tparse
 *          argument block to reference 'End of Stream'. It does NOT read a new
 *          source line.
 *
 * @rdesc   Tparse return status.
 * @parm    Pointer to Tparse argument block.
 */
unsigned long MOVE_TO_EOS (struct tpaarg *argblk_ptr)
{
    (*argblk_ptr).tokencntl =
        (*argblk_ptr).tokencntl + (*argblk_ptr).stringcntl;
    /* Expand token to end of line  */
    (*argblk_ptr).string_ptrpc =
        (*argblk_ptr).tokencntl + (*argblk_ptr).token_ptrpc;
    /* ...and expand string pointer */
    (*argblk_ptr).stringcntl = 0;
    return PSL_NORMAL;           /* ...and return success    */
}

/******************************************************************************
 * @func    This action routine is called on starting the parsing of a
 *          statement in the Definitions section, a command in the Code section
 *          or a label in the Code section. It resets the 'TOF' SCL source line
 *          number field. Each record written to a TOF file includes a 'srclin'
 *          parameter, taken from this field and containing the number of the
 *          first line of the corresponding SCL definition, command or label.
 *
 * @rdesc   Tparse return status.
 * @parm    Pointer to Tparse argument block.
 */
unsigned long RESET_TOF_SRCLIN (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;

    /*
     * .Reset 'Start of Command' Source Line Number (unless processing
     *  an INCLUDED file).
     */
    if ((*com_ptr).include_depthl == 0)
    {
        (*com_ptr).tof_srclinl = (*com_ptr).srclinl10[0];
    }

    return PSL_NORMAL;             /* ...and return success */
}

/************************************************************************/

/*    RESET_COMMAND_STATE - Reset command parsing flags for new command */
/*    ===================                           */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - TPARSE argument block pointer.     */
/*                                  */
/*    Outputs:  Command parsing linked list reset           */
/*      Command error flag cleared              */
/*                                  */
/*    This routine is called at the end of each command.            */
/*    The purpose of this routine is to reset all of the command    */
/*    specific data. This includes the operand parsing linked list  */
/*    and the command error flag.                   */
/*                                  */
/************************************************************************/

unsigned long   RESET_COMMAND_STATE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   oph     *oph_ptr;      /* Pointer to OPH       */
    struct   opd     *opd_ptr;      /* Pointer to OPD       */
    struct   array_blk *array_blk_ptr;  /* Array Block Pointer      */
    unsigned long    iret;          /* Status return variable   */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    First set the command erro flag to false.
     */
    (*tpa_ptr).error_in_commandn = PFALSE;

    /*
     *    ...then delete any remaining entries in the parsed operands linked list
     */
    do
    {
        iret = PSL_VM_REMOVE ((*tpa_ptr).parsed_operand_llps ,(char *) &oph_ptr
                              ,NULL ,NULL ,PSL_VM_POS_BEG);
        /* Remove the first OPH pointer */
        if (psl_test_status (iret)) /* If we found one      */
        {
            iret = PSL_VM_REMOVE ((*tpa_ptr).parsed_operand_llps
                                  ,(char *) &opd_ptr ,NULL ,NULL ,PSL_VM_POS_BEG);
            /* ...pickup associated OPD pointer */
            if (psl_test_status (iret)) /* ...if we found one       */
            {
                /* Free memory associated with the operand */
                FREE_OPH_OPD (oph_ptr, opd_ptr);
            }
        }
    }
    while (psl_test_status (iret));     /* Loop while not an error  */

    /*
     *    ...then delete any remaining entries in the parsed array linked list
     */
    do
    {
        iret = PSL_VM_REMOVE ((*tpa_ptr).parsed_array_llps,
                              (char *) &array_blk_ptr,
                              NULL ,NULL ,PSL_VM_POS_BEG);
        if (psl_test_status (iret))
        {
            FREE (array_blk_ptr);


#ifdef PSL_TRACING
            if (PSL_TEST_TRACE (PTRC_USR_USER5))
            {
                PSL_OUTPUT_TRACE ("FREE: array - '%p' - RESET\n",
                                  array_blk_ptr);
            }
#endif  /* PSL_TRACING */

        }
    }
    while (psl_test_status (iret));     /* Loop while not an error  */

    /*
     *    Show no command options specified and none current.
     */
    (*tpa_ptr).cmnd_optl = 0;
    (*tpa_ptr).cmnd_opt_specifiedl = 0;
    (*tpa_ptr).this_cmnd_optl = 0;

    /*
     *    ...and reset the array level to 0 to show back with the PRIMARY operands.
     */
    (*tpa_ptr).array_levell = 0;

    /*
     *  Reset expression nesting level.
     */
    (*tpa_ptr).expr_levell = 0;

    /*
     *    ...and return.
     */
    return PSL_NORMAL;          /* ...and return success    */
}


/************************************************************************/

/*    STEP_COND_DELIMITER - Step to the next conditional command delim. */
/*    ===================                           */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - TPARSE argument block pointer.     */
/*    Outputs:  TPARSE pointers updated to point to delimiter.          */
/*                                  */
/*    This routine is called when we find a section of conditional  */
/*    compilation code which we are going to IGNORE.            */
/*                                  */
/*    The purpose of this routine is to scan forward in the SCL file    */
/*    until we find a conditional compilation command delimiter or the  */
/*    end-of-file.                          */
/*                                  */
/*    The conditional command delimiters are #ELIF, #ELSE and #ENDIF    */
/*                                  */
/*    A slight complication is that we may nest the #IFDEF structures   */
/*    and so we should only terminate when we hit the delimiter at the  */
/*    same IFDEF nesting level as the one which we started at.      */
/*                                  */
/************************************************************************/
unsigned long STEP_COND_DELIMITER (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data */
    unsigned long    save_level;        /* Saved nesting level */
    char    *wrk_ptr;      /* Work pointer to char */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Save the current nesting level as this is the level at which we will
     *    require the delimiter.
     */
    save_level = (*tpa_ptr).ifdef_levell;

    /*
     *    GET_NEXT_LINE
     *    -------------
     *
     *    We branch here when we want to pick up another line to check.
     */
get_next_line:
    ;
    (void) SCL_GET_LINE (argblk_ptr);   /* Read in a new line */

    /*
     *    Check for end-of-file, returning if we have it.
     */
    (void) CHK_EOF (argblk_ptr);
    if ((*com_ptr).eof_flagl != 0) goto end;

    /*
     *    Look for a conditional command identifier ("#") at the beginning of the
     *    line (may only be preceded by horizontal tabs or spaces).
     */
    for (;(*argblk_ptr).stringcntl > 0;)
    {
        wrk_ptr = (*argblk_ptr).string_ptrpc;   /* Point to this character */
        (*argblk_ptr).string_ptrpc += 1;    /* Move to next character */
        (*argblk_ptr).stringcntl -= 1;      /* ... */
        (*argblk_ptr).token_ptrpc -= 1;     /* ... */
        (*argblk_ptr).tokencntl -= 1;       /* ... */
        if (*wrk_ptr == '#') goto got_cond_cmnd;
        if ((*wrk_ptr != ' ') && (*wrk_ptr != '\t')) goto get_next_line;
    }

    goto get_next_line;         /* Empty line, go get next line
     
    /*
     *    GOT_COND_CMND
     *    -------------
     *
     *    We have a conditional compilation command.
     *    Check to see if it is a terminator or another introducer and handle it
     *    accordingly.
     */
got_cond_cmnd:
    ;
    wrk_ptr += 1;           /* Advance work pointer to character after "#" */
    if (memcmp (wrk_ptr ,"IFDEF" ,5) == 0)
    {                   /* If another IFDEF command */
        (*tpa_ptr).ifdef_levell += 1;   /* ...increment nesting level   */
    }

    else if (memcmp (wrk_ptr ,"IFNDEF" ,5) == 0)
    {                   /* ...elseif another IFNDEF     */
        (*tpa_ptr).ifdef_levell += 1;   /* ...increment nesting level   */
    }

    else if (memcmp (wrk_ptr ,"ENDIF" ,5) == 0)
    {                   /* ...elseif an ENDIF       */
        if ((*tpa_ptr).ifdef_levell != save_level)
        {               /* ...but not at entry level    */
            (*tpa_ptr).ifdef_levell -= 1;
            /* ...decrement nesting level   */
        }
        else                /* ...but if at entry level...  */
        {
            goto end;           /* ...then branch to the end    */
        }
    }

    /*
     *    ELSE/ELIF commands are ignored unless we are at the entry level in which
     *    case we exit and allow the caller to deal with it.
     */
    else if ((memcmp (wrk_ptr ,"ELIF" ,4) == 0)
             ||       (memcmp (wrk_ptr ,"ELSE" ,4) == 0))
    {                   /* ...elseif an ELSE or ELIF    */
        if ((*tpa_ptr).ifdef_levell == save_level) goto end;
        /* ...exit if at entry level    */
    }

    /*
     *    All other commands beginning with the # character are invalid and must
     *    be rejected. We shall report them inline for nested levels or exit if
     *    at the top level for rejecting by the caller.
     */
    else                /* ...if any other it is invalid */
    {
        if ((*tpa_ptr).ifdef_levell != save_level)
        {               /* ...if not at entry level */
            REPORT_ERROR (argblk_ptr, SCL_INVCONDCMD);
            /* ...else report it        */
            goto get_next_line;     /* ...and branch for next line  */
        }
        else
        {
            goto end;
        }
    }

    goto get_next_line;

    /*
     *    END
     *    ---
     *
     *    At this point we have either reached a terminating conditional compilation
     *    command or the end of file so should now return.
     */
end:
    ;
    return PSL_NORMAL;          /* ...and return success    */
}


/*======================================================================*/
/*                                  */
/*    CHECK_FOR_DUPLICATES - Check for duplicate data definitions   */
/*    ====================                      */
/*                                  */
/*    Calling format:                           */
/*  duplicate = CHECK_FOR_DUPLICATES (argblk_ptr);          */
/*                                  */
/*    Where:                                */
/*  duplicate = Return Boolean, PTURE if a duplicate definition */
/*  argblk_ptr = Pointer to the TPARSE argument block       */
/*                                  */
/*    This routine is called prior to adding a data declaration to the  */
/*    appropriate SCL symbol table to determine if the symbol declared  */
/*    duplicates an already defined symbol.                 */
/*                                  */
/*    In most cases we simply check the symbol name as this must be     */
/*    unique. there is however an exception in that if we are allowing  */
/*    duplicate definitions and we detect that the symbol name is in    */
/*    use we need to thoroughly check the declaration and not treat it  */
/*    as an error if the declaration is IDENTICAL to the previous one.  */
/*                                  */
/*    Duplicate definitions are normally ONLY permitted in SQL included */
/*    data definitions.                         */
/*                                  */
/*----------------------------------------------------------------------*/

static Boolean  CHECK_FOR_DUPLICATES    (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   symtbl  *var_ptr;      /* Pointer to variables table   */
    struct   fattbl  *fattbl_ptr;       /* Pointer to Pipe symbol table */
    struct   thrtbl  *thrtbl_ptr;       /* Pointer to Thread Symbol Table */

    struct   timtbl  *tim_ptr;      /* Pointer to Timer Symbol Table */
    struct   tsttbl  *tsttbl_ptr;       /* Pointer to Test Symbol Table */

    struct   tstthrtbl
                *tstthrtbl_ptr;    /* Pointer to Test Thread Symbol Table*/

    unsigned long    isub;          /* Loop contro variable     */
    long    comp_result;       /* result of comparison     */
    Boolean    duplicate = PFALSE;    /* Duplicate flag       */
    Boolean    valid_duplicate = PTRUE;
    /* Duplicate is valid flag  */
    long    namlen;        /* Name length work variable    */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    We must check the different types of definition separately.
     */
    if ((*tpa_ptr).def_typl == DEF_TIMER)
{                   /* If we are defining a timer   */

        /*
         *    Scan the timers symbol table until we reach the end or find a match
         *    to this dataname.
         */
        tim_ptr = (*tpa_ptr).timer_begps;
        /* Timer symbol table       */
        for (isub=0; isub < (*tpa_ptr).tim_cntl; isub++, tim_ptr++)
        {
            if ((*tpa_ptr).def_nlengthl == strlen ((*tim_ptr).tim_namec68))
            {               /* If the lengths are the same  */
                comp_result = memcmp ((*tpa_ptr).def_namec64
                                      ,(*tim_ptr).tim_namec68
                                      ,(*tpa_ptr).def_nlengthl);
                /* Compare the 2 timers     */
                if (comp_result == 0) duplicate = PTRUE;
            }
        }
    }

    else if ((*tpa_ptr).def_typl == DEF_FILE)
    {                   /* ...else if a FILE definition */

        /*
         *    Scan the FILES symbol table until we reach the end or find a match
         *    to this dataname.
         */
        fattbl_ptr = (*tpa_ptr).fat_begps;
        /* FILES symbol table       */
        for (isub=0; isub < (*tpa_ptr).fat_cntl; isub++, fattbl_ptr++)
        {
            if ((*tpa_ptr).def_nlengthl == strlen ((*fattbl_ptr).fat_namec20))
            {               /* If the lengths are the same  */
                comp_result = memcmp ((*tpa_ptr).def_namec64
                                      ,(*fattbl_ptr).fat_namec20
                                      ,(*tpa_ptr).def_nlengthl);
                /* Compare the 2 Files      */
                if (comp_result == 0) duplicate = PTRUE;
            }
        }
    }

    else if ((*tpa_ptr).def_typl == DEF_PIPE)
    {                   /* ...else if a PIPE definition */

        /*
         *    Scan the PIPE symbol table until we reach the end or find a match
         *    to this dataname.
         */
        fattbl_ptr = (*tpa_ptr).pipe_begps;
        /* PIPE symbol table        */
        for (isub=0; isub < (*tpa_ptr).pipe_cntl; isub++, fattbl_ptr++)
        {
            if ((*tpa_ptr).def_nlengthl == strlen ((*fattbl_ptr).fat_namec20))
            {               /* If the lengths are the same  */
                comp_result = memcmp ((*tpa_ptr).def_namec64
                                      ,(*fattbl_ptr).fat_namec20
                                      ,(*tpa_ptr).def_nlengthl);
                /* Compare the 2 pipes      */
                if (comp_result == 0) duplicate = PTRUE;
            }
        }
    }

    else if ((*tpa_ptr).def_typl == DEF_TEST)
    {                   /* ...else if a TEST definition */

        /*
         *    Scan the TEST symbol table until we reach the end or find a name match.
         */
        tsttbl_ptr = (*tpa_ptr).tst_begps;
        /* TEST Symbol Table */
        for (isub=0; isub < (*tpa_ptr).tst_cntl; isub++, tsttbl_ptr++)
        {
            namlen = strlen ((*tsttbl_ptr).tst_namec68);

            if (namlen == (long)(*tpa_ptr).def_nlengthl)
            {               /* If the lengths are the same  */
                comp_result = memcmp ((*tpa_ptr).def_namec64
                                      ,(*tsttbl_ptr).tst_namec68
                                      ,namlen);
                /* Compare the 2 tests */
                if (comp_result == 0) duplicate = PTRUE;
            }
        }
    }

    else if ((*tpa_ptr).def_typl == DEF_TEST_THREAD)
    {                   /* ...else if TEST THREAD definition */

        /*
         *    Scan the TEST THREAD symbol table until we reach the end or find a
         *    name match.
         */
        tstthrtbl_ptr = (*tpa_ptr).tstthr_begps;
        /* TEST THREAD Symbol Table */
        for (isub=0; isub < (*tpa_ptr).tstthr_cntl; isub++, tstthrtbl_ptr++)
        {
            namlen = strlen ((*tstthrtbl_ptr).tstthr_namec68);

            if (namlen == (long)(*tpa_ptr).def_nlengthl)
            {               /* If the lengths are the same */
                comp_result = memcmp ((*tpa_ptr).def_namec64
                                      ,(*tstthrtbl_ptr).tstthr_namec68
                                      ,namlen);
                /* Compare the 2 test threads */
                if (comp_result == 0) duplicate = PTRUE;
            }
        }
    }

    else if ((*tpa_ptr).def_typl == DEF_THREAD)
    {                   /* ...else if a THREAD definition */

        /*
         *    Scan the THREAD symbol table until we reach the end or find a name match.
         */
        thrtbl_ptr = (*tpa_ptr).thr_begps;
        /* THREAD Symbol Table */
        for (isub=0; isub < (*tpa_ptr).thr_cntl; isub++, thrtbl_ptr++)
        {
            namlen = strlen ((*thrtbl_ptr).thr_namec20);

            if (namlen == (long)(*tpa_ptr).def_nlengthl)
            {               /* If the lengths are the same  */
                comp_result = memcmp ((*tpa_ptr).def_namec64
                                      ,(*thrtbl_ptr).thr_namec20
                                      ,namlen);
                /* Compare the 2 threads */
                if (comp_result == 0) duplicate = PTRUE;
            }
        }
    }

    else
    {                   /* ...else VARIABLE definition  */
        var_ptr = (*tpa_ptr).variable_begps;
        /* Variables symbol table   */
        for (isub=0; isub < (*tpa_ptr).var_cntl; isub++, var_ptr++)
        {
            if ((*tpa_ptr).def_nlengthl == (*var_ptr).var_nlengthl)
            {               /* If the lengths are the same  */
                comp_result = memcmp ((*tpa_ptr).def_namec64
                                      ,(*var_ptr).var_namec16
                                      ,(*var_ptr).var_nlengthl);
                /* Compare the 2 variables  */
                if (comp_result == 0)
                {
                    duplicate = PTRUE;
                    /* Name is duplicated       */

                    /*
                     *    Variables are more complicated then the other declarations here as there
                     *    is more to the definition. We should therefore check the rest of the
                     *    attributes of the variable already defined and that which we have just
                     *    defined and only if all of its attributes match should we consider it
                     *    to a valid duplicate.
                     */
                    if (((*var_ptr).var_addrl != (*tpa_ptr).def_addrl)
                            /* If different addressing mode */
                            ||  ((*var_ptr).var_typel != (*tpa_ptr).def_dtypl)
                            /* ...or data type      */
                            ||  ((*var_ptr).var_sizel != (*tpa_ptr).def_sizel)
                            /* ...or variable size      */
                            ||  ((*var_ptr).var_optl != (PSL_LONGWORD_T)(*tpa_ptr).def_optl)
                            /* ...or variable options   */
                            ||  ((*var_ptr).var_values_cntl!=(*tpa_ptr).def_values_cntl)
                            /* ...or Count of values in list */
                            ||  ((*var_ptr).var_low_rangel != (PSL_LONGWORD_T)(*tpa_ptr).def_low_rangel)
                            /* ...or low range value    */
                            ||  ((*var_ptr).var_high_rangel!=(PSL_LONGWORD_T)(*tpa_ptr).def_high_rangel)
                            /* ...or high range value   */
                            ||  ((*var_ptr).var_array_dimcntl!=(PSL_LONGWORD_T)(*tpa_ptr).array_dimcntl))
                        /* ...or differing array dimension */
                    {
                        valid_duplicate = PFALSE;
                    }
                }
            }
        }
    }

    /*
     *    If we have determined that we have an IDENTICAL DUPLICATE definition
     *    and we are not allowing duplicates then we must report an error before
     *    returning the duplicate flag to the caller.
     *
     *    We also report this error is the duplicate is not an exact duplicate
     *    (variables only) as in this case it is not valid.
     */
    if (duplicate)
    {
        if ((*tpa_ptr).ignore_duplicatesn && !valid_duplicate)
        {               /* If ignoring duplicates but this */
            /* is not a valid duplicate    */
            REPORT_ERROR (argblk_ptr, SCL_INVDUPDEF);
            /* ...then report conflicting variable*/
        }
        else if (!(*tpa_ptr).ignore_duplicatesn)
        {               /* ...else if not ignoring duplicates */
            REPORT_ERROR (argblk_ptr, SCL_DUPDEF);
            /* ...then report duplicate def */
        }
    }

    return duplicate;
}

/************************************************************************/

/*    ADD_FILE_SYMBOL - Add FILE definition to Variables symbol Table   */
/*    ===============                           */
/*                                  */
/*    Calling format:                           */
/*  ADD_FILE_SYMBOL (com_ptr ,fattbl_ptr);              */
/*                                  */
/*    Where:                                */
/*  com_ptr    = Pointer to the SCL common data structure       */
/*  fattbl_ptr = Pointer to the file access table entry for FILE    */
/*                                  */
/*    This routine is called to determine if the script is compiled     */
/*    with /DIAGNOSE or is an SQL mode script.              */
/*                                  */
/*    In these cases the FILE entry must be added to the external       */
/*    variables symbol table (that written for use by DEBUG/SQL and     */
/*    held in the SCP) in this situation a FILE is classed as a special */
/*    type of file.                         */
/*                                  */
/************************************************************************/

void    ADD_FILE_SYMBOL (struct scl_com *com_ptr
                         ,struct fattbl  *fattbl_ptr)

{
    /*
     *    Local data
     */
    struct   scl_tpa    *tpa_ptr;       /* TPARSE Data          */
    struct   scl_cmd    *scl_cmd_ptr;   /* Pointer to Command Line Struct */
    struct   scph       *header_ptr;    /* Pointer to header record     */
    struct   var_symtbl var_symtbl;     /* variables symbol table entry */
    unsigned long       iret;       /* Status return        */

    /*
     *    Copy pointers for ease of reference.
     */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* ...and TPARSE data pointer   */
    scl_cmd_ptr = (*com_ptr).cmd_ptrps; /* Pointer to command line data */
    header_ptr = (*tpa_ptr).header_begps;/* Pointer to header */

    /*
     *    If we are in SQL mode then we will need
     *    to maintain the external variables symbol table.
     *
     *    It is this table which we shall write to the SCP and which provides a
     *    lookup functionality in the Test Executer to match a File Ident with
     *    its index.
     *
     *    Add the FILE (null terminated) to the EXTERNAL VARIABLES symbol table.
     */
    if (((*header_ptr).scph_mode_b & MODE_IMPACT) != 0)
    {
        (void) strcpy (var_symtbl.symbol_namec20 ,(*fattbl_ptr).fat_namec20);
        /* ...copy name into symbol table */
        /*
         *    ...and add the other details about the symbol to the table.
         */
        var_symtbl.addrl = ADDR_FILE;   /* ...and addressing mode   */
        var_symtbl.typel = DTYP_FILE;   /* ...and variable data type    */
        var_symtbl.sizel = 0;       /* ...and variable size     */
        var_symtbl.offl = (*tpa_ptr).fat_cntl;
        /* ...and variable options  */

        /*
         *    ...and insert into the LL of variables.
         */
        iret = PSL_VM_INSERT ((*com_ptr).variables_ctxps
                              ,(char *) &var_symtbl
                              ,NULL
                              ,NULL
                              ,PSL_VM_POS_END);
        if (!psl_test_status (iret))    /* If error allocating list */
        {
            PSL_PUTDIAG ((*com_ptr).diag_ptrps
                         ,2
                         ,PSL_SYSERVERR
                         ,"PSL_VM_INSERT (ADD_FILE_SYMBOL)"
                         ,iret);
            PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        }
    }

    /*
     *    ...and return
     */
    return;
}
