/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*
 * scl_enc_subs_unix.c:
 * ====================
 *
 * This module contains UNIX specific SCL routines.
 */

/*
 *  Standard ANSI header files.
 */
#include <stdio.h>

/*
 *  PSL Macros.
 */
#include <psl_macros.h>
#include <vtest_macros.h>

/*
 *  Function Prototypes.
 */
#include <psl_rtndef.h>         /* PSL functions        */
#include "scl_rtndef.h"         /* SCL Routines         */

/*
 *  Structure definitions.
 */
#include <psl_tpaargdef.h>      /* TPARSE Argument Block    */
#include <psl_tpatabdef.h>      /* TPARSE Table         */
#include "scl_limits.h"         /* SCL limits           */
#include "scl_tbldef.h"         /* SCL Table Definitions    */
#include "scl_comdef.h"         /* Common Data          */
#include "scl_tpacom.h"         /* Tparse common Data       */
#include <scpdef.h>         /* Script Header Definitions*/

/*======================================================================*/
/*                                  */
/*    ENC_INTEGER - Encode an unsigned long value.          */
/*    ============                          */
/*                                  */
/*    Type: Encode routine                      */
/*    Status:   None                            */
/*    Pointers: Unchanged                       */
/*    Inputs:   value - long value to encode                */
/*      tpa_ptr - pointer to tparse data            */
/*    Outputs:  CODE_PTR = Object code pointer updated          */
/*                                  */
/*    This routine encodes an unsigned long into the object code buffer */
/*    and updates the code pointer to point past it.            */
/*                                  */
/*    This routine is PLATFORM specific as it depends on the ordering   */
/*    of the data.                          */
/*                                  */
/*    Integers in the range -2^31 - +2^31 may be encoded.       */
/*                                  */
/*----------------------------------------------------------------------*/

void    ENC_INTEGER (struct scl_tpa *tpa_ptr, long value)

{
    /*
     *    Local data
     */
    long  *int_ptr;         /* Work pointer to long     */
    void  *void_ptr;        /* Work pointer to void     */

    /*
     *    Simply copy the opcode into the script code buffer at the current
     *    point and update the code buffer offset.
     *
     *    Note that we have aligned the code pointer to a longword boundary and 
     *    may therefore convert the pointer to a pointer to long to enable us to
     *    copy the integer into the code buffer. We go via a VOID pointer to fool
     *    LINT which would otherwise report possible alignment problems.
     */
    void_ptr = (void *) (*tpa_ptr).code_ptrpc;
    /* Get void pointer to code */
    int_ptr = (long *) void_ptr;    /* ...to get long pointer   */
    *int_ptr = value;           /* Encode value at this offset  */
    (*tpa_ptr).code_ptrpc += sizeof (long);
    /* Update the code pointer  */
    return;             /* ...and return        */
}

/*======================================================================*/
/*                                  */
/*    ENC_INTEGER_AT - Encode integer value at the specified address    */
/*    ===============                           */
/*                                  */
/*    Type: Encode routine                      */
/*    Status:   None                            */
/*    Pointers: Unchanged                       */
/*    Inputs:   value - long value to encode                */
/*      addr_ptr - char pointer to address to save value at.    */
/*    Outputs:  None                            */
/*                                  */
/*    This routine encodes a long value into the object code buffer */
/*    at the address specified.                     */
/*                                  */
/*    This routine is PLATFORM specific as it depends on the ordering   */
/*    of the data.                          */
/*                                  */
/*    Integers in the range -2^31 - +2^31 may be encoded.       */
/*                                  */
/*----------------------------------------------------------------------*/

void    ENC_INTEGER_AT (char *addr_ptr, long value)

{
    /*
     *    Local data
     */
    long  *int_ptr;         /* Pointer to long      */
    void  *void_ptr;        /* Work pointer to void     */

    /*
     *    Simply copy the opcode into the script code buffer at the current
     *    point and update the code buffer offset.
     *
     *    Note that we have aligned the code pointer to a longword boundary and 
     *    may therefore convert the pointer to a pointer to long to enable us to
     *    copy the integer into the code buffer. We go via a VOID pointer to fool
     *    LINT which would otherwise report possible alignment problems.
     */
    void_ptr = (void *) addr_ptr;   /* Get work pointer to void */
    int_ptr = (long *) void_ptr;    /* Convert to pointer to long   */
    *int_ptr = value;           /* ...and copy in value.    */

    return;             /* ...and return        */
}

/*======================================================================*/
/*                                  */
/*    UNPACK_INTEGER - Decode integer value from the specified address  */
/*    ===============                           */
/*                                  */
/*    Type: Encode routine                      */
/*    Status:   None                            */
/*    Pointers: Unchanged                       */
/*    Inputs:   value - pointer to hold long decoded value      */
/*      addr_ptr - char pointer to address to decode at     */
/*    Outputs:  None                            */
/*                                  */
/*    This routine decodes a long value from the object code buffer */
/*    into the long value specified.                    */
/*                                  */
/*    This routine is PLATFORM specific as it depends on the ordering   */
/*    of the data.                          */
/*                                  */
/*    Integers in the range -2^31 - +2^31 may be decoded.       */
/*                                  */
/*----------------------------------------------------------------------*/

void    UNPACK_INTEGER (char *addr_ptr, long *val_ptr)

{
    /*
     *    Local data
     */
    long  *int_ptr;         /* Work pointer to long     */
    void  *void_ptr;        /* Work pointer to void     */

    /*
     *    Simply copy the opcode into the script code buffer at the current
     *    point and update the code buffer offset.
     *
     *    Note that we have aligned the code pointer to a longword boundary and 
     *    may therefore convert the pointer to a pointer to long to enable us to
     *    copy the integer into the code buffer. We go via a VOID pointer to fool
     *    LINT which would otherwise report possible alignment problems.
     */
    void_ptr = (void *) addr_ptr;   /* Get work poinetr to void */
    int_ptr = (long *) void_ptr;    /* Convert to pointer to long   */
    *val_ptr = *int_ptr;        /* ...and copy value from addr  */

    return;             /* ...and return        */
}
