/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// TParam.cpp
/**
 * Method defintitons of object to represent a parameter
 **/

#include "TParam.h"
#include "odl.h"
#include "cxxutils.h"

// Static Parameter name initialisations
const string TParam::stDelay		= "delay";
const string TParam::stConnid		= "conid";
const string TParam::stSessid		= "sesid";
const string TParam::stCmdid		= "cmdid";
const string TParam::stServer		= "server";
const string TParam::stUser		    = "user";
const string TParam::stPasswd		= "password";
const string TParam::stSql	    	= "sql";
const string TParam::stAutoFetch	= "autofetch";
const string TParam::stName		    = "name";
const string TParam::stValue		= "value";
const string TParam::stCount		= "count";
const string TParam::stCancel		= "cancel";
const string TParam::stMaxLong		= "maxlong";
const string TParam::stBatchSize	= "fbatch";
const string TParam::stIterations	= "iter";
const string TParam::stDefered		= "defer";
const string TParam::stAssign		= "asgn";
const string TParam::stFPrecision	= "fprecis";
const string TParam::stParseOnly	= "ponly";
const string TParam::stParse		= "prse";
const string TParam::stBinds		= "bnds";
const string TParam::stValues		= "vals";
const string TParam::stRow		    = "row";
const string TParam::stColumn		= "col";
const string TParam::stLogMsg		= "lmsg";
const string TParam::stCode		    = "code";
const string TParam::stAction		= "action";
const string TParam::stExpression	= "expr";
const string TParam::stDepth		= "depth";
const string TParam::stInitial		= "initial";
const string TParam::stEndValue		= "eValue";
const string TParam::stStep		    = "step";
const string TParam::stRetvar		= "retvar";
const string TParam::stOptions		= "options";
const string TParam::stDimensions	= "dimensions";
const string TParam::stIvalue		= "ivalue";
const string TParam::stLvalues		= "lValues";
const string TParam::stSvalues		= "sValues";
const string TParam::stSize		    = "size";
const string TParam::stFIdent		= "fident";
const string TParam::stFile		    = "file";
const string TParam::stDestination	= "destination";
const string TParam::stType		    = "type";
const string TParam::stStatusvar	= "statusvar";
const string TParam::stexpr		    = "expr";
const string TParam::stlabel		= "label";
const string TParam::stErrlab		= "errlab";
const string TParam::stTmolab		= "tmolab";
const string TParam::stGolab		= "golab";
const string TParam::stTarget		= "target";
const string TParam::stMode		    = "mode";
const string TParam::stFormat		= "format";
const string TParam::stParameters	= "parameters";
const string TParam::stTo		    = "to";
const string TParam::stString		= "string";

const string TParam::stFileId       = "filid";
const string TParam::stFileOperation= "operation";
const string TParam::stData         = "data";
const string TParam::stInto         = "into";
const string TParam::stRetCod       = "retcod";
const string TParam::stRetSts       = "retsts";
const string TParam::stTimer        = "timer";
const string TParam::stTimeOut      = "timeout";
const string TParam::stTDelay       = "tdelay";
const string TParam::stFrom         = "from";
const string TParam::stStatus       = "status";
const string TParam::stDescription  = "description";
const string TParam::stAccess       = "access";
const string TParam::stResult       = "result";
const string TParam::stScope        = "scope";
const string TParam::stRadix        = "radix";
const string TParam::stSrclin       = "srclin";
const string TParam::stInputArgType = "tpargument";
const string TParam::stDomain       = "domain";
const string TParam::stInputBlob    = "blob";
const string TParam::stOutputBlob   = "into";

/**
 * Constructor giving the parameter the rule that describes its
 * behaviour.
 **/
TParam::TParam(const TParamRule* prmRP)
        :   _prmRP(prmRP)
{}

/**
 * Destructor
 **/
TParam::~TParam()
{
    for(vector<TSegStr*>::const_iterator i = _segs.begin();
            i != _segs.end();
            i++) delete *i;
}

/**
 * Accessor for the atvf record rule that forms the basis of this
 * parameters rule.
 **/
SharedPtr<AtvfRecRule> TParam::innerRule() const
{
    return _prmRP->innerRule();
}

/**
 * Accessor for the param rule that describes this parameter
 **/
const TParamRule* TParam::rule() const
{
    return _prmRP;
}

/**
 * Accessor to find description of this parameter
 **/
string TParam::token() const
{
    return _prmRP->token();
}

/**
 * Accessor to find option associated with this parameter
 **/
const TOptions& TParam::options() const
{
    return _prmRP->options();
}

/**
 * Accessor to find configuration table associated with this parameter
 **/
const CConfigTable* TParam::config() const
{
    return _prmRP->config();
}

/**
 * Accessor to find if this parameter has a particular attribute.
 **/
bool TParam::hasAttr(const string& attr) const
{
    return _prmRP->hasAttr(attr);
}


/**
 * Accessor to get current value of this parameter as a string
 **/
string TParam::strVal(unsigned int idx) const
{
    if(_recs.size() > 0)
    {
        if(idx >= _recs.size()) idx = _recs.size() - 1;
        return _recs[idx]->value();
    }
    else
    { // no value assigned yet
        return _prmRP->defStrVal();
    }
}

/**
 * Accessor to get current value of this parameter as a string
 **/
int TParam::intVal(unsigned int idx) const
{
    if(_recs.size() > 0)
    {
        if(idx >= _recs.size()) idx = _recs.size() - 1;
        return _ints[idx];
    }
    else
    {
        return _prmRP->defIntVal();
    }
}

/**
 * Accessor to use this parameters rule to calculate what the
 * integer value of the supplied string is.
 **/
int TParam::intVal(const string& text) const
{
    return _prmRP->intVal(text);
}

/**
 * Accessor to find out if a text string is contained within one
 * of the values associated with this parameter.
 **/
bool TParam::findInVal(const string& text, bool matchCase) const
{
    if(hasAttr("nosearch")) return false;
    for(unsigned int idx = 0;
            idx < numVals();
            idx++)
    {
        if((strVal(idx).find(text) != string::npos) ||
                (!matchCase && caseIndValMatch(text, idx)))
        {
            return true;
        }
    }
    return false;
}



bool TParam::findInVal(CRegExp& regex, bool matchCase) const
{
    if(hasAttr(PATXT_NOSEARCH)) return false;
    for(unsigned int idx = 0; idx < numVals(); idx++)
    {
        if(find_in_string(strVal(idx), regex, matchCase) != string::npos)
        {
            return true;
        }
    }
    return false;
}


/**
 * Private accessor/helper method to perform case independent string
 * searching.
 **/
bool TParam::caseIndValMatch(const string& text, unsigned int idx) const
{
    static string textbit(" ");
    string str = strVal(idx);
    if(text.size() > str.size()) return false; // pint into half pint pot - No!

    char chr1 = text.at(0);
    if(chr1 != textbit.at(0))
    {
        textbit.resize(0);
        textbit.append(1, chr1);
        // This bit only works when ASCII
#ifndef __VAX
        if((chr1 < 91) && (chr1 > 64))
        { // If char is uppercase
            textbit.append(1, static_cast<char>(chr1+32)); // add lowercase copy,
        }

        else if((chr1 < 123) && (chr1 > 96))
        { // If char is lowercase
            textbit.append(1, static_cast<char>(chr1-32)); // add uppercase copy
        }

#else
        if((chr1 < 91) && (chr1 > 64))
        { // If char is uppercase
            textbit.append(1, (char)(chr1+32)); // add lowercase copy,
        }

        else if((chr1 < 123) && (chr1 > 96))
        { // If char is lowercase
            textbit.append(1, (char)(chr1-32)); // add uppercase copy
        }

#endif
    }

    for(string::size_type pos = str.find_first_of(textbit);
            pos <= (str.size() - text.size());
            pos = str.find_first_of(textbit, pos+1))
    {
        if(pos == string::npos) break; // no match poss
        string::size_type txtidx;
        for(txtidx = 1; txtidx < text.size(); txtidx++)
        {
            chr1 = str.at(pos+txtidx);
            char chr2 = text.at(txtidx);
            if((chr1 != chr2) &&
                    !((chr1 < 91) && (chr1 > 64) && (chr2 == chr1+32)) &&
                    !((chr1 < 123) && (chr1 > 96) && (chr2 == chr1-32)))
            {
                break; // no match this time
            }

        }
        if(txtidx == text.size()) return true; // All chars matched
        if(pos >= (str.size() - text.size())) break; // no match poss
    }

    return false;
}

/**
 * Accessor to get the segmented string of this parameter if it
 * exists. Returning NULL if there isn't one.
 **/
const TSegStr* TParam::segStr(unsigned int idx) const
{
    if(_segs.size() > 0)
    {
        if(idx >= _segs.size())
            idx = _segs.size() - 1;
        return _segs[idx];
    }
    else
    {
        return NULL;
    }
}

/**
 * Accessor to find out the number of values assigned to this parameter
 * If parameter does not have attribute 'repeatable' answer will always
 * be 0 or 1, (0 means no value assigned, will be defaulted)
 **/
unsigned int TParam::numVals() const
{
    return _recs.size();
}

/**
 * Mutator to give class a value that will have come from an
 * atvf style file.
 **/
AtvfRecord* TParam::setData(AtvfRecord* recP)
{
    if((_recs.size() > 0) &&
            !hasAttr("repeatable"))
    { // Not repeatable pop off old vals
        _recs.pop_back();
        _ints.pop_back();
        _segs.pop_back();
    }

    if(recP == NULL) recP = new AtvfRecord(innerRule());

    _recs.push_back(recP);
    string recval(recP->value());
    _ints.push_back(_prmRP->intVal(recval));
    _segs.push_back(_prmRP->segStr(recval));

    return recP;
}

/**
 * Mutator to set a value for the parameter.
 * AtvfRecord is created and returned if value never been set for
 * parameter before (Although the param keeps a reference to it
 * it is giving away ownership).  If NULL is returned it DOES NOT
 * signify an error it simply means and AtvfRecord already existed
 * and the value was assigned to this.
 * idx parameter ignored if rull shows this parameter to be non 
 * repeatable.
 **/
AtvfRecord* TParam::setVal(const string& val, unsigned int idx)
{
    // Only one element allowed if param isn't repeatable
    if((idx > 0) && !hasAttr("repeatable")) idx = 0;

    if(idx >= _recs.size())
    { // Stick new record on end
        AtvfRecord* recP = new AtvfRecord(innerRule());
        recP->value(val);
        setData(recP);
        return recP;
    }
    else
    { // Give existing record new value.
        _recs[idx]->value(val);
        _ints[idx] = _prmRP->intVal(val);
        if(_segs[idx] != NULL) delete _segs[idx];
        _segs[idx] = _prmRP->segStr(val);
        return NULL;
    }
}

/**
 * Friend of TParam used to write out its contents to an ostream
 **/
ostream& operator<<(ostream& os, const TParam& p)
{
    os << p.token() << " - " << p.strVal();
    if(p.numVals() == 0)
    {
        os << "(default)";
    }
    for(unsigned int i = 1;
            i < p.numVals();
            i++)
    {
        os << endl << "       " << p.strVal(i);
    }
    return os;
}
