/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// TRecord.cpp
/**
 * Method defintions and related for the TRecord class
 **/

#include <assert.h>

// Local includes
#include "TRecord.h"
#include "mdbexcept.h"
#include "resset.h"
#include "CECTofTxt.h"
#include "TContext.h"
#include "odl.h"

CFastLock *TRecord::m_pFastLock = NULL;

/**
 * Constructor
 **/
TRecord::TRecord(const TParamRule* headRuleP)
        :   _headPrmP(NULL),
        _headRecP(NULL),
        _delimRecP(NULL),
        _results(NULL)
{
    if(headRuleP!=NULL)
    {
        _headPrmP = new TParam(headRuleP);
        if(headRuleP->innerRule()->hasValue())
        {
            _prmidx.push_back(_headPrmP);
        }
    }
}

/**
 * Copy constructor
 **/
TRecord::TRecord(const TRecord& trec)
        :   _headPrmP(NULL),
        _headRecP(NULL),
        _delimRecP(NULL),
        _results(NULL)
{
    if(trec._headPrmP != NULL)
    {
        _headPrmP = new TParam((trec._headPrmP)->rule());
        _headRecP = new AtvfRecord((trec._headPrmP)->innerRule());
        _headPrmP->setData(_headRecP);
        if((_headRecP)->rule()->hasValue())
        {
            _headRecP->value(trec._headPrmP->strVal());
            _prmidx.push_back(_headPrmP);
        }
    }
    for(prmp_vect_t::const_iterator i = trec._prmidx.begin();
            i != trec._prmidx.end();
            i++)
    {
        if((*i) != trec._headPrmP)
        {
            TParam* prmP = new TParam((*i)->rule());
            for(unsigned int idx = 0;
                    idx < (*i)->numVals();
                    idx++)
            {
                AtvfRecord* recP = prmP->setVal((*i)->strVal(idx), idx);
                if(recP != NULL) _recs.push_back(recP);
            }
            _params[prmP->rule()->key()] = prmP;
            _prmidx.push_back(prmP);
        }
    }
}

/**
 * Destructor
 **/
TRecord::~TRecord()
{
    delete _headPrmP;
    delete _headRecP;
    delete _delimRecP;
    delete _results;
    // Loop through and delete all Atvf records and its list
    for(recp_cont_t::iterator i = _recs.begin();
            i != _recs.end();
            delete *(i++))
        ; // No 'for' content
    _recs.erase(_recs.begin(), _recs.end()); // no clear in g++
    if(!_params.empty())
    {
        for(prmp_cont_t::const_iterator i = _params.begin();
                i != _params.end();
                i++)
        {
            delete (*i).second;
        }
    }
    _params.erase(_params.begin(), _params.end()); // no clear in g++
}

Cresultset* TRecord::Results() const
{
    return _results;
}

void TRecord::SetResults(Cresultset* results)
{
    delete _results;
    _results = results;
}

/**
 * Accessor to get the token that describes this record type
 **/
string TRecord::token() const
{
    return (_headPrmP!=NULL)?_headPrmP->token():string("Unknown");
}

/**
 * Accessor to get any options this record type
 **/
const TOptions& TRecord::options() const
{
    return (_headPrmP!=NULL)?_headPrmP->options():_options;
}

/**
 * Accessor to find out if this is record that we know about
 **/
bool TRecord::isKnown() const
{
    return (_headPrmP==NULL)?false:true;
}

/**
 * Accessor to find out if this is record that we know about
 **/
bool TRecord::isType(const AtvfKey& key) const
{
    return (_headPrmP==NULL)?false:(key == _headPrmP->innerRule()->key());
}

/**
 * Accessor to find out if this record is a declaration record
 * ie. one not normally associated with the flow of the script
 * eg. a variable definition
 **/
bool TRecord::isDeclaration() const
{
    return ((_headPrmP!=NULL) &&
            _headPrmP->hasAttr("declaration")) ? true:false;
}

/**
 * Accessor to find a text string within the values of one of
 * the parameters, returns true if text found.  The index parameter
 * is the first parameter to search and will be set to the index
 * of the parameter containing the string if found and will be zeroed
 * if not found.  Parameters with the attribute 'nosearch' will
 * not be inspected, and if the record has this attribute the function
 * will always return false.  When performing a reverse search a
 * negative index  implies from the last parameter. 
 **/
bool TRecord::findInVal(const string& text,
                        int& prmidx,
                        bool matchCase,
                        bool forward) const
{
    if((_headPrmP != NULL) &&
            (_headPrmP->hasAttr(PATXT_NOSEARCH)))
    {
        prmidx = (forward ? 0 : -1);
        return false;
    }
    if(forward)
    {
        while(prmidx < static_cast<int>(_prmidx.size()))
        {
            if((prmidx >= 0) &&
                    (_prmidx[prmidx]->findInVal(text, matchCase))) return true;
            prmidx += 1;
        }
        prmidx = 0;
    }
    else // reverse search
    {
        if(prmidx < 0) prmidx = _prmidx.size() - 1; // set to end.
        while(prmidx >= 0)
        {
            if((prmidx < static_cast<int>(_prmidx.size())) &&
                    (_prmidx[prmidx]->findInVal(text, matchCase))) return true;
            prmidx -= 1;
        }
        prmidx = -1;
    }
    return false;
}

bool TRecord::findInVal(CRegExp& regex,
                        int& prmidx,
                        bool matchCase,
                        bool forward) const
{
    if((_headPrmP != NULL) &&
            (_headPrmP->hasAttr(PATXT_NOSEARCH)))
    {
        prmidx = (forward ? 0 : -1);
        return false;
    }
    if(forward)
    {
        while(prmidx < static_cast<int>(_prmidx.size()))
        {
            if((prmidx >= 0) &&
                    (_prmidx[prmidx]->findInVal(regex, matchCase))) return true;
            prmidx += 1;
        }
        prmidx = 0;
    }
    else // reverse search
    {
        if(prmidx < 0) prmidx = _prmidx.size() - 1; // set to end.
        while(prmidx >= 0)
        {
            if((prmidx < static_cast<int>(_prmidx.size())) &&
                    (_prmidx[prmidx]->findInVal(regex, matchCase))) return true;
            prmidx -= 1;
        }
        prmidx = -1;
    }
    return false;
}

/**
 * Accessor to save the contents out, to the given stream, in the atvf
 * transaction format.
 **/
void TRecord::save(ostream& out) const
{
    if(_headRecP != NULL) _headRecP->save(out);
    for(unsigned int i = 0; i < _recs.size(); i++)
    {
        (_recs[i])->save(out);
    }
}

/**
 * Execute whatever the contents of this record describe.
 * Do nothing in the base class , but will probably be overridden
 * to do something in derived classes.
 * Should return -1 on failure, 0 on success.
 * Other proposed meaningful values may become standard if required
 * such as -2 for execute this record again later and positive values
 * for script step to goto.
 * Although this is just an idea at the moment
 **/
int TRecord::execute(TContext& ctxt)
{
    if(ctxt.GetVerbose(WRNMSG))
        ctxt.log("WARNING: Object " + token() + " not defined (at least in that context).", WRNMSG);

    return 0; // Successfully done nothing!
}

//
// Pre execution setup, called prior to main record execution loop.
//	Do nothing in the base class , but may be overridden to do
//	something in derived classes.
//	Should return -1 on failure, 0 on success.

int TRecord::prepare(TScript* script, int objIndex, bool verbose)
{
    return 0;
}

/**
 * Accessor to save the delimiter line out, to the given stream.
 **/
void TRecord::saveDelim(ostream& out) const
{
    if(_delimRecP == NULL)
    {
        out << endl;
    }
    else
    {
        _delimRecP->save(out);
    }
}

/**
 * Accessor to return a pointer to the parameter of this record which
 * matches the supplied key. returns NULL if no such parameter exists.
 **/
const TParam* TRecord::param(const string& key) const
{
    AtvfKey atvfKey(key);

    if((_headPrmP != NULL) && (AtvfKey(_headPrmP->token()) == atvfKey))
        return _headPrmP;

    if(!_params.empty())
    {
        prmp_cont_t::const_iterator i = _params.find(atvfKey);
        if(i != _params.end())
        {
            return (*i).second;
        }
    }
    return NULL;
}

/**
 * Accessor to return a pointer to the parameter of this record which
 * has the given index.
 * Returns NULL if no such parameter exists.
 **/
const TParam* TRecord::param(unsigned int idx) const
{
    return (idx < numParams()) ? _prmidx[idx] :
           static_cast<TParam*>(NULL);
}

/**
 * Accessor to get the integer value of a parameter and perform any
 * variable substitution on it if it is needed.
 * Should probably throw exeption for none existent params, infers
 * doddgy ODL
 **/
int TRecord::intPrmVal(const string& key,
                       TContext& ctxt,
                       unsigned int idx) const
{
    const TParam* prmP = param(key);
    if(prmP == NULL) return 0; // No param with key return zero

    // Perform param checking and getting using context if needed
    const TSegStr* segP = prmP->segStr(idx);
    if(segP != NULL)
    {
        return prmP->intVal(segP->value(ctxt));
    }
    else
    {
        if(!prmP->hasAttr("expression"))
        {
            return prmP->intVal(idx);
        }
        else
        {
            //If not using TSegStr use CECTofTxt
            string toftxt=prmP->strVal(idx);
            TtIntExpr tce(ctxt.Script());
            if(tce.Parse(toftxt) == FALSE)
            {
                // Parsing error. Return 0.
                return 0;
            }
            else
            {
                return atoi(tce.Evaluate(ctxt).c_str());//convert answer to int
            }
        }

    }
}

/**
 * Accessor to get the integer value of a parameter and perform any
 * variable substitution on it if it is needed.
 * Should probably throw exeption for none existent params, infers
 * dodgy ODL
 **/
string TRecord::strPrmVal(const string& key,
                          TContext& ctxt,
                          unsigned int idx) const
{
    const TParam* prmP = param(key);
    if(prmP == NULL) return string();

    // Perform param checking and getting using context if needed
    const TSegStr* segP = prmP->segStr(idx);
    if(segP != NULL)
    {
        return segP->value(ctxt);
    }
    else
    {
        if(!prmP->hasAttr("expression"))
        {
            return prmP->strVal(idx);
        }
        else
        {
            //If not using TSegStr use CECTofTxt
            string toftxt=prmP->strVal(idx);
            TtChrExpr tce(ctxt.Script());
            if(tce.Parse(toftxt) == FALSE)
            {
                return string();
            }
            else
            {
	            return tce.Evaluate(ctxt);
            }
        }
        //old return straight value return prmP->strVal(idx);
    }

}

/**
 * Find out the number of parameters that this record has.
 **/
unsigned int TRecord::numParams() const
{
    return _prmidx.size();
}

/**
 * Produce a duplicate of this Transaction record.
 **/
TRecord* TRecord::duplicate() const
{
    return new TRecord(*this);
}

/**
 * Method to add a set of options to the record.
 * This is only really useful if we have no head parameter.
 **/
void TRecord::addOptions(const TOptions& opts)
{
    _options.add(opts);
    return;
}

/**
 * Method to add known parameters to this record
 **/
void TRecord::addParam(TParamRule* prP)
{
    TParam* prmP = new TParam(prP);
    _params[prP->key()] = prmP;
    _prmidx.push_back(prmP);
    return;
}

/**
 * Method to set the header part of this record
 **/
void TRecord::addHead(AtvfRecord* headP)
{
    if(_headPrmP != NULL)
    {
        headP = _headPrmP->setData(headP);
    }
    _headRecP = headP;
    return;
}

/**
 * Method to add a read in data (in the form of an Atvf
 * Record) to the record.  Only return false if the end of the
 * record has been reached, true all other times even if the data
 * does not belong with this record. Ownership assumed of all
 * Atvf records passed in.
 **/
bool TRecord::addData(AtvfRecord* recP)
{
    // If it was a blank record it signifies the end of the record
    if(recP->isBlank())
    {
        _delimRecP = recP;
        return false;
    }

    // Add the record to our list
    _recs.push_back(recP);

    // Link to, if one of parameters we are expecting
    if(!_params.empty())
    {
        prmp_cont_t::iterator i = _params.find(recP->key());
        if(i != _params.end())
        {
            ((*i).second)->setData(recP);
        }
    }

    return true;
}

void TRecord::deleteData(AtvfRecord* recP)
{
    for(recp_cont_t::iterator it = _recs.begin(); it != _recs.end(); it++)
    {
        if ((*it) == recP)
        {
            _recs.erase(it);
            delete recP;
            break;
        }
    }
}


/**
 * Mutator to set the value of a parameter within this record
 * If key is not for one of parameters of this record and exception
 * is thrown.
 **/
void TRecord::setPrmVal(const string& val, const string& key,
                        unsigned int idx)
{
    TParam* prmP = const_cast<TParam*>(param(key));
    if(prmP == NULL) throw mdbRuntimeError(key + " parameter doesn't exist");
    AtvfRecord* recP = prmP->setVal(val, idx);
    // If new record created add to list -
    if(recP != NULL)
    {
        if(prmP == _headPrmP) // This is the header record
        {
            _headRecP = recP;
        }
        else
        {
            _recs.push_back(recP);
        }
    }
    return;
}

/**
 * A friend global function to allow the TRecord class to
 * be written out to an ostream in textual form.
 **/
ostream& operator<<(ostream& os, const TRecord* rP)
{
    os << rP->token();
    if((rP->_headPrmP != NULL) && (rP->_headPrmP->innerRule()->hasValue()))
    {
        os << endl << "  value - " << rP->_headPrmP->strVal();
    }

    if(rP->isKnown())
    {
        for(prmp_cont_t::const_iterator i = (rP->_params).begin();
                i != (rP->_params).end();
                i++)
        {
            os << endl << "  " << *((*i).second);
        }
    }
    else // Don't know record attributes, just show records contained
    {
        if(rP->_headRecP!=NULL) os << endl << *(rP->_headRecP);
        for(unsigned int i = 0; i < (rP->_recs).size(); i++)
        {
            os << endl << "  " << *(rP->_recs[i]);
        }
        if(rP->_delimRecP!=NULL) os << endl << *(rP->_delimRecP);
    }

    return os;
}
