/* Copyright (C) 2001 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// SCLCompiler.cpp: implementation of the CSCLCompiler class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "SCLCompiler.h"
#include "RegFunctions.h"
#include "OstaDefs.h"
#include "depprmdef.h"          // SCL script dependendency file symbols

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CSCLCompiler::CSCLCompiler()
{
	m_strOutputFolder = GetCompilerOutputDir();
	CreateDirectory(m_strOutputFolder,NULL);
}

CSCLCompiler::~CSCLCompiler()
{

}

BOOL CSCLCompiler::CompileScript(const CString& strScript,BOOL bCompileDependencies)
{
	m_strErrorMsg.Empty();

	//Get output name
	CString strOutputPath,strDependPath;
	GenerateOutputPath(strScript,strOutputPath,strDependPath);
        
	//Get command line
	CString strCmdLine;
	GenerateCommandLine(strScript,strOutputPath,strCmdLine);

	//GetCurrent directory so we can restore it
	CString strCurDir;
	GetCurrentDirectory(MAX_PATH,strCurDir.GetBuffer(MAX_PATH));
	strCurDir.ReleaseBuffer();

	//Set current directory to src path
	int nSlashPos = strScript.ReverseFind('\\');
	if(nSlashPos!=-1)
		SetCurrentDirectory( strScript.Left(nSlashPos) );

	//Run compiler
    SECURITY_ATTRIBUTES saAttr;
    saAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
    saAttr.bInheritHandle = TRUE;
    saAttr.lpSecurityDescriptor = NULL;

	HANDLE hChildWrite=INVALID_HANDLE_VALUE;
	HANDLE hChildRead=INVALID_HANDLE_VALUE;

    // Create a pipe for the child's STDOUT.
    if(!CreatePipe (&hChildRead, &hChildWrite, &saAttr, 0))
	{
		m_strErrorMsg.Format("Failed to create compiler pipe (Error %d)",GetLastError());
		return false;
	}

    // Now create the child process to execute the command
	PROCESS_INFORMATION ProcInfo;
	STARTUPINFO StartInfo = {sizeof(StartInfo)};
	StartInfo.dwFlags = STARTF_USESTDHANDLES|STARTF_USESHOWWINDOW;
	StartInfo.hStdOutput = hChildWrite;
	StartInfo.hStdError = hChildWrite;
	StartInfo.wShowWindow = SW_HIDE;

	if (!CreateProcess (NULL
		                ,strCmdLine.GetBuffer(0)            // Command Line
                        ,NULL                            // Process Security attributes
                        ,NULL                            // Primary thread security attributes
                        ,TRUE							 // Inherit Handles
                        ,CREATE_NEW_CONSOLE              // Creation flags
                        ,NULL                            // Use Parents environment
                        ,NULL                            // Use Parents current directory
                        ,&StartInfo                      // STARTUPINFO pointer
                        ,&ProcInfo))                     // Receives Process Information
    {
		m_strErrorMsg.Format("Failed to start compiler. (Error %d)",GetLastError());
		return FALSE;
    }
    
    // Close the write end of the pipe before reading
    CloseHandle(hChildWrite);

    //Read from the pipe
	const DWORD dwBufSize = 4096;
    char szBuf[dwBufSize];

    // Read output from child, and write it to the output window edit box.
    DWORD dwRead=0;
	while(1)
    {
        if((!ReadFile(hChildRead, szBuf, dwBufSize-1, &dwRead, NULL)) || (dwRead == 0)) 
			break;
        
        m_strErrorMsg += CString(szBuf,dwRead);
    }

	//Check if compile succeeded 
	BOOL bCompileOK = IsUptodate(strScript,strOutputPath);
    if (!bCompileOK) return FALSE;

    // Read the SCL script dependency file (".scd") for the script in order to
    // identify all called scripts. Each called script is copied from
    // the Repository Source directory to the working directory. Note that
    // the check for called scripts needs to be recursive, in order to
    // account for scripts called by called scripts...
    //
    // The format of a called script dependency line is:
    //
    //      item_name|dep_type{|dep type specific parameter(s)}
    // e.g. MYSCRIPT|3|-|0
    //
    // For called script dependencies, the item name is the script name,
    // excluding path and file extension.
    //
    // Note: It would be better to have a dependency file class in order
    // to encapsulate the file handling code - rather than embed
    // knowledge of the file format here.

	if(bCompileDependencies)
	{
		try
		{
			CStdioFile fDep(strDependPath,CFile::modeRead);
		
            // Read lines from the SCL script dependency file.
			CString strDep;
			while(fDep.ReadString(strDep))
			{
                // See if this is a 'call script' dependency.
                int sep1 = strDep.Find(DEP_SEPCHR);
                int sep2 = strDep.Find(DEP_SEPCHR, sep1 + 1);
                if (sep2 == -1) sep2 = strDep.GetLength();

                CString strDepTyp = strDep.Mid(sep1 + 1, (sep2 - (sep1 + 1)));
                LPTSTR pchr = strDepTyp.GetBuffer(1);
                int iDepTyp = atoi(pchr);

                // If dependency file format error, report it.
                if ((sep1 == -1) || (iDepTyp == 0))
                {
				    m_strErrorMsg.Format("Failed to compile %s.HTP...\nDependency script format error.\nLine in error is:\n%s"
                        ,strScript, strDep);
		            return FALSE;
                }
                
                // If not a 'call script' dependency, loop to get next line.
                if (iDepTyp != DEP_TYP_CALSCR) continue;

                // Drop through to here if we have a 'called script' dependency.
                // to compile script (and its dependencies!).
                strDep = strDep.Left(sep1) + ".HTP";
			    if(!CompileScript(strDep,bCompileDependencies)) return FALSE;
			}

			fDep.Close();
		}
		catch(...)
		{
			return false;
		}

	    //Check if compile succeeded 
	    bCompileOK = IsUptodate(strScript,strOutputPath);
        if (!bCompileOK) return false;
    } // if(bCompileDependencies)

	//Restore directory
	SetCurrentDirectory(strCurDir);

    return true;
}

BOOL CSCLCompiler::GenerateCommandLine(const CString& strScript,const CString& strOutput,CString& strCmdLine)
{
	CString strExt;
    int nDotPos = strScript.ReverseFind('.');
	if(nDotPos != -1)
		strExt = strScript.Mid(nDotPos+1);

	CString strIncludeDir( GetRepositoryPath());
	strIncludeDir += strIncludeFolder;

	strCmdLine.Format("%sscl.exe -I \"%s\" -o \"%s\" \"%s\"",GetCompilerPath(),strIncludeDir,strOutput,strScript);
	return TRUE;
}	

BOOL CSCLCompiler::IsUptodate(const CString& strSrcPath,const CString& strOutputPath)
{
	//Compare date of source and output files
	//If source is newer than output file then 
	//the output is not uptodate
	CFileStatus fsSrc,fsOut;
	if(!CFile::GetStatus(strOutputPath,fsOut))
		return false;

	if(!CFile::GetStatus(strSrcPath,fsSrc))
		return false;

	return (fsSrc.m_mtime <= fsOut.m_mtime);
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Given a fully qualified script source file name, produce the fully
//          qualified names of the corresponding script object file (".TOF")
//          and script dependency file (".SCD"),e.g.:
//
//          Source:     C:\Program Files\OpenSTA\Repository\Scripts\Myscript.HTP
//          Object:     C:\Program Files\OpenSTA\Repository\Scripts\Myscript.TOF
//          Dependency: C:\Program Files\OpenSTA\Repository\Scripts\Myscript.SCD
//
// @parm    Pointer to fully qualified script source file name.
// @parm    Pointer to fully qualified script object file name.
// @parm    Pointer to fully qualified script dependency file name.
//
BOOL CSCLCompiler::GenerateOutputPath(const CString& strSrcPath,CString& strOutputPath,CString& strDependPath)
{
	strOutputPath = m_strOutputFolder;

	//Get the name of the src file (ie strip the path off)
	CString strSrcFile(strSrcPath);
	int nSlashPos = strSrcPath.ReverseFind('\\');
	if(nSlashPos != -1) strSrcFile.Delete(0,nSlashPos+1);
	
	//Append new filenames and extensions.
	int nDotPos = strSrcFile.ReverseFind('.');
	strOutputPath += strSrcFile.Left(nDotPos+1);
	strDependPath = strOutputPath;

    strOutputPath += "tof";
    strDependPath += "scd";

	return false;
}

CString CSCLCompiler::GetCompilerPath()
{
    // Get path to scl.exe
    CString strExePath(getenv("OPENSTA_HOME"));
	strExePath += "Engines\\";
	return strExePath;
}

void CSCLCompiler::SetOutputFolder(const CString& strFolder)
{
	if(CreateDirectory(strFolder,NULL))
		m_strOutputFolder = strFolder;
}
