/* Copyright (C) 2000 Cyrano SA

   This file is part of OpenSTA

   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.

   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html

   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

//========================================================================
//
// TEHttpRecords.cpp
// -----------------
//=========================================================================

#ifdef WIN32
#   pragma warning(disable:4786)
#   pragma warning(disable:4503)
#endif

#include "RegExp.h"
#include "TRecMake.h"
#include "TContext.h"
#include "tools.hpp"
#include "AsyncVirtualUser.h"
#include "Address.hpp"
#include "PString.hpp"
#include "PStringSpecialised.inl"
#include "TEHttpRecords.h"
#include "TEHttpRecords.inl"
#include "LogTrace.hpp"
#include <malloc.h>

static const char HTTP_NONE       [] = "None";
static const char HTTP_CONNECT    [] = "Connect";
static const char HTTP_REQUEST    [] = "Request";
static const char HTTP_DISCONNECT [] = "Disconnect";
static const char HTTP_RESPONSE	  [] = "Response";
static const char HTTP_SYNCHRONIZE[] = "Synchronize";
static const char HTTP_WAIT       [] = "Wait";


//
// TRcrdConnect::prepare
// =====================
//
// perform any preparation before conecting the threads
int TRcrdConnect::prepare(TScript* script, int objIndex, bool verbose) 
{
	const TParam* prm = param(TParam::stServer);
    if(prm == NULL) 
		return -1;
    string l_data = prm->strVal();
    _exprP = new TtChrExpr(script);

    // Encode the variable reference; report an error if encode fails.
    if (!(_exprP->Parse(l_data, 0)))
    {
		MSGERR("Connect : " + l_data +
			   " Failed during 'server' PREPARE with: " + _exprP->Reason());
		return -1;
    }
    return 0;
}

//
// TRcrdConnect::execute
// =====================
//
// Return ETofResult::TOF_RES_ASYNC_STARTED if must be called again to complete execution.
int TRcrdConnect::execute(TContext& ctxt)
{
    CAsyncVirtualUser& VU = dynamic_cast<CAsyncVirtualUser&> (ctxt);
    int nConid;
    CAsyncHTTPRequest* pConnection = NULL;

	ctxt.set_phase_text(PHASE_STRT_CONNECT);

    if(VU.GetTrace() && VU.GetTrace()->IsTrace("full"))
    {
        GetTraceStream(VU.GetTrace()) << "TRcrdConnect: entering step " << VU.GetTRecordCallStep() << ".\r\n";
        ReleaseTraceStream();
    }

    switch(VU.GetTRecordCallStep())
    {
    case 0:
        {
            // Synchronise on nConid in case it would still be in use (TOF anomaly).

        	// Get the variables from the tof
            nConid = intPrmVal(TParam::stConnid, ctxt);
            VU.SetCurrentConid(nConid);

            VU.SetTRecordCallStep(1);   // Set before the test to avoid run-conditions.
            if(VU.SynchroniseHTTPClient(nConid) == FALSE)
            {
                if(ctxt.GetVerbose(TRACEMSG))
                    ctxt.log("WARNING: CONNECTing with a connection ID that is still in use: " + int_to_string(nConid) + ". Synchronising.", TRACEMSG);

                return TOF_RES_ASYNC_STARTED;
            }
        }
        // No break since should continue if no need to synchronize.

    case 1:
        {
			// Eventually break the previous connection (if there's already a connection with that ID that 
            // received a Connection:close or if that connection was established to another server/port).

            nConid = VU.GetCurrentConid();

            string szServer;
            {
                szServer = _exprP->Evaluate(ctxt);
            }

            if(VU.GetTrace() && VU.GetTrace()->IsTrace("connection"))
            {
                GetTraceStream(VU.GetTrace()) << "TRcrdConnect: Connecting " << nConid << " to server " << szServer << ".\r\n";
                ReleaseTraceStream();
            }
            if(ctxt.GetVerbose(TRACEMSG))
                ctxt.log("CONNECT(" + int_to_string(nConid) + ") to proxy server " + szServer, TRACEMSG);

            // Get connection object, create new if necessary.
            pConnection = VU.ObtainRequestStruct(nConid);
            if(pConnection == NULL)
            {
                // All the request slots are already used, and was not able to allocate a new one. Fatal error.
                if(VU.GetTrace() && VU.GetTrace()->IsTrace("errors connection"))
                {
                    GetTraceStream(VU.GetTrace()) << "TRcrdConnect: Could not allocate a new connection.\r\n";
                    ReleaseTraceStream();
                }

                if(ctxt.GetVerbose(ERRMSG))
                    ctxt.log("Connection failed.", ERRMSG);
                VU.LogError("ERROR for VU " + VU.GetThreadID() + ": could not allocate a new connection");

                assert(0);
                return TOF_RES_FAILED;
            }

	        string szHost;
	        int	nPort;
	        unsigned int endServer = szServer.find(':');
	        if(endServer != string::npos)
	        {
		        szHost = szServer.substr(0, endServer).c_str();
		        nPort  = atoi(szServer.substr(endServer + 1, szServer.size() - endServer - 1).c_str());
	        }
            else
            {
                szHost = szServer;
                nPort  = 80;
            }

            // Store connection parameters into the VU for future use.
            VU.SetProxyServerName(szHost);
            VU.SetProxyServerPort(nPort);

            if(pConnection->IsTCPConnected())
            {
                // Disconnect if necessary.
                BOOL bMustDisconnect = FALSE;

                if(pConnection->GetConnectionType() == CAsyncHTTPRequest::CNX_DISCONNECTS)
                    bMustDisconnect = TRUE;
                else
                {
                    string  zCurrentServer;
                    DWORD   dwCurrentPort;

                    pConnection->GetConnParams(zCurrentServer, dwCurrentPort);

                    if(zCurrentServer != szHost || dwCurrentPort != (DWORD)nPort)
                    {
                        // Must connect to another server/port.
                        if(ctxt.GetVerbose(TRACEMSG))
                            ctxt.log("WARNING: CONNECTing with a connection ID that is still connected to another server: " + int_to_string(nConid) + ". Reconnecting.", TRACEMSG);

                        bMustDisconnect = TRUE;
                    }
                }

                if(bMustDisconnect)
                {
                    VU.SetTRecordCallStep(2);   // Will enter the "case 2" section next time execute is called.
                    if(VU.DisconnectClient(nConid) == TOF_RES_ASYNC_STARTED)    // Can't use CAsyncHTTPRequest::Disconnect().
                    {
                        // That case occurs when using SSL protocol.
                        return TOF_RES_ASYNC_STARTED;
                    }
                    else
                    {
                        pConnection->SetConnectionId(nConid);
                    }
                }
            }
        }
        // No break since should continue if no async. op.

    case 2:
        {
            // Perform the connection.

            nConid = VU.GetCurrentConid();

            if(pConnection == NULL)
                pConnection = VU.ObtainRequestStruct(nConid);
            assert(pConnection != NULL);

            // Get the values computed in case 1.
            string szHost = VU.GetProxyServerName();
	        int	nPort = VU.GetProxyServerPort();

            // Read the statistics verbose level flag. Tell the request structures what they're gonna have to do of the stats.
            pConnection->SetWatchStatsTo(VU.MustDumpStats());

            // Perform connection

	        // Note that a CONNECT object should not be asynchronous...
            DWORD dwRes = pConnection->Connect(szHost.c_str(), nPort, (IocpEvent*) pConnection);
            switch(dwRes)
	        {
	        case TOF_RES_OK:
		        {
			        // --- A Connect object implies that connection was made through a proxy
			        pConnection->SetUseProxyTo(TRUE);
			        ctxt.set_phase_text(HTTP_NONE);

			        return dwRes;
		        }

	        case TOF_RES_ASYNC_STARTED:
	        default:
                // Log client error
                if(VU.GetTrace() && VU.GetTrace()->IsTrace("errors connection"))
                {
                    GetTraceStream(VU.GetTrace()) << "TRcrdConnect: Connection Failed for " << nConid << ". Error is " << dwRes << ".\r\n";
                    ReleaseTraceStream();
                }

                if(ctxt.GetVerbose(ERRMSG))
                    ctxt.log("Connection ERROR for " + int_to_string(nConid) + ". Error is " + int_to_string((unsigned int)dwRes), ERRMSG);

                // Release the connection.
                VU.ReleaseRequest(pConnection);

		        return TOF_RES_FAILED;
	        };
        }   // Case 2.

    default:
        {
            assert(0);
            return TOF_RES_FAILED;
        }
    };

	ctxt.set_phase_text(HTTP_NONE);
    return TOF_RES_OK;
}

//
// TRcrdRequest::prepare
// =====================
//
// perform any preparation before sending the request
int TRcrdRequest::prepare(TScript* script, int objIndex, bool verbose) 
{
	const TParam* prm = param("uri");
    if(prm == NULL)
    {
		MSGERR("Request:PREPARE: could not find the \'uri\'");
		return -1;
    }
    string l_uri = prm->strVal();
    _uriP = new TtChrExpr(script);

    // Encode the variable reference; report an error if encode fails.
    if (!(_uriP->Parse(l_uri, 0)))
    {
		MSGERR("Request : " + l_uri +
			   " Failed during 'uri' PREPARE with: " + _uriP->Reason());
		return -1;
    }


	prm = param("header");
    if(prm == NULL) 
		return -1;
    string l_header = prm->strVal();
    _headerP = new TtValList(script);

    // Encode the variable reference; report an error if encode fails.
    if (!(_headerP->Parse(l_header, 0)))
    {
		MSGERR("Request : " + l_header +
			   " Failed during 'header' PREPARE with: " + _headerP->Reason());
		return -1;
    }


	prm = param("with");
    if(prm == NULL) 
		return -1;
    string l_with = prm->strVal();
    _withP = new TtValList(script);

    // Encode the variable reference; report an error if encode fails.
    if (!(_withP->Parse(l_with, 0)))
    {
		MSGERR("Request : " + l_with +
			   " Failed during 'with' PREPARE with: " + _withP->Reason());
		return -1;
    }


	prm = param("wout");
    if(prm == NULL) 
		return -1;
    string l_wout = prm->strVal();
    _woutP = new TtValList(script);

    // Encode the variable reference; report an error if encode fails.
    if (!(_woutP->Parse(l_wout, 0)))
    {
		MSGERR("Request : " + l_wout +
			   " Failed during 'wout' PREPARE with: " + _woutP->Reason());
		return -1;
    }


	prm = param("body");
    if(prm == NULL) 
		return -1;
    string l_body = prm->strVal();
    _bodyP = new TtChrExpr(script);

    // Encode the variable reference; report an error if encode fails.
    if (!(_bodyP->Parse(l_body, 0)))
    {
		MSGERR("Request : " + l_body +
			   " Failed during 'body' PREPARE with: " + _bodyP->Reason());
		return -1;
    }

    prm = param(TParam::stConnid);
    string l_conid = prm->strVal();
    _iConid = atoi(l_conid.c_str());


	prm = param("certificate");
    if(prm == NULL) 
    {
		MSGERR("Request : " + l_uri +
			   " Failed during 'certificate' PREPARE with: Parameter missing from ODL");
		return -1;
    }
	string l_cert = prm->strVal();
    _certName = new TtChrExpr(script);

    // Encode the variable reference; report an error if encode fails.
    if (!(_certName->Parse(l_cert, 0)))
    {
		MSGERR("Request : " + l_cert +
			   " Failed during 'certificate' PREPARE with: " + _certName->Reason());
		return -1;
    }

	prm = param(TParam::stRetCod);
    if(prm == NULL) 
		return -1;
    string szCode = prm->strVal();
    if(!szCode.empty())
    {
        _retCode = new TtVarRef(script);
        if (!(_retCode->Parse(szCode, 0)))
        {
		    MSGERR("Request : " + szCode +
			       " Failed during 'retCode' PREPARE with: " + _retCode->Reason());
		    return -1;
        }
    }

    return 0;
}

//
// TRcrdRequest::getRequestString
// ==============================
//
// translate the type into the actual Command string

const string &TRcrdRequest::getRequestString(ERequestType eType) const
{
    static const string zReqType[] = { "GET ", "HEAD ", "POST ", "CONNECT " };

    return zReqType[(int)eType%4];
}

//
// TRcrdRequest::execute
// =====================
//
// Execute a request.
int TRcrdRequest::execute(TContext& ctxt)
{
    CAsyncVirtualUser& VU = dynamic_cast<CAsyncVirtualUser&> (ctxt);

    string szStatus, szUri;
    CAsyncHTTPRequest* pConnection = NULL;
    int nConid;

    if(VU.GetTrace() && VU.GetTrace()->IsTrace("full"))
    {
        GetTraceStream(VU.GetTrace()) << "TRcrdRequest: entering step " << VU.GetTRecordCallStep() << ".\r\n";
        ReleaseTraceStream();
    }

    switch(VU.GetTRecordCallStep())
    {
    case 0:
        {
			// Wait for the last request executed on that connection to complete.

            ctxt.set_phase_text(HTTP_REQUEST);

            nConid = intPrmVal(TParam::stConnid, ctxt);
            VU.SetCurrentConid(nConid);

            if(VU.GetTrace() && VU.GetTrace()->IsTrace("request full"))
            {
                GetTraceStream(VU.GetTrace()) << "TRcrdRequest: executing on connection " << VU.GetManagerNumber() << '/' << nConid << ".\r\n";
                ReleaseTraceStream();
            }

            // Make sure that the last request on the specified connection is over.
            VU.SetTRecordCallStep(1);   // Set before the test to avoid run-conditions.
            if(VU.SynchroniseHTTPClient(nConid) == FALSE)
            {
                return TOF_RES_ASYNC_STARTED;
            }
        }
        // No break since should continue if no need to synchronize.

		// Perform connection if required.
        // Cases are:
        //  1/ TCP-connecting directly. => TCP-connect to remote server.
        //  2/ SSL-connecting directly. => TCP-connect to remote server + SSL-connect.
        //  3/ TCP-connecting through a proxy. => TCP-connect to proxy (performed by the CONNECT TOF object) if not already done by TOF-connect or connection to proxy was broken.
        //  4/ SSL-connecting through a proxy. => TCP-connect to proxy (performed by the CONNECT TOF object) + HTTP-CONNECT to remote server + SSL-connect.

    case 1:
        {
            // Previous request on that connection is now over.
            // Eventually disconnect if previous request received a Connection-close message from server.

            nConid = VU.GetCurrentConid();

            // Get a (new) request slot to work with.
            pConnection = VU.ObtainRequestStruct(nConid);

            if(pConnection == NULL)
            {
                // All the request slots are already used, and was not able to allocate a new one. Fatal error.
                if(VU.GetTrace() && VU.GetTrace()->IsTrace("connection errors"))
                {
                    GetTraceStream(VU.GetTrace()) << "TRcrdRequest: could not allocate new connection.\r\n";
                    ReleaseTraceStream();
                }

                if(ctxt.GetVerbose(ERRMSG))
                    ctxt.log("REQUEST ERROR: could not allocate new connection.", ERRMSG);

                VU.LogError("ERROR for VU " + VU.GetThreadID() + ": could not allocate a new connection");

                return TOF_RES_FAILED;
            }

            pConnection->SetWatchStatsTo(VU.MustDumpStats());

            // In case of HTTP 1.0 and if the selected connection has already been used, then server should have 
            // disconnected. Can close the connection.
            if(pConnection->GetConnectionType() == CAsyncHTTPRequest::CNX_DISCONNECTS)
            {
                if(pConnection->GetUseProxy())
                {
                    pConnection->SetWillReconnectThroughProxy(TRUE);
                }

                VU.SetTRecordCallStep(2);   // Will enter the "case 2" section next time execute is called.
                if(VU.DisconnectClient(nConid) == TOF_RES_ASYNC_STARTED)    // Can't use CAsyncHTTPRequest::Disconnect().
                {
                    // That case occurs when using SSL protocole. 
                    // The framework will then first call:
                    //  1- CAsyncSecureSocket::Close
                    //  2- CAsyncSecureSocket::CompleteClose (when SSL close dialog is over)
                    //  3- CAsyncTCPSocket::Close
                    //  4- CAsyncHTTPRequest::OnEvent
                    return TOF_RES_ASYNC_STARTED;
                }
                else
                {
                    pConnection->SetConnectionId(nConid);
                }
            }
        }
        // No break since should continue if did not perform an async. disconnect.

    case 2:
        {
            // Execute the HTTP-CONNECTION request against the proxy if required (using the connection 
            // established by the CONNECT TOF object), or TCP-connect directly to the remote server.

            nConid = VU.GetCurrentConid();

            // Get a (new) request slot to work with.
            if(pConnection == NULL)
                pConnection = VU.ObtainRequestStruct(nConid);

            assert(pConnection != NULL);

            {
                szUri = _uriP->Evaluate(ctxt);
            }

            // Check if that connection will be used for SSL requests LATER on
            // (for the moment, remain in standard TCP in case we need to perform a CONNECT request).
	        BOOL bSecureRequest;
	        if(find_in_string(szUri, "https://", FALSE, 0) != -1)
		        bSecureRequest = TRUE;
            else
                bSecureRequest = FALSE;

            // Parse the host:port string. In the HTTP/1.1 case, should find a 'Host' header field. Otherwise, parse the URI.
            string szRemoteServer, szPort;
            string szHeader; szHeader.reserve(512);

            BuildHeaderString(ctxt, szHeader);

            ExtractHostPort(ctxt, szHeader, szUri, szRemoteServer, szPort);

            if(szPort.empty())
            {
                if(bSecureRequest)
                    szPort = "443";
                else
                    szPort = "80";
            }

            // Store the values for later use.
            VU.SetHeaderText(szHeader);
            VU.SetRemoteServerName(szRemoteServer);
            VU.SetRemoteServerPort(szPort);
            VU.SetSecureModeUsed(bSecureRequest);
            VU.SetCurrentUri(szUri);

            // Check if the connection is established.
            if(pConnection->IsTCPConnected() == FALSE)
            {
			    if(VU.GetTrace() && VU.GetTrace()->IsTrace("request"))
                {
                    GetTraceStream(VU.GetTrace()) << "TRcrdRequest: Connecting " << nConid << " using connection slot number " << pConnection->GetRequestNumber() << ".\r\n";
                    ReleaseTraceStream();
                }

                pConnection->SetSecurityMode(CAsyncHTTPRequest::HTTP_NOT_SECURE);

                // Check if should first connect to a proxy server or directly to the remote server.
                if(pConnection->GetUseProxy() || pConnection->GetWillReconnectThroughProxy())
                {
                    // Connection to proxy, which was opened by a previous CONNECT TOF was lost. Reconnect.
                    if(pConnection->ReConnect() != 0)
                    {
                        // Note result cannot be TOF_RES_ASYNC_STARTED since won't connect to the proxy using SSL.
                        if(ctxt.GetVerbose(ERRMSG))
                            ctxt.log("REQUEST: ERROR re-connecting to the proxy.", ERRMSG);
                        VU.LogError("ERROR for VU " + VU.GetThreadID() + ": failed to re-connect to proxy.");
                        return TOF_RES_FAILED;
                    }
                    pConnection->SetWillReconnectThroughProxy(FALSE);
                    pConnection->SetUseProxyTo(TRUE);
                }
                else
                {
                    // Is not yet connected to the web server or the connection was lost.
                    WORD wPort = (WORD)atoi(szPort.c_str());

                    // TCP-connect to the remote server, whether this is to later perform a secure request or not.
                    if(pConnection->Connect(szRemoteServer.c_str(), (WORD)wPort, (IocpEvent*) pConnection) != 0)
                    {
                        if(ctxt.GetVerbose(ERRMSG))
                            ctxt.log("REQUEST: ERROR connecting to " + szRemoteServer + ':' + szPort, ERRMSG);
                        VU.LogError("ERROR for VU " + VU.GetThreadID() + ": failed to connect to " + szRemoteServer + ':' + szPort);
                        return TOF_RES_FAILED;
                    }
                }
            }

            // In case of SSL proxy tunneling, and if not sent any requests yet, send a HTTP CONNECT request.
            if(bSecureRequest && pConnection->NeedConnectToProxy())
            {
                // Is connected to the proxy server since the 'Connect' TOF object. Should now HTTP-CONNECT to the remote server.

                // Build the CONNECT request.
                string szRequestText; szRequestText.reserve(255);

                string szURI = "CONNECT " + szRemoteServer + ':' + szPort + " HTTP/1.1";
                szRequestText = szURI + "\r\nUser-Agent: Mozilla/4.0 (compatible; MSIE 5.0; Windows NT; DigExt)\r\nHost: "
                                + szRemoteServer + ':' + szPort + "\r\nContent-Length: 0\r\nPragma: no-cache\r\n\r\n";

                if(VU.GetTrace() && VU.GetTrace()->IsTrace("request"))
                {
                    GetTraceStream(VU.GetTrace()) << "TRcrdRequest: Executing (size is " << szRequestText.size() << "):\r\n" << szRequestText << ".\r\n";
                    ReleaseTraceStream();
                }

                if(ctxt.GetVerbose(TRACEMSG))
                    ctxt.log("REQUEST: Connecting the proxy to the remote server to enable SSL session establishment.", TRACEMSG);

                pConnection->GetStats().SetUrl(szURI);

                pConnection->SetRequestText(szRequestText);

                DWORD dwRes = pConnection->InitiateCONNECT(NO_RESULT);
                if(dwRes != 0)
                {
                    if(ctxt.GetVerbose(ERRMSG))
                        ctxt.log("REQUEST: ERROR executing the CONNECT request through the proxy. Error is: " + int_to_string((LONG)dwRes), ERRMSG);
                    VU.LogError("ERROR for VU " + VU.GetThreadID() + ": failed to execute the CONNECT request through the proxy. Error is: " + int_to_string((LONG)dwRes));
                    return TOF_RES_FAILED;
                }

                // Wait for the CONNECT request to complete.
                VU.SetTRecordCallStep(3);   // Will enter the "case 3" section next time execute is called.
                if(VU.SynchroniseHTTPClient(nConid) == FALSE)
                {
                    // Download on that connection is not over yet. Cannot continue.
                    return TOF_RES_ASYNC_STARTED;
                }
            }
        }
        // No break since should continue if did not start an HTTP-CONNECT request.

    case 3:
        {
            // In the case of https requests, execute the SSL connection sequence.

            nConid = VU.GetCurrentConid();

            if(pConnection == NULL)
            {
                // Get a (new) request slot to work with.
                pConnection = VU.ObtainRequestStruct(nConid);

                if(pConnection == NULL)
                {
                    // All the request slots are already used, and was not able to allocate a new one. Fatal error.
                    if(VU.GetTrace() && VU.GetTrace()->IsTrace("connection errors"))
                    {
                        GetTraceStream(VU.GetTrace()) << "TRcrdRequest: could not allocate new connection.\r\n";
                        ReleaseTraceStream();
                    }

                    VU.LogError("ERROR for VU " + VU.GetThreadID() + ": failed to allocate a new connection.");
                    return TOF_RES_FAILED;
                }
            }
            // else moved directly from case 2 to case 3. Thus, already acquired a connection slot.

            if(szUri.empty())
            {
                szUri = VU.GetCurrentUri();
            }

	        BOOL bSecureRequest = VU.GetSecureModeUsed();
	        if(bSecureRequest)
            {
                pConnection->SetSecurityMode(CAsyncHTTPRequest::HTTP_SECURE);

                // Check if a certificate name is provided.
                string szCertificateName;
                {
            		szCertificateName = _certName->Evaluate(ctxt);
                }
                pConnection->SetCertificateName(szCertificateName);
            }
            else
            {
                pConnection->SetSecurityMode(CAsyncHTTPRequest::HTTP_NOT_SECURE);
            }

            if(bSecureRequest == TRUE)
            {
                // Perform a SSL connect.

                string szRemoteServer = VU.GetRemoteServerName();
                string szPort = VU.GetRemoteServerPort();

                assert(szRemoteServer.empty() == FALSE);
                assert(szPort.empty() == FALSE);

                // Perform SSL connection if required.
                if(pConnection->IsConnected() == FALSE)
                {
                    VU.SetTRecordCallStep(4);   // In case Connect returns TOF_RES_ASYNC_STARTED, will enter the "case 4" section 
                                                // next time execute is called.
                    DWORD dwConnRes = pConnection->Connect(szRemoteServer.c_str(), atoi(szPort.c_str()), (IocpEvent*) pConnection);
				    switch(dwConnRes)
				    {
				    case WSAEISCONN:
				    case TOF_RES_OK:
                        // Will move on to case 4.
					    break;

				    case TOF_RES_ASYNC_STARTED:
                        {
					        return TOF_RES_ASYNC_STARTED;
                        }

				    default:
					    {
						    // If error
						    ctxt.set_phase_text(HTTP_NONE);
						    // Log client error
                            if(VU.GetTrace() && VU.GetTrace()->IsTrace("errors"))
                            {
                                GetTraceStream(VU.GetTrace()) << "ERROR: Connection to " << szRemoteServer << " on port " << szPort << " failed.\r\n";
                                ReleaseTraceStream();
                            }

                            if(ctxt.GetVerbose(ERRMSG))
    			    	    	ctxt.log("ERROR: Connection to " + szRemoteServer + " on port " + szPort + " failed.", ERRMSG);      

						    VU.ReleaseRequest(pConnection->GetRequestNumber());

    					    return TOF_RES_FAILED;
					    }
				    };
                }
            }
            // else don't need to connect in SSL.
        }
        // No break since should continue if we're performing a standard TCP connection.

	case 4:
        {
            // (SSL) Connection is over. Execute the request.

            nConid = VU.GetCurrentConid();

            // --- Get the variables from the tof.
			ERequestType    eType   = (ERequestType)intPrmVal(TParam::stType, ctxt);
            int		        nOptions= intPrmVal(TParam::stOptions, ctxt);

            string szBody;
            {
                szBody = _bodyP->Evaluate(ctxt);
            }

            if(szUri.empty())
                szUri = VU.GetCurrentUri();

            if(ctxt.GetVerbose(TRACEMSG))
                ctxt.log("REQUEST(" + int_to_string(nConid) + "): "	+ getRequestString(eType) + szUri, TRACEMSG);

	        // ------------------------------
            // Format headers

            if(pConnection == NULL)
            {
                // Get a (new) request slot to work with.
                pConnection = VU.ObtainRequestStruct(nConid);

                if(pConnection == NULL)
                {
                    // All the request slots are already used, and was not able to allocate a new one. Fatal error.
                    if(VU.GetTrace() && VU.GetTrace()->IsTrace("request errors"))
                    {
                        GetTraceStream(VU.GetTrace()) << "TRcrdRequest: could not allocate new connection.\r\n";
                        ReleaseTraceStream();
                    }

                    VU.LogError("ERROR for VU " + VU.GetThreadID() + ": failed to allocate a new connection.");
                    return TOF_RES_FAILED;
                }
            }
            // else moved directly from case 2 to case 3. Thus, already acquired a connection slot.

	        string szHeader = VU.GetHeaderText();
            assert(szHeader.empty() == FALSE);

			string newBody;
            if(eType == HTTP_REQUEST_POST)
            {
				if (szBody.length())
				{
					Ascii2Binary(szBody, newBody);
				}

                // In case this is a POST, must make sure the optional parameter "Content-Length" is fine.
                DWORD dwStartPos = strinfind(szHeader.data(), szHeader.size()-1, "content-length", sizeof "content-length" -1);
                if(dwStartPos != ULONG_MAX)
                {
                    // Replace the size value.
                    dwStartPos = szHeader.find_first_not_of(": \t", dwStartPos+sizeof("content-length")-1);
                    DWORD dwEndPos = szHeader.find_first_of("\r\n \t", dwStartPos);

                    CHAR lenStr [20];
                    sprintf(lenStr, "%d", newBody.size());

                    szHeader.replace(dwStartPos, dwEndPos-dwStartPos, lenStr);
                }
                else
                {
                    // Add the right header field.
		            char buf[30];
		            sprintf(buf, "Content-Length: %d\r\n", newBody.size());
                    string::size_type CRLF_CRLFpos = szHeader.rfind("\r\n\r\n");
                    if(CRLF_CRLFpos != string::npos)
                        szHeader.insert(CRLF_CRLFpos + 2, buf);
                    else
		                szHeader.append(buf + string("\r\n"));
                }
            }
            else
                newBody = szBody;


            // ------------------------------
            // Perform the request.

            if(szUri.find("HTTP/1.0") != string::npos 
                && pConnection->GetSecurityMode() != CAsyncHTTPRequest::HTTP_SECURE
                && pConnection->GetUseProxy() == FALSE)
            {
                // Should expect to re-connect before executing the next request.
                pConnection->SetConnectionType(CAsyncHTTPRequest::CNX_DISCONNECTS);
            }
            else
            {
                // HTTP/1.1 as well as Secure mode usually uses Keep-alive.
                pConnection->SetConnectionType(CAsyncHTTPRequest::CNX_KEEP_ALIVE);
            }

            pConnection->GetStats().SetUrl(getRequestString(eType) + szUri);

            if(pConnection->GetSecurityMode() == CAsyncHTTPRequest::HTTP_SECURE || pConnection->GetUseProxy() == FALSE)
            {
                // Remove the http://server_name part from szUri if it exists.
                string::size_type nPos = szUri.find("http");
                if(nPos != string::npos)
                {
                    nPos += sizeof "http"-1;
                    if(szUri[nPos] == 's')
                        nPos++;
                    // Skip server_name.
                    nPos = szUri.find_first_of('/', nPos+3);    // Skip the '://' sequence.
                    if(nPos != string::npos)
                    {
                        szUri.erase(0, nPos);
                    }
                }
            }

            string szRequestText; szRequestText.reserve(512);
            szRequestText = getRequestString(eType) + szUri + "\r\n" + szHeader + newBody;
            pConnection->SetRequestText(szRequestText);

            if(VU.GetTrace() && VU.GetTrace()->IsTrace("request"))
            {
                GetTraceStream(VU.GetTrace()) << "TRcrdRequest: Executing (size is "
                    << szRequestText.size() << "):\r\n" << szRequestText << ".\r\n"
                    << ((_eResultsRequired==NO_RESULT)?
                        "Won't parse results.\r\n"
                        :(
                            (_eResultsRequired==HEADER_ONLY)?
                                "Will parse header only.\r\n"
                                :"Will parse the body.\r\n"
                         )
                       );
                ReleaseTraceStream();
            }

            pConnection->SetTimerName(strPrmVal(TParam::stTimer, ctxt));
	        if(!(pConnection->GetTimerName().empty()))
		        ctxt.StartTimer(pConnection->GetTimerName());

            DWORD dwReqRes;

            switch(eType)
            {
            case HTTP_REQUEST_GET:  // GET
                dwReqRes = pConnection->InitiateGET(_eResultsRequired);
                break;

            case HTTP_REQUEST_HEAD: // HEAD
                dwReqRes = pConnection->InitiateHEAD(_eResultsRequired);
                break;

            case HTTP_REQUEST_POST: // POST
                dwReqRes = pConnection->InitiatePOST(_eResultsRequired);
                break;

            default:
                // ERROR. Unsupported request type!?!
                assert(0);
                return TOF_RES_FAILED;
                break;
            };

            if(dwReqRes != 0)
            {
                if(ctxt.GetVerbose(TRACEMSG))
                {
                    LPVOID lpMsgBuf;
                    if(FormatMessage( 
                        FORMAT_MESSAGE_ALLOCATE_BUFFER | 
                        FORMAT_MESSAGE_FROM_SYSTEM | 
                        FORMAT_MESSAGE_IGNORE_INSERTS,
                        NULL,
                        dwReqRes,
                        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
                        (LPTSTR) &lpMsgBuf,
                        0,
                        NULL 
                    ) == 0)
                    {
                        char buff[30];
                        sprintf(buff, "%#lX", dwReqRes);

                        ctxt.log("ERROR: REQUEST(" + int_to_string(nConid) + "): " + getRequestString(eType) + szUri + " failed with error " + buff, TRACEMSG);
                        VU.LogError("ERROR for VU " + VU.GetThreadID() + ": REQUEST(" + int_to_string(nConid) + "): " + getRequestString(eType) + szUri + " failed with error " + buff);
                    }
                    else
                    {
                        ctxt.log("ERROR: REQUEST(" + int_to_string(nConid) + "): " + getRequestString(eType) + szUri + " failed with error " + (LPCTSTR)lpMsgBuf, TRACEMSG);
                        VU.LogError("ERROR for VU " + VU.GetThreadID() + ": REQUEST(" + int_to_string(nConid) + "): " + getRequestString(eType) + szUri + " failed with error " + (LPCTSTR)lpMsgBuf);
                        LocalFree(lpMsgBuf);
                    }
                }

                return TOF_RES_FAILED;
            }

            // Eventually waits for the end of the request if the result status of the request is required before executing further.
            // Used if the error code of the request is required.
            szStatus    = strPrmVal(TParam::stRetSts, ctxt);

            if(_retCode || szStatus.empty() == FALSE)
            {
                // Synchronise on that request.
                VU.SetTRecordCallStep(5);   // If SynchroniseHTTPClient returns FALSE then will enter the "case 2" 
                                            // section next time execute is called.
                if(VU.SynchroniseHTTPClient(nConid) == FALSE)
                {
                    // Download on that connection is not over yet. Cannot continue.
                    return TOF_RES_ASYNC_STARTED;
                }
            }
            else
            {
                // No need to wait.
                VU.SetTRecordCallStep(0);   // Reset call phase for the next call to an async TRecord::execute.
                ctxt.set_phase_text(HTTP_NONE);

                return TOF_RES_OK;
            }
        }
//        break;    No break since may continue when information about the results is required.

    case 5:
        {
            // Request over. Parse the retcode and status values. Only when asked to.

            nConid = VU.GetCurrentConid();

            if(szStatus.empty())
                szStatus = strPrmVal(TParam::stRetSts, ctxt);

            // If some variables names are set, then request must compute the values.
            if(!(szStatus.empty() && _retCode == NULL))
            {
                if(pConnection == NULL)
                {
                    pConnection = VU.ObtainRequestStruct(nConid);

                    if(pConnection == NULL)
                    {
                        // All the request slots are already used, and was not able to allocate a new one. Fatal error.
                        if(VU.GetTrace() && VU.GetTrace()->IsTrace("request errors"))
                        {
                            GetTraceStream(VU.GetTrace()) << "TRcrdRequest: could not allocate new connection.\r\n";
                            ReleaseTraceStream();
                        }

                        VU.LogError("ERROR for VU " + VU.GetThreadID() + ": failed to allocate a new connection.");
                        return TOF_RES_FAILED;
                    }
                }

                CStatistics &stats = pConnection->GetStats();
	            if(_retCode)
                {
                    VU.set_variable(_retCode->Varnam(),
                        int_to_string((int)stats.GetHttpReplyCode()),
                        _retCode->Evaluate(ctxt));
                }

                if(szStatus.empty()   == FALSE)
                {
                    if(stats.GetHttpReplyCode() < 400)
                        VU.set_variable(szStatus, int_to_string(1));  // OK.
                    else
                        VU.set_variable(szStatus, int_to_string(0));  // Error.
                }
            }

            VU.SetTRecordCallStep(0);   // Reset call phase for the next call to an async TRecord::execute.
            ctxt.set_phase_text(HTTP_NONE);

            return TOF_RES_OK;
        }
        break;

    default:
        {
            nConid = VU.GetCurrentConid();

            ctxt.set_phase_text(HTTP_NONE);
            if(VU.GetTrace() && VU.GetTrace()->IsTrace("request errors"))
            {
                GetTraceStream(VU.GetTrace()) << "ERROR: entered default case in TRcrdRequest for connection " << nConid << ".\r\n";
                ReleaseTraceStream();
            }
            assert(0);

            return TOF_RES_FAILED;
        }
    };
}


void TRcrdRequest::BuildHeaderString(TContext& ctxt, string &szFormattedHeader) const
{
    list<string> headerList;
    list<string> headerWithList;
    list<string> headerWithoutList;

    {
        _headerP->  EvaluateValues(ctxt, headerList);
        _withP->    EvaluateValues(ctxt, headerWithList);
	    _woutP->    EvaluateValues(ctxt, headerWithoutList);
    }

	string szWithHeader;
	string szWoutHeader;
    szWithHeader.reserve(255);
    szWoutHeader.reserve(255);

	list<string>::iterator j;
	for(j = headerList.begin(); j != headerList.end(); j++) 
	{
        string &str = *j;
        if(str[str.size()-1] == '\n')
            szFormattedHeader.append(str.data(), str.size()-1); // Remove the \n.
        else
            szFormattedHeader.append(str);
		szFormattedHeader.append("\r\n");
	}

	for(j = headerWithList.begin(); j != headerWithList.end(); j++) 
	{
        string &str = *j;
        if(str[str.size()-1] == '\n')
            szWithHeader.append(str.data(), str.size()-1); // Remove the \n.
        else
            szWithHeader.append(str);
		szWithHeader.append("\r\n");
	}

	for(j = headerWithoutList.begin(); j != headerWithoutList.end(); j++) 
	{
        string &str = *j;
        if(str[str.size()-1] == '\n')
            szWoutHeader.append(str.data(), str.size()-1); // Remove the \n.
        else
            szWoutHeader.append(str);
		szWoutHeader.append("\r\n");
	}

	string_lf2crlf(szFormattedHeader);
	string_lf2crlf(szWithHeader);
	string_lf2crlf(szWoutHeader);

    // Remove the "with" header field
	string::size_type offset = 0;
	string::size_type endLine = 0;
	string::size_type beginField = 0;
	string::size_type endField = 0;

	while ( (offset != string::npos) && ((endField = szWithHeader.find_first_of(" :", offset)) != string::npos) )
	{
		beginField = find_in_string(szFormattedHeader, szWithHeader.substr(offset, endField - offset), false, 0);
		endLine = szFormattedHeader.find("\r\n", beginField);
		if ( (beginField != string::npos) && (endLine != string::npos) )
			szFormattedHeader.erase(beginField, endLine - beginField + 2);
		offset = szWithHeader.find("\r\n", offset);
		offset += 2;
	}

	// Add the "with" header fields
    szFormattedHeader.append(szWithHeader);

	// And remove the "wout" ones
	offset = endLine = beginField = endField = 0;
	while ( (offset != string::npos) && ((endField = szWoutHeader.find_first_of(" \r\n", offset)) != string::npos) )
	{
		beginField = find_in_string(szFormattedHeader, szWoutHeader.substr(offset, endField - offset), false, 0);
		endLine = szFormattedHeader.find("\r\n", beginField);
		if ( (beginField != string::npos) && (endLine != string::npos) )
			szFormattedHeader.erase(beginField, endLine - beginField + 2);
		offset = szWoutHeader.find("\r\n", offset);
		offset += 2;
	}

	// Add an empty line
    szFormattedHeader.append("\r\n");
}

// Build the header string, then parse the server and port strings.
// In HTTP/1.1, should parse the 'Host' header field. Otherwise, should parse the URI string.
// szPort will be empty if no explicit communication port was found.
void TRcrdRequest::ExtractHostPort(TContext& ctxt, const string &szHeader, const string &szUri, string &szRemoteServer, string &szPort) const
{
    // Find the HTTP://... format.
    DWORD dwPos = szUri.find("://");
    assert(dwPos != string::npos);
    dwPos += 3;

    // Found the host name. Search for the port.
    DWORD dwCurs = dwPos;
    while(szUri[dwCurs] != ':' && szUri[dwCurs] != '\r' && szUri[dwCurs] != '\0' && szUri[dwCurs] != '/')
        dwCurs++;

    szRemoteServer.assign(&szUri[dwPos], dwCurs-dwPos);

    if(szUri[dwCurs] == ':')
    {
        // Read the communication port value.
        dwPos = ++dwCurs;
        while(szUri[dwCurs] != '/' && szUri[dwCurs] != '\r' && szUri[dwCurs] != '\0')
            dwCurs++;
        szPort.assign(&szUri[dwPos], dwCurs-dwPos);
    }
    else
    {
        // No explicit port. Return an empty string.
        szPort.erase();
    }
}

//
// Ascii2Binary
// ============
//
// Convert an ascii body to its binary value or generate a body with given size
//
int TRcrdRequest::Ascii2Binary(const string &inBuf, string &outBuf) const
{
	outBuf.reserve(inBuf.size());

    // if inBuf begin with "binary("
    if (begins_with(inBuf, "BINARY("))
    {
        // extract the body length
        int beginLen = strlen("binary(");
        int endLen = inBuf.find_last_not_of(')', beginLen);
        string szLen = inBuf.substr(beginLen, endLen - beginLen);
        int length = atoi(szLen.c_str());
        outBuf.assign(length, 'A');
        return 0;
    }

    // if inBuf does not begin with "0x", its not a binary body
    if (!begins_with(inBuf, "0X"))
    {
        outBuf = inBuf;
        return 0;
    }

    int bufLen = inBuf.size();

    char tmp[2];
    int hexcode;
    for (int i = 2; i < bufLen; i += 2)
    {
        tmp[0] = inBuf[i];
        tmp[1] = inBuf[i+1];
        sscanf(tmp, "%x", &hexcode);
        outBuf.append(1, (char)hexcode);
    }

    return 0;
}

//
// TRcrdDisconnect::prepare
// ========================
//
// perform any preparation before disconnecting
int TRcrdDisconnect::prepare(TScript* script, int objIndex, bool verbose) 
{
    return 0;
}

//
// TRcrdDisconnect::execute
// ========================
//
// execute the disconnect

int TRcrdDisconnect::execute(TContext& ctxt)
{
    CAsyncVirtualUser& VU = dynamic_cast<CAsyncVirtualUser&> (ctxt);

    unsigned nConid = intPrmVal(TParam::stConnid, ctxt);

    if(VU.GetTrace() && VU.GetTrace()->IsTrace("full"))
    {
        GetTraceStream(VU.GetTrace()) << "TRcrdDisconnect: entering step " << VU.GetTRecordCallStep() << ".\r\n";
        ReleaseTraceStream();
    }

    switch(VU.GetTRecordCallStep())
    {
    case 0: // Must synchronize on the pending IOs.
        {
            ctxt.set_phase_text(HTTP_DISCONNECT);
	        unsigned nCancel = intPrmVal(TParam::stOptions, ctxt);

            if(ctxt.GetVerbose(TRACEMSG))
            {
                if(nCancel == 0)
                    ctxt.log("DISCONNECT(" + int_to_string(nConid) + ')', TRACEMSG);
                else
                    ctxt.log("DISCONNECT(" + int_to_string(nConid) + "), cancelling the current request(s)", TRACEMSG);
            }

            VU.SetTRecordCallStep(1);   // If SynchroniseHTTPClient returns FALSE then will enter the "case 1" 
                                        // section next time execute is called.
            if(nCancel == 0)
            {
                // Should not cancel the request(s).
                if(nConid == 0)
                {
                    // Synchronise on all clients.
                    if(VU.GetTrace() && VU.GetTrace()->IsTrace("request"))
                    {
                        GetTraceStream(VU.GetTrace()) << "TRcrdDisconnect: Synchronising on all connections for " << VU.GetManagerNumber() << ".\r\n";
                        ReleaseTraceStream();
                    }

                    if(VU.SynchroniseAllHTTPClients() == FALSE)
                    {
                        return TOF_RES_ASYNC_STARTED;
                    }
                    else
                    {
                        // No pending requests.
                    }
                }
                else
                {
                    // Synchronise on that specific client.
                    if(VU.GetTrace() && VU.GetTrace()->IsTrace("request"))
                    {
                        GetTraceStream(VU.GetTrace()) << "TRcrdDisconnect: Synchronising on " << VU.GetManagerNumber() << '/' << nConid << ".\r\n";
                        ReleaseTraceStream();
                    }

                    if(VU.SynchroniseHTTPClient(nConid) == FALSE)
                    {
                        return TOF_RES_ASYNC_STARTED;
                    }
                }
            }
            else
            {
                // Should finish with the pending IO but not receive all the data.
                if(nConid == 0)
                {
                    // Cancel all queries.
                    if(VU.GetTrace() && VU.GetTrace()->IsTrace("request"))
                    {
                        GetTraceStream(VU.GetTrace()) << "TRcrdDisconnect: Cancelling all the requests on " << VU.GetManagerNumber() << ".\r\n";
                        ReleaseTraceStream();
                    }

                    VU.CancelAllQueries();

                    // Synchronise on all clients.
                    if(VU.SynchroniseAllHTTPClients() == FALSE)
                    {
                        return TOF_RES_ASYNC_STARTED;
                    }
                }
                else
                {
                    // Cancel that query.
                    if(VU.GetTrace() && VU.GetTrace()->IsTrace("request"))
                    {
                        GetTraceStream(VU.GetTrace()) << "TRcrdDisconnect: Cancelling request " << VU.GetManagerNumber() << '/' << nConid << ".\r\n";
                        ReleaseTraceStream();
                    }

                    VU.CancelQuery(nConid);

                    // Synchronise on that request.
                    if(VU.SynchroniseHTTPClient(nConid) == FALSE)
                    {
                        return TOF_RES_ASYNC_STARTED;
                    }
                }
            }
        }
        // No break since should walk through case 1.

    case 1: // The client(s) is (are) over.
        {
            DWORD dwRes;

            VU.SetTRecordCallStep(2);   // In the case of secure connection, disconnect is a two steps process.
            if(nConid == 0)
            {
                // Disconnect all CAsyncHTTPRequest s.
                dwRes = VU.DisconnectAllClients();
            }
            else
            {
                // Disconnect that CAsyncHTTPRequest.
                dwRes = VU.DisconnectClient(nConid);
            }

            if(dwRes != 0)
            {
                if(dwRes == TOF_RES_ASYNC_STARTED)
                {
                    if(VU.GetTrace() && VU.GetTrace()->IsTrace("connection"))
                    {
                        GetTraceStream(VU.GetTrace()) << "TRcrdDisconnect: Waiting for an async disconnect to complete.\r\n";
                        ReleaseTraceStream();
                    }
                    return TOF_RES_ASYNC_STARTED;
                }
                else
                {
                    if(VU.GetTrace() && VU.GetTrace()->IsTrace("errors"))
                    {
                        GetTraceStream(VU.GetTrace()) << "TRcrdDisconnect: could not disconnect connection " << nConid << ".\r\n";
                        ReleaseTraceStream();
                    }

                    if(ctxt.GetVerbose(ERRMSG))
                        ctxt.log("ERROR DISCONNECTING (" + int_to_string(nConid) + ')', ERRMSG);

                    return TOF_RES_OK;
                    // return TOF_RES_FAILED;   This is not considered as a relevant error.
                }
            }

            break;  // We're over.
        }

    case 2: // The client disconnection is over.
        {
            // Nothing to do.
            if(ctxt.GetVerbose(TRACEMSG))
                ctxt.log("DISCONNECTED (" + int_to_string(nConid) + ')', TRACEMSG);

            break;
        }

    default:
        assert(0);
        break;
    };

    ctxt.set_phase_text(HTTP_NONE);
    VU.SetTRecordCallStep(0);   // Reset call phase for the next call to an async TRcrdDisconnect::execute.

    return TOF_RES_OK;
}

//
// TRcrdHttpResponse::prepare
// ========================
//
// perform any preparation before reading the response
int TRcrdHttpResponse::prepare(TScript* script, int objIndex, bool verbose)
{
	const TParam* prmP = param(TParam::stData);
	_responseType = prmP->intVal();

	prmP = param(TParam::stConnid);
    string l_conid = prmP->strVal();

	_iConid = atoi(l_conid.c_str());
    for (string::size_type i=0; i<l_conid.size(); i++)
	{
		if (!isdigit(l_conid[i]))
		{
			_iConid = -1 ;
			break ;
		}
	}

    // Get variable string and expression string from 'name' parameters.
    prmP = param(TParam::stInto);
    string l_name = prmP->strVal();
	
    // Create variable reference object, for variable to be set.
    _varP = new TtVarRef(script);
	
    // Encode the variable reference; report an error if encode fails.
    if (!(_varP->Parse(l_name, 0)))
    {
		MSGERR("SET VARIABLE: " + l_name + " Failed during 'into' PREPARE with: " + _varP->Reason());
		return -1;
    }
	
    // Get pointer to TRecord for variable - as we need information about the variable
    // to process the expression string.
    TRecord* trecP = (*script).GetVardef(_varP->Varnam());
    if (trecP == NULL) // If not a defined variable
    {
		MSGERR("SET VARIABLE: " + l_name + " Failed during PREPARE with: Variable '" + _varP->Varnam() + "' not defined");
		return -1;
    }

    // Get variable options mask so that we can check the variable type and so
    // determine what type of expression we need (integer or character).
    prmP = trecP->param(TParam::stOptions);
    if (prmP == NULL) // If no 'options'  parameter (should always be present).
    {
		MSGERR("SET VARIABLE: " + l_name + " Failed during PREPARE with: Variable options missing");
		return -1;
    }

    int options = prmP->intVal();   // Variable options mask

    // Create an expression object (of the appropriate type)
    if (_exprP) delete _exprP;	// Delete any existing object first
	

    if (options & VTCHAR)
    {
		_exprP = new TtChrExpr(script);
    }
    else if (options & VTINT)
    {
        _exprP = new TtIntExpr(script);
    }
    else // Inrecognised variable type.
    {
		MSGERR("SET VARIABLE: " + l_name + " Failed during PREPARE with: Unrecognised variable type");
		return -1;
    }

    // Keep the variable size
    prmP = trecP->param(TParam::stSize);
    if(prmP == FALSE)
    {
        // TODO: ERR
        return -1;
    }
    _varSize = prmP->intVal();   // Variable length.

	prmP = param("identifier");
    if(prmP == NULL) 
		return -1;
    string l_ident = prmP->strVal();
    _identP = new TtChrExpr(script);

    // Encode the variable reference; report an error if encode fails.
    if (!(_identP->Parse(l_ident, 0)))
    {
		MSGERR("Connect : " + l_ident +
			   " Failed during 'identifier' PREPARE with: " + _identP->Reason());
		return -1;
    }

	if (_responseType == HEADER_VAR)
    {
		unsigned int newIndex = objIndex;
		while (newIndex > 0)
		{
			newIndex--;
			TRecord * prevRecord = script->rec_at_index(newIndex);
			if (prevRecord && prevRecord->isType(TRcrdRequest::KEY))
			{
				TRcrdRequest *prevRequestRecord = (TRcrdRequest *)prevRecord;
				if ( (prevRequestRecord->_iConid == _iConid) || (prevRequestRecord->_iConid == 0) )
				{
                    // Eventually upgrade the store level to HEADER_ONLY.
                    if(prevRequestRecord->_eResultsRequired == NO_RESULT)
                        prevRequestRecord->_eResultsRequired = HEADER_ONLY;
					break;
				}
			}
		}
		return 0 ;
    }

	// if the conid is -1, build a new map.
	if (_iConid == -1)
	{
		_varAddressMap[l_name] = _identP ;
		_identP = NULL ;
		return 0 ;
	}
	else
	{
		int newIndex = objIndex;
		while (newIndex > 0)
		{
			newIndex--;
			TRecord * prevRecord = script->rec_at_index(newIndex) ;
			if (prevRecord && prevRecord->isType(TRcrdRequest::KEY))
			{
				TRcrdRequest *prevRequestRecord = (TRcrdRequest *)prevRecord;
				if ( (prevRequestRecord->_iConid == _iConid) || (prevRequestRecord->_iConid == 0) )
				{
                    // Upgrade the store level to ALL_DATA.
                    prevRequestRecord->_eResultsRequired = ALL_DATA;
					break;
				}
			}
		}

		// Try to find a previous Response Record with the same conid 
		// and the same responseType
		while (objIndex > 0)
		{
			objIndex -- ;
			TRecord * prevRecord = script->rec_at_index(objIndex) ;
			// check that the previous record is also a TRcrdHttpResponse Record
			// with type==BODY
			if (prevRecord->isType(KEY))
			{
				TRcrdHttpResponse * prevResponseRecord = (TRcrdHttpResponse *)prevRecord ; 
				if (prevResponseRecord->_iConid == _iConid && 
					prevResponseRecord->_responseType == _responseType)
				{
					_addressQueryList = prevResponseRecord->_addressQueryList ;
					_varAddressMap = prevResponseRecord->_varAddressMap ;
					prevResponseRecord->_addressQueryList.clear() ;
					prevResponseRecord->_varAddressMap.clear() ;
					break ;
				}
			}
			else 
				break ;
		}
	}

	// If the address begins by " and finishes by " and 
	// there are no other " in the address, remove the "
	// else has to evaluate the address in execute later
	string address = l_ident ;
	if (!address.empty() && address[0] == '"' && address.find("\"", 1) == address.size() - 1)
		address = address.substr(1, address.size() - 2);
	else
	{
		_varAddressMap[l_name] = _identP ;
		_identP = NULL ;
		return 0 ;
	}

	if (!address.empty())
	{
		AddressQuery * aq = new AddressQuery(l_name, address.c_str());
		// Normally, it should be a valid address, if not, will be evaluated in
		// the next
		if (aq->IsValidAddress())
		{
			_addressQueryList.push_back(aq) ;
			delete _identP ;
			_identP = NULL ;
		}
		else
		{
			delete aq ;
			_varAddressMap[l_name] = _identP ;
			_identP = NULL ;
			return 0 ;
		}
	}
	else
	{
		_varAddressMap[l_name] = _identP ;
		_identP = NULL ;
	}
    return 0;
}



//
// TRcrdHttpResponse::execute
// ========================
//
// Wait for the response and parse the variables values.

template <class ACCESSOR_TYPE>
BOOL ParseBodyBuffers(CDataBuffHandler &messageBuffer, BOOL bContainContinueHeader, const AddressQueryList &aql, PhParser<ACCESSOR_TYPE>::AddressValueList &avl, unsigned long* pulErrorLine, unsigned long* pulErrorCol);

template <class ACCESSOR_TYPE>
BOOL StoreResults(TContext &ctxt, PhParser<ACCESSOR_TYPE>::AddressValueList &avl);

int TRcrdHttpResponse::execute(TContext& ctxt)
{
    int nType = intPrmVal(TParam::stData, ctxt);  // Location of the data: HEADER or BODY.

    if(nType == BODY_VAR && _addressQueryList.empty() && _varAddressMap.empty())
    {
        // Nothing to parse. Lload instructions were concatenated into a single one. Parsing will be performed in one of the next Lloads.
		return TOF_RES_OK;
    }

    CAsyncVirtualUser& VU = dynamic_cast<CAsyncVirtualUser&> (ctxt);

    int nConid;

    if(VU.GetTrace() && VU.GetTrace()->IsTrace("full"))
    {
        GetTraceStream(VU.GetTrace()) << "TRcrdHttpResponse: entering step " << VU.GetTRecordCallStep() << ".\r\n";
        ReleaseTraceStream();
    }

    switch(VU.GetTRecordCallStep())
    {
    case 0: // Must synchronize on the pending IOs.
        {
            ctxt.set_phase_text(HTTP_RESPONSE);

            nConid = intPrmVal(TParam::stConnid, ctxt);
            VU.SetCurrentConid(nConid);

            // Synchronise on that request.
            VU.SetTRecordCallStep(1);   // In the case of secure connection, disconnect is a two steps process.
            if(VU.SynchroniseHTTPClient(nConid) == FALSE)
            {
                // Download on that connection is not over yet. Cannot continue.
                return TOF_RES_ASYNC_STARTED;
            }
        }
        // No break since should continue if no need to synchronize.

    case 1:
        {
            string szVarVal;
            nConid = VU.GetCurrentConid();

            CAsyncHTTPRequest* pConnection = VU.ObtainRequestStruct(nConid);
            if(pConnection == NULL)
            {
                if(VU.GetTrace() && VU.GetTrace()->IsTrace("errors"))
                {
                    GetTraceStream(VU.GetTrace()) << "TRcrdHttpResponse: Invalid connection id(" << nConid << ").\r\n";
                    ReleaseTraceStream();
                }

                if(ctxt.GetVerbose(ERRMSG))
                    ctxt.log("HTTPRESPONSE: Invalid connection id(" + int_to_string(nConid) + ')', ERRMSG);
                VU.LogError("ERROR for VU " + VU.GetThreadID() + ": Invalid connection id(" + int_to_string(nConid) + ')');
                return TOF_RES_FAILED;
	        }

            CDataBuffHandler &messageBuff = pConnection->GetRequestResults();

            if(messageBuff.IsEmpty() == TRUE)
            {
                // ERROR: the data is not available (connection was broken inbetween ...).
                if(VU.GetTrace() && VU.GetTrace()->IsTrace("errors"))
                {
                    GetTraceStream(VU.GetTrace()) << "TRcrdHttpResponse: No data available for connection id(" << nConid << ").\r\n";
                    ReleaseTraceStream();
                }

                if(ctxt.GetVerbose(ERRMSG))
                    ctxt.log("HTTPRESPONSE: No data available for connection id(" + int_to_string(nConid) + ')', ERRMSG);
                VU.LogError("ERROR for VU " + VU.GetThreadID() + ": No data available for connection id(" + int_to_string(nConid) + ')');
                return TOF_RES_FAILED;
            }

            BOOL bFoundContinueHeader = pConnection->FoundContinueHeader();

            if(nType == HEADER_VAR)
            {
		        string szName = strPrmVal(TParam::stInto, ctxt);
                string szIdent;
			    szIdent = _identP->Evaluate(ctxt);

                if(VU.GetTrace() && VU.GetTrace()->IsTrace("full"))
                {
                    GetTraceStream(VU.GetTrace()) << "TRcrdHttpResponse: (" << nConid << "): Parsing header contents. Name: " << szName << " Identity: " << szIdent << ".\r\n";
                    ReleaseTraceStream();
                }

                if(ctxt.GetVerbose(TRACEMSG))
                    ctxt.log("HTTPRESPONSE(" + int_to_string(nConid) + "): Parsing header contents. Name: " + szName + " Identity: " + szIdent, TRACEMSG);

                if(szIdent.empty() == TRUE)
                {
                    // Return up to _dwVarLength characters from the header in szVarVal.
                    DumpHeader(messageBuff, szVarVal, _varSize, bFoundContinueHeader);
                }
                else
                {
                    // Parse header/cookie value matching expression szIdent.
                    ParseHeaderValue(messageBuff, szIdent, bFoundContinueHeader, szVarVal);
                }

                if(ctxt.GetVerbose(TRACEMSG))
                    ctxt.log("HTTPRESPONSE(" + int_to_string(nConid) + "): Setting " + szName + " to \'" + szVarVal + '\'', TRACEMSG);

                // Store the header results.
				if(ctxt.set_variable(szName, szVarVal) == -1)
			    {
					// Error storing the results. Print statistics before exiting

                    string msg = "Set variable failed(" + szName + ')';

                    if(VU.GetTrace() && VU.GetTrace()->IsTrace("errors"))
                    {
                        GetTraceStream(VU.GetTrace()) << msg << ".\r\n";
                        ReleaseTraceStream();
                    }

                    ctxt.log(msg, ERRMSG); throw mdbParseException(msg);
			    }
            }
            else    // nType == BODY_VAR.
            {
                // Handle multiple-address parsing at a time.
	            TtChrExpr *pIdent;

	            AddressQueryList tmpAddressQueryList = _addressQueryList;
	            AddressQueryList newAddressQueryList;

                try
                {
	                // Evaluate the variable addresses.
                    // Addresses that need parsing will be added to newAddressQueryList; 'empty' addresses will be evaluated there.
	                for (variableAddressMap::iterator itr = _varAddressMap.begin(); itr != _varAddressMap.end(); itr ++)
	                {
		                pIdent = itr->second ;
		                if (pIdent)
		                {
			                string address(pIdent->Evaluate(ctxt));
			                if(address.empty())
			                {
    			                // The address is empty, get all the document.
                                if(VU.GetTrace() && VU.GetTrace()->IsTrace("full"))
                                {
                                    GetTraceStream(VU.GetTrace()) << "TRcrdHttpResponse: (" << nConid << "): Parsing header contents. Name: " << itr->first << " Identity: " << address << ").\r\n";
                                    ReleaseTraceStream();
                                }

                                if(ctxt.GetVerbose(TRACEMSG))
				                    ctxt.log("HTTPRESPONSE(" + int_to_string(nConid) + "): Name: " + itr->first + " Identity: " + address, TRACEMSG);

                                DumpBody(messageBuff, szVarVal, _varSize, bFoundContinueHeader);
				                if(ctxt.set_variable(itr->first, szVarVal) == -1)
				                {
					                // Error storing the results. Print statistics before exiting

                                    string msg = "Set variable failed(" + itr->first + ')';

                                    if(VU.GetTrace() && VU.GetTrace()->IsTrace("errors"))
                                    {
                                        GetTraceStream(VU.GetTrace()) << msg << ".\r\n";
                                        ReleaseTraceStream();
                                    }

                                    ctxt.log(msg, ERRMSG); throw mdbParseException(msg);
				                }
			                }
                            else
                            {
			                    // Need to use the parser. Put the addressQuery in the addressQueryList.
			                    AddressQuery *newAddressQuery  = new AddressQuery(itr->first, address.c_str());
			                    newAddressQueryList.push_back(newAddressQuery) ;
                            }
		                }
                        else
                        {
                            assert(0);
                        }
	                }

                    // Merge the static and variable address lists.
                    tmpAddressQueryList.insert(tmpAddressQueryList.end(), newAddressQueryList.begin(), newAddressQueryList.end());
                    if(!tmpAddressQueryList.empty())
                    {
                        BOOL bRes;
                        unsigned long ulErrorLine, ulErrorCol;

                        // Call the HTML parser on the address list.
                        if(pConnection->GetMessageLengthType() == CAsyncHTTPRequest::MSG_NON_IDENTITY)
                        {
                            // Data in chunk format.
	                        PhParser<CDataBuffListChunkAccessor>::AddressValueList avl;
                            bRes = ParseBodyBuffers<CDataBuffListChunkAccessor> (   pConnection->GetRequestResults(), 
                                                                                    pConnection->DoesContainContinueHeader(), 
                                                                                    tmpAddressQueryList,
                                                                                    avl,
                                                                                    &ulErrorLine,
                                                                                    &ulErrorCol);
                            if(bRes == TRUE)
                            {
                                bRes = StoreResults<CDataBuffListChunkAccessor> (ctxt, avl);
                            }
                            else
                            {
                                if(ctxt.GetVerbose(ERRMSG))
                                    ctxt.log("HTTPRESPONSE: HTML parser error for connection (" + int_to_string(nConid) + ") on line " + int_to_string((long)ulErrorLine) + ", column " + int_to_string((long)ulErrorCol), ERRMSG);
                                assert(0);
                            }
                        }
                        else
                        {
                            // Standard HTTP data.
	                        PhParser<CDataBuffListAccessor>::AddressValueList avl;
                            bRes = ParseBodyBuffers<CDataBuffListAccessor> (    pConnection->GetRequestResults(), 
                                                                                pConnection->DoesContainContinueHeader(), 
                                                                                tmpAddressQueryList, 
                                                                                avl,
                                                                                &ulErrorLine,
                                                                                &ulErrorCol);
                            if(bRes == TRUE)
                            {
                                bRes = StoreResults<CDataBuffListAccessor> (ctxt, avl);
                            }
                            else
                            {
                                if(ctxt.GetVerbose(ERRMSG))
                                    ctxt.log("HTTPRESPONSE: HTML parser error for connection (" + int_to_string(nConid) + ") on line " + int_to_string((long)ulErrorLine) + ", column " + int_to_string((long)ulErrorCol), ERRMSG);
                                assert(0);
                            }
                        }
                    }
                }
                catch(mdbParseException &)
                {
                    // Perform cleanup: delete the dynamicaly evaluated variables.
	                for (AddressQueryList::iterator newitr = newAddressQueryList.begin(); newitr != newAddressQueryList.end(); newitr ++)
		                delete *newitr;

                    throw;
                }
            }   // End of BODY_VAR section.

            break;
        }

    default:
        {
            assert(0);
        }
    };

    ctxt.set_phase_text(HTTP_NONE);

	return TOF_RES_OK;
}

void SkipHeader(CDataBuffListAccessor &iter, BOOL bFoundContinueHeader)
{
    // First skip the CONTINUE header if exists.
    if(bFoundContinueHeader)
    {
        do
        {
            iter.SkipCRLF();
        }
        while(iter.IsEndOfBuffs() == FALSE && iter.IsCRLF() == FALSE);
        iter.SkipCRLF();
    }

    // Skip the real header.
    do
    {
        iter.SkipCRLF();
    }
    while(iter.IsEndOfBuffs() == FALSE && iter.IsCRLF() == FALSE);
    iter.SkipCRLF();
}

// Return up to dwVarLength characters from the header or body (depending on nDataArea) in szVarVal.
// WARNING: in the case of headers, returns only the real header (skip the CONTINUE header if applicable).
void TRcrdHttpResponse::DumpHeader(CDataBuffHandler &messageBuff, string &szVarVal, DWORD dwVarLength, BOOL bFoundContinueHeader) const
{
    CDataBuffListAccessor iter (&messageBuff.GetWholeBuffList());
    CDataBuffListAccessor iterEnd (iter);

    if(bFoundContinueHeader)
    {
        SkipHeader(iterEnd, FALSE);
    }

    // Compute header length.
    SkipHeader(iterEnd, FALSE);
    DWORD dwHeaderLen = iterEnd-iter;

    iter.BuildString(szVarVal, __min(dwVarLength, dwHeaderLen));
}

void TRcrdHttpResponse::DumpBody(CDataBuffHandler &messageBuff, string &szVarVal, DWORD dwVarLength, BOOL bFoundContinueHeader) const
{
    CDataBuffListAccessor iter (&messageBuff.GetWholeBuffList());

    // Skip the header.
    SkipHeader(iter, bFoundContinueHeader);

    // Build the result.
    iter.BuildString(szVarVal, dwVarLength);
}

template <class ACCESSOR_TYPE>
BOOL ParseBodyBuffers(CDataBuffHandler &messageBuffer, BOOL bContainContinueHeader, const AddressQueryList &aql, PhParser<ACCESSOR_TYPE>::AddressValueList &avl, unsigned long* pulErrorLine, unsigned long* pulErrorCol)
{
    // Move to the begining of the body section (use generic iterator).
    CDataBuffListAccessor iterSkipHeader(&messageBuffer.GetWholeBuffList());
    SkipHeader(iterSkipHeader, bContainContinueHeader);

    // Parse. Use either chunk or regular iterator.
    ACCESSOR_TYPE iterData;
    iterData.MoveIterTo(iterSkipHeader);

    PhLexer<ACCESSOR_TYPE>  lexer   (iterData);
    PhParser<ACCESSOR_TYPE> parser  (lexer, TRUE);

    return parser.Parse(aql, avl, pulErrorLine, pulErrorCol);
}


template <class ACCESSOR_TYPE>
BOOL StoreResults(TContext &ctxt, PhParser<ACCESSOR_TYPE>::AddressValueList &avl)
{
    string szVarVal;

	// Associate each addressValue to each variable.
	for(PhParser<ACCESSOR_TYPE>::AddressValueList::iterator avlItr = avl.begin(); avlItr != avl.end(); avlItr ++)
	{
        // Build the request result.
        ACCESSOR_TYPE &iterResultBeg = avlItr->pcBegin_;

        if(iterResultBeg != NULL)
        {
            if(iterResultBeg.BuildString(szVarVal, avlItr->ulSize_) == FALSE)
            {
                assert(0);
                return FALSE;
            }
        }

		// Get the variable name associated with the AddressQuery.
		const AddressQuery * aq = avlItr->pAq_ ;
		string vName = aq->strVariable_ ;	

        if(ctxt.GetVerbose(TRACEMSG))
        	ctxt.log("HTTPRESPONSE: Name: " + vName + " Identity: " + (aq->address_).ToString() + " Value:" + szVarVal, TRACEMSG);

        // Store the results.
		int ret = ctxt.set_variable(vName, szVarVal);
		if(ret == -1)
		{
			// Error storing the results. Print statistics before exiting.
            string msg = "Set variable failed(" + vName + ')';
			ctxt.log(msg, ERRMSG); throw mdbParseException(msg);
		}
	}

    return TRUE;
}

//
// Find a cookie value.
// This function expects there is only one cookie per "Set-Cookie" line in the header
// (ie: GET ...
//      Set-Cookie: name1 = value1
//      Set-Cookie: name2 = value2
// )
//
BOOL TRcrdHttpResponse::ParseCookieValue(CDataBuffHandler &messageBuff, const string &zName, BOOL bFoundContinueHeader, string &zCookieVal) const
{
    zCookieVal.erase();

    // Use regular expressions to find the cookie name.
    string zModifName = '^' + zName;
    string::size_type i = 0;
    while ( (i = zModifName.find_first_of('.', i)) != string::npos )
    {
        zModifName.insert(i, 1, '\\');
        i += 2;
    }
    i = 0;
    while ( (i = zModifName.find_first_of('*', i)) != string::npos )
    {
        zModifName.insert(i, 1, '.');
        i += 2;
    }
    i = 0;
    while ( (i = zModifName.find_first_of('?', i)) != string::npos )
    {
        zModifName.replace(i, 1, 1, '.');
        i += 2;
    }

    // Add '=' at end of cookie's name to differentiate cookies like "foo"
    // from those like "foofoo"
    // TODO: is that possible that we find tabs or spaces before '='?
    zModifName += '=';

    CRegExp regExp;
    regExp.RegComp(zModifName.c_str());

    CDataBuffListAccessor iter (&messageBuff.GetWholeBuffList());

    // First skip the CONTINUE header if exists.
    if(bFoundContinueHeader == TRUE)
    {
        SkipHeader(iter, FALSE);    // Set FALSE to skip only the first header.
    }

    while(iter.IsEndOfBuffs() == FALSE && iter.IsCRLF() == FALSE)
    {
        // Iterate through the header lines.
        // Skip white spaces (that might not be necessary, but we never know ...).
        if(iter.SkipWS() == FALSE) { assert(0); return FALSE; /*Found nothing but W.S.*/ }

        if(iter.MatchFirstWordToLower("set-cookie", sizeof "set-cookie" - 1) == TRUE)
	    {
            iter.Skip(sizeof "set-cookie" - 1);

		    // Skip the column.
            if(iter.SkipChar(':') == FALSE) { assert(0); return FALSE; }

		    // Skip the LWS.
            if(iter.SkipWS() == FALSE)      { assert(0); return FALSE; }

            // Build the cookie name string (+ the equal sign) and check if it matches.
            string zFoundCookieName;
            CDataBuffListAccessor base = iter;

            if(iter.SkipChar('=') == FALSE)                                                 { assert(0); return FALSE; }

            base.BuildString(zFoundCookieName, iter-base);

            BOOL bFound = (regExp.RegFind(zFoundCookieName.c_str()) != -1);

		    // Skip the LWS.
            if(iter.SkipWS() == FALSE)                                                      { assert(0); return FALSE; }

            // Copy the value.
            if(bFound)
            {
                // Return the found value.
                base = iter;
                if(iter.SkipTillAny(";\r\n", sizeof ";\r\n" - 1) == FALSE)                  { assert(0); return FALSE; }
                base.BuildString(zCookieVal, iter-base);
                zCookieVal = zFoundCookieName + zCookieVal;
                return TRUE;
            }
            else
            {
                // Skip to next header line.
                if(iter.SkipCRLF() == FALSE)                                                { assert(0); return FALSE; }
            }
	    }
        else
        {
            // Not a "Set-Cookie" line. Move to next line.
            iter.SkipCRLF();
        }
    }   // Parsing header lines.

#ifdef __DLLSA__

    // If this cookie-name is not available, show the others Set-Cookie fields (only for unique replay).
    g_pLogTrace->AddTrace(TRCCAT_RUNNING, TRCLEV_LOW, "Cannot find cookie " + zName + ", must be one of the following:");

    CDataBuffListAccessor iterDumpCookies(&messageBuff.GetWholeBuffList());

    // First skip the CONTINUE header if exists.
    if(bFoundContinueHeader == TRUE)
    {
        SkipHeader(iterDumpCookies, FALSE); // Set FALSE to skip only the first header.
    }

    while(iterDumpCookies.IsEndOfBuffs() == FALSE && iterDumpCookies.IsCRLF() == FALSE)
    {
        // Iterate through the header lines.
        // Skip white spaces (that might not be necessary, but we never know ...).
        if(iterDumpCookies.SkipWS() == FALSE) { assert(0); /*Found nothing but W.S.*/ }
        if(iterDumpCookies.MatchFirstWordToLower("set-cookie", sizeof "set-cookie" - 1) == TRUE)
	    {
            // Add end of line to msg.
            iterDumpCookies.SkipChar(':');
            iterDumpCookies.SkipWS();
            DWORD dwEOL;
            if(iterDumpCookies.FindAny("\r\n;,", 4, dwEOL) == FALSE)  { assert(0); break; }
            string zCookie;
            iterDumpCookies.BuildString(zCookie, dwEOL);
            g_pLogTrace->AddTrace(TRCCAT_RUNNING, TRCLEV_LOW, zCookie);
        }

        iterDumpCookies.SkipCRLF();
    }
    g_pLogTrace->AddTrace(TRCCAT_RUNNING, TRCLEV_LOW, "-- end of the available cookies list --");

#endif  __DLLSA__

    zCookieVal = zName + '=';

    return FALSE;  // Not found.
}

// Browse through the header searching for a header-field definition.
void TRcrdHttpResponse::ParseHeaderValue(CDataBuffHandler &messageBuff, const string &zIdent, BOOL bFoundContinueHeader, string &zHeaderValue) const
{
    DWORD dwPos = strinfind(zIdent.c_str(), zIdent.size() -1, "set-cookie", sizeof "set-cookie"-1);
    if(dwPos != ULONG_MAX)
    {
        // Requested to parse a cookie data from messageBuff.

        // Evaluate the cookie name.
        string::size_type beginName = zIdent.find_first_not_of(", ", (string::size_type) dwPos + strlen("Set-Cookie"));

		if(beginName == string::npos)
        {
            assert(0);
            zHeaderValue.erase();
			return ;
        }

		string::size_type endName = zIdent.find_first_of(" \r\n", beginName);

		if(endName == string::npos)
			endName = zIdent.size();

		string zCookieName = zIdent.substr(beginName, endName - beginName);

        // Parse the cookie value.
        if(ParseCookieValue(messageBuff, zCookieName, bFoundContinueHeader, zHeaderValue) == FALSE)
        {
            // TODO: error.
        }
    }
    else
    {
        // Parse header value matching expression zIdent.
        CDataBuffListAccessor iter (&messageBuff.GetWholeBuffList());

        // First skip the CONTINUE header if exists.
        if(bFoundContinueHeader == TRUE)
        {
            SkipHeader(iter, FALSE);    // Set FALSE to skip only the first header.
        }

        // Iterate through the header lines.
        while(iter.IsCRLF() == FALSE && iter.IsEndOfBuffs() == FALSE)   // Not the end of the header section.
        {
            // Skip white spaces (that might not be necessary, but we never know ...).
            if(iter.SkipWS() == FALSE)  { assert(0); break; /* TODO: Found nothing but W.S. Error !!!*/ }

            if(iter.MatchFirstWordToLower(zIdent.c_str(), zIdent.size() - 1) == TRUE)
	        {
                // Found the right header field. Return the associated value.
                if(iter.Skip(zIdent.size()) == FALSE)           { assert(0); break; }
                if(iter.SkipChar(':') == FALSE)                 { assert(0); break; }
                if(iter.SkipWS() == FALSE)                      { assert(0); break; }

                CDataBuffListAccessor tmpIter = iter;

                if(tmpIter.SkipCRLF() == FALSE)                 { assert(0); break; }

                if(iter.BuildString(zHeaderValue, tmpIter-iter-2) == FALSE) { assert(0); /* (Should not happen) */ }

                return;
            }
            else
            {
                // Move to next header-field.
                if(iter.SkipCRLF() == FALSE)                    { assert(0); break; }
            }
        }

        zHeaderValue.erase();    // TODO : not found.
    }
}


//
// TRcrdSynchronize::prepare
// ========================
//
// prepare
int TRcrdSynchronize::prepare(TScript* script, int objIndex, bool verbose)
{
    const TParam* prmP = param(TParam::stTimeOut);
    if (prmP == NULL) 
		return 0;
    int tgtvar = prmP->intVal();

	if (tgtvar == 0)
	    return 0;
    _labelIndex = PrepareGoto(this, TParam::stTmolab, script, true);
    return 0;
}


//
// TRcrdSynchronize::execute
// ========================
//
// read the response on all connections
int TRcrdSynchronize::execute(TContext& ctxt)
{
    CAsyncVirtualUser& VU = dynamic_cast<CAsyncVirtualUser&> (ctxt);

    if(VU.GetTrace() && VU.GetTrace()->IsTrace("full"))
    {
        GetTraceStream(VU.GetTrace()) << "TRcrdSynchronize: entering step " << VU.GetTRecordCallStep() << ".\r\n";
        ReleaseTraceStream();
    }

    switch(VU.GetTRecordCallStep())
    {
    case 0:
        {
            if(ctxt.GetVerbose(TRACEMSG))
                ctxt.log("SYNCHRONIZE", TRACEMSG);

            ctxt.set_phase_text(HTTP_SYNCHRONIZE);

            int nTimeout    = intPrmVal(TParam::stTimeOut, ctxt);
            int nGoto       = intPrmVal(TParam::stTmolab, ctxt);

            // Synchronise on that VU's requests.
            VU.SetTRecordCallStep(1);
            if(nTimeout == 0)
            {
                // Either value not specified or should wait indefinitely.
                if(VU.SynchroniseAllHTTPClients() == FALSE)
                {
                    // Downloads on that VU are not over yet. Must wait.
                    return TOF_RES_ASYNC_STARTED;
                }
            }
            else
            {
                // ERROR: do not care for time-outs with Thread pool version.
                assert(0);
            }
        }
        // No break since should move to case 1 if all the users are already synchronized.

    case 1:
        {
            if(ctxt.GetVerbose(TRACEMSG))
                ctxt.log("All requests synchronised.", TRACEMSG);

            ctxt.set_phase_text(HTTP_NONE);

            break;
        }

    default:
        {
            assert(0);
        }
    };

    VU.SetTRecordCallStep(0);   // Reset call phase for the next call to an async TRecord::execute.

	return TOF_RES_OK;
}

int TRcrdWait::execute(TContext& ctxt)
{
    CAsyncVirtualUser& VU = dynamic_cast<CAsyncVirtualUser&> (ctxt);

    switch(VU.GetTRecordCallStep())
    {
    case 0:
        {
            ctxt.set_phase_text(PHASE_WAIT);

            unsigned nDelay = intPrmVal(TParam::stTDelay, ctxt);

            if(ctxt.GetVerbose(TRACEMSG))
                ctxt.log("WAIT : Delay: " + int_to_string(nDelay), TRACEMSG);

            if(VU.GetTrace() && VU.GetTrace()->IsTrace("sync"))
            {
                GetTraceStream(VU.GetTrace()) << "TRcrdWait: Initiating a wait operation for user " << VU.GetManagerNumber() << ". Duration is " << nDelay << ".\r\n";
                ReleaseTraceStream();
            }

            VU.SetTRecordCallStep(1);   // Will enter the "case 1" section next time execute is called.
            if(VU.StartEventTimer(nDelay) == FALSE)
            {
                assert(0);
                return TOF_RES_FAILED;
            }

            return TOF_RES_ASYNC_STARTED;  // Should let the framework call us again when the timer op. is completed.
        }

    case 1:
        {
            if(VU.GetTrace() && VU.GetTrace()->IsTrace("sync"))
            {
                GetTraceStream(VU.GetTrace()) << "TRcrdWait: Waking up after a wait operation for user " << VU.GetManagerNumber() << ".\r\n";
                ReleaseTraceStream();
            }

            ctxt.set_phase_text(PHASE_NONE);

            VU.SetTRecordCallStep(0);   // Reset call phase for the next call to an async TRecord::execute.

            return TOF_RES_OK;   // Move to next TOF object.
        }

    default:
        VU.SetTRecordCallStep(0);   // Reset call phase for the next call to an async TRecord::execute.
        ctxt.set_phase_text(PHASE_NONE);
        assert(0);

        return TOF_RES_FAILED;
    };
}

//
// TRcrdAuthentication::prepare
// ============================
//
// Prepare the record before the authentication step.
int TRcrdAuthentication::prepare(TScript* script, int objIndex, bool verbose)
{
    const TParam* prm = param(TParam::stType);
    if(prm == NULL) 
		return -1;
    _eAuthProtocol = (AuthProtocol)prm->intVal();

    // Authentication method:
    prm = param(TParam::stInputArgType);
    if(prm == NULL) 
		return -1;
    _eAuthMethod = (AuthMethod)prm->intVal();

    if(_eAuthMethod == USER_CREDENTIALS_METH)
    {
        // User:
        prm = param(TParam::stUser);
        if(prm == NULL) 
		    return -1;
        _pUser = new TtChrExpr(script);
        if(!(_pUser->Parse(prm->strVal(), 0)))
        {
		    MSGERR("Authentication step failed during 'user' PREPARE with: " + _pUser->Reason());
		    return -1;
        }

        // Password:
        prm = param(TParam::stPasswd);
        if(prm == NULL) 
		    return -1;
        _pPassword = new TtChrExpr(script);
        if(!(_pPassword->Parse(prm->strVal(), 0)))
        {
		    MSGERR("Authentication step failed during 'password' PREPARE with: " + _pPassword->Reason());
		    return -1;
        }

        // Domain:
        prm = param(TParam::stDomain);
        if(prm == NULL) 
		    return -1;
        _pDomain = new TtChrExpr(script);
        if(!(_pDomain->Parse(prm->strVal(), 0)))
        {
		    MSGERR("Authentication step failed during 'domain' PREPARE with: " + _pDomain->Reason());
		    return -1;
        }
    }
    else if(_eAuthMethod == BLOB_METH)
    {
        // Blob:
        prm = param(TParam::stInputBlob);
        if(prm == NULL) 
		    return -1;
        _pInputBlob = new TtChrExpr(script);
        if(!(_pInputBlob->Parse(prm->strVal(), 0)))
        {
		    MSGERR("Authentication step failed during 'blob' PREPARE with: " + _pInputBlob->Reason());
		    return -1;
        }
    }

    // Output blob:
	prm = param(TParam::stOutputBlob);
    if(prm == NULL) 
		return -1;
    string szBlob = prm->strVal();
    if(szBlob.empty())
    {
		MSGERR("Authentication step failed during 'into' PREPARE: the target variable is missing.");
		return -1;
    }

    _pOutputBlob = new TtVarRef(script);
    if(!(_pOutputBlob->Parse(szBlob, 0)))
    {
		MSGERR("Authentication step failed during 'into' PREPARE with: " + _pOutputBlob->Reason());
		return -1;
    }

    return 0;
}

//
// TRcrdAuthentication::execute
// ============================
//
// Execute an authentication step.
int TRcrdAuthentication::execute(TContext& ctxt)
{
    CAsyncVirtualUser& VU = dynamic_cast<CAsyncVirtualUser&>(ctxt);
    SecurityFunctionTable* pSecFuncTbl = CAsyncSecureSocket::GetSecurityFunctionTable();
    string strOutputBlob;
    switch(_eAuthProtocol)
    {
        case BASIC_AUTH:
            strOutputBlob = "BASIC ";
            break;

        case NTLM_AUTH:
            strOutputBlob = "NTLM ";
            break;

        case NEGOTIATE_AUTH:
            strOutputBlob = "Negotiate ";
            break;
    }

    if(_eAuthMethod == USER_CREDENTIALS_METH || _eAuthMethod == CURRENT_USER_METH)
    {
        // This is step 1 of the authentication process.

        string strUser, strPassword, strDomain;
        SEC_WINNT_AUTH_IDENTITY AuthIdentity = {0};
        SEC_WINNT_AUTH_IDENTITY *pAuthIdentity;
        if(_eAuthMethod == CURRENT_USER_METH)
        {
            // This form should never be used for BASIC authentication.
            if(_eAuthProtocol == BASIC_AUTH)
            {
                VU.FreeAuthenticationInfo();
                MSGERR("Basic authentication in incompatible with CURRENT USER. Explicit user/password should be used.");
                return -1;
            }
            pAuthIdentity = NULL;   // Use default credentials.
        }
        else // _eAuthMethod == USER_CREDENTIALS_METH
        {
            // Evaluate user credentials.
            strUser     = _pUser->Evaluate(ctxt);
            strPassword = _pPassword->Evaluate(ctxt);
            strDomain   = _pDomain->Evaluate(ctxt);

            AuthIdentity.User           = (unsigned char *)(const char*)strUser.c_str();
            AuthIdentity.UserLength     = (unsigned long)strUser.size();
            AuthIdentity.Password       = (unsigned char *)(const char*)strPassword.c_str();
            AuthIdentity.PasswordLength = (unsigned long)strPassword.size();
            if(!strDomain.empty())
            {
                AuthIdentity.Domain       = (unsigned char *)(const char*)strDomain.c_str();
                AuthIdentity.DomainLength = (unsigned long)strDomain.size();
            }
            AuthIdentity.Flags = SEC_WINNT_AUTH_IDENTITY_ANSI;
            pAuthIdentity = &AuthIdentity;
        }

        if(_eAuthProtocol == BASIC_AUTH)
        {
            // BASIC authentication only requires step 1.

            string strOut;
            if(!strDomain.empty())
                strUser = strDomain + '\\' + strUser;
            strUser += ':' + strPassword;
            DWORD dwBytesEncoded;
            char *pcEncoded = uuencode((const BYTE *)strUser.c_str(), (DWORD)strUser.size(), &dwBytesEncoded);
            strOutputBlob += pcEncoded;
            delete [] pcEncoded; pcEncoded = NULL;
        }
        else if(_eAuthProtocol == NTLM_AUTH || _eAuthProtocol == NEGOTIATE_AUTH)
        {
            // NTLM authentication requires steps 1 and 2.
            const char* zPackage = NTLM_AUTH ? "NTLM" : "Negotiate";

            TimeStamp Lifetime;
            CredHandle hCred;
            SECURITY_STATUS ss = pSecFuncTbl->AcquireCredentialsHandle(
                NULL,                   // New principal
                (char*)zPackage,        // Package name
                SECPKG_CRED_OUTBOUND,
                NULL,                   // Logon ID
                pAuthIdentity,          // Auth Data
                NULL,                   // Get key func
                NULL,                   // Get key arg
                &hCred,
                &Lifetime);

            if(ss == SEC_E_OK)
            {
                SecPkgInfo* pSecPkgInfo;
                ss = pSecFuncTbl->QuerySecurityPackageInfo((char*)zPackage, &pSecPkgInfo);
                if(ss == SEC_E_OK)
                {
                    DWORD dwMaxToken = pSecPkgInfo->cbMaxToken;
                    pSecFuncTbl->FreeContextBuffer(pSecPkgInfo);
                    SecBufferDesc OutBuffDesc = {0};
                    SecBuffer     OutSecBuff  = {0};
                    void* pBuff = _alloca((size_t)dwMaxToken);

                    // Prepare out buffer.
                    OutBuffDesc.ulVersion = 0;
                    OutBuffDesc.cBuffers  = 1;
                    OutBuffDesc.pBuffers  = &OutSecBuff;

                    OutSecBuff.cbBuffer   = dwMaxToken;
                    OutSecBuff.BufferType = SECBUFFER_TOKEN;
                    OutSecBuff.pvBuffer   = pBuff;

                    CtxtHandle hContext = {0};
                    ULONG ulContextAttributes;
                    ss = pSecFuncTbl->InitializeSecurityContext(
                        &hCred,
                        NULL,
                        NULL,
                        0,
                        0,
                        SECURITY_NATIVE_DREP,
                        NULL,
                        0,
                        &hContext,
                        &OutBuffDesc,
                        &ulContextAttributes,
                        &Lifetime);
                    if(!FAILED(ss))
                    {
                        if(ss == SEC_I_COMPLETE_NEEDED || ss == SEC_I_COMPLETE_AND_CONTINUE)
                        {
                            ss = pSecFuncTbl->CompleteAuthToken(&hContext, &OutBuffDesc);
                            if(FAILED(ss))
                            {
                                pSecFuncTbl->DeleteSecurityContext(&hContext);
                                pSecFuncTbl->FreeCredentialsHandle(&hCred);
                                return -1;
                            }
                        }
                        DWORD dwBytesEncoded;
                        char *pcEncoded = uuencode(
                            (const BYTE *)OutSecBuff.pvBuffer,
                            OutSecBuff.cbBuffer,
                            &dwBytesEncoded);
                        strOutputBlob += pcEncoded;
                        delete [] pcEncoded; pcEncoded = NULL;

                        if(VU.GetTrace() && VU.GetTrace()->IsTrace("full"))
                        {
                            GetTraceStream(VU.GetTrace())
                                << "TRcrdAuthentication::execute (step 1): output blob is:\""
                                << strOutputBlob
                                << "\".\r\n";
                            ReleaseTraceStream();
                        }

                        // Cache the credentials handle, security context and
                        // max token in the virtual user, since they will be
                        // needed in the next authentication step.
                        VU.SetAuthenticationInfo(hCred, hContext, dwMaxToken);
                    }
                    else
                    {
                        pSecFuncTbl->FreeCredentialsHandle(&hCred);
                        if(ss == SEC_E_LOGON_DENIED)
                            MSGERR("Authentication: logon denied.");
                        else
                            MSGERR("Authentication failed.");
                        return -1;
                    }
                }
                else
                    return -1;
            }
            else
                return -1;
        }
    }
    else if(_eAuthMethod == BLOB_METH)
    {
        // This is step 2 of the authentication process.
        // This step should never be used for BASIC authentication.
        if(_eAuthProtocol == BASIC_AUTH)
        {
            VU.FreeAuthenticationInfo();
            MSGERR("Basic authentication in incompatible with input blobs. User/password should be used.");
            return -1;
        }

        // Evaluate blob.
        string strBlob = _pInputBlob->Evaluate(ctxt);

        if(VU.GetTrace() && VU.GetTrace()->IsTrace("full"))
        {
            GetTraceStream(VU.GetTrace())
                << "TRcrdAuthentication::execute (step 2): input blob is:\""
                << strBlob
                << "\".\r\n";
            ReleaseTraceStream();
        }

        // Skip the BASIC or NTLM or KERBEROS and the following spaces.
        string::size_type idx = 0;
        char c;
        while((c = strBlob[idx]) != ' ' && c != '\t')
            ++idx;
        while((c = strBlob[idx]) == ' ' || c == '\t')
            ++idx;
        strBlob.erase(0, idx);

        // Decode the blob.
        DWORD dwDecoded;
        BYTE* pDecoded = uudecode(strBlob.c_str(), &dwDecoded);

        // Retrieve cached authentication infos.
        CredHandle hCred;
        CtxtHandle hContext;
        DWORD dwMaxToken;
        VU.GetAuthenticationInfo(hCred, hContext, dwMaxToken);

        // Build the second blob.
        SecBufferDesc OutBuffDesc = {0};
        SecBuffer     OutSecBuff  = {0};
        SecBufferDesc InBuffDesc  = {0};
        SecBuffer     InSecBuff   = {0};
        void* pBuff = _alloca((size_t)dwMaxToken);

        // Prepare out buffer.
        OutBuffDesc.ulVersion = 0;
        OutBuffDesc.cBuffers  = 1;
        OutBuffDesc.pBuffers  = &OutSecBuff;

        OutSecBuff.cbBuffer   = dwMaxToken;
        OutSecBuff.BufferType = SECBUFFER_TOKEN;
        OutSecBuff.pvBuffer   = pBuff;

        // Prepare in buffer.
        InBuffDesc.ulVersion = 0;
        InBuffDesc.cBuffers  = 1;
        InBuffDesc.pBuffers  = &InSecBuff;

        InSecBuff.cbBuffer   = dwDecoded;
        InSecBuff.BufferType = SECBUFFER_TOKEN;
        InSecBuff.pvBuffer   = (void*)pDecoded;

        ULONG ulContextAttributes;
        TimeStamp Lifetime;
        SECURITY_STATUS ss;
        CFastLock* pSChannelBugLock = CAsyncSecureSocket::GetChannelBugLock();
        if(pSChannelBugLock)
        {
            CLock lock(pSChannelBugLock);

            ss = pSecFuncTbl->InitializeSecurityContext(
                &hCred,
                &hContext,
                NULL,   // Might need to be a valid value for kerberos.
                0,
                0,
                SECURITY_NATIVE_DREP,
                &InBuffDesc,
                0,
                &hContext,
                &OutBuffDesc,
                &ulContextAttributes,
                &Lifetime);
        }
        else
        {
            ss = pSecFuncTbl->InitializeSecurityContext(
                &hCred,
                &hContext,
                NULL,   // Might need to be a valid value for kerberos.
                0,
                0,
                SECURITY_NATIVE_DREP,
                &InBuffDesc,
                0,
                &hContext,
                &OutBuffDesc,
                &ulContextAttributes,
                &Lifetime);
        }

        delete [] pDecoded; pDecoded = NULL;
        if(!FAILED(ss))
        {
            if(ss == SEC_I_COMPLETE_NEEDED || ss == SEC_I_COMPLETE_AND_CONTINUE)
            {
                ss = pSecFuncTbl->CompleteAuthToken(&hContext, &OutBuffDesc);
                if(FAILED(ss))
                {
                    VU.FreeAuthenticationInfo();
                    return -1;
                }
            }
            DWORD dwBytesEncoded;
            char *pcEncoded = uuencode(
                (const BYTE *)OutSecBuff.pvBuffer,
                OutSecBuff.cbBuffer,
                &dwBytesEncoded);
            strOutputBlob += pcEncoded;
            delete [] pcEncoded; pcEncoded = NULL;

            if(VU.GetTrace() && VU.GetTrace()->IsTrace("full"))
            {
                GetTraceStream(VU.GetTrace())
                    << "TRcrdAuthentication::execute (step 2): output blob is:\""
                    << strOutputBlob
                    << "\".\r\n";
                ReleaseTraceStream();
            }

            // Free the cached authentication infos.
            VU.FreeAuthenticationInfo();
        }
        else
        {
            VU.FreeAuthenticationInfo();
            if(ss == SEC_E_LOGON_DENIED)
            {
                MSGERR("Authentication: logon denied.");
                if(VU.GetTrace() && VU.GetTrace()->IsTrace("full"))
                {
                    GetTraceStream(VU.GetTrace())
                        << "TRcrdAuthentication::execute: logon denied.\r\n";
                    ReleaseTraceStream();
                }
            }
            else
            {
                MSGERR("Authentication failed.");
                if(VU.GetTrace() && VU.GetTrace()->IsTrace("full"))
                {
                    GetTraceStream(VU.GetTrace())
                        << "TRcrdAuthentication::execute: logon failed. Error is "
                        << (DWORD)ss
                        << ".\r\n";
                    ReleaseTraceStream();
                }
            }
            return -1;
        }
    }

    // Set the output variable.
    VU.set_variable(_pOutputBlob->Varnam(), strOutputBlob, _pOutputBlob->Evaluate(ctxt));

    return 0;   // Success.
}

/* 2 Arrays for uuencode / uudecode: */
const int pr2six[256]={
    64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,
    64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,62,64,64,64,63,
    52,53,54,55,56,57,58,59,60,61,64,64,64,64,64,64,64,0,1,2,3,4,5,6,7,8,9,
    10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,64,64,64,64,64,64,26,27,
    28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50,51,
    64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,
    64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,
    64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,
    64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,
    64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,
    64,64,64,64,64,64,64,64,64,64,64,64,64
};
const char six2pr[64]={
    'A','B','C','D','E','F','G','H','I','J','K','L','M',
    'N','O','P','Q','R','S','T','U','V','W','X','Y','Z',
    'a','b','c','d','e','f','g','h','i','j','k','l','m',
    'n','o','p','q','r','s','t','u','v','w','x','y','z',
    '0','1','2','3','4','5','6','7','8','9','+','/'
};

// uuencode function. The caller must free the return value after use.
char* TRcrdAuthentication::uuencode(const BYTE* pbin, DWORD dwBytes, DWORD* pdwEncoded)
{
    unsigned char* outptr;
    char* pReturn;
    unsigned int i;

    pReturn = new char[dwBytes + ((dwBytes + 3) / 3) + 4];
    outptr = (unsigned char *)pReturn;

    for(i=0; i<dwBytes; i += 3)
    {
        *(outptr++) = six2pr[*pbin >> 2];
        *(outptr++) = six2pr[((*pbin << 4) & 060)   | ((pbin[1] >> 4) & 017)];
        *(outptr++) = six2pr[((pbin[1] << 2) & 074) | ((pbin[2] >> 6) & 03)];
        *(outptr++) = six2pr[pbin[2] & 077];

        pbin += 3;
    }

    if(i == dwBytes+1)
        outptr[-1] = '=';
    else if(i == dwBytes+2)
    {
        outptr[-1] = '=';
        outptr[-2] = '=';
    }

    *outptr = '\0';
    *pdwEncoded = (char*)outptr - pReturn;

    return pReturn;
}

// uudecode function. The caller must free the return value after use.
BYTE* TRcrdAuthentication::uudecode(const char* pcInputBuf, DWORD* pdwDecoded)
{
    int iDecoded;
    const char *pcIn = pcInputBuf;
    unsigned char *pcOut;
    int iBytes;
    BYTE* pReturn;

    // Skip leading whitespace.
    while(*pcInputBuf == ' ' || *pcInputBuf == '\t')
        pcInputBuf++;

    pcIn = pcInputBuf;
    while(pr2six[*(pcIn++)] <= 63);
    iBytes = pcIn - pcInputBuf - 1;
    iDecoded = ((iBytes + 3) / 4) * 3;

    pReturn = new BYTE[iDecoded + 1];
    *pdwDecoded = iDecoded;

    pcOut = (unsigned char *)pReturn;
    pcIn = pcInputBuf;

    while(iBytes > 0)
    {
        *(pcOut++) = (unsigned char) (pr2six[*pcIn]   << 2 | pr2six[pcIn[1]] >> 4);
        *(pcOut++) = (unsigned char) (pr2six[pcIn[1]] << 4 | pr2six[pcIn[2]] >> 2);
        *(pcOut++) = (unsigned char) (pr2six[pcIn[2]] << 6 | pr2six[pcIn[3]]);
        pcIn += 4;
        iBytes -= 4;
    }

    if(iBytes & 03)
    {
        if(pr2six[pcIn[-2]] > 63)
            iDecoded -= 2;
        else
            iDecoded -= 1;
    }

    pReturn[iDecoded] = '\0';

    return pReturn;
}
