/* Copyright (C) 2000 Cyrano SA

   This file is part of OpenSTA

   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.

   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html

   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

//========================================================================
//
// TEHttpRecords.h
// ---------------
//=========================================================================

#ifndef CYR_TEHTTPRECORDS_H
#define CYR_TEHTTPRECORDS_H

#include <string>

#ifdef WIN32
#pragma warning(disable:4786)
using namespace std;
#endif

#include <TRecord.h>

#include "MemoryBufferHandler.hpp"
#include "HTMLParser/PhParse.hpp"
#include "PhParseSpecialised.tpp"
enum { HEADER_VAR=0, BODY_VAR=1 };
typedef map<std::string, TtChrExpr *> variableAddressMap;

///////////////////////////////////////////////////////////////////////////////
// @class   TRcrdConnect | This class is used to Connect in an HTTP environment
//
// @base    public | TRecord.
//
class TRcrdConnect : public TRecord
{
// @access Public Members.
public:
    // @cmember Set variable record ATVF key.
    static const AtvfKey KEY;

    // @cmember Constructor.
    TRcrdConnect (const TParamRule* headRuleP) : TRecord(headRuleP)
        , _exprP(NULL)
    {}

    // @cmember Destructor.
    virtual ~TRcrdConnect()
    {
        if(_exprP)
            delete _exprP;
    }

    // @cmember Execute record in current context and return status.
    virtual int execute (TContext&);
    // @cmember Prepare record for execution and return status.
    virtual int prepare (TScript* script, int objIndex, bool verbose=false);

    // @cmember Duplicate object.
    virtual TRecord* duplicate() const
    {
        // TODO: this method is not currently being called.
        // As soon as a copy constructor becomes available,
        // remove this comment, remove the assertion and
        // the return NULL, and uncomment the following:
        // return new TRcrdConnect(*this);
        assert(0); return NULL;
    }
    // @cmember
    static const bool _RegisterCreator();

// @access Private Members
private:
    // @cmember
    static bool _registeredCreator;

	TtChrExpr* _exprP;

private:
    //// Forbidden
    TRcrdConnect();
    TRcrdConnect(const TRcrdConnect&);
    operator=(const TRcrdConnect&);
}
; // TRcrdConnect


/**
 * Definition of Request Record class
 **/
class TRcrdRequest : public TRecord
{
public:
    static const AtvfKey KEY;
  // Constructors & Destructor
    TRcrdRequest(const TParamRule* headRuleP) : TRecord(headRuleP), 
                                        _uriP(NULL), _headerP(NULL), _withP(NULL), 
                                        _woutP(NULL), _bodyP(NULL), _certName(NULL),
                                        _retCode(NULL),
                                        _eResultsRequired(NO_RESULT),
                                        _iConid(0)
										{}

    // @cmember Destructor.
    virtual ~TRcrdRequest ()
    {
        if (_uriP)  delete _uriP;
        if(_headerP)delete _headerP; 
        if(_withP)  delete _withP;
        if(_woutP)  delete _woutP; 
        if(_bodyP)  delete _bodyP;
		if(_certName) delete _certName;
		if(_retCode)  delete _retCode;
    }

    // Accessors
    virtual int execute(TContext&);
    // @cmember Prepare record for execution and return status.
    virtual int prepare (TScript* script, int objIndex, bool verbose=false);
    // @cmember Duplicate object.

    virtual TRecord* duplicate() const
    {
        // TODO: this method is not currently being called.
        // As soon as a copy constructor becomes available,
        // remove this comment, remove the assertion and
        // the return NULL, and uncomment the following:
        // return new TRcrdRequest(*this);
        assert(0); return NULL;
    }
    static const bool _RegisterCreator();

  // Mutators

    typedef enum {HTTP_REQUEST_GET=0, HTTP_REQUEST_HEAD=1, HTTP_REQUEST_POST=2, /*HTTP_REQUEST_CONNECT=3, */HTTP_REQUEST_NONE=4} ERequestType;

    const std::string &getRequestString(ERequestType eType) const; 
    void SetRequiredResults(const ERequiredResults res) { _eResultsRequired = res; }
    ERequiredResults GetRequiredResults()               const { return _eResultsRequired; }

protected:
    //// Helpers.
    void BuildHeaderString(TContext& ctxt, std::string &szFormattedHeader) const;
    void ExtractHostPort(TContext& ctxt, const std::string &szHeader, const std::string &szUri, std::string &szRemoteServer, std::string &szPort) const;
    int	Ascii2Binary(const std::string &inBuf, std::string &outBuf) const;

private:
    static bool _registeredCreator;
	TtChrExpr* _uriP;
	TtValList* _headerP;
	TtValList* _withP;
	TtValList* _woutP;
	TtChrExpr* _bodyP;
	TtChrExpr* _certName;
    TtVarRef*  _retCode;    // For HTTP return code.

public:
    int _iConid;
    ERequiredResults _eResultsRequired;

private:
    //// Forbidden
    TRcrdRequest();
    TRcrdRequest(const TRcrdRequest&);
    operator=(const TRcrdRequest&);
}
; // TRcrdRequest

/**
 * Definition of Disconnect Record class
 **/
class TRcrdDisconnect : public TRecord
{
public:
    static const AtvfKey KEY;
  // Constructors & Destructor
    TRcrdDisconnect(const TParamRule* headRuleP) : TRecord(headRuleP)
    {}
  // Accessors
    virtual int execute(TContext&);
    // @cmember Prepare record for execution and return status.
    virtual int prepare (TScript* script, int objIndex, bool verbose=false);
    // @cmember Duplicate object.
    virtual TRecord* duplicate() const
    {
        return new TRcrdDisconnect(*this);
    }
    static const bool _RegisterCreator();

private:
    static bool _registeredCreator;
}
; // TRcrdDisc

/**
 * Definition of HttpResponse Record class
 **/
class TRcrdHttpResponse : public TRecord
{
public:
	static const AtvfKey KEY;
    // Constructors & Destructor
    TRcrdHttpResponse(const TParamRule* headRuleP) : TRecord(headRuleP)
                    , _varP(NULL), _exprP(NULL), _identP(NULL), _varSize(0)
    {}

    // @cmember Destructor.
    virtual ~TRcrdHttpResponse()
    {
        if(_varP) delete _varP;
        if(_exprP) delete _exprP; 
        if(_identP) delete _identP;
		for (AddressQueryList::iterator itr = _addressQueryList.begin(); itr != _addressQueryList.end(); itr ++)
			delete *itr;
		for (variableAddressMap::iterator itrMap = _varAddressMap.begin(); itrMap != _varAddressMap.end(); itrMap ++)
			delete itrMap->second;
    }

  // Accessors
    virtual int execute(TContext&);
    // @cmember Prepare record for execution and return status.
    virtual int prepare (TScript* script, int objIndex, bool verbose=false);

    // @cmember Duplicate object.
    virtual TRecord* duplicate() const
    {
        // TODO: this method is not currently being called.
        // As soon as a copy constructor becomes available,
        // remove this comment, remove the assertion and
        // the return NULL, and uncomment the following:
        // return new TRcrdHttpResponse(*this);
        assert(0); return NULL;
    }
    // @cmember
    static const bool _RegisterCreator();

  // Mutators

private:
    int _varSize;   // variable size

    //// Helpers
    BOOL    ParseCookieValue(CDataBuffHandler &messageBuff, const string &szName,   BOOL bFoundContinueHeader, string &zCookieVal) const;
    void    ParseHeaderValue(CDataBuffHandler &messageBuff, const string &szIdent,  BOOL bFoundContinueHeader, string &zHeaderValue) const;
    void    DumpHeader      (CDataBuffHandler &messageBuff, string &szVarVal, DWORD dwVarLength, BOOL bFoundContinueHeader) const;
    void    DumpBody        (CDataBuffHandler &messageBuff, string &szVarVal, DWORD dwVarLength, BOOL bFoundContinueHeader) const;

    // @cmember Destination label index in script
    int	_labelIndex;
    // @cmember Pointer to set variable 'name' object.
    TtVarRef* _varP;
    // @cmember Pointer to set variable 'expression' object.
    CECTofTxt* _exprP;
    // @cmember Pointer to set variable 'identity' object.
	TtChrExpr* _identP;
    // @cmember
    static bool _registeredCreator;

public:
	int _iConid ;
	int _responseType ;				// header or body
	AddressQueryList _addressQueryList ; // List of compiled addresses 
	variableAddressMap	_varAddressMap ;	// Map of HTML address in ascii mode
									// These address cannot be converted in the
									// prepare phase because they are not constant.	
									// key = variable, value = expression

private:
    //// Forbidden
    TRcrdHttpResponse();
    TRcrdHttpResponse(const TRcrdHttpResponse&);
    operator=(const TRcrdHttpResponse&);
}
; // TRcrdHttpResponse


/**
 * Definition of Synchronize Record class
 **/
class TRcrdSynchronize : public TRecord
{
public:
    static const AtvfKey KEY;
  // Constructors & Destructor
    TRcrdSynchronize(const TParamRule* headRuleP) : TRecord(headRuleP)
    {}
  // Accessors
    virtual int execute(TContext&);
    // @cmember Prepare record for execution and return status.
    virtual int prepare (TScript* script, int objIndex, bool verbose=false);
    // @cmember Duplicate object.
    virtual TRecord* duplicate() const
    {
        return new TRcrdSynchronize(*this);
    }
    static const bool _RegisterCreator();

protected:
    int _labelIndex;    // Destination label index in script

private:
    static bool _registeredCreator;

private:
    //// Forbidden
    TRcrdSynchronize();
}
;

/**
 * Definition of Authentication Record class
 **/
class TRcrdAuthentication : public TRecord
{
    enum AuthProtocol {BASIC_AUTH = TOF_AUTTYP_BASIC, NTLM_AUTH = TOF_AUTTYP_NTLM, KERBEROS_AUTH = TOF_AUTTYP_KERBEROS, NEGOTIATE_AUTH = TOF_AUTTYP_NEGOTIATE};
    enum AuthMethod   {CURRENT_USER_METH = TOF_AUTIN_CURRENT, USER_CREDENTIALS_METH = TOF_AUTIN_CREDENTIALS, BLOB_METH = TOF_AUTIN_BLOB};

public:
    static const AtvfKey KEY;

    // Constructor.
    TRcrdAuthentication(const TParamRule* headRuleP) : TRecord(headRuleP), 
        _eAuthProtocol(BASIC_AUTH), _eAuthMethod(CURRENT_USER_METH),
        _pUser(NULL), _pPassword(NULL), _pDomain(NULL), _pInputBlob(NULL),
        _pOutputBlob(NULL)
	{}

    // @cmember Destructor.
    virtual ~TRcrdAuthentication()
    {
        if(_pUser)      delete _pUser;
        if(_pPassword)  delete _pPassword;
        if(_pDomain)    delete _pDomain;
        if(_pInputBlob) delete _pInputBlob;
        if(_pOutputBlob)delete _pOutputBlob;
    }

    // Accessors
    virtual int execute(TContext&);
    // @cmember Prepare record for execution and return status.
    virtual int prepare (TScript* script, int objIndex, bool verbose=false);
    // @cmember Duplicate object.

    virtual TRecord* duplicate() const
    {
        // TODO: this method is not currently being called.
        // As soon as a copy constructor becomes available,
        // remove this comment, remove the assertion and
        // the return NULL, and uncomment the following:
        // return new TRcrdRequest(*this);
        assert(0); return NULL;
    }
    static const bool _RegisterCreator();

  // Mutators

protected:
    //// Helpers.
    static char* uuencode(const BYTE* pbin, DWORD dwBytes, DWORD* pdwEncoded);
    static BYTE* uudecode(const char* pcInputBuf, DWORD* pdwDecoded);

private:
    static bool _registeredCreator;
    AuthProtocol _eAuthProtocol;
    AuthMethod   _eAuthMethod;
    TtChrExpr    *_pUser, *_pPassword, *_pDomain, *_pInputBlob;
    TtVarRef     *_pOutputBlob;

private:
    //// Forbidden
    TRcrdAuthentication();
    TRcrdAuthentication(const TRcrdAuthentication&);
    operator=(const TRcrdAuthentication&);
}
; // TRcrdAuthentication

#endif
