/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/


////////////////////////////////////////////////////////////////////////////////
// HTTP thread pool client implementation file.
////////////////////////////////////////////////////////////////////////////////


#include "stdafx.h"

#include <assert.h>

#include "Trace.hpp"
#include "ErrorCodes.hpp"
#include "AsyncHTTPRequest.hpp"
#include "asyncrequestmanager.hpp"  // Only to display the manager number.
#include "Tools.hpp"
#include "TRecord.h"    // Just to get the ETofResult definition.
#include "AsyncVirtualUser.h"
#include "TEMainStructure.hpp"

#ifdef USE_PERF_DATA
#   include "PerfCounterKeys.hpp"
#endif

#define CR '\r'
#define LF '\n'
#define OpenErrorStream()   {std::ostringstream errStm; errStm
#define CloseErrorStream()  g_pTExecuter->LogError(errStm.str().c_str());}

extern CMainExecuterStructure *g_pTExecuter;   // Defined and initialised in main.cpp.


#ifdef USE_PERF_DATA
// Static members initialisation.
LONG *CAsyncHTTPRequest::s_plPendingQueriesPerfCtr  = NULL;
LONG *CAsyncHTTPRequest::s_plQueryRatePerfCtr       = NULL;
LONG *CAsyncHTTPRequest::s_plOutThroughPutPerfCtr   = NULL;
LONG *CAsyncHTTPRequest::s_plInThroughPutPerfCtr    = NULL;
#endif

in_addr CAsyncHTTPRequest::s_localAddress;
CSharedMemBuffs *CAsyncHTTPRequest::s_pSSLBuffs = NULL;
DWORD CAsyncHTTPRequest::s_dwMinSSLBuffSize = 0;

// Construction / Destruction

CAsyncHTTPRequest::CAsyncHTTPRequest(CIOCP& iocp, IAsyncRequestEvent& requestEventListener, DWORD dwRequestNumber, CTrace* pTrace)
	:   CAsyncRequest(iocp, requestEventListener, dwRequestNumber, pTrace),
        m_messageLengthType(MSG_UNKNOWN_TYPE),
        m_bIsHeaderFullyParsed(FALSE), m_bIsChunkFullyReceived(FALSE),
        m_eResultsRequired(NO_RESULT), m_dwBytesRead(0), 
        m_requestType(RQ_GET), m_dwContentLengthLeft(0),
        m_dwConnectionId((DWORD)-1), m_bUseProxy(FALSE),
        m_messageBuffIter(&m_messageBuffer.GetWholeBuffList()),
        m_bMessageBufferChanged(TRUE), m_bContainContinueHeader(FALSE),
        m_cnxType(CNX_UNKNOWN), m_bIsTunnelingThroughProxy(FALSE),
        m_securityMode(HTTP_NOT_SECURE), m_bWillReconnectThroughProxy(FALSE)
{
	m_dwState = CLIENT_INITIAL;
}

CAsyncHTTPRequest::~CAsyncHTTPRequest()
{
    m_messageBuffer.ClearBuffer();

    // Eventually return the SSL buffer to the common pool if not already done.
    PBYTE pLastBuff = static_cast<CAsyncSecureSocket*>(m_pDevice)->CAsyncSecureSocket::GetReadBuffer();
    if(pLastBuff != NULL)
        s_pSSLBuffs->ReturnDataBuffer(pLastBuff);

    if(m_pDevice)
    {
        delete m_pDevice;
        m_pDevice = NULL;
    }
}

// Static methods.

BOOL CAsyncHTTPRequest::InitClass(DWORD dwMaxConcurrentSSLReq, DWORD dwSSLGrowingBuffCount)
{
    CSharedMemBuffs::InitClass();

#ifdef USE_PERF_DATA
    s_plPendingQueriesPerfCtr   = g_pPerformance->GetCounterValue(PRFCNT_OPENSTA_THREADPOOL_PENDINGQUERIES);
    s_plQueryRatePerfCtr        = g_pPerformance->GetCounterValue(PRFCNT_OPENSTA_THREADPOOL_QUERYRATE);
    s_plOutThroughPutPerfCtr    = g_pPerformance->GetCounterValue(PRFCNT_OPENSTA_OUTGOING_THROUGHPUT);
    s_plInThroughPutPerfCtr     = g_pPerformance->GetCounterValue(PRFCNT_OPENSTA_INCOMING_THROUGHPUT);
#endif

    char hostname[256];
    if(gethostname(hostname, 256) != 0)
        return FALSE;

    struct hostent FAR *host = gethostbyname(hostname);
    if(host)
    {
        s_localAddress = *((struct in_addr*)((host->h_addr_list)[0]));
    }

    // Create the buffers for use internally by CAsyncSecureSocket.
    DWORD dwRes = CAsyncSecureSocket::GetReadBufferMinSize(s_dwMinSSLBuffSize);
    if(dwRes != 0)
        return FALSE;

    s_pSSLBuffs = new CSharedMemBuffs(dwMaxConcurrentSSLReq, s_dwMinSSLBuffSize, dwSSLGrowingBuffCount);
    if(s_pSSLBuffs == NULL)
        return FALSE;

    return TRUE;
}

void CAsyncHTTPRequest::UninitClass()
{
    if(s_pSSLBuffs)
        delete s_pSSLBuffs;
    s_pSSLBuffs = NULL;
}

// User methods

void CAsyncHTTPRequest::StartTimer()
{
    if(m_bWatchStats == TRUE)
    {
        m_statistics.OnRequestStart();
    }
}

void CAsyncHTTPRequest::EndTimer()
{
    if(m_bWatchStats == TRUE)
    {
        m_statistics.OnRequestOver();
    }
}

void CAsyncHTTPRequest::InitParseStructures()
{
    m_bMessageBufferChanged     = TRUE;
	m_bIsHeaderFullyParsed	    = FALSE;
	m_bIsChunkFullyReceived     = FALSE;
	m_messageLengthType		    = MSG_UNKNOWN_TYPE;
    m_bContainContinueHeader    = FALSE;

    m_messageBuffer.ClearBuffer();
    m_messageBuffIter.Rewind();

	m_chunkStep = CK_READ_SIZE;
}


///////////////////////////////////////////////////////////////////////////////
//
// CONNECTION METHODS
//
///////////////////////////////////////////////////////////////////////////////

// Connect to the server.
// In case of secured connection, that method is asynchronous and will be called twice. The second time, parameters will be
// useless.
DWORD CAsyncHTTPRequest::Connect(LPCSTR zServer, WORD wPort, IocpEvent* pEventTarget)
{
    // Store connection info. for use by the ReConnect method.
    DWORD dwRes = 0;
    m_bUseProxy = FALSE;    // Will be set to TRUE later on if current method was called by TRcrdConnect::execute.

    if(m_securityMode == HTTP_SECURE)
    {
        CAsyncSecureSocket* pSecureSocket = static_cast<CAsyncSecureSocket*>(m_pDevice);

        // Performing secured connection.
        if(m_dwState != CLIENT_CONNECTING)
        {
            // Called first time only to start the asynchronous connection.
            m_zServer = zServer;
            m_dwPort = (DWORD) wPort;
            m_pEventTarget = pEventTarget;

            m_dwState = CLIENT_CONNECTING;

            if(m_pTrace && m_pTrace->IsTrace("connection"))
            {
                GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::Connect: Performing an SSL connection on " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId
                    << "with socket object " << std::hex << (DWORD)pSecureSocket << std::dec << " to " << zServer << " on port " << wPort << ".\r\n";
                ReleaseTraceStream();
            }

            PBYTE pLastBuff;
            pSecureSocket->SetReadBuffer(s_pSSLBuffs->AllocDataBuffer(), s_dwMinSSLBuffSize, pLastBuff);
            assert(pLastBuff == NULL);

            if(pLastBuff != NULL)
                s_pSSLBuffs->ReturnDataBuffer(pLastBuff);

            dwRes = static_cast<CAsyncSecureSocket*>(m_pDevice)->CAsyncSecureSocket::Open(zServer, (DWORD)wPort, pEventTarget);

		    if(dwRes == ERROR_IO_PENDING)
			    return TOF_RES_ASYNC_STARTED;
        }
        else
        {
            // Called when the SSL connection has been performed.
            if(m_pTrace && m_pTrace->IsTrace("connection"))
            {
                GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::Connect: SSL connection finished on " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId
                    << " on socket object " << std::hex << (DWORD)pSecureSocket << std::dec << "\r\n";
                ReleaseTraceStream();
            }
        }
    }
    else
    {
        // Performing standard TCP connection.
        m_zServer = zServer;
        m_dwPort = (DWORD) wPort;
        m_pEventTarget = pEventTarget;

        if(m_pTrace && m_pTrace->IsTrace("connection"))
        {
            GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::Connect: Performing a connection on " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId
                        << " to " << zServer << " on port " << wPort << ".\r\n";
            ReleaseTraceStream();
        }
        dwRes = static_cast<CAsyncTCPSocket*>(m_pDevice)->CAsyncTCPSocket::Open(zServer, (DWORD)wPort, pEventTarget);
    }

    if(dwRes != 0)
    {
        if(m_pTrace && m_pTrace->IsTrace("connection errors"))
        {
            GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::Connect: ERROR " << dwRes << " while connecting " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
            ReleaseTraceStream();
        }
        CAsyncVirtualUser& vu = static_cast<CAsyncVirtualUser&>(m_requestEventListener);
        OpenErrorStream() << "ERROR " << dwRes << " for VU " << vu.GetThreadID()
            << " while establishing "
            << (m_securityMode == HTTP_SECURE ? "secure " : "")
            << "connection " 
            << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ".";
        CloseErrorStream();
    }
    else
    {
        m_dwState = CLIENT_CONNECTED;

        if(m_bWatchStats == TRUE)
        {
            CAsyncVirtualUser& vu = static_cast<CAsyncVirtualUser&>(m_requestEventListener);
            m_statistics.SetUserId(vu.GetTaskGroupID(), vu.GetExternalVUID());
        }
    }

	return dwRes;
}

// Re-connect to the same server following a connection breaking from the server side.
// Note that method will be asynchronous when using SSL.
DWORD CAsyncHTTPRequest::ReConnect()
{
    if(m_pTrace && m_pTrace->IsTrace("connection full"))
    {
        GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::ReConnect: ReConnecting " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId
                    << " to " << m_zServer << " on port " << m_dwPort << "\r\n";
        ReleaseTraceStream();
    }

    DWORD dwRes;

    if(m_securityMode == HTTP_SECURE)
    {
        assert(0);  // Not managed yet.
    }
    else
        dwRes = static_cast<CAsyncTCPSocket*>(m_pDevice)->CAsyncTCPSocket::Open(m_zServer.c_str(), m_dwPort, m_pEventTarget);

    if(dwRes != 0)
    {
        if(m_pTrace && m_pTrace->IsTrace("connection errors"))
        {
            GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::ReConnect: ERROR re-connecting " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId
                        << " to " << m_zServer << " on port " << m_dwPort << "\r\n";
            ReleaseTraceStream();
        }
        OpenErrorStream() << "ERROR " << dwRes << " while re-connecting " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId
                          << " to " << m_zServer << " on port " << m_dwPort << ".";
        CloseErrorStream();
    }

    m_dwState = CLIENT_CONNECTED;
    m_bIsTunnelingThroughProxy = FALSE;

	return dwRes;
}

// Disconnect from the server.
// In case of secured connection, that method is asynchronous and will be called twice.
DWORD CAsyncHTTPRequest::Disconnect()
{
    DWORD dwRes = 0;

    assert(m_pDevice->DidOverlappedIoComplete() == TRUE);

    if(m_securityMode == HTTP_SECURE)
    {
        // Breaking an SSL connection.
        dwRes = SSLDisconnect();
	    if(dwRes == ERROR_IO_PENDING)
		    return ERROR_IO_PENDING;
    }
    else
    {
        // Breaking a standard TCP connection.
        dwRes = TCPDisconnect();
    }

    OnRequestClosed();

    return dwRes;
}

DWORD CAsyncHTTPRequest::TCPDisconnect()
{
    if(m_pTrace && m_pTrace->IsTrace("connection"))
    {
        GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::Disconnect: Disconnecting " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ".\r\n";
        ReleaseTraceStream();
    }

    if(static_cast<CAsyncTCPSocket*>(m_pDevice)->CAsyncTCPSocket::IsOpen())
    {
        return static_cast<CAsyncTCPSocket*>(m_pDevice)->CAsyncTCPSocket::Close();
    }
    else
        return 0;
}

DWORD CAsyncHTTPRequest::SSLDisconnect()
{
    DWORD dwRes;
    CAsyncSecureSocket* pSecureSocket = static_cast<CAsyncSecureSocket*>(m_pDevice);
    if(m_dwState != CLIENT_DISCONNECTING)
    {
        if(m_pTrace && m_pTrace->IsTrace("connection"))
        {
            GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::Disconnect: Disconnecting " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId
                << " SSL connection on secure socket object " << std::hex << (DWORD)pSecureSocket << std::dec << ".\r\n";
            ReleaseTraceStream();
        }

        m_dwState = CLIENT_DISCONNECTING;

        // Close the secure socket, whatever its state.
        dwRes = pSecureSocket->Close();
    }
    else
    {
        if(m_pTrace && m_pTrace->IsTrace("connection"))
        {
            GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::Disconnect: Disconnected " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId
                << " SSL connection on secure socket object " << std::hex << (DWORD)pSecureSocket << std::dec << ".\r\n";
            ReleaseTraceStream();
        }
        // Note the TCP connection has already been closed by CAsyncSecureSocket.

        // Return the SSL buffer to the common pool.
        PBYTE pLastBuff;
        pSecureSocket->SetReadBuffer(NULL, s_dwMinSSLBuffSize, pLastBuff);
        assert(pLastBuff != NULL);

        if(pLastBuff != NULL)
            s_pSSLBuffs->ReturnDataBuffer(pLastBuff);
    }

    return dwRes;
}

// Are we connected to the server?
// Cases are:
//  1/ Using direct TCP connection to a server.
//  2/ Using direct SSL connection to a server (should test the TCP connection + the SSL connection).
//  3/ Using TCP connection through a proxy (implies a TCP connection to the proxy and a HTTP-connect to the remote server).
//  4/ Using SSL connection through a proxy (implies a TCP connection to the proxy + a HTTP-connect to the remote server + a SSL-connect).
BOOL CAsyncHTTPRequest::IsConnected() const
{
    if(m_securityMode == HTTP_SECURE)
    {
        // Testing a secure connection.
        return static_cast<CAsyncSecureSocket*>(m_pDevice)->CAsyncSecureSocket::IsOpen()
                && (m_bUseProxy ? m_bIsTunnelingThroughProxy : TRUE);
    }
    else
    {
        // Testing a standard TCP connection.
        return static_cast<CAsyncTCPSocket*>(m_pDevice)->CAsyncTCPSocket::IsOpen()
                /*&& (m_bUseProxy ? m_bIsTunnelingThroughProxy : TRUE)*/;
    }
}


///////////////////////////////////////////////////////////////////////////////
//
// REQUEST METHODS
//
///////////////////////////////////////////////////////////////////////////////

// Used by InitiateGET, InitiateHEAD and InitiatePOST.
// bSignal must be set to FALSE if used to resend a query
DWORD CAsyncHTTPRequest::InitiateGenericRequest(ERequiredResults eResultsRequired, BOOL bSignal)
{
    DWORD dwRes = 0;

    InterlockedExchange((LPLONG)&m_bCancel, (LONG)FALSE);
	if(m_dwState == CLIENT_INITIAL || m_dwState == CLIENT_CONNECTED)
	{
        m_dwHTTPBytes = 0;
	    m_eResultsRequired = eResultsRequired;

	    InitParseStructures();

        if(bSignal)
        {
            OnNewRequest();

#ifdef USE_PERF_DATA
#   ifdef LOCK_COUNTERS
            InterlockedIncrement(s_plPendingQueriesPerfCtr);
            InterlockedIncrement(s_plQueryRatePerfCtr);
#   else
            (*s_plPendingQueriesPerfCtr)++;
            (*s_plQueryRatePerfCtr)++;
#   endif
#endif
        }

        // Start the timer if required to.
        StartTimer();

        if(m_pTrace && m_pTrace->IsTrace("request"))
        {
            GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::InitiateGenericRequest: Initiating a Write operation on connection " << static_cast<CAsyncRequestManager&>(m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ". " << static_cast<CAsyncRequestManager&>(m_requestEventListener).GetPendingRequestCount() << " requests are pending.\r\n";
            if(m_securityMode == HTTP_SECURE)
                *pOstream << " Secure socket object is " << std::hex << (DWORD)(static_cast<CAsyncSecureSocket*>(m_pDevice)) << std::dec << ".\r\n";
            ReleaseTraceStream();
        }

        m_dwState = CLIENT_REQUEST_SENT;

        if(m_securityMode == HTTP_SECURE)
            dwRes = static_cast<CAsyncSecureSocket*>(m_pDevice)->CAsyncSecureSocket::Write(m_szCurrentRequest.c_str(), m_szCurrentRequest.size());
        else
            dwRes = static_cast<CAsyncTCPSocket*>(m_pDevice)->CAsyncTCPSocket::Write(m_szCurrentRequest.c_str(), m_szCurrentRequest.size());

        if(dwRes != 0)
        {
            if(dwRes == WSAECONNRESET || dwRes == WSAENETDOWN || dwRes == WSAENETRESET || dwRes == WSAENOTCONN || dwRes == WSAESHUTDOWN || dwRes == WSAECONNABORTED)
            {
                // Connection has been broken inbetween (possibly because we are talking to an HTTP 1.0 server and we didn't know yet). Must re-connect.
                if(m_securityMode != HTTP_SECURE)
                {
                    if(m_pTrace && m_pTrace->IsTrace("request"))
                    {
                        GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::InitiateGenericRequest: WSASend returned " << dwRes << ". Reconnecting and initiating a Write operation AGAIN on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                        ReleaseTraceStream();
                    }

                    ReConnect();

                    m_dwState = CLIENT_REQUEST_SENT;

                    // Can now try the Write (send) operation again.
                    if(m_securityMode == HTTP_SECURE)
                        dwRes = static_cast<CAsyncSecureSocket*>(m_pDevice)->CAsyncSecureSocket::Write(m_szCurrentRequest.c_str(), m_szCurrentRequest.size());
                    else
                        dwRes = static_cast<CAsyncTCPSocket*>(m_pDevice)->CAsyncTCPSocket::Write(m_szCurrentRequest.c_str(), m_szCurrentRequest.size());
                }
            }
        }

        if(dwRes != 0)
        {
            if(m_pTrace && m_pTrace->IsTrace("request errors"))
            {
                GetTraceStream(m_pTrace) << "ERROR sending request on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ". Returning error " << dwRes << "\r\n";
                ReleaseTraceStream();
            }
            OpenErrorStream() << "ERROR " << dwRes << " while sending request on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ".";
            CloseErrorStream();

            // Still in error. Reset. Note the connection can be a secured one. Violently close the connection.
            m_securityMode = HTTP_NOT_SECURE;   // Force TCP disconnect.
            Disconnect();

            if(bSignal)
            {
                OnRequestCancelled();
            }
        }
    }
    else
    {
        dwRes = (DWORD)OPENSTA_E_BAD_STATE;

        if(m_pTrace && m_pTrace->IsTrace("request errors"))
        {
            GetTraceStream(m_pTrace) << "ERROR: connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << " already in use. Working step is " << m_dwState << "\r\n";
            ReleaseTraceStream();
        }
    }

	return dwRes;
}

// Initiates a GET request.
DWORD CAsyncHTTPRequest::InitiateGET(ERequiredResults eResultsRequired)
{
    if(m_pTrace && m_pTrace->IsTrace("request"))
    {
        GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::InitiateGET: Sending a GET request on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
        ReleaseTraceStream();
    }

    m_requestType = RQ_GET;

    return InitiateGenericRequest(eResultsRequired, TRUE);
}

// Initiates a POST request.
DWORD CAsyncHTTPRequest::InitiatePOST(ERequiredResults eResultsRequired)
{
    if(m_pTrace && m_pTrace->IsTrace("request"))
    {
        GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::InitiatePOST: Sending a POST request on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
        ReleaseTraceStream();
    }

    m_requestType = RQ_POST;

    return InitiateGenericRequest(eResultsRequired, TRUE);
}

// Initiates a HEAD request.
DWORD CAsyncHTTPRequest::InitiateHEAD(ERequiredResults eResultsRequired)
{
    if(m_pTrace && m_pTrace->IsTrace("request"))
    {
        GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::InitiateHEAD: Sending a HEAD request on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
        ReleaseTraceStream();
    }

    m_requestType = RQ_HEAD;

    return InitiateGenericRequest(eResultsRequired, TRUE);
}

// Initiates a CONNECT request.
DWORD CAsyncHTTPRequest::InitiateCONNECT(ERequiredResults eResultsRequired)
{
    if(m_pTrace && m_pTrace->IsTrace("request"))
    {
        GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::InitiateCONNECT: Sending a CONNECT request on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
        ReleaseTraceStream();
    }

    m_requestType = RQ_CONNECT;

    return InitiateGenericRequest(eResultsRequired, TRUE);
}

///////////////////////////////////////////////////////////////////////////////
//
// I/O COMPLETION CALLBACKS
//
///////////////////////////////////////////////////////////////////////////////

void CAsyncHTTPRequest::OnEvent(DWORD dwNumberOfBytesTransferred, BOOL bIsIo, DWORD dwEventType)
{
    if(bIsIo == FALSE)
    {
        if(m_pTrace && m_pTrace->IsTrace("full"))
        {
            SYSTEMTIME st;
            GetLocalTime(&st);
            GetTraceStream(m_pTrace) << '[' << st.wMinute << ':' << st.wSecond << ':' << st.wMilliseconds << "] CAsyncHTTPRequest::OnEvent: received an IO of type " << dwEventType << " on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ".\r\n";
            ReleaseTraceStream();
        }

        switch(dwEventType)
        {
        case CAsyncSecureSocket::EVENT_TYPE_MESSAGE_RECEIVED:
            {
                // Signals that receival of data initiated by a Read command completed.
                // Will continue to the 'Handling IOCP events' section.
                break;
            }

        case CAsyncSecureSocket::EVENT_TYPE_MESSAGE_SENT:
            {
                // Signals that sending of data initiated by a Write command completed.
                // Will continue to the 'Handling IOCP events' section.
                break;
            }

        case CAsyncSecureSocket::EVENT_TYPE_PLEASE_PROVIDE_CLIENT_CERTIFICATE:
            {
                static_cast<CAsyncSecureSocket*>(m_pDevice)->CAsyncSecureSocket::SetClientCertificateName(m_szCertificateName.c_str());

                return; // Return to CAsyncSecureSocket::MakeClientCertificateCredentials.
            }

        case CAsyncSecureSocket::EVENT_TYPE_CONNECTION_ESTABLISHED:
            {
                // Signals the end of the SSL connection initiated by CAsyncHTTPRequest::Connect.
                Connect(NULL, 0, NULL);

                // Resume TOF processing.
                static_cast<CAsyncVirtualUser&>(m_requestEventListener).ProcessTof();   // No need to handle the results.

                return;
            }

        case CAsyncSecureSocket::EVENT_TYPE_CONNECTION_CLOSED:
            {
                // Signals the end of the SSL + TCP disconnection.
                // Two cases:
                //  1/ initiated by CAsyncHTTPRequest::Disconnect.
                //  2/ initiated by the remote server which disconnected the SSL connection (occur when called CAsyncSecuredSocket::Read).
                //      Two more cases then:
                //      2a/ Followed a "Connection: close" header and thus, was predictable.
                //      2b/ Happened unexpectedly (for instance, server timed-out).
                if(m_dwState == CLIENT_DISCONNECTING)
                {
                    // Case 1: must complete SSL disconnect.
                    Disconnect();
                }
                else
                {
                    if(m_cnxType == CNX_DISCONNECTS)
                    {
                        // Case 2a.
                    }
                    else
                    {
                        // Case 2b. Must TCP-close the connection.
                        if(m_pTrace && m_pTrace->IsTrace("connection"))
                        {
                            GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEvent: server broke the " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << " SSL connection.\r\n";
                            ReleaseTraceStream();
                        }
                    }

                    // TCP-disconnect.
                    m_securityMode = HTTP_NOT_SECURE;
                    m_cnxType = CNX_DISCONNECTS;    // Will disconnect before next request.

                    OnRequestFinished();
                }
                return;
            }

        default:
            {
                assert(0);
                if(m_pTrace && m_pTrace->IsTrace("errors"))
                {
                    GetTraceStream(m_pTrace) << "ERROR: CAsyncHTTPRequest::OnEvent: received an unexpected user-defined event type " << dwEventType << " on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ".\r\n";
                    ReleaseTraceStream();
                }
                return;
            }
        };  // switch(dwEventType).
    }   // bIsIo == FALSE.

    // Handling IOCP events.

	m_dwBytesRead = dwNumberOfBytesTransferred;

    assert(m_dwConnectionId !=  ULONG_MAX);

    if(m_pTrace && m_pTrace->IsTrace("full"))
    {
        SYSTEMTIME st;
        GetLocalTime(&st);
        GetTraceStream(m_pTrace) << '[' << st.wMinute << ':' << st.wSecond << ':' << st.wMilliseconds << "] CAsyncHTTPRequest::OnEvent: OnEvent thread " << GetCurrentThreadId() << " received " << dwNumberOfBytesTransferred 
                                << " bytes on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId
                                << ".\r\n";
        ReleaseTraceStream();
    }

    if(m_dwConnectionId == ULONG_MAX)
    {
        // Connection was broken inbetween.
        return;
    }

    switch(m_dwState)
	{
	case CLIENT_REQUEST_SENT :
		{
            if(m_pTrace && m_pTrace->IsTrace("request"))
            {
                GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEvent: Framework acknowledged request sent on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                ReleaseTraceStream();
            }

#ifdef USE_PERF_DATA
#   ifdef LOCK_COUNTERS
            InterlockedExchangeAdd(s_plOutThroughPutPerfCtr, dwNumberOfBytesTransferred);
#   else
            *s_plOutThroughPutPerfCtr += dwNumberOfBytesTransferred;
#   endif
#endif
            m_dwHTTPBytes += dwNumberOfBytesTransferred;

            if(m_bCancel == TRUE)
            {
                OnRequestCancelled();
            }
            else
            {
			    // Get ready to accept data from the server.
			    m_dwState = CLIENT_RECEIVING;

                if(m_pTrace && m_pTrace->IsTrace("request"))
                {
                    GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEvent: Initiating a Read operation on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                    if(m_securityMode == HTTP_SECURE)
                        *pOstream << " Secure socket object is " << std::hex << (DWORD)(static_cast<CAsyncSecureSocket*>(m_pDevice)) << std::dec << ".\r\n";
                    ReleaseTraceStream();
                }

                DWORD dwRes;

                WSABUF *aWsaBuf = m_messageBuffer.AllocWsaBuffList();
                assert(aWsaBuf != NULL);
                if(m_securityMode == HTTP_SECURE)
                    dwRes = static_cast<CAsyncSecureSocket*>(m_pDevice)->CAsyncSecureSocket::Read(aWsaBuf, CDataBuffHandler::WSABUF_COUNT);
                else
                    dwRes = static_cast<CAsyncTCPSocket*>(m_pDevice)->CAsyncTCPSocket::Read(aWsaBuf, CDataBuffHandler::WSABUF_COUNT);

			    if(dwRes != 0)
			    {
                    if(dwRes == WSAENETDOWN || dwRes == WSAENETRESET || dwRes == WSAESHUTDOWN || dwRes == WSAECONNABORTED || dwRes == WSAECONNRESET || dwRes == WSAEDISCON || dwRes == WSAENOTCONN)
                    {
                        // Connection has been broken inbetween (possibly because we are talking to an HTTP 1.0 server). Must re-connect.
                        if(m_securityMode != HTTP_SECURE)
                        {
                            ReConnect();

                            if(m_szCurrentRequest.size() != 0)
                            {
                                // Can now try the Write (send) operation again.
                                switch(m_requestType)
                                {
                                case RQ_GET :
                                    if(m_pTrace && m_pTrace->IsTrace("request"))
                                    {
                                        GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEvent: Sending again a GET request after a reconnect on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                                        ReleaseTraceStream();
                                    }

                                    dwRes = InitiateGenericRequest(m_eResultsRequired, FALSE);
                                    break;

                                case RQ_POST :
                                    if(m_pTrace && m_pTrace->IsTrace("request"))
                                    {
                                        GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEvent: Sending again a POST request after a reconnect on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                                        ReleaseTraceStream();
                                    }

                                    dwRes = InitiateGenericRequest(m_eResultsRequired, FALSE);
                                    break;

                                case RQ_HEAD :
                                    if(m_pTrace && m_pTrace->IsTrace("request"))
                                    {
                                        GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEvent: Sending again a HEAD request after a reconnect on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                                        ReleaseTraceStream();
                                    }

                                    dwRes = InitiateGenericRequest(m_eResultsRequired, FALSE);
                                    break;

                                case RQ_CONNECT :
                                    if(m_pTrace && m_pTrace->IsTrace("request"))
                                    {
                                        GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEvent: Sending again a CONNECT request after a reconnect on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                                        ReleaseTraceStream();
                                    }

                                    dwRes = InitiateGenericRequest(m_eResultsRequired, FALSE);
                                    break;
                                };
                            }

                            if(dwRes != 0)
                            {
                                // Cannot execute that request again. However, must keep on working.
                                if(m_pTrace && m_pTrace->IsTrace("request errors"))
                                {
                                    GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEvent: ERROR while initiating a READ operation after sending a request (cannot send the request again). Connection ID is " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ", error number is " << dwRes << "\r\n";
                                    ReleaseTraceStream();
                                }
                                OpenErrorStream() << "ERROR " << dwRes << " while initiating a READ operation after sending a request (cannot send the request again). Connection ID is " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ".";
                                CloseErrorStream();
                            }
                        }
                        else
                        {
                            if(m_pTrace && m_pTrace->IsTrace("request errors"))
                            {
                                GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEvent: ERROR while initiating a READ operation after sending a request. Connection ID is " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ", error number is " << dwRes << "\r\n";
                                ReleaseTraceStream();
                            }
                            OpenErrorStream() << "ERROR " << dwRes << " while initiating a READ operation after sending a request. Connection ID is " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ".";
                            CloseErrorStream();
                        }
                        // Continue with the following in order to clean up.
                    }
                    else    // Unknown error code.
                    {
                        if(m_pTrace && m_pTrace->IsTrace("request errors"))
                        {
                            GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEvent: ERROR while initiating a READ operation after sending a request. Connection ID is " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ", error number is " << dwRes << "\r\n";
                            ReleaseTraceStream();
                        }
                        OpenErrorStream() << "ERROR " << dwRes << " while initiating a READ operation after sending a request. Connection ID is " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ".";
                        CloseErrorStream();
                    }
                }

                if(dwRes != 0)
                {
                    // Still in error after the recovery attempt. Give up.

                    // Do not disconnect since the connection info may be used by OnRequestFinished or by another TOF waiting for that request.
                    OnRequestFinished();
                    return;
                }
            }
		}
		break;

	case CLIENT_RECEIVING :
		{
            if(m_pTrace && m_pTrace->IsTrace("request"))
            {
                GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEvent: Framework signaled data received on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                ReleaseTraceStream();
            }

            if(dwNumberOfBytesTransferred == 0)
            {
                // The server disconnected.
                if(m_pTrace && m_pTrace->IsTrace("request"))
                {
                    GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEvent: zero-sized data buffer. Means we'll have to reconnect connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                    ReleaseTraceStream();
                }
            }
            else
            {
#ifdef USE_PERF_DATA
#   ifdef LOCK_COUNTERS
                InterlockedExchangeAdd(s_plInThroughPutPerfCtr, dwNumberOfBytesTransferred);
#   else
                *s_plInThroughPutPerfCtr += dwNumberOfBytesTransferred;
#   endif
#endif

                m_dwHTTPBytes += dwNumberOfBytesTransferred;
                if(m_bWatchStats == TRUE)
                {
                    m_statistics.IncDataFlow(dwNumberOfBytesTransferred);
                }

			    // Store the received buffer
			    ProcessReadData();
            }

			if(IsDataTransferComplete() == FALSE)
			{
                if(m_bCancel == TRUE)
                {
                    OnRequestCancelled();
                }
                else
                {
				    // Read more data
                    if(m_pTrace && m_pTrace->IsTrace("request"))
                    {
                        GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEvent: Initiating a Read operation on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                        ReleaseTraceStream();
                    }

                    DWORD dwRes;

                    // In the case of SSL, can't use scatter-gather.
                    WSABUF *aWsaBuf = m_messageBuffer.AllocWsaBuffList();
                    assert(aWsaBuf != NULL);

                    // In case we're reading data that we know we won't process, set the buffer pointer to NULL to avoid buffer copying.
                    if(m_messageLengthType != MSG_CONTENT_LENGTH// We don't know (yet) the size of the data we're gonna receive.
                        || m_eResultsRequired == ALL_DATA       // Need all the data (eg for parsing purposes).
                        || (m_eResultsRequired == HEADER_ONLY && m_bIsHeaderFullyParsed == FALSE)   // Not yet read the whole header.
                        )
                    {
                        // Read the data.
                        if(m_securityMode == HTTP_SECURE)
                            dwRes = dynamic_cast<CAsyncSecureSocket*>(m_pDevice)->CAsyncSecureSocket::Read(aWsaBuf, CDataBuffHandler::WSABUF_COUNT);
                        else
                            dwRes = dynamic_cast<CAsyncTCPSocket*>(m_pDevice)->CAsyncTCPSocket::Read(aWsaBuf, CDataBuffHandler::WSABUF_COUNT);
                    }
                    else
                    {
                        // Tell the communication layer to skip the data.
                        if(m_securityMode == HTTP_SECURE)
                            dwRes = static_cast<CAsyncSecureSocket*>(m_pDevice)->CAsyncSecureSocket::Read((LPSTR)NULL, m_dwContentLengthLeft);
                        else
                            dwRes = static_cast<CAsyncTCPSocket*>(m_pDevice)->CAsyncTCPSocket::Read(aWsaBuf, CDataBuffHandler::WSABUF_COUNT);
                    }

			        if(dwRes != 0)
			        {
                        if(dwRes == WSAENETDOWN || dwRes == WSAENETRESET || dwRes == WSAESHUTDOWN || dwRes == WSAECONNABORTED || dwRes == WSAECONNRESET || dwRes == WSAEDISCON || dwRes == WSAENOTCONN)
                        {
                            // Connection has been broken inbetween (possibly because we are talking to an HTTP 1.0 server). Must re-connect.
                            if(m_securityMode != HTTP_SECURE)
                            {
                                ReConnect();

                                if(m_szCurrentRequest.size() != 0)
                                {
                                    // Can now try the Write (send) operation again.
                                    switch(m_requestType)
                                    {
                                    case RQ_GET :
                                        if(m_pTrace && m_pTrace->IsTrace("request"))
                                        {
                                            GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEvent: Sending again a GET request after a reconnect on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                                            ReleaseTraceStream();
                                        }

                                        dwRes = InitiateGET(m_eResultsRequired);
                                        break;

                                    case RQ_POST :
                                        if(m_pTrace && m_pTrace->IsTrace("request"))
                                        {
                                            GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEvent: Sending again a POST request after a reconnect on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                                            ReleaseTraceStream();
                                        }

                                        dwRes = InitiatePOST(m_eResultsRequired);
                                        break;

                                    case RQ_HEAD :
                                        if(m_pTrace && m_pTrace->IsTrace("request"))
                                        {
                                            GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEvent: Sending again a HEAD request after a reconnect on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                                            ReleaseTraceStream();
                                        }

                                        dwRes = InitiateHEAD(m_eResultsRequired);
                                        break;

                                    case RQ_CONNECT :
                                        if(m_pTrace && m_pTrace->IsTrace("request"))
                                        {
                                            GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEvent: Sending again a CONNECT request after a reconnect on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                                            ReleaseTraceStream();
                                        }

                                        dwRes = InitiateGenericRequest(m_eResultsRequired, FALSE);
                                        break;
                                    };
                                }

                                if(dwRes != 0)
                                {
                                    // Cannot execute that request again. However, must keep on working.
                                    if(m_pTrace && m_pTrace->IsTrace("request errors"))
                                    {
                                        GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEvent: ERROR while initiating a READ operation after sending a request (cannot send the request again). Connection ID is " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ", error number is " << dwRes << "\r\n";
                                        ReleaseTraceStream();
                                    }
                                    OpenErrorStream() << "ERROR " << dwRes << " while initiating a READ operation after sending a request (cannot send the request again). Connection ID is " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ".";
                                    CloseErrorStream();

                                }
                            }
                            else
                            {
                                if(m_pTrace && m_pTrace->IsTrace("request errors"))
                                {
                                    GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEvent: ERROR " << dwRes << " while initiating a READ operation while in the \"receiving\" state. Connection is " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                                    ReleaseTraceStream();
                                }
                                OpenErrorStream() << "ERROR " << dwRes << " while initiating a READ while in the \"receiving\" state. Connection is " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ".";
                                CloseErrorStream();
                            }
                        }
                        else
                        {
                            // Unknown error code.
                            if(m_pTrace && m_pTrace->IsTrace("request errors"))
                            {
                                GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEvent: ERROR " << dwRes << " while initiating a READ operation while in the \"receiving\" state. Connection is " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                                ReleaseTraceStream();
                            }
                            OpenErrorStream() << "ERROR " << dwRes << " while initiating a READ while in the \"receiving\" state. Connection is " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ".";
                            CloseErrorStream();
                        }
				    }

			        if(dwRes != 0)
			        {
                        // Still in error after the recovery attempt. Give up.

                        // Do not disconnect since the connection info may be used by OnRequestFinished or by another TOF waiting for that request.
                        OnRequestFinished();

                        return;
                    }
                }
			}
			else
            {
				// Transfer over. Signal end of processing and get ready for another GET/POST/HEAD.
				m_dwState = CLIENT_CONNECTED;

                if(m_pTrace)
                {
                    if(m_pTrace->IsTrace("request"))
                    {
                        GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEvent: Reception is over on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                        ReleaseTraceStream();
                    }

                    if(m_pTrace->IsTrace("Results"))
                    {
                        GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEvent: Results are:\r\n" << m_messageBuffer << "\r\n";
                        ReleaseTraceStream();
                    }
                }

                OnRequestFinished();
			}
		}
		break;

	/*
    case CLIENT_INITIAL :
    case CLIENT_CONNECTING :
    case CLIENT_CONNECTED:*/
	default :
        {
            // Unwanted states.
            assert(0);
            if(m_pTrace && m_pTrace->IsTrace("request errors"))
            {
                GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEvent: ERROR: Unexpected state ";
                switch(m_dwState)
                {
                case CLIENT_INITIAL:    *pOstream << "CLIENT_INITIAL";      break;
                case CLIENT_CONNECTING: *pOstream << "CLIENT_CONNECTING";   break;
                case CLIENT_CONNECTED:  *pOstream << "CLIENT_CONNECTED";    break;
                default:                *pOstream << m_dwState;
                };
                *pOstream << ". Connection is " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                ReleaseTraceStream();
            }
		    break;
        }
	};
}

void CAsyncHTTPRequest::OnEventFailed(DWORD /*dwNumberOfBytesTransferred*/, BOOL bIsIo, DWORD dwEventType, DWORD dwError)
{
    if(bIsIo == FALSE)
    {
        switch(dwEventType)
        {
        case CAsyncSecureSocket::EVENT_TYPE_CONNECTION_ESTABLISHED:
            {
                CAsyncVirtualUser& vu = static_cast<CAsyncVirtualUser&>(m_requestEventListener);
                if(m_pTrace && m_pTrace->IsTrace("request errors"))
                {
                    GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEventFailed: ERROR "
                        << dwError << " for VU " << vu.GetThreadID()
                        << " while establishing a secure connection on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                    ReleaseTraceStream();
                }
                OpenErrorStream() << "ERROR " << dwError << " for VU " << vu.GetThreadID()
                    << " while establishing a secure connection on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ".";
                CloseErrorStream();

                TContext *pCtxt = dynamic_cast<TContext*>(&m_requestEventListener);
                if(pCtxt->GetVerbose(ERRMSG))
                {
                    pCtxt->log("Connection error for VU " + vu.GetThreadID()
                        + " while establishing a secure connection on connection "
                        + int_to_string((long)((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber())
                        + '/' + int_to_string((long)m_dwConnectionId), ERRMSG);
                }

                // Hard disconnect.
                m_cnxType = CNX_DISCONNECTS;
                Disconnect();

                static_cast<CAsyncVirtualUser&>(m_requestEventListener).ProcessTof();   // No need to handle the results.
                return;
            }

        case CAsyncSecureSocket::EVENT_TYPE_CONNECTION_CLOSED:
            {
                if(m_pTrace && m_pTrace->IsTrace("request errors"))
                {
                    GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEventFailed: ERROR " << dwError << " while closing a secured connection on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                    ReleaseTraceStream();
                }
                OpenErrorStream() << "ERROR " << dwError << " while closing a secured connection on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ".";
                CloseErrorStream();

                // Signals the end of the SSL + TCP disconnection.
                // Two cases:
                //  1/ initiated by CAsyncHTTPRequest::Disconnect.
                //  2/ initiated by the remote server which disconnected the SSL connection. Occurs when called CAsyncSecuredSocket::Read.
                if(m_dwState == CLIENT_DISCONNECTING)
                {
                    // Case 1: must complete SSL disconnect.
                    Disconnect();
                }
                else
                {
                    // Case 2: must TCP-close the connection.
                    if(m_pTrace && m_pTrace->IsTrace("connection"))
                    {
                        GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEvent: server broke the " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << " SSL connection.\r\n";
                        ReleaseTraceStream();
                    }

                    m_dwState = CLIENT_DISCONNECTING;
                    Disconnect();

                    OnRequestFinished();
                }
                return;
            }

        case CAsyncSecureSocket::EVENT_TYPE_MESSAGE_RECEIVED:
            {
                if(m_pTrace && m_pTrace->IsTrace("request errors"))
                {
                    GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEventFailed: ERROR " << dwError << " while receiving data on a secured connection on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                    ReleaseTraceStream();
                }
                OpenErrorStream() << "ERROR " << dwError << " while receiving data on a secured connection on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ".";
                CloseErrorStream();

                // Will disconnect before using that connection during the next request.
                m_cnxType = CNX_DISCONNECTS;

                OnRequestFinished();
                return;
            }

        case CAsyncSecureSocket::EVENT_TYPE_MESSAGE_SENT:
            {
                if(m_pTrace && m_pTrace->IsTrace("request errors"))
                {
                    GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEventFailed: ERROR " << dwError << " while sending data on a secured connection on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                    ReleaseTraceStream();
                }
                OpenErrorStream() << "ERROR " << dwError << " while sending data on a secured connection on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ".";
                CloseErrorStream();

                // Will disconnect before using that connection during the next request.
                m_cnxType = CNX_DISCONNECTS;

                OnRequestFinished();
                return;
            }

        default:
            {
                // Unexpected event (routed badly?).
                assert(0);
                if(m_pTrace && m_pTrace->IsTrace("errors"))
                {
                    GetTraceStream(m_pTrace) << "ERROR: CAsyncHTTPRequest::OnEventFailed: received an unexpected user-defined event type " << dwEventType << " on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ".\r\n";
                    ReleaseTraceStream();
                }

                return;
            }
        };  // switch(dwEventType).
    }
    else
    {
        CAsyncVirtualUser& vu = static_cast<CAsyncVirtualUser&>(m_requestEventListener);
        if(m_pTrace && m_pTrace->IsTrace("errors"))
        {
            GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::OnEventFailed: IO failed for VU "
                << vu.GetThreadID() << " on connection "
                << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber()
                << '/' << m_dwConnectionId
                << " with error " << dwError
                << ".\r\n";
            ReleaseTraceStream();
        }
        OpenErrorStream() << "IO failed for VU " << vu.GetThreadID() << " on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << " with error " << dwError << ".";
        CloseErrorStream();

        TContext *pCtxt = dynamic_cast<TContext*>(&m_requestEventListener);
        if(pCtxt->GetVerbose(ERRMSG))
        {
            pCtxt->log("IO failed for VU " + vu.GetThreadID() + " on connection "
                + int_to_string((long)((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber())
                + '/' + int_to_string((long)m_dwConnectionId) + " with error " + int_to_string((long)dwError)
                , ERRMSG);
        }

        OnRequestFinished();
    }
}

void CAsyncHTTPRequest::OnRequestFinished()
{
    EndTimer();
    m_dwState = CLIENT_CONNECTED;

#ifndef CYRANO_TP_STANDALONE
    if(m_bWatchStats == TRUE)
    {
        // Write the statistics down to disk.
        m_statistics.DumpData();
    }
#endif

#ifdef USE_PERF_DATA
#   ifdef LOCK_COUNTERS
    InterlockedDecrement(s_plPendingQueriesPerfCtr);
#   else
    (*s_plPendingQueriesPerfCtr)--;
#   endif
#endif

    switch (m_eResultsRequired)
    {
    case NO_RESULT:
        m_messageBuffer.ClearBuffer();
        break;

    case HEADER_ONLY:
        // TODO: should clear everything but the header.
        break;
    };

    // Otherwise, buffers will be cleared on the next request execution on that connection.
    CAsyncRequest::OnRequestFinished();
}

void CAsyncHTTPRequest::OnRequestCancelled()
{
    EndTimer();
    m_dwState = CLIENT_CONNECTED;

    if(m_pTrace && m_pTrace->IsTrace("requests"))
    {
        GetTraceStream(m_pTrace) << "CAsyncHTTPRequest: " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << " cancelled connection " << m_dwConnectionId << ".\r\n";
        ReleaseTraceStream();
    }

#ifndef CYRANO_TP_STANDALONE
    if(m_bWatchStats == TRUE)
    {
        // Write the statistics down to disk.
        m_statistics.DumpData();
    }
#endif

#ifdef USE_PERF_DATA
#   ifdef LOCK_COUNTERS
    InterlockedDecrement(s_plPendingQueriesPerfCtr);
#   else
    (*s_plPendingQueriesPerfCtr)--;
#   endif
#endif

    switch (m_eResultsRequired)
    {
    case NO_RESULT:
        m_messageBuffer.ClearBuffer();
        break;

    case HEADER_ONLY:
        // TODO: should clear everything but the header.
        break;
    };

    // Otherwise, buffers will be cleared on the next request execution on that connection.
    CAsyncRequest::OnRequestCancelled();
}

void CAsyncHTTPRequest::OnRequestClosed()
{
    // Update the state machine.
    m_bIsTunnelingThroughProxy = FALSE;
    m_dwState       = CLIENT_INITIAL;
    m_securityMode  = HTTP_NOT_SECURE;
    m_bUseProxy     = FALSE;
    m_cnxType       = CNX_UNKNOWN;
    SetConnectionId(ULONG_MAX);
    m_messageBuffer.ClearBuffer();

    CAsyncRequest::OnRequestClosed();   // Note: not needed all the time, especially when SSL signals 
                                        // a server-disconnect. Could then avoid signalling the VU.
}


///////////////////////////////////////////////////////////////////////////////
//
// DATA PROCESSING METHODS
//
///////////////////////////////////////////////////////////////////////////////

// If m_eResultsRequired is TRUE then adds the received buffer to the m_messageBuffer buffer.
// However, always parse data in order to find the end of message.
// Handle the chunk coding info.
void CAsyncHTTPRequest::ProcessReadData()
{
    assert(m_messageLengthType != MSG_CONTENT_LENGTH || m_dwBytesRead <= m_dwContentLengthLeft);

    // Store the received data (archive the received buffers from WSABUF list to CDataBuffHandler).
    m_messageBuffer.ReturnWsaBuffList(m_dwBytesRead);

    // Reset temporary indicators.
    m_bNoMeaningFullDataLeft = FALSE;

    // Adjust the iterator in case it was left in a bad position because reached the end of buffer or there
    // was no buffer yet.
    if(m_messageBuffIter.AdjustIterator() == FALSE)
        assert(0);

    do  // while(BufferParsingOver() == TRUE)
	{
		if(m_bIsHeaderFullyParsed == FALSE)
		{
			// Header-parse not complete. Keep on parsing.

            if(m_pTrace && m_pTrace->IsTrace("request"))
            {
                GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::ProcessReadData: processing header for connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                ReleaseTraceStream();
            }

			if(m_requestType == RQ_HEAD)
			{
				m_messageLengthType = MSG_NO_BODY;

				// Not interested in the remaining header info.
				if(StoreHeader() == TRUE)
				{
					m_bIsHeaderFullyParsed = TRUE;

                    if(m_pTrace && m_pTrace->IsTrace("request"))
                    {
                        GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::ProcessReadData: header fully processed for connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                        ReleaseTraceStream();
                    }
				}
			}
			else
			{
                CDataBuffListAccessor iterParam, iterEndOfParam;
    			EHeaderFieldType headerFieldType;
				while(m_bIsHeaderFullyParsed == FALSE && ParseHeaderLine(headerFieldType, iterParam, iterEndOfParam) == TRUE)
				{
					// Found a whole line.
					switch(headerFieldType)
					{
					case FT_UNUSED :
						// That header field is of no use for us.
						break;

					case FT_CONTENT_LENGTH :
                        {
						    if(m_messageLengthType != MSG_NO_BODY && m_messageLengthType != MSG_NON_IDENTITY)
						    {
							    // iterParam points to the body-length string.
                                if(iterParam.ReadDWORD(m_dwContentLengthLeft) == FALSE)
                                {
                                    // Error ... Should not happen.
                                    assert(0);
                                }

                                if(m_pTrace && m_pTrace->IsTrace("request"))
                                {
                                    GetTraceStream(m_pTrace) << "On connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << " : Result is delimited by CONTENT-LENGTH=" << m_dwContentLengthLeft << "\r\n";
                                    ReleaseTraceStream();
                                }

							    m_messageLengthType = MSG_CONTENT_LENGTH;
						    }
						    // MSG_NO_BODY hides Chunked and Content-Length;
						    // Chunked hides Content-Length.
						    break;
                        }

					case FT_TRANSFER_ENCODING :
                        {
						    if(m_messageLengthType != MSG_NO_BODY)
						    {
                                if(iterParam.ContainString("chunked", 7, iterEndOfParam, FALSE) == TRUE)
							    {
								    m_messageLengthType = MSG_NON_IDENTITY;

                                    if(m_pTrace && m_pTrace->IsTrace("request"))
                                    {
                                        GetTraceStream(m_pTrace) << "On connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << " : Result is delimited by Chunk format.\r\n";
                                        ReleaseTraceStream();
                                    }
							    }
							    else
							    {
                                    if(m_pTrace && m_pTrace->IsTrace("request errors"))
                                    {
                                        GetTraceStream(m_pTrace) << "On connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << " : ERROR: Result should be delimited by Chunk format?\r\n";
                                        ReleaseTraceStream();
                                    }
							    }
						    }
						    // MSG_NO_BODY hides Chunked and Content-Length
						    break;
                        }

					case FT_MESSAGE :
                        {
                            // Skip the version digit.
                            if(iterParam.SkipTillWS() == FALSE)
                                assert(0);

                            // Move to the message status number.
                            if(iterParam.SkipWS() == FALSE)
                                assert(0);

                            DWORD dwStatusCode;
                            if(iterParam.ReadNextWord(dwStatusCode) == FALSE)
                                assert(0);

#ifdef __DLLSA__
                            TContext *pCtxt = dynamic_cast<TContext*>(&m_requestEventListener);
                            if(pCtxt->GetVerbose(TRACEMSG))
                            {
                                pCtxt->log("Connection " + int_to_string((long)m_dwConnectionId) + ": receiving results with status " + int_to_string((long)dwStatusCode), TRACEMSG);
                            }
#endif // __DLLSA__

                            // Update the reply code in the statistics.
                            m_statistics.SetReplyCode((WORD)dwStatusCode);

                            if(dwStatusCode == 100)
                            {
                                // Received a "continue" message. Must expect the real message to arrive pretty soon ...
                                m_messageLengthType = MSG_CONTINUE;
                                m_bContainContinueHeader = TRUE;

                                if(m_pTrace && m_pTrace->IsTrace("request"))
                                {
                                    GetTraceStream(m_pTrace) << "On connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << " : Result is a CONTINUE message.\r\n";
                                    ReleaseTraceStream();
                                }
                            }
                            else
                            {
                                if(dwStatusCode/100 == 1 || dwStatusCode == 204 || dwStatusCode == 304 || (m_requestType == RQ_CONNECT && dwStatusCode < 400))
						        {
							        m_messageLengthType = MSG_NO_BODY;

                                    if(m_pTrace && m_pTrace->IsTrace("request"))
                                    {
                                        GetTraceStream(m_pTrace) << "On connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << " : Result is a message (" << dwStatusCode << ") and has no body.\r\n";
                                        ReleaseTraceStream();
                                    }
						        }
                                else
                                {
							        m_messageLengthType = MSG_UNKNOWN_TYPE;
                                }
                            }

                            // In the case of a TCP-CONNECT request, must check if proxy server managed to connect to remote server
                            // and should not expect a message body.
                            if(m_requestType == RQ_CONNECT && dwStatusCode == 200)
                                m_bIsTunnelingThroughProxy = TRUE;

						    break;
                        }

					case FT_END_OF_HEADER :
                        {
						    m_bIsHeaderFullyParsed = TRUE;
						    break;
                        }

                    case FT_PROXY_CONNECTION :
                    case FT_CONNECTION :
                        {
                            if(iterParam.MatchFirstWordToLower("close", sizeof "close"-1) == TRUE)
                            {
                                // TODO: WARNING: what happens if we're in Secure mode or through a proxy?
                                // The server says it will close the connection after receiving data.
                                m_cnxType = CNX_DISCONNECTS;

                                if(m_pTrace && m_pTrace->IsTrace("request"))
                                {
                                    GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::ProcessReadData: received a CONNECTION: close for connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                                    ReleaseTraceStream();
                                }
                            }
                            else
                            {
                                if(iterParam.MatchFirstWordToLower("keep-alive", sizeof "keep-alive"-1) == TRUE)
                                {
                                    // TODO: WARNING: what happens if we're in Secure mode or through a proxy?
                                    m_cnxType = CNX_KEEP_ALIVE;

                                    if(m_pTrace && m_pTrace->IsTrace("request"))
                                    {
                                        GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::ProcessReadData: reveived a CONNECTION: keep-alive for connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                                        ReleaseTraceStream();
                                    }
                                }
                            }
                            break;
                        }

                    case FT_CONTENT_TYPE :
                        {
                            // TODO: exclude other not-parseable document types.
                            if(iterParam.MatchFirstWordToLower("image/", sizeof "image/" -1) == TRUE
                                || iterParam.MatchFirstWordToLower("audio/", sizeof "audio/" -1) == TRUE
                                )
                            {
                                // This is not an HTML page. Won't need to parse the contents. Note that if we're here, m_eResultsRequired equals to ALL_DATA and we're in SSL mode.
                                m_eResultsRequired = HEADER_ONLY;
                            }
                            break;
                        }

                    default:
                        assert(0);
					};
				}

                if(m_bIsHeaderFullyParsed && m_messageLengthType == MSG_CONTINUE)
                {
                    // Parsed the entire header fields for a CONTINUE message. Must now parse a standard header.
                    // Warning: if ever a server sends a CONTINUE message with a body, this will fail!!!
                    m_bIsHeaderFullyParsed = FALSE;
                }
			}
		}
		else	// if(m_bIsHeaderFullyParsed == TRUE)
		{
			// Parsing the message body.

			switch(m_messageLengthType)
			{
			case MSG_NON_IDENTITY :
				{
					// Expecting chunk format.

                    CDataBuffListAccessor backupIter;
                    backupIter = m_messageBuffIter;
                    BOOL bFoundChunkInf = TRUE;

					switch(m_chunkStep)
					{
					case CK_READ_SIZE :
						{
							// The first word of the next line should be the size of that chunk.

                            if(m_pTrace && m_pTrace->IsTrace("request"))
                            {
                                GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::ProcessReadData: reading chunk size for connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                                ReleaseTraceStream();
                            }

                            CDataBuffListAccessor iterBegLine, iterEndLine;

							if(ParseLine(iterBegLine, iterEndLine) == TRUE)
                            {
                                if(iterBegLine.ReadHexDWORD(m_dwCurrentChunkSizeLeft) == FALSE)
                                {
                                    // Error: managed to parse the line but not the chunk size.
                                    if(m_pTrace && m_pTrace->IsTrace("errors"))
                                    {
                                        GetTraceStream(m_pTrace) << "On connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << " : PROTOCOL ERROR: Did not find a chunk size where it should be.\r\n";
                                        ReleaseTraceStream();
                                    }
                                    assert(0);
                                }

                                if(m_pTrace && m_pTrace->IsTrace("request"))
                                {
                                    GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::ProcessReadData: chunk size is " << m_dwCurrentChunkSizeLeft << " for connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                                    ReleaseTraceStream();
                                }

								if(m_dwCurrentChunkSizeLeft == 0)
								{
									m_chunkStep = CK_READ_TRAILER;
								}
								else
								{
									m_chunkStep = CK_READ_DATA;
								}

                                // Note the chunk extension is skipped by ParseLine at the same time as the end of the line.
							}
                            else
                                bFoundChunkInf = FALSE; // Not enough inf. Wait for the next buffers.

							break;
						}

					case CK_READ_DATA :
						{
							// We are in the chunk buffer.

                            if(m_pTrace && m_pTrace->IsTrace("request"))
                            {
                                GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::ProcessReadData: reading chunk data for connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                                ReleaseTraceStream();
                            }

							if(StoreChunk() == TRUE)
							{
								// Reached the end of the current chunk.
								m_chunkStep = CK_READ_END_OF_DATA;
							}
							// Else keep on reading.

							break;
						}

					case CK_READ_END_OF_DATA :
						{
							// Reads the CRLF sequence following each chunked data buffer.
                            if(m_pTrace && m_pTrace->IsTrace("request"))
                            {
                                GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::ProcessReadData: reading end of chunk for connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
                                ReleaseTraceStream();
                            }

                            if(m_messageBuffIter.SkipCRLF() == TRUE)
                            {
								m_chunkStep = CK_READ_SIZE;
							}
                            else
                                bFoundChunkInf = FALSE; // Not enough inf. Wait for the next buffers.

							break;
						}

					case CK_READ_TRAILER :
						{
							// Skip the trailer, eventually containing several entity-headers.
                            if(m_pTrace && m_pTrace->IsTrace("request"))
                            {
                                GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::ProcessReadData: reading chunk footer for connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ".\r\n";
                                ReleaseTraceStream();
                            }

                            while(m_bIsChunkFullyReceived == FALSE && bFoundChunkInf == TRUE)
                            {
                                // Check if CRLF or LF forms the begining of the line (marks the end of the trailer).
                                if(*m_messageBuffIter == '\r')
                                {
                                    CHAR nextChar;
                                    if(m_messageBuffIter.NextChar(nextChar) == TRUE)
                                    {
                                        if(nextChar == '\n')
                                        {
                                            // End of trailer marks the end of message parsing.
								            m_bIsChunkFullyReceived = TRUE;
								            m_dwState = CLIENT_CONNECTED;
                                        }
                                    }
                                    else
                                    {
                                        bFoundChunkInf = FALSE;    // Not enough inf. Wait for the next buffer.
                                        continue;
                                    }
                                }
                                else
                                {
                                    if(*m_messageBuffIter == '\n')
                                    {
                                        // End of trailer marks the end of message parsing.
								        m_bIsChunkFullyReceived = TRUE;
								        m_dwState = CLIENT_CONNECTED;
                                    }
                                }

                                if(m_bIsChunkFullyReceived == FALSE)
                                {
                                    // We're on one of the trailer's entity-header lines. Skip it.
                                    if(m_messageBuffIter.SkipCRLF() == TRUE)
                                        bFoundChunkInf = FALSE;    // Not enough inf. Wait for the next buffer.
                                }
                            }

							break;
						}
					};

                    if(bFoundChunkInf == FALSE)
                    {
                        // Move the iterator back to the begining. Will try again when the next buffers will come.
                        m_messageBuffIter = backupIter;
                        m_bNoMeaningFullDataLeft = TRUE;
                    }

					break;
				}

			case MSG_NO_BODY :
				{
					// ERROR since nothing should follow the header.
                    if(m_pTrace && m_pTrace->IsTrace("errors"))
                    {
                        GetTraceStream(m_pTrace) << "On connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << " : ERROR: Nothing should follow the body.\r\n";
                        ReleaseTraceStream();
                    }
                    OpenErrorStream() << "On connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << " : ERROR: Nothing should follow the body.";
                    CloseErrorStream();

					break;
				}

			case MSG_CONTENT_LENGTH :
				{
                    if(m_pTrace && m_pTrace->IsTrace("request"))
                    {
                        GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::ProcessReadData: processing a content-length body for connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ". Storing: " << m_messageBuffIter.ComputeDataLeft() << "\r\n";
                        ReleaseTraceStream();
                    }

                    StoreMessage();
					break;
				}

            case MSG_UNKNOWN_TYPE :
                {
                    // Entered in cases like Proxy-Connection or header fields missing. End of request indicated by a disconnect.

                    StoreMessage();
                    break;
                }

            case MSG_CONTINUE :
			default :
				{
					// Unknown
                    if(m_pTrace && m_pTrace->IsTrace("request"))
                    {
                        GetTraceStream(m_pTrace) << "On connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << " : Don't know the message length type.\r\n";
                        ReleaseTraceStream();
                    }

					if(m_eResultsRequired == ALL_DATA)
					{
						// TODO
					}

					break;
				}
			};  // switch(m_messageLengthType)
		}   // Message body parsing.
	}
    while(BufferParsingOver() == FALSE);
}

BOOL CAsyncHTTPRequest::IsDataTransferComplete()
{
	if(m_dwBytesRead == 0)
	{
		// Empty buffer case: transmission is complete.
		return TRUE;
	}
	else
	{
		switch(m_messageLengthType)
		{
		case MSG_NO_BODY :
			return m_bIsHeaderFullyParsed;

		case MSG_NON_IDENTITY :		// Note : "Transfer-Encoding" has priority on "Content-length".
			return IsLastChunk();

		case MSG_CONTENT_LENGTH :
			return CheckContentLength();

		case MSG_UNKNOWN_TYPE :
			// Buffer does not contain the whole header. Can't make a decision. Will have to wait till the server disconnects.
			return FALSE;

        case MSG_CONTINUE :
            return FALSE;

		default :
            /* Unexpected value. Bug */
            assert(0);
            if(m_pTrace && m_pTrace->IsTrace("request errors"))
            {
                GetTraceStream(m_pTrace) << "On connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << " : ERROR: Do not know how to evalute the end of transfer.\r\n";
                ReleaseTraceStream();
            }
            OpenErrorStream() << "On connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << " : ERROR: Do not know how to evalute the end of transfer.";
            CloseErrorStream();
			return FALSE;
		};
	}
}

BOOL CAsyncHTTPRequest::IsLastChunk()
{
	return m_bIsChunkFullyReceived;
}

BOOL CAsyncHTTPRequest::CheckContentLength()
{
    if(m_pTrace && m_pTrace->IsTrace("full"))
    {
        GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::CheckContentLength remaining " << m_dwContentLengthLeft << " bytes to download on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
        ReleaseTraceStream();
    }

	return m_dwContentLengthLeft == 0;
}

// Parse Header lines.
// Returns FALSE if line not complete.
// Note : a header line can be folded into multiple lines if the continuation line begins with a space or horizontal tab. (Cf. page 16)
BOOL CAsyncHTTPRequest::ParseHeaderLine(EHeaderFieldType &headerFieldType, CDataBuffListAccessor &iterParam, CDataBuffListAccessor &iterEndOfParam)
{
    if(m_pTrace && m_pTrace->IsTrace("request"))
    {
        string str; DWORD dwCharCount;
        CDataBuffListAccessor localIter = m_messageBuffIter;
        localIter.FindAny("\r\n", 2, dwCharCount);
        m_messageBuffIter.BuildString(str, dwCharCount);
        GetTraceStream(m_pTrace) << str << "\r\n";
        ReleaseTraceStream();
    }

    CDataBuffListAccessor iterInitial;
    iterInitial = m_messageBuffIter;

    // Skip white spaces (that might not be necessary, but we never know ...).
    if(m_messageBuffIter.SkipWS() == FALSE)
    {
        m_messageBuffIter = iterInitial;
        return FALSE;   // Found nothing but W.S.
    }

    if(m_messageBuffIter.MatchFirstWordToLower("content-length", sizeof "content-length" - 1) == TRUE)
	{
        m_messageBuffIter.Skip(sizeof "content-length" - 1);
		headerFieldType = FT_CONTENT_LENGTH;
	}
	else
	{
		if(m_messageBuffIter.MatchFirstWordToLower("transfer-encoding", sizeof "transfer-encoding" - 1) == TRUE)
		{
            m_messageBuffIter.Skip(sizeof "transfer-encoding" - 1);
			headerFieldType = FT_TRANSFER_ENCODING;
		}
		else
		{
            //// Check if this is a MESSAGE-header line.
            if(m_messageBuffIter.MatchFirstWordToLower("http/1.", sizeof "http/1." - 1) == TRUE)
			{
                m_messageBuffIter.Skip(sizeof "http/1." - 1);
				headerFieldType = FT_MESSAGE;
			}
            else
            {
		        if(m_messageBuffIter.MatchFirstWordToLower("connection", sizeof "connection" - 1) == TRUE)
		        {
                    m_messageBuffIter.Skip(sizeof "connection" - 1);
			        headerFieldType = FT_CONNECTION;
		        }
                else
                {
                    if(m_messageBuffIter.MatchFirstWordToLower("proxy-connection", sizeof "proxy-connection" - 1) == TRUE)
                    {
                        m_messageBuffIter.Skip(sizeof "proxy-connection" - 1);
			            headerFieldType = FT_PROXY_CONNECTION;
                    }
                    else
                    {
                        // If m_eResultsRequired is different from ALL_DATA then won't need the content-type info since we 
                        // already know that we don't need all the data.
                        // If m_securityMode is different from HTTP_SECURE, then knowing the content-type won't help us 
                        // enhance the performance since it will at best tell us that we just need the headers. This is usefull
                        // for SSL (because we'll save the buffers-copying overhead) but in TCP we won't be able to release the
                        // buffers in OnRequestFinished (which is the only use of m_eResultsRequired) because we cannot 
                        // release only the body buffers.
                        if(m_eResultsRequired == ALL_DATA && m_securityMode == HTTP_SECURE 
                            && m_messageBuffIter.MatchFirstWordToLower("content-type", sizeof "content-type" - 1) == TRUE)
		                {
                            m_messageBuffIter.Skip(sizeof "content-type" - 1);
			                headerFieldType = FT_CONTENT_TYPE;
		                }
			            else // if(m_eResultsRequired != ALL_DATA), decided to skip the data early in the "prepare" phase.
			            {
				            headerFieldType = FT_UNUSED;
			            }
                    }
                }
            }
		}
	}

	// Search for the param and CRLF pos.

	if(headerFieldType == FT_UNUSED)
	{
		// Not interested in that line. Check if it is the end of header and skip the next CRLF.

        if(m_messageBuffIter.IsCRLF() == TRUE)
        {
            // Current line is empty. We found the end of header statement.
			headerFieldType = FT_END_OF_HEADER;

            m_messageBuffIter.SkipCRLF();   // Do not care for the returned value: may be FALSE 
                                            // if no body section. However, called IsCRLF first 
                                            // which is enough + next access to that iterator
                                            // will be "IsEndOfBuff" (thus don't care if it's unstable).

			return TRUE;
        }
        else
        {
            // Skip unused header.
            if(m_messageBuffIter.SkipCRLF() == FALSE)
            {
			    // Could not find the CRLF sequence for that line. Wait for the next read.
                m_messageBuffIter = iterInitial;
			    return FALSE;
            }
        }
    }
	else	// Recognised a header field.
	{
		// Skip the useless bytes.

        if(headerFieldType != FT_MESSAGE)
		{
			// Skip the column.
            if(m_messageBuffIter.SkipChar(':') == FALSE)
            {
                m_messageBuffIter = iterInitial;
                return FALSE;
            }

		    // Skip the LWS.
            if(m_messageBuffIter.SkipWS() == FALSE)
            {
                m_messageBuffIter = iterInitial;
                return FALSE;
            }
        }

        iterParam = m_messageBuffIter;


        // Move iterEndOfParam to the end of the parameter string. The parameter 
        // string may be split into several lines.
        BOOL bKeepOn = TRUE;
        while(bKeepOn)
        {
            if(m_messageBuffIter.SkipCRLF() == FALSE)
		    {
			    // This was a truncated line. Wait for the next buffer.
                m_messageBuffIter = iterInitial;
			    return FALSE;
		    }
		    else
		    {
			    // Found the CRLF sequence.
                iterEndOfParam = m_messageBuffIter;
		    }

            // Check if next line belongs to the current header field.
            if(IsWS(*iterEndOfParam) == FALSE)
                bKeepOn = FALSE;
        }
	}

	return TRUE;
}

BOOL CAsyncHTTPRequest::StoreChunk()
{
    DWORD dwSkipped = m_messageBuffIter.Skip(m_dwCurrentChunkSizeLeft);
    m_dwCurrentChunkSizeLeft -= dwSkipped;

    if(m_pTrace && m_pTrace->IsTrace("request"))
    {
        GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::StoreMessage: Stored " << dwSkipped << " bytes for connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ".\r\n";
        ReleaseTraceStream();
    }

    return m_dwCurrentChunkSizeLeft == 0;
}

BOOL CAsyncHTTPRequest::SkipChunk()
{
    DWORD dwSkipped = m_messageBuffIter.Skip(m_dwCurrentChunkSizeLeft);
    m_dwCurrentChunkSizeLeft -= dwSkipped;

    if(m_pTrace && m_pTrace->IsTrace("request"))
    {
        GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::StoreMessage: Skipped " << dwSkipped << " bytes for connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ".\r\n";
        ReleaseTraceStream();
    }

	return m_dwCurrentChunkSizeLeft == 0;
}

BOOL CAsyncHTTPRequest::StoreMessage()
{
    if(m_messageLengthType == MSG_CONTENT_LENGTH)
    {
        DWORD dwSkipped = m_messageBuffIter.Skip(m_dwContentLengthLeft);
        m_dwContentLengthLeft -= dwSkipped;

        if(m_pTrace && m_pTrace->IsTrace("request"))
        {
            GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::StoreMessage: Stored " << dwSkipped << " bytes for connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ".\r\n";
            ReleaseTraceStream();
        }
        if(m_pTrace && m_pTrace->IsTrace("errors"))
        {
            DWORD dwLeft = m_messageBuffIter.ComputeDataLeft();
            if(dwLeft != 0)
            {
                GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::StoreMessage: ERROR. On connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ": Remaining " << dwLeft << " bytes despite reached CONTENT-LENGTH.\r\n";
                ReleaseTraceStream();
            }
        }
        assert(m_messageBuffIter.ComputeDataLeft() == 0);

	    return m_dwContentLengthLeft == 0;
    }
    else
    {
        // Do not know how much data we should receive.
        m_messageBuffIter.SkipEnd();
        return FALSE;
    }
}

BOOL CAsyncHTTPRequest::SkipMessage()
{
    DWORD dwSkipped = m_messageBuffIter.Skip(m_dwContentLengthLeft);
    m_dwCurrentChunkSizeLeft -= dwSkipped;

    if(m_pTrace && m_pTrace->IsTrace("request"))
    {
        GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::StoreMessage: Skipped " << dwSkipped << " bytes for connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << ".\r\n";
        ReleaseTraceStream();
    }

    return m_dwContentLengthLeft == 0;
}

// Returns TRUE if the whole buffer has been parsed:
// - read all the bytes from the last receive
// - no chunks left to read for the whole message
BOOL CAsyncHTTPRequest::BufferParsingOver()
{
	if(
        m_dwBytesRead == 0                                                              // Server disconnected.
		|| (m_bIsHeaderFullyParsed == TRUE && m_messageLengthType == MSG_NO_BODY)       // Header received and no need to wait for body.
		|| m_bIsChunkFullyReceived == TRUE                                              // No more chunks to receive (found the empty chunk).
        || m_messageBuffIter.IsEndOfBuffs()                                             // Iterator is at the end of the data buffers.
        || (m_messageLengthType == MSG_CONTENT_LENGTH && m_dwContentLengthLeft == 0)    // Received all the expected body data.
        || m_bIsHeaderFullyParsed == FALSE                                              // Did not receive the whole header fields.
                                                                                        // Must wait for next buffers.
        || m_bNoMeaningFullDataLeft == TRUE                                             // Still some bytes to parse, but not enough.
        )
	{
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}

// Returns FALSE if header not completely read.
BOOL CAsyncHTTPRequest::SkipHeader()
{
    CDataBuffListAccessor iterBegLine, iterEndLine;
	while(ParseLine(iterBegLine, iterEndLine))
	{
		if(iterEndLine-iterBegLine == 0)
		{
			return TRUE;
		}
	}

	return FALSE;
}

// Returns FALSE if header not completely read.
BOOL CAsyncHTTPRequest::StoreHeader()
{
    CDataBuffListAccessor iterBegLine, iterEndLine;
    BOOL bEndsWithEmptyLine;
	while(ParseLine(iterBegLine, iterEndLine, bEndsWithEmptyLine))
	{
		if(iterEndLine-iterBegLine == 2)
		{
			return TRUE;        // Buffer contains (but does not end with) an empty line.
		}
	}

    return bEndsWithEmptyLine;  // Buffer ends with empty line.
}

// Returns the current line and moves cursor to the next line.
// The length returned does NOT include the CRLF sequence.
// Returns FALSE if line not complete or if could not move to the character that follows the LF.
// Means that if a buffer finishes with LF, the return value will be FALSE despite the LF.
// In this last case, bEndsWithEmptyLine is set to TRUE to confirm that the line is empty.
BOOL CAsyncHTTPRequest::ParseLine(CDataBuffListAccessor & begLine, CDataBuffListAccessor &endLine, BOOL& bEndsWithEmptyLine)
{
    begLine = m_messageBuffIter;

    bEndsWithEmptyLine = m_messageBuffIter.IsCRLF();

    if(m_messageBuffIter.SkipCRLF() == TRUE)
    {
        endLine = m_messageBuffIter;
        return TRUE;
    }
    else
        return FALSE;
}

// Returns the current line and moves cursor to the next line.
// The length returned does NOT include the CRLF sequence.
// Returns FALSE if line not complete or if could not move to the character that follows the LF.
// Means that if a buffer finishes with LF, the return value will be FALSE despite the LF.
BOOL CAsyncHTTPRequest::ParseLine(CDataBuffListAccessor & begLine, CDataBuffListAccessor &endLine)
{
    begLine = m_messageBuffIter;

    if(m_messageBuffIter.SkipCRLF() == TRUE)
    {
        endLine = m_messageBuffIter;
        return TRUE;
    }
    else
        return FALSE;
}

void CAsyncHTTPRequest::CancelRequest()
{
    if(m_dwState == CLIENT_RECEIVING)
    {
        if(m_pTrace && m_pTrace->IsTrace("request"))
        {
            GetTraceStream(m_pTrace) << "CAsyncHTTPRequest::CancelRequest(): Cancellation requested while receiving on connection " << ((CAsyncRequestManager&)m_requestEventListener).GetManagerNumber() << '/' << m_dwConnectionId << "\r\n";
            ReleaseTraceStream();
        }
    }

    CAsyncRequest::CancelRequest();
}
