/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#include "stdafx.h"
#include "AsyncVirtualUserController.h"
#include "CORBARepository.h"

#ifndef CYRANO_TP_STANDALONE
#   include "CECSemaphore.h"
#endif

#include "TEMainStructure.hpp"

extern CMainExecuterStructure *g_pTExecuter;   // Defined and initialised in main.cpp.

extern CORBAServer*    pCServer;

CFastLock *CAsyncVirtualUser::s_pCacheSl = NULL;

// ----------------------------------------------------------------------------
//
//  CScriptContextData
//
// ----------------------------------------------------------------------------

CScriptContextData::CScriptContextData(DWORD dwParallelism)
{
    m_pTaskScript           = NULL;
    m_dwTRecordCallStep     = 0;
}


// ----------------------------------------------------------------------------
//
//  CAsyncVirtualUser
//
// ----------------------------------------------------------------------------

CAsyncVirtualUser::CAsyncVirtualUser()
    : m_pTimerQueue(NULL), m_dwTestCaseFail(0), m_dwTestCasePass(0),
      m_lCurTaskIteration(0), m_lCurTaskGroupIteration(0), m_bFirstTaskIteration(TRUE),
      m_dwExternalVUID(0), m_bHasAuthenticationInfo(FALSE), m_dwMaxToken(0)
#ifndef CYRANO_TP_STANDALONE
    , m_pCurrentSemaphore(NULL)
#endif
{
    assert(s_pCacheSl != NULL);
    ZeroMemory(&m_hCred, sizeof(m_hCred));
    ZeroMemory(&m_hContext, sizeof(m_hContext));
}

CAsyncVirtualUser::~CAsyncVirtualUser()
{
    if(m_bHasAuthenticationInfo)
    {
        CAsyncSecureSocket::GetSecurityFunctionTable()->DeleteSecurityContext(&m_hContext);
        CAsyncSecureSocket::GetSecurityFunctionTable()->FreeCredentialsHandle(&m_hCred);
    }
#ifndef CYRANO_TP_STANDALONE
    if(m_pCurrentSemaphore)
        delete m_pCurrentSemaphore;
#endif
}

void CAsyncVirtualUser::LogError(const std::string& strError) const
{
    g_pTExecuter->LogError(strError.c_str());
}

DWORD CAsyncVirtualUser::Init(CThreadPoolController& controller, CTimerQueue* pTimerQueue, DWORD dwManagerNumber, DWORD dwParallelism, CTrace* pTrace)
{
    DWORD dwError = 0;

    m_dwParallelism     = dwParallelism;
    m_execPhase         = EXEC_INIT;

    m_pTimerQueue = pTimerQueue;

    dwError = CAsyncHTTPRequestManager::Init(controller, dwManagerNumber, dwParallelism, pTrace);

    return dwError;
}

BOOL CAsyncVirtualUser::InitClass()
{
    assert(s_pCacheSl == NULL);

    if(s_pCacheSl == NULL)
        s_pCacheSl = new CFastLock();

    if(s_pCacheSl == NULL)
        return FALSE;
    else
        return TRUE;
}

void CAsyncVirtualUser::UninitClass()
{
    if(s_pCacheSl != NULL)
    {
        delete s_pCacheSl;
        s_pCacheSl = NULL;
    }
}

void CAsyncVirtualUser::Initialise(SharedPtr<TaskGroupDefinition>& pTaskGroup)
{
    m_pTaskGroup = pTaskGroup;

    // Initialize external VU ID.
    m_dwExternalVUID = (DWORD)InterlockedIncrement((LPLONG)&m_pTaskGroup->m_lVirtualUserID);

    // Set up thread ID.
    char zNum[32];
    wsprintf(zNum, "%ld-%ld", m_pTaskGroup->m_iTaskGroupNumber+1, m_dwExternalVUID);
    m_ThreadID = zNum;

#ifndef CYRANO_TP_STANDALONE
    m_szTestName    = g_pTExecuter->GetTestName();
#endif

    set_state(MDB_SRSIG);   // Into TContext.
}

// Run thread task(s)
// Can be called by kernel thread start routine, must be mt safe
// Code got inspired from "int Cmdbthread::run(void)" and "ThreadTask::run"

// Executes TOF sequences. Ends when the TOF needs to wait for an IO or sync. op.
// Gets called again when IO is over.
// Handles the recursive call of scripts, finishing the calling scripts execution when
// the called script is over.
// Used both internally and by CAsyncSecureSocket in case SSL Connect/Disconnect are executed.

ETofResult CAsyncVirtualUser::ProcessTof()
{
    if(GetTrace() && GetTrace()->IsTrace("full"))
    {
        GetTraceStream(GetTrace()) << "CAsyncVirtualUser::ProcessTof: thread " << GetCurrentThreadId() << " processing user " << m_dwManagerNumber << ".\r\n";
        ReleaseTraceStream();
    }

    // Process current thread state.
    m_execPhase = EXEC_RUNNING;
    ETofResult res;
    BOOL bContinue = TRUE;
    
    do
    {
        BOOL bSuccessfulUser = TRUE;
        res = CURR_SCRIPT_CTXT_DATA->m_pTaskScript->GetScript()->run(this);

        switch(res)
        {
        case TOF_RES_FINISHED:
            return TOF_RES_FINISHED;

        case TOF_RES_FAILED :
            {
                // Aborting.
                m_execPhase = EXEC_OVER;

                // Signal we are over.
                CleanUp();  // Cf. TContext or derived.

                int iOldTask = m_pCurTask - m_pTaskGroup->m_tasks.begin();
                CAsyncVirtualUserController* pController = (CAsyncVirtualUserController*)m_pController;
                pController->IncFailedUsers();
                bSuccessfulUser = FALSE;

                res = TOF_RES_FINISHED;
                // Fall through.
            }

        case TOF_RES_KILL_THREAD :
            {
                // Must terminate all scripts.
                while(m_ContextDataStack.size() > 1)
                {
                    res = EndScript(res);
                }

                if(bSuccessfulUser)
                {
                    // End of that VU simulation.
                    ((CAsyncVirtualUserController*)m_pController)->IncSuccessfullUsers();
                }

                bContinue = FALSE;

                res = EndScript(res);

                break;
            }

        case TOF_RES_ASYNC_STARTED :
            {
                // IO under process. Let the framework do its job and call us again.
                return TOF_RES_ASYNC_STARTED;
            }

        case TOF_RES_OVER :
            {
                // End-of-Executable-Records for the current thread. Remaining records are procedure definitions.
                // If this is the base-script, then signal the controller we are over with our TOF list.

                m_execPhase = EXEC_OVER;

                // End TOF processing.
                size_t ctsize = m_ContextDataStack.size();
                if(ctsize == 1)
                {
                    // End of that VU simulation.
                    ((CAsyncVirtualUserController*)m_pController)->IncSuccessfullUsers();
                    bContinue = FALSE;
                }
                else
                {
                    // End of current script. Will continue with the calling one.
                }

                if(ctsize > 0)
                {
                    ETofResult res2 = EndScript(res);
                    if(res2 != TOF_RES_ASYNC_STARTED && m_ContextDataStack.size() == 0)
                        return TOF_RES_FINISHED;
                    else
                        res = res2;
                }

                break;
            }

/*
        case TOF_RES_OK :
        case TOF_SKIP_TO_NEXT :
            {
                // Never used or error since processed in TScript::run().
            }
*/
        default :
            {
                // Error: should have been processed in TScript::run.
                assert(0);
                break;
            }
        };
    }
    while(bContinue == TRUE);

    return res;
}


void CAsyncVirtualUser::OnRequestFinished(DWORD dwRequestSlotNumber)
{
    ProcessEndOfRequest(dwRequestSlotNumber, FALSE);
}

void CAsyncVirtualUser::OnRequestCancelled(DWORD dwRequestSlotNumber)
{
    ProcessEndOfRequest(dwRequestSlotNumber, TRUE);
}

// --------------------------------------------------------------------------------
// Called when a request is over.
// If a synchronisation operation was initiated then the TOF processing for that
// VU was waiting for us, thus must enter ProcessTof again.
// Otherwise, the TOF list is being processed in an other thread.
// --------------------------------------------------------------------------------
void CAsyncVirtualUser::ProcessEndOfRequest(DWORD dwRequestSlotNumber, BOOL bWasCancelled)
{
    assert(dwRequestSlotNumber != ULONG_MAX);

    DWORD           dwLocPendingRequests;
    EVuExecPhase    locExecPhase;
    DWORD           dwHTTPBytes;

    {
        CLock guard (m_sl);

        // Code based on CAsyncRequestManager::OnRequestFinished.
        --m_dwPendingRequests; // Code based on CAsyncRequestManager::OnRequestFinished.

        CAsyncHTTPRequest* pHTTPRequest = (CAsyncHTTPRequest*)m_requests[dwRequestSlotNumber];
        string zTimerName = pHTTPRequest->GetTimerName();
        dwHTTPBytes       = pHTTPRequest->GetHTTPBytes();
        if(!zTimerName.empty())
            StopTimer(zTimerName);

        dwLocPendingRequests = m_dwPendingRequests;

        locExecPhase = m_execPhase;
    }

    CAsyncVirtualUserController* pController = (CAsyncVirtualUserController*)m_pController;
    pController->OnRequestComplete(m_pTaskGroup->m_iTaskGroupNumber, dwHTTPBytes);

    // Handle synchronisation requests, and move further.
    switch(locExecPhase)
    {
    case EXEC_HTTP_SYNC :
        {
            // Last TOF object required to wait for a specific HTTP request to complete.
            if(((CAsyncHTTPRequest*)m_requests[dwRequestSlotNumber])->GetConnectionId() == m_dwWaitedConnId)
            {
                // This is THE request we were waiting for. Get back to work. Usually, will process the current request's results.
                if(GetTrace() && GetTrace()->IsTrace("full"))
                {
                    GetTraceStream(GetTrace()) << "CAsyncVirtualUser::OnRequest" << (bWasCancelled?"Cancelled":"Finished") << ": connection " << m_dwManagerNumber << '/' << m_dwWaitedConnId << " is over and was waited upon.\r\n";
                    ReleaseTraceStream();
                }

                ProcessTof();   // No need to handle the results.
            }
            else
            {
                if(GetTrace() && GetTrace()->IsTrace("full"))
                {
                    GetTraceStream(GetTrace()) << "CAsyncVirtualUser::OnRequest" << (bWasCancelled?"Cancelled":"Finished") << ": user " << m_dwManagerNumber << " waiting for " << m_dwWaitedConnId << ". " << dwLocPendingRequests << " pending.\r\n";
                    ReleaseTraceStream();
                }
            }
            // Otherwise, don't care for that request's result.
            break;
        }

    case EXEC_ALL_HTTP_SYNC :
        {
            // Last TOF object required to wait for all the HTTP requests to complete.
            if(dwLocPendingRequests == 0)
            {
                // This was the last executing request in the list.
                if(GetTrace() && GetTrace()->IsTrace("full"))
                {
                    GetTraceStream(GetTrace()) << "CAsyncVirtualUser::OnRequest" << (bWasCancelled?"Cancelled":"Finished") << ": user " << m_dwManagerNumber << " releasing the sync. mechanism.\r\n";
                    ReleaseTraceStream();
                }

                ProcessTof();   // No need to handle the results.
            }
            else
            {
                if(GetTrace() && GetTrace()->IsTrace("full"))
                {
                    GetTraceStream(GetTrace()) << "CAsyncVirtualUser::OnRequest" << (bWasCancelled?"Cancelled":"Finished") << ": user " << m_dwManagerNumber << " waiting for " << dwLocPendingRequests << " more requests to complete.\r\n";
                    ReleaseTraceStream();
                }
            }
            break;
        }

    default :
        {
            // Nothing particular to do.
            if(GetTrace() && GetTrace()->IsTrace("full"))
            {
                GetTraceStream(GetTrace()) << "CAsyncVirtualUser::OnRequest" << (bWasCancelled?"Cancelled":"Finished") << ": connection " << m_dwManagerNumber << '/' << ((CAsyncHTTPRequest*)m_requests[dwRequestSlotNumber])->GetConnectionId() << " is over.\r\n";
                ReleaseTraceStream();
            }
            break;
        }
    };
}

void CAsyncVirtualUser::OnEvent(DWORD /*dwNumberOfBytesTransferred*/, BOOL bIsIo, DWORD dwEventType)
{
    if(bIsIo == TRUE)
    {
        if(GetTrace() && GetTrace()->IsTrace("errors"))
        {
            GetTraceStream(m_pTrace) << "CAsyncVirtualUser::OnEvent: ERROR. Client " << m_dwManagerNumber << " received an unexpected IO.\r\n";
            ReleaseTraceStream();
        }
    }
    else
    {
        switch(dwEventType)
        {
        case CAsyncVirtualUser::EVENT_TYPE_VIRTUAL_USER_START:
            {
                if(GetTrace() && GetTrace()->IsTrace("full"))
                {
                    GetTraceStream(GetTrace()) << "CAsyncVirtualUser::OnEvent: client " << m_dwManagerNumber << " is starting.\r\n";
                    ReleaseTraceStream();
                }

                OnStartProcessing();
                break;
            }

        case IocpEvent::EVENT_TYPE_TIMER:
            {
                if(GetTrace() && GetTrace()->IsTrace("full"))
                {
                    GetTraceStream(GetTrace()) << "CAsyncVirtualUser::OnEvent: client " << m_dwManagerNumber << " received a timer event.\r\n";
                    ReleaseTraceStream();
                }

                OnTimerFinished();
                break;
            }

        default:
            if(GetTrace() && GetTrace()->IsTrace("errors"))
            {
                GetTraceStream(GetTrace()) << "CAsyncVirtualUser::OnEvent: client " << m_dwManagerNumber << " received an unsupported event: " << dwEventType << ".\r\n";
                ReleaseTraceStream();
            }
        };
    }
}

void CAsyncVirtualUser::OnEventFailed(DWORD dwNumberOfBytesTransferred, BOOL bIsIo, DWORD dwEventType, DWORD dwError)
{
    if(GetTrace() && GetTrace()->IsTrace("errors"))
    {
        GetTraceStream(GetTrace()) << "CAsyncVirtualUser::OnEventFailed: ERROR. Client " << m_dwManagerNumber << " failed.\r\n";
        ReleaseTraceStream();
    }

    if(bIsIo == TRUE)
    {
        if(GetTrace() && GetTrace()->IsTrace("errors"))
        {
            GetTraceStream(m_pTrace) << "CAsyncVirtualUser::OnEventFailed: ERROR. Client " << m_dwManagerNumber << " received an unexpected IO.\r\n";
            ReleaseTraceStream();
        }
    }
    else
    {
        // Execute despite an error was received.
        switch(dwEventType)
        {
            case IocpEvent::EVENT_TYPE_TIMER:
                OnTimerFinished();
                break;

            default:
                if(GetTrace() && GetTrace()->IsTrace("errors"))
                {
                    GetTraceStream(GetTrace()) << "CAsyncVirtualUser::OnEventFailed: client " << m_dwManagerNumber << " received an unsupported event: " << dwEventType << ".\r\n";
                    ReleaseTraceStream();
                }
        };
    }
}

void CAsyncVirtualUser::OnStartProcessing()
{
    if(GetTrace() && GetTrace()->IsTrace("Startup"))
    {
        GetTraceStream(GetTrace()) << "CAsyncVirtualUser::OnStartProcessing: client " << m_dwManagerNumber << " is starting.\r\n";
        ReleaseTraceStream();
    }

    CAsyncVirtualUserController* pController = (CAsyncVirtualUserController*)m_pController;

    // Increment the number of active VUs on the task group.
    pController->IncTaskGroupVUs(m_pTaskGroup);

    // Initialize the task group iteration count.
    m_lCurTaskGroupIteration = 1;

    // Initialize the task iteration count.
    m_lCurTaskIteration = 1;

    // Get the first task.
    m_pCurTask = pController->GetFirstTask(m_pTaskGroup);

    m_execPhase = EXEC_RUNNING;
    time_t t;
    time(&t);
    m_dwStartTime = t;

    StartScript(m_pCurTask->m_strScriptName);
}

void CAsyncVirtualUser::OnTimerFinished()
{
    switch(m_execPhase)
    {
        case EXEC_WAIT_DELAY:
            ((CAsyncVirtualUserController*)m_pController)->OnUserEnterTask(*this, m_pTaskGroup->m_iTaskGroupNumber, m_pCurTask - m_pTaskGroup->m_tasks.begin());
            // Fall through.
        case EXEC_WAIT_TIMER:
            ProcessTof();   // Nb: results handled in ProcessTof().
            break;
        default:
            assert(0);      // Internal error.
    }
}

void CAsyncVirtualUser::OnTimerCancelled()
{
    if(m_execPhase == EXEC_WAIT_TIMER)
    {
        ProcessTof();   // Nb: results handled in ProcessTof().
    }
    else
    {
        // Strange: How come we were signaled?
        assert(0);
    }
}

// --------------------------------------------------------------------------------
// An async (SSL) disconnect happened. Must resume TOF processing.
// --------------------------------------------------------------------------------
void CAsyncVirtualUser::OnRequestClosed(DWORD dwRequestNumber)
{
    CLock lock(m_sl);

    // Call base class' implementation.
    CAsyncRequestManager::OnRequestClosed(dwRequestNumber);

    if(m_dwDisconnectToWait > 0)
        --m_dwDisconnectToWait;

    if(m_execPhase == EXEC_WAIT_DISCONNECTS)
    {
        if(m_dwDisconnectToWait == 0)
        {
            if(GetTrace() && GetTrace()->IsTrace("connect sync"))
            {
                GetTraceStream(GetTrace()) << "EXEC WAIT DISCONNECTS: all connections disconnected.\r\n";
                ReleaseTraceStream();
            }

            ProcessTof();   // No need to handle the results.
        }
    }
}


///////////////////////////////////////////////////////////////////////////////
////
////  Code which purpose is HTTP request slot selection / allocation / connection.
////
///////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------------
// GetRequestFromConnId
//
// As connection IDs are incremented each time a new connection is established,
// that method relates a connection ID to the connection slot it was assigned at
// connection time.
//
// Protected method. Serialization must be handled by calling methods.
// --------------------------------------------------------------------------------
CAsyncHTTPRequest* CAsyncVirtualUser::GetRequestFromConnId(DWORD dwConnectionId) const
{
    assert(dwConnectionId != ULONG_MAX);

    // Search for the right slot.
    DWORD dwSize = m_requests.size();
    for(DWORD dwIndice=0 ; dwIndice<dwSize ; ++dwIndice)
    {
        if(((CAsyncHTTPRequest*)m_requests[dwIndice])->GetConnectionId() == dwConnectionId)
        {
            return (CAsyncHTTPRequest*)m_requests[dwIndice];
        }
    }

    return NULL;
}

// --------------------------------------------------------------------------------
// GetNotConnectedRequest
//
// Find a free request. If none is available, then return NULL.
// Protected method. Serialization must be handled by calling methods.
// --------------------------------------------------------------------------------
CAsyncRequest* CAsyncVirtualUser::GetNotConnectedRequest()
{
    // Return the first request found whose processing is over and for which results are no more used.
    DWORD dwReqId = CAsyncRequestManager::GetAvailableRequest();

    if(dwReqId == ULONG_MAX)
    {
        // Allocate a new request. Would not be used if TOF was recorded from an HTTP-compliant browser (max 4 connections).
        CAsyncHTTPRequest* pNewReq = new CAsyncHTTPRequest(m_pController->GetIocp(), *this, m_requests.size(), m_pTrace);
        pNewReq->Init();

        AddRequest(pNewReq);

        return pNewReq;
    }
    else
    {
        return m_requests[dwReqId];
    }
}

// --------------------------------------------------------------------------------
// ObtainRequestStruct
//
// Retrieve or allocate a request slot and assign it to connection ID dwConnectionId. 
// Called by TRcrdConnect and TRcrdRequest.
// --------------------------------------------------------------------------------
CAsyncHTTPRequest* CAsyncVirtualUser::ObtainRequestStruct(DWORD dwConnectionId)
{
    CAsyncHTTPRequest* pNewReq;

    {
        CLock lock(m_sl);

        // First check if that connection has already been allocated.
        pNewReq = GetRequestFromConnId(dwConnectionId);
        if(pNewReq == NULL)
        {
            // No connection allocated for that connection ID. Allocate a new one.
            pNewReq = (CAsyncHTTPRequest*)GetNotConnectedRequest();
            if(pNewReq != NULL)
            {
                pNewReq->SetConnectionId(dwConnectionId);
            }
            else
            {
                // Unexpected error.
                assert(0);
                return NULL;
            }
        }
    }

    if(GetTrace() && GetTrace()->IsTrace("connection"))
    {
        GetTraceStream(GetTrace()) << "CAsyncVirtualUser::ObtainRequestStruct associated request slot " << pNewReq->GetRequestNumber() << " to connection ID " << dwConnectionId << "\r\n";
        ReleaseTraceStream();
    }

    return pNewReq;
}

// --------------------------------------------------------------------------------
// ReleaseRequest
//
// Release the connection slot associated with dwConnectionId.
// --------------------------------------------------------------------------------
void CAsyncVirtualUser::ReleaseRequest(DWORD dwConnectionId)
{
    CLock lockReq (m_sl);

    CAsyncHTTPRequest* pReq = GetRequestFromConnId(dwConnectionId);
    if(pReq != NULL)
    {
        ReleaseRequest(pReq);
    }
    else
    {
        assert(0);
    }
}

// --------------------------------------------------------------------------------
// ReleaseRequest
//
// Release the connection slot associated with the connection slot pointed at by
// pReq.
// --------------------------------------------------------------------------------
void CAsyncVirtualUser::ReleaseRequest(CAsyncHTTPRequest* pReq)
{
    assert(pReq != NULL);
    CLock lock(m_sl);

    pReq->OnRequestClosed();
}

// --------------------------------------------------------------------------------
// --------------------------------------------------------------------------------
DWORD CAsyncVirtualUser::DisconnectClient(DWORD dwConnectionId)
{
    CLock lock(m_sl);

    // Search the slot owned by dwConnectionId.
    CAsyncHTTPRequest* pReq = GetRequestFromConnId(dwConnectionId);
    if(pReq != NULL)
    {
        if(     pReq->GetStatus() != CAsyncHTTPRequest::CLIENT_INITIAL
            &&  pReq->GetStatus() != CAsyncHTTPRequest::CLIENT_CONNECTED)
        {
            // Should not happen (must first terminate the request). Otherwise, chances are status changed inbetween.
            // May result in events from a previous request being considered as results of the next one?
            if(GetTrace() && GetTrace()->IsTrace("errors"))
            {
                GetTraceStream(GetTrace()) << "DisconnectClient: ERROR. Wrong request status: " << pReq->GetStatus() << ".\r\n";
                ReleaseTraceStream();
            }

            assert(0);
        }

        m_dwDisconnectToWait = 1;

        if(pReq->Disconnect() == ERROR_IO_PENDING)
        {
            m_execPhase = EXEC_WAIT_DISCONNECTS;

            return TOF_RES_ASYNC_STARTED;
        }
        else
            return TOF_RES_OK;
    }
    else
    {
        // WARNING: Did not find that connectId. Probably did a Disconnect on a connection that does not exist.
        if(GetTrace() && GetTrace()->IsTrace("connection"))
        {
            GetTraceStream(GetTrace()) << "WARNING: DISCONNECT CLIENT: could not find a connection with id " << dwConnectionId << ". Already disconnected?\r\n";
            ReleaseTraceStream();
        }

        return TOF_RES_OK;
    }
}

// --------------------------------------------------------------------------------
// --------------------------------------------------------------------------------
DWORD CAsyncVirtualUser::DisconnectAllClients()
{
    CLock lock(m_sl);

    m_dwDisconnectToWait = 0;

    CAsyncHTTPRequest* pCurrRequest;

    for(DWORD dw=0 ; dw<m_requests.size() ; ++dw)
    {
        pCurrRequest = (CAsyncHTTPRequest*)m_requests[dw];
        if(pCurrRequest->GetConnectionId() != ULONG_MAX)
        {
            if(    pCurrRequest->GetStatus() != CAsyncHTTPRequest::CLIENT_INITIAL
                && pCurrRequest->GetStatus() != CAsyncHTTPRequest::CLIENT_CONNECTED)
            {
                // Should not happen (must first terminate the request). Otherwise, chances are status changed inbetween.
                // May result in events from a previous request being considered as results of the next one?
                if(GetTrace() && GetTrace()->IsTrace("errors"))
                {
                    GetTraceStream(GetTrace()) << "DisconnectAllClients: ERROR. Wrong request status: " << pCurrRequest->GetStatus() << ".\r\n";
                    ReleaseTraceStream();
                }

                assert(0);
            }

            if(GetTrace() && GetTrace()->IsTrace("connection"))
            {
                GetTraceStream(GetTrace()) << "DisconnectAllClients: disconnecting connection " << pCurrRequest->GetConnectionId() << ".\r\n";
                ReleaseTraceStream();
            }

            ++m_dwDisconnectToWait;

            if(pCurrRequest->Disconnect() == ERROR_IO_PENDING)
            {
                m_execPhase = EXEC_WAIT_DISCONNECTS;
            }
        }
    }

    if(m_execPhase == EXEC_WAIT_DISCONNECTS)
        return TOF_RES_ASYNC_STARTED;
    else
        return TOF_RES_OK;
}

///////////////////////////////////////////////////////////////////////////////
//
//  Code which purpose is synchronisation.
//
///////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------------
// Return FALSE if connection dwClientConnId is still busy (thus caller must wait), 
// or TRUE if it's free or not created yet. Note CLIENT_CONNECTED is not a busy 
// status.
// --------------------------------------------------------------------------------
BOOL CAsyncVirtualUser::SynchroniseHTTPClient(DWORD dwClientConnId)
{
    assert(dwClientConnId != ULONG_MAX);
    BOOL bIsUnderProcess = FALSE;

    CLock lock(m_sl);

    // Search the slot owned by dwConnectionId.
    CAsyncHTTPRequest* pReq = GetRequestFromConnId(dwClientConnId);
    if(pReq != NULL)
    {
        CAsyncHTTPRequest::EClientStateType status = pReq->GetStatus();

        if(    status != CAsyncHTTPRequest::CLIENT_INITIAL
            && status != CAsyncHTTPRequest::CLIENT_CONNECTED)
        {
            if(GetTrace() && GetTrace()->IsTrace("sync"))
            {
                GetTraceStream(GetTrace()) << "SYNC HTTP CLIENT Initiating a wait operation on connection " << dwClientConnId << "\r\n";
                ReleaseTraceStream();
            }

            m_dwWaitedConnId = dwClientConnId;
            m_execPhase = EXEC_HTTP_SYNC;

            return FALSE;
        }
    }

    // dwClientConnId is sleeping.
    return TRUE;
}

// --------------------------------------------------------------------------------
// Return TRUE if no request is busy anymore.
// --------------------------------------------------------------------------------
BOOL CAsyncVirtualUser::SynchroniseAllHTTPClients()
{
    CLock lock(m_sl);

    if(m_dwPendingRequests == 0)
    {
        if(GetTrace() && GetTrace()->IsTrace("request"))
        {
            GetTraceStream(GetTrace()) << "SYNC ALL HTTP CLIENTS: all requests completed.\r\n";
            ReleaseTraceStream();
        }
        return TRUE;
    }
    else
    {
        if(GetTrace() && GetTrace()->IsTrace("request"))
        {
            GetTraceStream(GetTrace()) << "SYNC ALL HTTP CLIENTS: " << m_dwPendingRequests << " requests under process.\r\n";
            ReleaseTraceStream();
        }
        m_execPhase = EXEC_ALL_HTTP_SYNC;
        return FALSE;
    }
}

// --------------------------------------------------------------------------------
// Ask the timer object to warn us when the specified delay is elapsed.
// --------------------------------------------------------------------------------
BOOL CAsyncVirtualUser::StartEventTimer(DWORD dwNbMilliSeconds)
{
    status = MDB_SLEEP;

    assert(m_execPhase == EXEC_RUNNING || m_execPhase == EXEC_INIT);

    m_execPhase = EXEC_WAIT_TIMER;
    if(GetTrace() && GetTrace()->IsTrace("sync"))
    {
        GetTraceStream(GetTrace()) << "START TIMER Initiating a " << dwNbMilliSeconds*timefac/100 << " milliseconds wait operation for user " << m_dwManagerNumber << ".\r\n";
        ReleaseTraceStream();
    }
    m_pTimerQueue->Push((IocpEventDispatcher*) this, dwNbMilliSeconds*timefac/100);

    return TRUE;
}

void CAsyncVirtualUser::CancelQuery(DWORD dwConnectionId)
{
    // Search the slot owned by dwConnectionId.
    CLock lock(m_sl);
    RequestsType::iterator itEnd = m_requests.end();
    for(RequestsType::iterator it = m_requests.begin(); it != itEnd; ++it)
    {
        CAsyncHTTPRequest *pCurrReq = static_cast<CAsyncHTTPRequest*>(*it);
        if(pCurrReq->GetConnectionId() == dwConnectionId)
        {
            pCurrReq->CancelRequest();

            return;
        }
    }
}

///////////////////////////////////////////////////////////////////////////////
//
//  Code for script execution.
//
///////////////////////////////////////////////////////////////////////////////

void CAsyncVirtualUser::OnNoteText(const std::string& strNote)
{
    m_noteText = strNote;
}

// Inspired from TEThread::StartScript.
ETofResult CAsyncVirtualUser::StartScript(const string& fullscriptname)
{
    string scriptname = strip_extension(filename_base(fullscriptname));

    m_ContextDataStack.push(new CScriptContextData(m_dwParallelism));
    m_ScriptNameStack.push(scriptname);
    set_file(scriptname);

    if(GetTrace() && GetTrace()->IsTrace("full"))
    {
        GetTraceStream(m_pTrace) << "CAsyncVirtualUser::StartScript: pushed new script context for script '"
            << scriptname << "'. Stack depth = " << m_ContextDataStack.size() << ".\r\n";
        ReleaseTraceStream();
    }

    m_ContextDataStack.top()->exit_called_flag = FALSE;   // TODO : not in use anymore?

#ifndef CYRANO_TP_STANDALONE
    // Register script for Variable factory.
    CECInetInfo inet;
    CORBA::Object_var obj = pCServer->FindObjectWithName(VarFactory::CLASS_NAME,
        inet.GetIPAddressAsString(g_pTExecuter->GetRepositoryHost()).c_str());
    VarFactory_var VF = VarFactory::_narrow( obj.inout());
    CORBA::String_var cb_scriptname = CORBA::string_dup(ScriptName().c_str());
    VF->RegisterScript(cb_scriptname.in());
#endif

    // If a task duration has been specified, set the end timestamp.
    if(m_pCurTask->m_lDuration)
    {
        FILETIME ftNow;
        GetSystemTimeAsFileTime(&ftNow);
        m_qwEndTask = ftNow.dwHighDateTime;
        m_qwEndTask <<= 32;
        m_qwEndTask |= ftNow.dwLowDateTime;
        unsigned __int64 qwDuration = UInt32x32To64(m_pCurTask->m_lDuration * 1000, 10000L);    // Convert duration from milliseconds to hundreds of nanoseconds.
        m_qwEndTask += qwDuration;
    }

    return BeginTask(fullscriptname);
}

ETofResult CAsyncVirtualUser::ReStartScript()
{
    m_ContextDataStack.top()->exit_called_flag = FALSE;   // TODO : not in use anymore?
    return BeginTask(m_ScriptNameStack.top());
}

ETofResult CAsyncVirtualUser::BeginTask(const std::string &szFileName)
{
    ETofResult res = TOF_RES_OK;

    SetParamList(TtValList());  // Actual param list is empty for top-level script!

    // Get parameter data.
    const list<CECExprDsc>* varExprListP = m_paramsInfo.ExprListP();

    // Clear parameter list.
    m_Params.clear();

    if(varExprListP->empty() == FALSE)
    {
        BOOL eval = m_paramsInfo.EvaluateValues(*this, m_Params);
        if(!eval)
        {
            string strNm = "Unable to Evaluate Parameter values for Script '" + szFileName + "'";
            g_pTExecuter->LogError(strNm.c_str());
            throw mdbParseException("Unable to Evaluate Parameter values for Script");
        }
    }

    if(CreateTask(szFileName) == FALSE)
    {
        // Error. Cannot start VU.
        assert(0);
        return res;
    }

    // The first iteration of the first task is not preceded by a delay.
    // A zero delay means no delay.
    if(m_bFirstTaskIteration
        ||  m_pCurTask->m_bFixedDelay
                && m_pCurTask->m_delay.m_iFixedDelay == 0
        || !m_pCurTask->m_bFixedDelay
                && m_pCurTask->m_delay.m_variableDelay.m_iMinDelay == 0
                && m_pCurTask->m_delay.m_variableDelay.m_iMaxDelay == 0)
    {
        // Execute without delay.
        m_bFirstTaskIteration = FALSE;
        ((CAsyncVirtualUserController*)m_pController)->OnUserEnterTask(*this, m_pTaskGroup->m_iTaskGroupNumber, m_pCurTask - m_pTaskGroup->m_tasks.begin());
        res = ProcessTof();   // Nb: results handled in ProcessTof().
    }
    else
    {
        // Generate a delay!
        DWORD dwNbMilliSeconds;
        if(m_pCurTask->m_bFixedDelay)
            dwNbMilliSeconds = m_pCurTask->m_delay.m_iFixedDelay * 1000;
        else
        {
            DWORD dwMin = m_pCurTask->m_delay.m_variableDelay.m_iMinDelay*1000;
            DWORD dwMax = m_pCurTask->m_delay.m_variableDelay.m_iMaxDelay*1000;
            srand(GetTickCount());
            dwNbMilliSeconds = (rand() << 15) | rand(); // 30 bit random number.
            dwNbMilliSeconds = dwMin + (dwNbMilliSeconds % (dwMax - dwMin + 1));
        }

        status = MDB_SLEEP;
        m_execPhase = EXEC_WAIT_DELAY;
        if(GetTrace() && GetTrace()->IsTrace("full"))
        {
            GetTraceStream(GetTrace()) << "Initiating a " << dwNbMilliSeconds << " milliseconds wait operation for user " << m_dwManagerNumber
                << " before executing task '" << m_pCurTask->m_strScriptName << "'...\r\n";
            ReleaseTraceStream();
        }
        m_pTimerQueue->Push((IocpEventDispatcher*) this, dwNbMilliSeconds);
    }

    return res;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc Call another script within this thread
// @parm  ScriptName 
// @parm  parameters to be passed to script
// @rdesc -1 for failure
//
// Modifies the calling environment in the context in order to handle the new script.
// Current environment is stored in order to allow to continue with the calling script
// when this one is over.

// Inspired by TEThread::CallScript.
int CAsyncVirtualUser::CallScript(const string& name)
{
    // Get parameter data before we stack up.
    const list<CECExprDsc>* varExprListP = m_paramsInfo.ExprListP();

    // Clear parameter list.
    m_Params.clear();

    if(varExprListP->empty() == FALSE)
    {
        BOOL eval = m_paramsInfo.EvaluateValues(*this, m_Params);
        if(!eval)
        {
            string strNm = "Unable to Evaluate Parameter values for Call Script '" + name + "'";
            g_pTExecuter->LogError(strNm.c_str());
            throw mdbParseException("Unable to Evaluate Parameter values for Call Script");
        }
    }

    // Stack up member data.
    CScriptContextData* pData   = new CScriptContextData(m_dwParallelism);
    CScriptContextData* pTop    = CURR_SCRIPT_CTXT_DATA;
    pData->phase                = pTop->phase;
    pData->phasetxt             = "Start New Script";
    pData->flags                = pTop->flags;
    pData->oi                   = -1;

    // Add threadwide variables from previous varlist to this one.
    pData->varlist.add_vars(pTop->varlist.get_threadwide_vars());
    
    m_ContextDataStack.push(pData);
    m_ScriptNameStack.push(name);

    if(GetTrace() && GetTrace()->IsTrace("full"))
    {
        GetTraceStream(m_pTrace) << "CAsyncVirtualUser::CallScript: pushed new script context for script '"
            << name << "'. Stack depth = " << m_ContextDataStack.size() << ".\r\n";
        ReleaseTraceStream();
    }

    if(CreateTask(name) == FALSE)
        return TOF_RES_FAILED;

    // get parameter data
    // run task, Note: this calls the StartScript and EndScript methods in 
    // derived TContext classes for specific setup/clear down
    int iRet;   // Return value, -1 indicates an error.

    try
    {
        iRet = ProcessTof();
    }
    catch(const DetachThreadException& d)//if thread detatching pass on up to top
    {
        throw d;
    }
    catch(...)
    {
        // Catch all exceptions for now and treat as pass on to error handling to ensure neater closedown.
        iRet = -1;
    }

    return iRet;
}

BOOL CAsyncVirtualUser::CreateTask(const string &szFileName)
{
    if(GetVerbose(AUDITMSG))
        log("Start Script", AUDITMSG);

    // Get new file based upon szFileName.

    // Bodge to stop unnecessary checking out of files from repository.

    // Adjust file name.
    string szScriptName = strip_extension(filename_base(szFileName));

#ifndef CYRANO_TP_STANDALONE
    string tmpString = g_pTExecuter->GetCurrentDir() + FN_PATHSEP + szScriptName + ".tof";
#else
    string tmpString = g_pTExecuter->GetTofFileDir() + FN_PATHSEP + szScriptName + ".tof";
#endif

    set_file(tmpString);    // Push latest FULL file name onto stack.

    string ScriptFilename;

    // Put a mutex lock around the code to retrieve the file.
    CLock guard(s_pCacheSl);

#ifdef CYRANO_TP_STANDALONE
    ScriptFilename = tmpString;
#else
    if(((CAsyncVirtualUserController*)m_pController)->get_cache().doesTaskExist(tmpString) == FALSE)
    {
        if(GetTrace() && GetTrace()->IsTrace("Debug-Generic"))
        {
            GetTraceStream(m_pTrace) << "CAsyncVirtualUser::CreateTask: Get the File: " << szFileName << ".\r\n";
            ReleaseTraceStream();
        }

        try
        {
            ScriptFilename = szScriptName + ".tof";
            g_pTExecuter->GetInjectorControl()->GetTofFile(ScriptFilename.c_str());
            TCHAR zPath[MAX_PATH];
            GetModuleFileName(NULL, zPath, sizeof(zPath));
            ScriptFilename = filename_dir(zPath) + '\\' + ScriptFilename;
        }
        catch(...)
        {
            if(GetTrace() && GetTrace()->IsTrace("full"))
            {
                GetTraceStream(GetTrace()) << "Unable to retrieve File: " << szFileName << ".\r\n";
                ReleaseTraceStream();
            }
            string strMsg = "Unable to retrieve File: '" + szFileName + "' from repository.";
            g_pTExecuter->LogError(strMsg.c_str());
            return FALSE;
        }
    }
    else
    {
        ScriptFilename = tmpString;
    }
#endif

    if(ScriptFilename.empty())
    {
        if(GetTrace() && GetTrace()->IsTrace("full"))
        {
            GetTraceStream(GetTrace()) << "Unable to retrieve File: " << szFileName << ".\r\n";
            ReleaseTraceStream();
        }
        string strMsg = "Unable to retrieve File: '" + szFileName + "' from repository or cache.";
        g_pTExecuter->LogError(strMsg.c_str());
        return FALSE;
    }


    //// Create new task
    try
    {
        // Create new task, replacing current.
        CURR_SCRIPT_CTXT_DATA->m_pTaskScript = ((CAsyncVirtualUserController*)m_pController)->get_cache().makeTask(ScriptFilename);
        assert(CURR_SCRIPT_CTXT_DATA->m_pTaskScript != NULL);
        SetScript(CURR_SCRIPT_CTXT_DATA->m_pTaskScript->GetScript());
    }
    catch(...)
    {
        if(GetVerbose(ERRMSG))
            log("Error creating new Script Task for " + ScriptFilename, ERRMSG);
        string strMsg = "Error creating new Script Task for " + ScriptFilename;
        g_pTExecuter->LogError(strMsg.c_str());

        return FALSE;
    }


    //// Initialise the current script status. Used to be done in TScript::run.
    // fill out the context with this scripts variable definitions
    declarations(CURR_SCRIPT_CTXT_DATA->m_pTaskScript->GetScript()->GetTrecDefs());

    // Open any external datafiles
    //pCtxt.open_datafiles();   // TODO?

    // Clean out the OpenObjects stack
    while (!OpenObjects().empty())
    {
        OpenObjects().pop();
    }

    ClearMask(~CTXT_DEFAULT);

    // Get record count for script. This must be done at least once
    // after file is loaded to initialize record holder iterator array.
    SetScriptLength(CURR_SCRIPT_CTXT_DATA->m_pTaskScript->GetScript()->length());
    SetScriptPos(0);

    return TRUE;
}


// Called when we are over with the called script.
// Set the environment for the calling script again.
// TODO : look at TEThread::EndScript() + ExitCommand.
ETofResult CAsyncVirtualUser::EndScript(ETofResult res)
{
    if(GetVerbose(AUDITMSG))
        log("End Script", AUDITMSG);

    if(GetTrace() && GetTrace()->IsTrace("full"))
    {
        GetTraceStream(GetTrace()) << "VU " << m_dwManagerNumber
            << " entering CAsyncVirtualUser::EndScript() for script '" << GetCurrScriptName()
            << "'\r\n";
        ReleaseTraceStream();
    }

    if(!m_ContextDataStack.top()->exit_called_flag)
    	ExitCommand(FALSE, "");

    DWORD dwContextStackSize = m_ContextDataStack.size();
    if(dwContextStackSize == 1)
    {
        // The highest level script is over.
        m_pController->OnAllRequestsFinished(m_dwManagerNumber);

        // Get the current task's number.
        int iOldTask = m_pCurTask - m_pTaskGroup->m_tasks.begin();
        CAsyncVirtualUserController* pController = (CAsyncVirtualUserController*)m_pController;

        // If the task group has been disabled (as is done typically when
        // the Test Manager asks for a task group to stop), then stop processing
        // the task group.
        // NOTE: the current implementation implies that the task group stops
        // only after all virtual users running it have completed their current
        // task iteration.
        // If the virtual user has failed or was aborted, then it must also exit.
        if(!m_pTaskGroup->m_bTaskGroupEnabled || res == TOF_RES_FINISHED)
        {
            if(GetTrace() && GetTrace()->IsTrace("full"))
            {
                GetTraceStream(GetTrace()) << "CAsyncVirtualUser::EndScript: Task Group execution interrupted for VU " << m_dwManagerNumber << " because the task group is disabled.\r\n";
                ReleaseTraceStream();
            }
            pController->OnUserExitTask(*this,
                m_pTaskGroup->m_iTaskGroupNumber, iOldTask);
#ifndef CYRANO_TP_STANDALONE
            ExitTaskGroup(tgstop_stopped);
#else
            ExitTaskGroup();
#endif
            if(GetTrace() && GetTrace()->IsTrace("full"))
            {
                GetTraceStream(m_pTrace) << "CAsyncVirtualUser::EndScript: popping (0) script context from stack for script '"
                    << GetCurrScriptName() << "'. Stack depth will be " << m_ContextDataStack.size()-1 << "\r\n";
                ReleaseTraceStream();
            }

            delete CURR_SCRIPT_CTXT_DATA;
            m_ContextDataStack.pop();
            m_ScriptNameStack.pop();
        }
        else
        {
            // Increment the task iteration counter.
            ++m_lCurTaskIteration;

            BOOL bFinishTask = FALSE;
            if(m_pCurTask->m_lDuration)
            {
                // Execution of the task is limited by a duration.
                FILETIME ftNow;
                GetSystemTimeAsFileTime(&ftNow);
                unsigned __int64 qwNow = ftNow.dwHighDateTime;
                qwNow <<= 32;
                qwNow |= ftNow.dwLowDateTime;

                bFinishTask = (qwNow >= m_qwEndTask);

                if(GetTrace() && GetTrace()->IsTrace("full") && bFinishTask)
                {
                    GetTraceStream(GetTrace()) << "VU " << m_dwManagerNumber
                        << ": stopping Task '" << m_pCurTask->m_strScriptName
                        << "' after " << m_lCurTaskIteration-1 << " iteration(s) because duration has expired.\r\n";
                    ReleaseTraceStream();
                }
            }
            else
            {
                // Execution of the task is limited by an iteration count.
                bFinishTask = (m_lCurTaskIteration > m_pCurTask->m_lIterations);

                if(GetTrace() && GetTrace()->IsTrace("full") && bFinishTask)
                {
                    GetTraceStream(GetTrace()) << "VU " << m_dwManagerNumber
                        << ": stopping Task '" << m_pCurTask->m_strScriptName
                        << "' after " << m_lCurTaskIteration-1 << " iteration(s) because all iterations have been performed.\r\n";
                    ReleaseTraceStream();
                }
            }

            // Increment the task iteration count.
            if(!bFinishTask)
            {
                // Must perform the next task iteration.
                pController->OnUserExitTask(*this,
                    m_pTaskGroup->m_iTaskGroupNumber, iOldTask);
                res = ReStartScript();
            }
            else
            {
                //// Move on to the next task.

                // Initialize the task iteration count.
                m_lCurTaskIteration = 1;

                // Get the next task to execute in the task group.
                if(pController->GetNextTask(m_pTaskGroup, m_pCurTask))
                {
                    if(GetTrace() && GetTrace()->IsTrace("full"))
                    {
                        GetTraceStream(GetTrace()) << "CAsyncVirtualUser::EndScript: moving to next task: " << m_pCurTask->m_strScriptName << "\r\n";
                        ReleaseTraceStream();
                    }

                    if(GetTrace() && GetTrace()->IsTrace("full"))
                    {
                        GetTraceStream(m_pTrace) << "CAsyncVirtualUser::EndScript: popping (1) script context from stack for script '"
                            << GetCurrScriptName() << "'. Stack depth will be " << m_ContextDataStack.size()-1 << ".\r\n";
                        ReleaseTraceStream();
                    }

                    delete CURR_SCRIPT_CTXT_DATA;
                    m_ContextDataStack.pop();
                    m_ScriptNameStack.pop();

                    pController->OnUserExitTask(*this,
                        m_pTaskGroup->m_iTaskGroupNumber, iOldTask);
                    res = StartScript(m_pCurTask->m_strScriptName);
                }
                else
                {
                    // We reached the end of the task group.
                    ++m_lCurTaskGroupIteration;

                    BOOL bFinish = FALSE;
#ifndef CYRANO_TP_STANDALONE
                    TaskGroupStopReason eReason;
#endif
                    if(m_pTaskGroup->m_lTaskGroupDuration > 0)
                    {
                        // The end of task group execution is determined by the duration.
                        // The task group iteration limit is ignored.
                        FILETIME ftNow;
                        GetSystemTimeAsFileTime(&ftNow);
                        unsigned __int64 qwNow = ftNow.dwHighDateTime;
                        qwNow <<= 32;
                        qwNow |= ftNow.dwLowDateTime;

                        bFinish = (qwNow >= m_pTaskGroup->m_qwEndTime);
#ifndef CYRANO_TP_STANDALONE
                        if(bFinish)
                            eReason = tgstop_timeout;
#endif

                        if(GetTrace() && GetTrace()->IsTrace("full") && bFinish)
                        {
                            GetTraceStream(GetTrace()) << "VU " << m_dwManagerNumber << ": stop Task Group because duration expired.\r\n";
                            ReleaseTraceStream();
                        }
                    }
                    else
                    {
                        // Only the task group iteration count limits execution.
                        bFinish = (m_lCurTaskGroupIteration > m_pTaskGroup->m_lNumTaskGroupIterations);
#ifndef CYRANO_TP_STANDALONE
                        if(bFinish)
                            eReason = tgstop_completed;
#endif

                        if(GetTrace() && GetTrace()->IsTrace("full") && bFinish)
                        {
                            GetTraceStream(GetTrace()) << "VU " << m_dwManagerNumber << ": stop Task Group because all iterations performed.\r\n";
                            ReleaseTraceStream();
                        }
                    }

                    if(!bFinish)
                    {
                        // Must perform the next task group iteration:
                        // loop around to the first task.
                        m_pCurTask = pController->GetFirstTask(m_pTaskGroup);

                        if(GetTrace() && GetTrace()->IsTrace("full"))
                        {
                            GetTraceStream(GetTrace()) << "CAsyncVirtualUser::EndScript: Starting task group iteration " << m_lCurTaskGroupIteration << "\r\n";
                            ReleaseTraceStream();
                        }

                        if(GetTrace() && GetTrace()->IsTrace("full"))
                        {
                            GetTraceStream(m_pTrace) << "CAsyncVirtualUser::EndScript: popping (2) script context from stack for script '"
                                << GetCurrScriptName() << "'. Stack depth will be " << m_ContextDataStack.size()-1 << "\r\n";
                            ReleaseTraceStream();
                        }

                        delete CURR_SCRIPT_CTXT_DATA;
                        m_ContextDataStack.pop();
                        m_ScriptNameStack.pop();

                        pController->OnUserExitTask(*this,
                            m_pTaskGroup->m_iTaskGroupNumber, iOldTask, TRUE);
                        res = StartScript(m_pCurTask->m_strScriptName);
                    }
                    else
                    {
                        pController->OnUserExitTask(*this,
                            m_pTaskGroup->m_iTaskGroupNumber, iOldTask, TRUE);
#ifndef CYRANO_TP_STANDALONE
                        ExitTaskGroup(eReason);
#else
                        ExitTaskGroup();
#endif
                        if(GetTrace() && GetTrace()->IsTrace("full"))
                        {
                            GetTraceStream(m_pTrace) << "CAsyncVirtualUser::EndScript: popping (3) script context from stack for script '"
                                << GetCurrScriptName() << "'. Stack depth will be " << m_ContextDataStack.size()-1 << "\r\n";
                            ReleaseTraceStream();
                        }

                        delete CURR_SCRIPT_CTXT_DATA;
                        m_ContextDataStack.pop();
                        m_ScriptNameStack.pop();
                    }
                }
            }
        }
    }
    else    // m_ContextDataStack.size() > 1
    {
        // We are inside a nested script (a Call Script command).
        // Return to the calling script.

        // Do not delete the first item, for algorithm reasons (Cf. TContext...).
        CScriptContextData* pCtxtData = CURR_SCRIPT_CTXT_DATA;

        // Pop member data.
        if(GetTrace() && GetTrace()->IsTrace("full"))
        {
            GetTraceStream(m_pTrace) << "CAsyncVirtualUser::EndScript: popping (4) script context from stack for script '"
                << GetCurrScriptName() << "'. Stack depth will be " << m_ContextDataStack.size()-1 << "\r\n";
            ReleaseTraceStream();
        }
        delete pCtxtData;
        m_ContextDataStack.pop();
        m_ScriptNameStack.pop();

        // Set m_execPhase to running because that's the calling script's value.
        m_execPhase = EXEC_RUNNING;
    }

    return res;
}

#ifndef CYRANO_TP_STANDALONE
void CAsyncVirtualUser::ExitTaskGroup(TaskGroupStopReason eReason)
#else
void CAsyncVirtualUser::ExitTaskGroup()
#endif
{
    // This VU finished executing the task group.
    // Release the requests.
    if(GetTrace() && GetTrace()->IsTrace("full"))
    {
        GetTraceStream(GetTrace()) << "CAsyncVirtualUser::ExitTaskGroup: Task Group execution complete for VU " << m_dwManagerNumber << ".\r\n";
        ReleaseTraceStream();
    }

    {
        CLock lock(m_sl);
        for(RequestsType::iterator it = m_requests.begin(); it != m_requests.end(); ++it)
        {
            if(((CAsyncHTTPRequest*)*it)->GetConnectionId() != ULONG_MAX)
                ReleaseRequest((CAsyncHTTPRequest*)*it);
        }
    }

    // If all VUs are done, the task group is over.
    CAsyncVirtualUserController* pController = (CAsyncVirtualUserController*)m_pController;
    pController->OnVirtualUserOver(m_dwManagerNumber);
    if(pController->DecTaskGroupVUs(m_pTaskGroup))
    {
        if(GetTrace() && GetTrace()->IsTrace("full"))
        {
            GetTraceStream(GetTrace()) << "CAsyncVirtualUser::ExitTaskGroup: Task Group execution complete for all VUs.\r\n";
            ReleaseTraceStream();
        }

        // Notify the controller.
        // After this notification, the virtual user should not attempt to access
        // the task group any more.
#ifndef CYRANO_TP_STANDALONE
        pController->OnTaskGroupComplete(m_pTaskGroup->m_iTaskGroupNumber, eReason);
#else
        pController->OnTaskGroupComplete(m_pTaskGroup->m_iTaskGroupNumber);
#endif
    }
}

///////////////////////////////////////////////////////////////////////////////
// Evaluate parameters for Call Script Entry Command
// string of comma separated variable names to be set
// 0 for success
int CAsyncVirtualUser::Entry(const TtValList &varPrmListP)
{
    // Get pointer to parameter descriptor list for formal parameters.
    const list<CECExprDsc>* varExprListP = varPrmListP.ExprListP();

    // Get pointer to parameter descriptor list for actual parameters.
    const list<CECExprDsc>* exprListP = m_paramsInfo.ExprListP();
    list<CECExprDsc>::const_iterator dscItr = exprListP->begin();

    // Iterate through the formal parameter list, evaluating the actual parameter
    // values and setting the corresponding parameter variables.
    string prmVal;  // Evaluated parameter value
    string varIdx;  // Variable index
    list<string>::const_iterator ParamValIter = m_Params.begin();
    for (list<CECExprDsc>::const_iterator varDscItr = varExprListP->begin();
                                          varDscItr != varExprListP->end();
                                        ++varDscItr)
    {
        // Get pointer to the descriptor for the actual parameter and
        // report an error if there isn't one.
        // USER_AGENT and USE_PAGE_TIMERS are special cases which always
        // require a value!
        if (dscItr == exprListP->end())
        {
            if(   varDscItr->VarNam() != "USER_AGENT"
               && varDscItr->VarNam() != "USE_PAGE_TIMERS")
                throw mdbParseException("Required Parameter Missing");// Return error if no actual parameter
        }
        else
        {
            CECExprDsc prmValDsc = *dscItr;			// Get actual parameter descriptor
            ++dscItr;					            // Point to next actual parameter

            // Get pointer to the descriptor for the formal parameter and
            // return failure, if the formal and actual parameters do NOT have the same data type.
            if (prmValDsc.DTyp() != varDscItr->DTyp()) 
                throw mdbParseException("Paramter type mismatch");
        }

        // Put the actual parameter value into the corresponding parameter variable.
        list<CECOperand>::const_iterator varExprItr = varDscItr->ExprIt();
        if (!(varPrmListP.EvaluateVarIdx(*this, &varExprItr, &varIdx))) 
            throw mdbParseException("Failed to Evaluate");

        if(ParamValIter != m_Params.end())
        {
            prmVal = *ParamValIter;     // Corresponding value in list
            ++ParamValIter;
        }
        else if(varDscItr->VarNam() == "USER_AGENT")
            prmVal = "DEFAULT";
        else if(varDscItr->VarNam() == "USE_PAGE_TIMERS")
            prmVal = (m_pTaskGroup->m_bUsePageTimers) ? "1" : "0";
        else
            throw mdbParseException("Missing value for paramater " + varDscItr->VarNam());

        set_variable(varDscItr->VarNam(), prmVal, varIdx);   // Set parameter variable	

    } // End of for

    return 0;
}

BOOL CAsyncVirtualUser::TryAcquireSemaphore(const string &s_name, unsigned l_scope)
{
#ifndef CYRANO_TP_STANDALONE

    if(m_szCurrentSemaphoreName != s_name)
    {
        if(m_pCurrentSemaphore != NULL)
            delete m_pCurrentSemaphore;

        m_pCurrentSemaphore = new CECSemaphore (s_name, l_scope);
    }

    BOOL bRes = m_pCurrentSemaphore->tryacquire();
    if(bRes == TRUE)
    {
        // No need to keep trace of that semaphore.
        delete m_pCurrentSemaphore;
        m_pCurrentSemaphore = NULL;
    }

    return bRes;

#else

    return TRUE;

#endif
}


//// --------------------------------------------------------------------------------
//// TContext virtual methods.
//// --------------------------------------------------------------------------------
void CAsyncVirtualUser::CleanUp()
{
}

//Call thread type specific method for setting note Text.
void CAsyncVirtualUser::SetNoteText(const string& msg)
{
    m_noteText = msg;
}

string CAsyncVirtualUser::get_test_name()
{
    return m_szTestName;
}


// Clear up this Script on exit from script with/without keep alive options set
int CAsyncVirtualUser::ExitCommand(bool keepalive, const string& retvar)
{
    if(m_ContextDataStack.top()->exit_called_flag == TRUE)
	    return 0;   // Have already exited don't try and do it again

#ifndef CYRANO_TP_STANDALONE

    // Deregister ScriptObject from VariableFactories
    CECInetInfo inet;
    CORBA::Object_var obj = pCServer->FindObjectWithName(VarFactory::CLASS_NAME,
        inet.GetIPAddressAsString(g_pTExecuter->GetRepositoryHost()).c_str());
    VarFactory_var VF = VarFactory::_narrow( obj.inout());

    if(!CORBA::is_nil(VF))
    {
        CORBA::String_var cb_scriptname = CORBA::string_dup(ScriptName().c_str());
        // crashes !?!:   VF->DeregisterScript(cb_scriptname.in(), keepalive);

        TContext::ExitCommand(keepalive, retvar);
        m_ContextDataStack.top()->exit_called_flag = true;  // Have explicitly called exit in Script
    }
    else
    {
        if(GetTrace())
        {
            GetTraceStream(GetTrace()) << "Variable Factory unreachable.\r\n";
            ReleaseTraceStream();
        }
    }
    return 0;   //Got here without throwing an exception so must be successful

#else
    m_ContextDataStack.top()->exit_called_flag = true;  // Have explicitly called exit in Script
    return 0;

#endif
}


///////////////////////////////////////////////////////////////////////////////
// Terminates the current test case, as either passed or failed depending
// on the status of the flag passed in.
//
// @rdesc   integer 0 if the test case is terminated successfully, otherwise
//	    false.
// @parm    int | result | Flag determining the state of termination
//	    (TOF_TCTRES_PASS/FAIL).
// @parm    description of TestCase passed or failed (taken from start test case
//	    record at prepare time).
int CAsyncVirtualUser::TCterminate(int result, const string& desc)
{
	if(result == TOF_TCTRES_FAIL)
	    ++m_dwTestCaseFail;
	else
	    ++m_dwTestCasePass;

	return TContext::TCterminate(result, desc);
}

string CAsyncVirtualUser::getThreadVariable(const string& strName, const string& strIndex) const
{
    string strRet;

    // Lookup the variable.
    string strInternalName = strName + '_' + strIndex;
    ThreadVariables_t::const_iterator pStrVal = m_threadVariables.find(strInternalName);
    if(pStrVal != m_threadVariables.end())
        strRet = pStrVal->second;

    return strRet;
}

int CAsyncVirtualUser::setThreadVariable(const string& strName, const string& strIndex, const string& strValue)
{
    // Lookup the variable.
    string strInternalName = strName + '_' + strIndex;
    ThreadVariables_t::iterator pStrVal = m_threadVariables.find(strInternalName);
    if(pStrVal == m_threadVariables.end())
    {
        // Create the variable.
        m_threadVariables.insert(make_pair(strInternalName, strValue));
    }
    else
    {
        // Update the variable's value.
        pStrVal->second = strValue;
    }

    return 0;
}

// Read the statistics verbose level flag. Tell the request structures what they're gonna have to do of the stats.
BOOL CAsyncVirtualUser::MustDumpStats() const
{
    return m_pTaskGroup->m_iLoggingLevel == TaskGroupDefinition::loglev_high
        || m_pTaskGroup->m_iLoggingLevel == TaskGroupDefinition::loglev_low && m_dwExternalVUID <= 10;
}
