/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef _OPENSTA_PERFORMANCE_H_
#define _OPENSTA_PERFORMANCE_H_

////////////////////////////////////////////////////////////////////////////////
// OpenSTA
//
// Performance counters definition file.
//
// Return values: unless otherwise stated, a DWORD return value is a
// GetLastError() error code.
////////////////////////////////////////////////////////////////////////////////

#include "Registry.hpp"
#include<vector>
#include<list>
#include<algorithm>
#include<iosfwd>

#ifdef _DEBUG
#   include<malloc.h>
#endif

namespace OpenSTA_Performance {

/////////////////// Performance object / counter static map ////////////////////

#define BEGIN_PERFORMANCE(name) static const CPerfMapType name[]={
#define PERF_OBJECT(dwKey,zName,zHelp,dwDetailLevel,                    \
    dwMaxInstances,dwMaxInstanceNameSize,dwDefaultCounterKey)           \
    {FALSE,dwKey,zName,zHelp,dwDetailLevel,                             \
    0,0,                                                                \
    TRUE,dwMaxInstances,dwMaxInstanceNameSize,dwDefaultCounterKey},
#define PERF_OBJECT_NO_INSTANCE(dwKey,zName,zHelp,dwDetailLevel,dwDefaultCounterKey)\
    {FALSE,dwKey,zName,zHelp,dwDetailLevel,                             \
    0,0,                                                                \
    FALSE,PERF_NO_INSTANCES,0,dwDefaultCounterKey},
#define PERF_COUNTER(dwKey,zName,zHelp,dwDetailLevel,dwType,            \
    lDefaultScale)                                                      \
    {TRUE,dwKey,zName,zHelp,dwDetailLevel,                              \
    dwType,lDefaultScale,                                               \
    0,0,0,0},
#define END_PERFORMANCE };
#define MAP_SIZE(name) (sizeof(name)/sizeof(name[0]))

struct CPerfMapType
{
    BOOL    m_bIsCounter;           // TRUE if item is a counter; FALSE if item is an object.

    // Attributes common to objects and counters:
    DWORD   m_dwKey;                // The key by which the object / counter is identified by the source.
    PCWSTR  m_zName;                // The name of the object / counter.
    PCWSTR  m_zHelp;                // The help text for this object / counter.
    DWORD   m_dwDetailLevel;        // Level of detail of object / counter.

    // Counter attributes:
    DWORD   m_dwType;               // Counter type.
    LONG    m_lDefaultScale;        // Default scale for counter.

    // Object attributes:
    BOOL    m_bSupportsInstances;   // Does the object support instances?
    DWORD   m_dwMaxInstances;       // Maximum number of instances supported for this object.
    DWORD   m_dwMaxInstanceNameSize;// Maximum size of instance name (characters).
    DWORD   m_dwDefaultCounterKey;  // Key of default counter.
};

/////////////////// Performance object / counter definitions ////////////////////

// This class encapsulates a PERF_COUNTER_BLOCK structure.
// To simplify counter management, all counters are stored as 64-bit counters.
struct CPerfCounterBlock
{
    PERF_COUNTER_BLOCK* m_pPcb;
    DWORD               m_dwNumCounters;

    // Construct.
    CPerfCounterBlock(DWORD dwNumCounters) : m_pPcb(NULL), m_dwNumCounters(dwNumCounters) {}

    // Initializes the PERF_COUNTER_BLOCK structure.
    void InitializeStruct(PERF_COUNTER_BLOCK* pPcb)
    {
        DWORD dwSize = InternalGetSize(m_dwNumCounters);
        ZeroMemory((PVOID)pPcb, dwSize);
        pPcb->ByteLength = dwSize;

        m_pPcb = pPcb;
    }

    // Get the size of the structure in bytes.
    DWORD GetSize() const
    {return InternalGetSize(m_dwNumCounters);}

    // Counter block size for a given number of counters.
    static DWORD GetSize(DWORD dwNumCounters) {return InternalGetSize(dwNumCounters);}

    // Returns a counter offset given its index (zero-based value identifying
    // the counter within its object).
    static DWORD IndexToOffset(DWORD dwIndex) {return sizeof(DWORD) + dwIndex * sizeof(__int64);}

    // Get a counter value by offset.
    const BYTE* operator[](DWORD dwOffset) const {return ((const BYTE*)m_pPcb) + dwOffset;}
    BYTE* operator[](DWORD dwOffset) {return ((BYTE*)m_pPcb) + dwOffset;}

#ifdef _DEBUG
    void Dump(std::ostream& os);
#endif

private:
    static DWORD InternalGetSize(DWORD dwNumCounters)
    {return sizeof(PERF_COUNTER_BLOCK) + dwNumCounters * sizeof(__int64);}
};

// This class defines the attributes common to a performance object and a
// performance counter.
struct CPerfCommon
{
    DWORD   m_dwKey;                // The key by which the object / counter is identified.
    PCWSTR  m_zName;                // The name of the object / counter.
    PCWSTR  m_zHelp;                // The help text for this object / counter.
    DWORD   m_dwNameTitleIndex;     // The object / counter number.
    DWORD   m_dwHelpTitleIndex;     // The object / counter help number.
    DWORD   m_dwDetailLevel;        // Level of detail of object / counter.
};

// This class defines a performance counter and encapsulates a PERF_COUNTER_DEFINITION
// structure.
struct CPerfCounter : public CPerfCommon
{
    DWORD   m_dwType;               // Counter type.
    LONG    m_lDefaultScale;        // Default scale for counter.
    DWORD   m_dwIndex;              // Index of the counter in the object's list.
    PERF_COUNTER_DEFINITION* m_pPcd;

    // Initializes the PERF_COUNTER_DEFINITION structure.
    void InitializeStruct(PERF_COUNTER_DEFINITION* pPcd)
    {
        ZeroMemory(pPcd, sizeof(PERF_COUNTER_DEFINITION));

        pPcd->ByteLength            = sizeof(PERF_COUNTER_DEFINITION);
        pPcd->CounterNameTitleIndex = m_dwNameTitleIndex;
        pPcd->CounterHelpTitleIndex = m_dwHelpTitleIndex;
        pPcd->DefaultScale          = m_lDefaultScale;
        pPcd->DetailLevel           = m_dwDetailLevel;
        pPcd->CounterType           = m_dwType;
        pPcd->CounterSize           = 8;
        pPcd->CounterOffset         = CPerfCounterBlock::GetSize(m_dwIndex);

        m_pPcd = pPcd;
    }

#ifdef _DEBUG
    void Dump(std::ostream& os);
#endif
};

// This class encapsulates a PERF_INSTANCE_DEFINITION structure.
struct CPerfInstance
{
    typedef std::list<const CPerfInstance*> InstanceSons;

    InstanceSons              m_sons;               // The instances which are sons of the current instance.
    BOOL                      m_bUsed;              // TRUE if the instance definition is used.
    PERF_INSTANCE_DEFINITION* m_pPid;
    CPerfCounterBlock         m_cb;
    DWORD                     m_dwInstanceIndex;    // Replicated instance index: zero-based index in the object's
                                                    // instance vector.
    DWORD                     m_dwParentObjectTitleIndex; // Replicated parent object's title index (object number),
                                                          // or 0 if no parent.
    CPerfInstance*            m_pParentInstance;    // NULL if no parent instance.
    PCWSTR                    m_zInstanceName;      // Name of the instance. NULL if a unique id is used instead.
    LONG                      m_lUniqueId;          // Unique id of the instance. N/A if m_zInstanceName is not NULL.
    const DWORD               m_dwMaxNameSize;      // Max name size in characters.


    // Construct with instance name.
    CPerfInstance(DWORD dwInstanceIndex, PCWSTR zInstanceName, DWORD dwMaxNameSize, DWORD dwNumCounters,
        DWORD dwParentObjectTitleIndex = 0, CPerfInstance* pParentInstance = NULL)
        : m_bUsed(FALSE), m_pPid(NULL), m_dwInstanceIndex(dwInstanceIndex), m_dwMaxNameSize(dwMaxNameSize),
          m_pParentInstance(pParentInstance), m_zInstanceName(zInstanceName), m_lUniqueId(0), m_cb(dwNumCounters),
          m_dwParentObjectTitleIndex(dwParentObjectTitleIndex) {}

    // Construct with unique id.
    CPerfInstance(DWORD dwInstanceIndex, LONG lUniqueId, DWORD dwMaxNameSize, DWORD dwNumCounters,
        DWORD dwParentObjectTitleIndex = 0, CPerfInstance* pParentInstance = NULL)
        : m_bUsed(FALSE), m_pPid(NULL), m_dwInstanceIndex(dwInstanceIndex), m_dwMaxNameSize(dwMaxNameSize),
          m_pParentInstance(pParentInstance), m_zInstanceName(NULL), m_lUniqueId(lUniqueId), m_cb(dwNumCounters),
          m_dwParentObjectTitleIndex(dwParentObjectTitleIndex) {}

    // Initializes a PERF_INSTANCE_DEFINITION structure.
    void InitializeStruct(PERF_INSTANCE_DEFINITION* pPid);

    // Get the size of the structure in bytes.
    DWORD GetSize() const {return InternalGetSize(m_pPid->NameLength);}

    // Get the size of the structure including the size of the counter block.
    DWORD GetSizeWithCounterBlock() const {return GetSize() + m_cb.GetSize();}

    // Get the maximum size of the instance with the counter block.
    static DWORD GetSizeWithCounterBlock(DWORD dwCounters, DWORD dwMaxNameSize)
    {return InternalGetSize((dwMaxNameSize + 1) * sizeof(WCHAR)) + CPerfCounterBlock::GetSize(dwCounters);}

#ifdef _DEBUG
    void Dump(std::ostream& os);
#endif

private:
    static DWORD InternalGetSize(DWORD dwInstanceNameSizeInBytes)
    {
        // Return the size of the PERF_INSTANCE_DEFINITION structure, plus the
        // instance name, plus a possible padding to the next 32-bit boundary.
        if((dwInstanceNameSizeInBytes & 2) == 2)
            dwInstanceNameSizeInBytes += 2;
        return sizeof(PERF_INSTANCE_DEFINITION) + dwInstanceNameSizeInBytes;
    }

    // Forbidden.
    operator=(const CPerfInstance&);
};

// This class defines a performance object and encapsulates a PERF_OBJECT_TYPE
// structure.
struct CPerfObject : public CPerfCommon
{
    typedef std::vector<CPerfInstance*> Instances;
    typedef std::vector<CPerfCounter*>  Counters;

    CPerfObject() : m_pDefaultCounter(NULL), m_pPot(NULL), m_cb(0) {}
    ~CPerfObject();

    // Get the maximum size occupied by the object, its
    // instances and counters as the block formed by its
    // PERF_OBJECT_TYPE, PERF_COUNTER_DEFINITION,
    // PERF_INSTANCE_DEFINITION and PERF_COUNTER_BLOCK
    // structures. Take into account the padding necessary
    // to align the next object block on an 8-byte boundary.
    DWORD GetMaxStructSize() const {return GetStructSize(m_dwMaxInstances);}

    // Get the size occupied by the object. The difference with
    // GetMaxStructSize() is that the calculations are performed
    // for a given number of instances.
    DWORD GetStructSize(DWORD dwNumInstances) const;

    // Get the size of the PERF_OBJECT_TYPE structure in bytes.
    static DWORD GetSize() {return sizeof(PERF_OBJECT_TYPE);}

    // After the counters have been added (m_counters is initialized),
    // and the default counter (m_pDefaultCounter) has been set,
    // this method initializes the PERF_OBJECT_TYPE structure.
    // It returns a valid address for the next PERF_OBJECT_TYPE structure,
    // assuming that the object uses its maximum size.
    PERF_OBJECT_TYPE* InitializeStruct(PERF_OBJECT_TYPE* pPot);

    // Add / Delete instance. To be called after InitializeStruct.
    CPerfInstance* AddInstance(LONG lUniqueID, DWORD dwParentObjectTitleIndex = 0,
        CPerfInstance* pParentInstance = NULL);
    CPerfInstance* AddInstance(PCWSTR zInstanceName, DWORD dwParentObjectTitleIndex = 0,
        CPerfInstance* pParentInstance = NULL);
    void DeleteInstance(CPerfInstance* pInstance);

    // Get a pointer to a 32-bit counter value.
    PLONG GetCounterValue(DWORD dwKey, CPerfInstance* pInstance = NULL);

#ifdef _DEBUG
    void Dump(std::ostream& os);
#endif

    BOOL    m_bSupportsInstances;   // Does the object support instances?
    DWORD   m_dwMaxInstances;       // Maximum number of instances supported.
    DWORD   m_dwMaxInstanceNameSize;// Maximum size of instance name (characters).
    Counters m_counters;            // Counters associated to this object.
    CPerfCounter* m_pDefaultCounter;// Object's default counter.
    PERF_OBJECT_TYPE* m_pPot;
    CPerfCounterBlock m_cb;         // Counter block (for objects that don't
                                    // support instances).
    Instances m_id;                 // Collection of instances of the object (for
                                    // objects that support instances).
};

///////////////////////////////////////////////////////////////////////////////////////////

// This class implements a cross-process exclusive lock.
class CCrossProcessLock
{
public:
    CCrossProcessLock(HANDLE hMutex) : m_hMutex(hMutex)
    {WaitForSingleObject(m_hMutex, INFINITE);}
    ~CCrossProcessLock()
    {ReleaseMutex(m_hMutex);}

private:
    HANDLE m_hMutex;

    // The following prototypes make the class stack-only:
    void* operator new(size_t);
    void* operator new[](size_t);
};

// This class implements core Windows NT/2000 performance counter management, used
// by performance counter source and client objects (see classes below).
// - Shared memory management.
// 
// A shared memory segment is created and initialized to hold the all the runtime
// data structures, as if a client application had requested all possible counters,
// with the maximum number of instances present.
// When a client application actually requests some counters, the data will be copied
// from the shared memory to the client buffer, and the total byte length will be
// adjusted.
class CPerfCore
{
public:
    //// --------------------------------------------------------------------------------
    //// Construction / Destruction.
    //// --------------------------------------------------------------------------------
    CPerfCore(PCWSTR zSourceName, PCWSTR zOpenFuncName, PCWSTR zCollectFuncName,
        PCWSTR zCloseFuncName);
    ~CPerfCore();

    // Parses the object / counter map and creates the oject / counter definitions.
    // m_dwFirstCounter, m_dwFirstHelp, m_dwLastCounter and m_dwLastHelp must already
    // be initialized.
    DWORD ParseMap(const CPerfMapType* const pMap, DWORD dwMapSize);

    // Adjust the object/counter title indexes so that they start at the actual first
    // counter value. This must be done after ParseMap(), and when the performance DLL
    // has been successfully installed.
    DWORD AdjustTitleIndexes();

protected:
    //// --------------------------------------------------------------------------------
    //// Types.
    //// --------------------------------------------------------------------------------
    typedef std::vector<CPerfObject*> PerfObjects;

    //// --------------------------------------------------------------------------------
    //// Data.
    //// --------------------------------------------------------------------------------
    HANDLE          m_hMutex;           // To protect the shared memory against concurrent
                                        // access.
    HANDLE          m_hMap;             // Handle of file-mapping object for shared memory.
    PBYTE           m_pbView;           // Mapped view for shared memory.
    PerfObjects     m_po;               // Collection of performance objects created after parsing
                                        // the object / counter map.
    DWORD           m_dwFirstCounter;
    DWORD           m_dwFirstHelp;
    DWORD           m_dwLastCounter;
    DWORD           m_dwLastHelp;
    PCWSTR          m_zSourceName;      // Name of the performance counter source. This will be
                                        // used to create a subkey under HKLM\SYSTEM\CurrentControlSet\Services.
    PCWSTR          m_zOpenFuncName;    // Name of the Open function exported from performance DLL.
    PCWSTR          m_zCollectFuncName; // Name of the Collect function exported from performance DLL.
    PCWSTR          m_zCloseFuncName;   // Name of the Close function exported from performance DLL.

    //// --------------------------------------------------------------------------------
    //// Helpers.
    //// --------------------------------------------------------------------------------

    // Determine the size needed for the shared memory.
    DWORD CalculateMaxSize() const;

    // Builds and initializes the shared memory.
    DWORD BuildSharedMemory(DWORD dwSize, BOOL bIsPerfSource = TRUE);
};

// This class implements Windows NT/2000 performance counter management:
// - Performance objects/counters installation
// - Performance objects/counters uninstallation
// - Performance counter values production.
class CPerfSource : public CPerfCore
{
public:
    //// --------------------------------------------------------------------------------
    //// Construction / Destruction.
    //// --------------------------------------------------------------------------------
    CPerfSource(PCWSTR zSourceName, PCWSTR zOpenFuncName, PCWSTR zCollectFuncName,
        PCWSTR zCloseFuncName)
        : CPerfCore(zSourceName, zOpenFuncName, zCollectFuncName, zCloseFuncName) {}

    //// --------------------------------------------------------------------------------
    //// Static methods.
    //// --------------------------------------------------------------------------------

    // Install the performance DLL into the registry. If it was already installed, then
    // this function first uninstalls the DLL, then performs a fresh installation.
    DWORD Install(PCWSTR zPerfDllPath);

    // Uninstalls the performance DLL from the registry.
    DWORD Uninstall();

    // After the map has been parsed, and AdjustTitleIndexes() has been called,
    // call this method to create and initialize the shared memory.
    // After this function is called successfully, you're ready to go!
    DWORD Initialize();

    // Get a pointer to a 32-bit counter value.
    // For objects supporting instances, adding or deleting an instance invalidates
    // the pointer returned from this function. Otherwise, it's useful to cache this
    // pointer to access a counter value, because this function is cpu consuming.
    PLONG GetCounterValue(DWORD dwCounterKey, CPerfInstance* pInstance = NULL);

    // Add / Delete instance. To be called after Initialize.
    CPerfInstance* AddInstance(DWORD dwObjectKey, LONG lUniqueID, DWORD dwParentObjectTitleIndex = 0,
        CPerfInstance* pParentInstance = NULL);
    CPerfInstance* AddInstance(DWORD dwObjectKey, PCWSTR zInstanceName, DWORD dwParentObjectTitleIndex = 0,
        CPerfInstance* pParentInstance = NULL);
    void DeleteInstance(DWORD dwObjectKey, CPerfInstance* pInstance);

#ifdef _DEBUG
    void Dump(std::ostream& os);
#endif

private:
    //// --------------------------------------------------------------------------------
    //// Helpers.
    //// --------------------------------------------------------------------------------

    // Add counter/help info to the registry.
    DWORD AddCounterHelpInfo(CRegistryKey& reg, DWORD& dwTitleIndex, BOOL bIsHelp = FALSE);

    // Delete counter/help info from the registry. dwLastTitleIndex is updated to reflect the
    // actual new last title index after the deletion.
    DWORD DeleteCounterHelpInfo(CRegistryKey& reg, DWORD dwFirstTitleIndex, DWORD& dwLastTitleIndex,
        BOOL bIsHelp = FALSE);

    // Get a pointer to an object given by its key.
    CPerfObject* GetObject(DWORD dwObjectKey)
    {
        CPerfObject* pObject = NULL;
        for(PerfObjects::iterator pObj = m_po.begin(); pObj != m_po.end(); ++pObj)
        {
            if((*pObj)->m_dwKey == dwObjectKey)
            {
                pObject = *pObj;
                break;
            }
        }
        return pObject;
    }

    // Get the number of digits of a decimal number.
    static DWORD GetDigitCount(DWORD dw);

    //// --------------------------------------------------------------------------------
    //// Forbidden.
    //// --------------------------------------------------------------------------------
    CPerfSource(const CPerfSource&);
    operator=(const CPerfSource&);
};

// This class implements a performance counter client able to consume data produced by
// CPerfCounterSource.
class CPerfClient : public CPerfCore
{
public:
    //// --------------------------------------------------------------------------------
    //// Construction / Destruction.
    //// --------------------------------------------------------------------------------
    CPerfClient(PCWSTR zSourceName, PCWSTR zOpenFuncName, PCWSTR zCollectFuncName,
        PCWSTR zCloseFuncName)
        : CPerfCore(zSourceName, zOpenFuncName, zCollectFuncName, zCloseFuncName) {}

    // After the map has been parsed, and AdjustTitleIndexes() has been called,
    // call this method to create the shared memory and initialize everything.
    // After this function is called successfully, you're ready to go!
    DWORD Initialize();

    // Data collection.
    DWORD Collect(PCWSTR zObjectNumbers, PVOID* ppData, PDWORD pdwBytes, PDWORD pdwObjectTypes);

private:
    //// --------------------------------------------------------------------------------
    //// Helpers.
    //// --------------------------------------------------------------------------------

    // Actually perform collection.
    DWORD DoCollect(PCWSTR zObjectNumbers, PVOID* ppData, PDWORD pdwBytes, PDWORD pdwObjectTypes);

    struct ObjToCollect
    {
        DWORD m_dwObjTitleIndex;
        BOOL  m_bToCollect;
    };

    // Mark an object "to be collected".
    void MarkToCollect(ObjToCollect* paOc, DWORD dwObjTitleIndex);

    // Get an object by title index.
    PERF_OBJECT_TYPE* GetObject(DWORD dwObjTitleIndex);

private:
    //// --------------------------------------------------------------------------------
    //// Forbidden.
    //// --------------------------------------------------------------------------------
    CPerfClient(const CPerfClient&);
    operator=(const CPerfClient&);
};

#ifdef _DEBUG
std::ostream& ToAnsi(std::ostream& os, LPCWSTR z);
#endif

}   // OpenSTA_Performance

#endif  // _OPENSTA_PERFORMANCE_H_
