/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef __GATEWAY__
#define __GATEWAY__


//==================================================================< Pragmas >
#ifdef  WIN32
#  pragma warning(disable : 4710)
#endif  // WIN32


//=================================================================< Includes >
#ifndef __DEFINITION__
#  include "definition.h"
#endif  // __DEFINITION__

#ifndef __EXCEPTION__
#  include <exception.hpp>
#endif  // __EXCEPTION__

#ifndef __MUTEX__
#  include <mutex.hpp>
#endif  // __MUTEX__

#ifndef __SERVICE__
#  include <service.hpp>
#endif  // __SERVICE__

#ifndef __MTHREAD__
#  include <mthread.hpp>
#endif  // __MTHREAD__

#ifndef __LOGFILE__
#  include <logfile.hpp>
#endif  // __LOGFILE__

#ifndef __SOCKET__
#  include <socket.hpp>
#endif  // __SOCKET__

#ifndef __MTSAFE__
#  include <mtsafe.hpp>
#endif  // __MTSAFE__

#ifndef __XALLOC__
#  include "xalloc.hpp"
#endif  // __XALLOC__

#include <list>
using std::list;
//======================================================< Definition of class >

class CGGatewayInfo : public CTMTSafe
{
    //-----------------------------------------------------------------------------
public:                     // Constructor(s) and destructor.
    CGGatewayInfo ()
: bAdministration_ (false), listenPort_ (0), bTargetAddressAutoDetection_ (true)
    { }
    ;
    CGGatewayInfo (const bool bAdministration, const string& targetAddress, const port_t listenPort, const string& library, const bool bTargetAddressAutoDetection)
: bAdministration_ (bAdministration), targetAddress_ (targetAddress), listenPort_ (listenPort), library_ (library), bTargetAddressAutoDetection_ (bTargetAddressAutoDetection || targetAddress.empty () )
    { }
    ;
    CGGatewayInfo (const CGGatewayInfo& gatewayInfo)
: bAdministration_ (gatewayInfo.bAdministration_), targetAddress_ (gatewayInfo.targetAddress_), listenPort_ (gatewayInfo.listenPort_), library_ (gatewayInfo.library_), bTargetAddressAutoDetection_ (gatewayInfo.bTargetAddressAutoDetection_)
    { }
    ;
    ~CGGatewayInfo ()
    { }
    ;

public:                     // Get methodes.
    bool                    getIsAdministrationListen (void) const
    {
        return bAdministration_ ;
    }
    const string&       getTargetAddress (void) const
    {
        return targetAddress_ ;
    }
    port_t                  getPort (void) const
    {
        return listenPort_ ;
    }
    const string&       getLibrary (void) const
    {
        return library_ ;
    }
    bool                    getTargetAddressAutoDetectionMode (void) const
    {
        return bTargetAddressAutoDetection_ ;
    }

    //-----------------------------------------------------------------------------
private:                    // Members.
    bool                    bAdministration_ ;
    string              targetAddress_ ;
    port_t                  listenPort_ ;
    string              library_ ;
    bool                    bTargetAddressAutoDetection_ ;
}
; // CGGatewayInfo

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////// Class CGTaskException //
///////////////////////////////////////////////////////////////////////////////
// Class used to keep tasks exception.
class CGTaskException : public CTException
{
    //-----------------------------------------------------------------------------
public:                     // Constructor(s) and destructor.
    CGTaskException (const string& clMsg, const string& clFilename, short sLineno, long lNumber = -1L)
: CTException (clMsg, clFilename, sLineno, lNumber)
    { }

    CGTaskException(CGTaskException & cg)
:CTException(cg)
    {}

public:                     // Operators.
    void                    operator= (const CGTaskException& exception)
    {
        CTException::operator= (exception) ;
    }

    //-----------------------------------------------------------------------------
}

; // CGTaskException

///////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////// End of Class CGTaskException //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////// Class CGTask //
///////////////////////////////////////////////////////////////////////////////
class CGTasksList ;
class CGTask ;
typedef vector<CGTask *> CGTaskStlList ;

class CGTask : public CTMTSafe
{
    //-----------------------------------------------------------------------------
public:                         // Types.
    enum enumTaskType
    {
        TASK_CONTROL,
        TASK_ADMINISTRATION_LISTEN,
        TASK_ADMINISTRATION_CLIENT,
        TASK_GATEWAY_LISTEN,
        TASK_GATEWAY_CLIENT
    } ;

public:                         // Constructor(s) and destructor.
    CGTask (void* pInitInfo, CGTask& father, CGTasksList& tasksList, enumTaskType eType, int iTaskId, bool bToBeCreated = true) ;
    ~CGTask () ;

public:                         // Operators.
    CGTask&                     operator = (const CGTask& task) ;
    bool                        operator == (const CGTask& task)
    {
        return (pThread_ == task.pThread_) ;
    } ;

public:                         // Controls methodes.
    void                        initializationTerminated (bool bInitializationIsOk = true) ;
    void                        setData (void* pData)
    {
        pData_ = pData;
    } ;
    void                        kill (bool bRemoveFromFather = false, bool bRemoveFromTasksList = false, bool bChildrenAlso = false) ;
    void                        join (void) ;
    void                        suspend (bool bChildrenAlso = false) ;
    void                        resume (bool bChildrenAlso = false) ;
    void                        reinitialize (bool bChildrenAlso = false) ;
    void                        reinitialized (void)
    {
        bToBeReinitialized_ = false ;
    } ;
    void                        invalidate (void)
    {
        delete pThread_ ;
        pThread_ = NULL ;
    } ;
    void                        setFather (CGTask* pFather)
    {
        pFather_ = pFather ;
    } ;

public:                         // Get methodes.
    const void*                 getInitInfo (void) const
    {
        return pInitInfo_ ;
    } ;
    const void*                 getData (void) const
    {
        return pData_ ;
    } ;
    CGTask*                     getFather (void) const
    {
        return pFather_ ;
    } ;
    CGTasksList&                getTasksList (void) const
    {
        return tasksList_ ;
    } ;
    CGTaskStlList&              getChildrenList (void)
    {
        return childrenList_ ;
    } ;
    CTThread*                   getThread (void) const
    {
        return pThread_ ;
    } ;
    const enumTaskType          getType (void) const
    {
        return eType_ ;
    } ;
    bool                        initializationIsTerminated (void) const
    {
        return bInitializationTerminated_ ;
    } ;
    bool                        initializationIsOk (void) const
    {
        return bInitializationIsOk_ ;
    } ;
    bool                        hasToBeReinitialized (void) const
    {
        return bToBeReinitialized_ ;
    } ;
    int                         getTaskId (void) const
    {
        return iTaskId_ ;
    } ;
    const string&           getTaskName (void) const
    {
        return name_ ;
    }
    bool                        isValid (void) const
    {
        return (pThread_ != NULL) ;
    } ;

    //-----------------------------------------------------------------------------
private:    // Members.
    void*                       pInitInfo_ ;
    void*                       pData_ ;
    CGTask*                     pFather_ ;
    CGTasksList&                tasksList_ ;
    CGTaskStlList               childrenList_ ;
    CTThread*                   pThread_ ;
    void*                       pStatus_ ;
    enumTaskType                eType_ ;
    bool                        bInitializationTerminated_ ;
    bool                        bInitializationIsOk_ ;
    bool                        bToBeReinitialized_ ;
    int                         iTaskId_ ;
    int                         iLevel_ ;
    string                  name_ ;
    static const char*          szTaskType_[] ;

    //-----------------------------------------------------------------------------
}

; // CGTask

///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////// End of Class CGTask //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////// Class CGTasksList //
///////////////////////////////////////////////////////////////////////////////
class CGTasksList : public CTMonitor<CTMutex>, virtual public CTMTSafe
{
    //-----------------------------------------------------------------------------
public:                     // Types.
    typedef bool (*fnTestTask) (CGTask*, void*) ;

public:                     // Constructor(s) and destructor.
    CGTasksList ()
: iNbTasks_ (0), conditionTaskInitialize_ (mutexTaskInitialize_)
    { }
    ;
    ~CGTasksList ()
    { }
    ;

public:                     // Controls methodes.
    void                    addCurrentTask (CGTask::enumTaskType eType, CGTask& father) ;
    void                    newTask (CGTask::enumTaskType eType, void* pInitInfo, CGTask& father) ;
    void                    killTask (int iIndex, bool bChildrenAlso = false) ;
    void                    killAll (void) ;
    void                    joinTask (int iIndex) ;
    void                    joinAll (void) ;
    void                    suspendTask (int iIndex, bool bChildrenAlso = false) ;
    void                    suspendAll (void) ;
    void                    resumeTask (int iIndex, bool bChildrenAlso = false) ;
    void                    resumeAll (void) ;
    void                    reinitializeTask (int iIndex, bool bChildrenAlso = false) ;

public:                     // Get methodes.
    CGTaskStlList&          getTasksList (void)
    {
        return tasksList_ ;
    } ;
    int                     getNbTasks (void)
    {
        return iNbTasks_ ;
    } ;
    CTMutex&                getMutexTaskInitialize (void)
    {
        return mutexTaskInitialize_ ;
    } ;
    CTCondition&            getConditionTaskInitialize (void)
    {
        return conditionTaskInitialize_ ;
    } ;
    CGTask&                 getCurrentTask (void) ;
    int                     index (fnTestTask testTask, void* pCompInfo) ;

    //-----------------------------------------------------------------------------
private:                    // Members.
    CGTaskStlList           tasksList_ ;
    int                     iNbTasks_ ;
    CTMutex                 mutexTaskInitialize_ ;
    CTCondition             conditionTaskInitialize_ ;

    //-----------------------------------------------------------------------------
}

; // CGTasksList

///////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////// End of Class CGTasksList //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////// Class CGMain //
///////////////////////////////////////////////////////////////////////////////
class CGMain : public CTService
{
    //-----------------------------------------------------------------------------
public:                     // Types.
    enum enumServiceCommand
    {
        SRV_START = 0,
        SRV_CONSOLE,
        SRV_SERVICE,
        SRV_INSTALL,
        SRV_UNINSTALL,
        SRV_STOP,
        SRV_PAUSE,
        SRV_CONTINUE,
        SRV_HELP
    } ;
    enum enumOption
    {
        OPT_CONSOLE = 0,
        OPT_SERVICE,
        OPT_INSTALL,
        OPT_UNINSTALL,
        OPT_STOP,
        OPT_PAUSE,
        OPT_CONTINUE,
        OPT_HELP,
        OPT_ADMINISTRATION_LISTEN_PORT,
        OPT_REUSE_PORT,
        OPT_CLIENT_CONFIG,
        OPT_LOGFILE,
        OPT_TRACE,
        OPT_LIBARG,
        OPT_LAST
    } ;

public:                     // Constructor(s) and destructor.
    CGMain (const string& serviceDisplayName, int argc, char* argv[]) ;
    ~CGMain () ;

public:                     // Controls methodes.
    void                    parse (int argc, char* argv[]) ;
    void                    go (void) ;
    void                    help (void) ;
    void                    signalInterrupt (int iSig) ;
    void                    abort (void) ;

public:                     // User methodes.
    void                    init (void) ;
    void                    main (void) ;
    void                    exit (void) ;
    void                    controlNotify (enumServiceControl) ;

public:                     // Get methodes.
    bool&                   getTraceMode (void)
    {
        return bTraceMode_ ;
    } ;
    CTOLogFileList&         getLogFileList (void)
    {
        return logFileList_ ;
    } ;
    bool&                   waitForTasks (void)
    {
        return bWaitTasks_ ;
    } ;
    string&             getCommandLibrary (void)
    {
        return commandLibrary_ ;
    } ;
    string&             getLibraryCommandLineArguments (void)
    {
        return libraryCmdLineArg_ ;
    } ;
    bool                    canReuseAddress (void)
    {
        return bReuseAddress_ ;
    } ;

    //-----------------------------------------------------------------------------
private:                    // Members.
    string              programName_ ;
    string              serviceDisplayName_ ;
    enumServiceCommand      eCommand_ ;
    bool                    bTraceMode_ ;
    string              logFileName_ ;
    ofstream                logFile_ ;
    CTOLogFileList          logFileList_ ;
    static const char*      azOption_[OPT_LAST] ;
    string              adminPort_ ;
    string              clientConfig_ ;
    CTMutex                 mutexContinue_ ;
    CTCondition             conditionContinue_ ;
    CGTasksList             tasksList_ ;
    bool                    bWaitTasks_ ;
    string              commandLibrary_ ;
    string              libraryCmdLineArg_ ;
    bool                    bReuseAddress_ ;

    //-----------------------------------------------------------------------------
}

; // CGMain

///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////// End of Class CGMain //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////// Globals Functions //
///////////////////////////////////////////////////////////////////////////////

//=====================================================================< Data >
extern CGMain*          GpProcessMain ;
extern CGTask           GnullTask ;
extern CGTasksList      GnullTaskList ;

//===============================================================< Prototypes >
extern void administrationListenTask (CGTask& task) ;
extern void administrationClientTask (CGTask& task) ;
extern void gatewayListenTask (CGTask& task) ;
extern void gatewayClientTask (CGTask& task) ;


///////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////// End of Globals Functions //
///////////////////////////////////////////////////////////////////////////////


#endif  // __GATEWAY__
